/*
 *    flattening/sorting routines for AC terms
 */
#include "defs.h"
#include "term_types.h"
#include "functions.h"

#if defined(DEBUG)
#include <stdio.h>
#endif

TERM_LIST *merge_sort();

/*
 *	flatten a term
 */
void flatten(TERM *t)
{
  TERM_LIST *p, *active;
  int count;
  TERM_TYPE type = t->type;
  
  ASSERT(type != AC_COMPRESSED, "flatten() - argument not in normal form");
  if (type == VARIABLE || type == CONSTANT)
    {
      return;
    }
#if defined(DEBUG)
  printf("flattenning subterms of t...\n");
#endif
  for (p = t->rest.f.arg_list; p; p = p->next_arg)
    {
      flatten(p->arg);
    }
  if (type != AC_NORMAL)
    {
      return;
    }

  count = 0;
  active = NULL;
  for (p = t->rest.f.arg_list; p; p = p->next_arg)
    {
      if (same_head(p->arg,t))
	{
	  count += p->arg->rest.f.list_len;
	  if (active != NULL)
	    active->next_arg = p->arg->rest.f.arg_list;
	  else
	    t->rest.f.arg_list = p->arg->rest.f.arg_list;
	  active = p->arg->rest.f.arg_tail;
	  active->next_arg = p->next_arg;
	  FREE(p->arg);
	  FREE(p);
	  p = active;	/* subtle - but correct! */
	}
      else
	{
	  count++;
	  active = p;
	}
    }
  t->rest.f.arg_tail = active;
  t->rest.f.list_len = count;
}

void ac_compress(t)
TERM *t;
{
  BOOL ac;
  TERM_LIST *p, *p2;
  AC_LIST *q;
  TERM *c;

#if defined(DEBUG)
  printf("checking t->type...\n");
#endif

  switch (t->type)
    {
    case VARIABLE:
    case CONSTANT:
      return;
    case FUNCTION:
      ac = FALSE;
      break;
    case AC_NORMAL:
    case COMMUTATIVE:
      ac = TRUE;
      break;
    case AC_COMPRESSED:
    case COM_COMPRESSED:
      ASSERT(FALSE, "ac_compress() - argument not in normal form");
    }

#if defined(DEBUG)
  printf("compressing subterms of t...\n");
#endif

  for (p = t->rest.f.arg_list; 
       p != NULL; 
       p = p->next_arg)
    ac_compress(p->arg);

#if defined(DEBUG)
  printf("compressing at top...\n");
#endif

  if (ac) 
    {
#if defined(DEBUG)
      printf("sorting subterms of t=");
      print_term_eker(t);
      printf("...\n");
#endif
      p = merge_sort(t->rest.f.arg_list, t->rest.f.list_len);
#if defined(DEBUG)
      printf("compressing list of subterms...\n");
#endif
      t->rest.a.arg_count = t->rest.f.list_len;
      t->rest.a.ac_list = q = MALLOC(AC_LIST);
      q->arg = c = p->arg;
      q->mult = 1;
      p2 = p->next_arg;
      FREE(p);
      for(p = p2; p; p = p2)
	{
	  if (term_cmp(c, p->arg) == 0)
	    {
	      (q->mult)++;
	      destroy_term(p->arg);
	    }
	  else
	    {
	      q = (q->next_ac = MALLOC(AC_LIST));
	      q->arg = (c = p->arg);
	      q->mult = 1;
	    }
	  p2 = p->next_arg;
	  FREE(p);
	}
      q->next_ac = NULL;
/*      t->rest.a.ac_tail = q; */ /* do we ever use the tail? no !*/
      if (t->type == AC_NORMAL)
	{
	  t->type = AC_COMPRESSED;
	}
      else
	{
	  t->type = COM_COMPRESSED;
	}
    }
}

/*
 *	Sort argument lists in descending order
 */
TERM_LIST *merge_sort(p, len)
TERM_LIST *p;
int len;
{
  TERM_LIST *p2, *q, *base;
  int i, l;

#if defined(DEBUG)
  printf("spliting list (len %d)...\n",len);
#endif
  if(len <= 1)
    return(p);
  l = len / 2;
  for(q = p, i = l - 1; i > 0; i--)
    q = q->next_arg;
  p2 = q->next_arg;
  q->next_arg = NULL;
#if defined(DEBUG)
  printf("sorting first sublist...\n");
#endif
  p = merge_sort(p, l);
#if defined(DEBUG)
  printf("sorting second sublist...\n");
#endif
  p2 = merge_sort(p2, len - l);
#if defined(DEBUG)
  printf("merging sublists...\n");
#endif
  if(term_cmp(p->arg, p2->arg) >= 0){
    q = p;
    p = p->next_arg;
    if(p == NULL){
      q->next_arg = p2;
      return(q);
    }
  }
  else{
    q = p2;
    p2 = p2->next_arg;
    if(p2 == NULL){
      q->next_arg = p;
      return(q);
    }
  }
  base = q;
  for(;;){
    if(term_cmp(p->arg, p2->arg) >= 0){
      q = (q->next_arg = p);
      p = p->next_arg;
      if(p == NULL){
        q->next_arg = p2;
        break;
      }
    }
    else{
      q = (q->next_arg = p2);
      p2 = p2->next_arg;
      if(p2 == NULL){
        q->next_arg = p;
        break;
      }
    }
  }
  return(base);
}

/*
 *	Compare flattened/sorted terms using lexicographic order for
 *	argument list of free function symbols and multiset order for
 *	argument lists of AC function symbols.
 */

int term_cmp(TERM *t1, TERM *t2)
{
  TERM_TYPE type1 = t1->type, type2 = t2->type;
  TERM_LIST *p1, *p2;
  AC_LIST *q1, *q2;
  int r;
  
  if(((int) type1) < ((int) type2)) return -1;
  if(((int) type1) > ((int) type2)) return 1;
  if(t1->sym < t2->sym) return -1;
  if(t1->sym > t2->sym) return 1;
  if (type1 == CONSTANT || type1 == VARIABLE) return 0;
  if (type1 == FUNCTION)
    { /* lexicographic ordering on subterms */
      for (p1 = t1->rest.f.arg_list, 
	   p2 = t2->rest.f.arg_list; ;
	   p1 = p1->next_arg, 
	   p2 = p2->next_arg)
	{
	  if (p1 == NULL) return p2 == NULL ? 0 : -1;
	  if (p2 == NULL) return 1;
	  r = term_cmp(p1->arg, p2->arg);
	  if (r != 0) return r;
	}
    }
  else
    { /* lexicographic ordering on subterms */
      for (q1 = t1->rest.a.ac_list, 
	   q2 = t2->rest.a.ac_list; ;
	   q1 = q1->next_ac, 
	   q2 = q2->next_ac)
	{
	  if (q1 == NULL) return q2 == NULL ? 0 : -1;
	  if (q2 == NULL) return 1;
	  r = term_cmp(q1->arg, q2->arg);
	  if (r != 0) return r;
          if (q1->mult < q2->mult) return -1;
          if (q1->mult > q2->mult) return 1;
    }
  }
}

void destroy_term(TERM *t)
{
  TERM_LIST *p1, *p2;
  AC_LIST *q1, *q2;

#if defined(DEBUG)
  printf ("destroy_term: t->type = %d\n",t->type);
#endif

  switch (t->type)
    {
    case VARIABLE:
    case CONSTANT:
      break;
    case FUNCTION:
    case AC_NORMAL:
      for (p1 = t->rest.f.arg_list; 
	   p1 != NULL ; 
	   p1 = p2)
	{
	  destroy_term(p1->arg);
	  p2 = p1->next_arg;
	  FREE(p1);
	}
      break;
    case AC_COMPRESSED:
    case COM_COMPRESSED:
      for (q1 = t->rest.a.ac_list; 
	   q1 != NULL; 
	   q1 = q2)
	{
	  destroy_term(q1->arg);
	  q2 = q1->next_ac;
	  FREE(q1);
	}
      break;
    default:
      ASSERT(FALSE, "destroy_term(): unknown term type");
  }
  FREE(t);
}

