(***************************************************************************

Module for cache

CiME Project - Dmons research team - LRI - Universit Paris XI

$Id: cache.ml,v 1.4 2002/07/01 12:01:36 contejea Exp $

***************************************************************************)





type ('a,'b) hash_item =
  {
    first : 'a;
    second : 'b;
    priority_index : int
  }

and priority_item =
  {
    mutable next : int;
    mutable prev : int;
    mutable hash_index : int
  }

and ('a,'b) cache =
  {
    hash_table : ('a,'b) hash_item list array;
    priority_queue : priority_item array 
  }
;;


(*

(create n) returns a create cache of size n

*)

let create n =

  let queue = Array.create (n+1) {next = 0; prev = 0; hash_index = -1 }
  in
    
    for i=0 to n do
      queue.(i) <- {next = i+1; prev = i-1; hash_index = -1 }
    done;
    queue.(0).prev <- n;
    queue.(n).next <- 0;
    {
      hash_table = Array.create n [];
      priority_queue = queue
    }
;;

(*

(unlink q i) remove index i from priority queue q

*)

let unlink q i =
  let i_prev = q.(i).prev
  and i_next = q.(i).next
  in
  q.(i_prev).next <- i_next;
  q.(i_next).prev <- i_prev
;;

(*

(link q i h) links again index i in priority queue q, at the
first place

*)


let link q i h =
  let first = q.(0).next
  in
  q.(i).next <- first;
  q.(i).prev <- 0;
  q.(i).hash_index <- h;
  q.(0).next <- i;
  q.(first).prev <- i
;;

(*

(add x y c) adds the pair (x,y) in cache c. If there is no room left
in c, the earliest accessed pair of c is removed.

If there was already a pair (x,z) for some z in c, it is removed.

*)


let add x y c =
  let queue = c.priority_queue
  and table = c.hash_table
  in
  let h = (Hashtbl.hash x) mod (Array.length table)
  and oldest_hash_index = ref (-1)
  and oldest_index = queue.(0).prev
  in

  let rec remove_oldest_from_list = function
    [] -> failwith "Cache.add"
  | ({first = _; second = _; priority_index = i} as first)::l ->
      if i=oldest_index 
        then l
	else first::(remove_oldest_from_list l)

  and add_in_list = function

    [] -> 
      oldest_hash_index := queue.(oldest_index).hash_index;
      unlink queue oldest_index;
      link queue oldest_index h;
      [{first = x; second = y; priority_index = oldest_index}]

  | ({first = x'; second = _; priority_index = i} as first)::l ->
      if x'=x 
        then begin
      	     unlink queue i;
	     link queue i h;
             {first = x; second = y; priority_index = i}::l
  	     end
        else first::(add_in_list l)

  in

  table.(h) <- add_in_list table.(h) ;
  if !oldest_hash_index>=0 
  then table.(!oldest_hash_index) <- 
     remove_oldest_from_list table.(!oldest_hash_index);
;;



(*


(find x c) returns y such that (x,y) belongs to c. This pair will be
marked as the more recently accessed pair.

find raises exception Not_found if no such pair List.exists. 

*)


let find x c =
  let queue = c.priority_queue
  and table = c.hash_table
  in
  let h = (Hashtbl.hash x) mod (Array.length table)
  
  in

  let rec find_in_list = function

    [] -> raise Not_found
      
  | ({first = x'; second = y'; priority_index = i} as first)::l ->
      if x'=x
        then begin
	     unlink queue i;
	     link queue i h;
             y'
	     end
	else find_in_list l

  in

  find_in_list table.(h)
;;


let int_int_find (x1,x2 as x) c =
  let queue = c.priority_queue
  and table = c.hash_table
  in
  let h = (Hashtbl.hash x) mod (Array.length table)
  
  in

  let rec find_in_list = function

    [] -> raise Not_found
      
  | ({first = (x'1,x'2); second = y'; priority_index = i} as first)::l ->
      if x'1==x1 && x'2==x2
        then begin
	     unlink queue i;
	     link queue i h;
             y'
	     end
	else find_in_list l

  in

  find_in_list table.(h)
;;


let iter f c =

  let rec find_and_apply i = function
      [] -> 
      	failwith "Cache.iter"
    | {first = x; second = y; priority_index=j}:: l ->
      	if i=j 
      	  then f x y
	  else find_and_apply i l

  in

  let queue = c.priority_queue
  and table = c.hash_table
  in
  let i = ref queue.(0).next in
  while !i<>0 do
    begin
    let h = queue.(!i).hash_index
    in if h>=0 then find_and_apply !i table.(h)
    end;
    i := queue.(!i).next
  done
;;



(*

[function1 n f] is a function that calls [f] with caching of
arguments, in a cache of size [n]

*)

let function1 n f =
  let c = create n in
    fun x ->
      try
	find x c 
      with Not_found ->
	let y = f x in
	  add x y c;
	  y
;;


let collision_rate c =
  let h = c.hash_table  
  and count = ref 0
  and rate = ref 0.0
  in
  for i=0 to (Array.length h)-1 do
    let n = float(List.length h.(i))
    in 
    rate := !rate +. n *. (n+.1.0) /. 2.0;
    count := !count + (List.length h.(i))
  done;
  !rate /.  float(!count)
;;

let max_hash_length c =
  let h = c.hash_table  
  and count = ref 0
  in
  for i=0 to (Array.length h)-1 do
    let n = List.length h.(i)
    in 
    if n > !count
      then count := n
  done;
  !count
;;

    
let list_of_cache c =

  let l = ref []
  in
  let collect x y = l := (x,y)::!l
  in 
  iter collect c;
  !l
;;

  
(*

some tests

*)

(*


let foo = create 3;;

add "a" 1 foo;;

list_of_cache foo;;  (* should return ["a", 1] *)

add "b" 2 foo;;

list_of_cache foo;;  (* should return ["a", 1; "b", 2] *)

add "c" 3 foo;;

list_of_cache foo;;  (* should return ["a", 1; "b", 2; "c", 3] *)

find "b" foo;;       (* should return 2 *)

list_of_cache foo;;  (* should return ["a", 1; "c", 3; "b", 2] *)

find "a" foo;;       (* should return 1 *)

list_of_cache foo;;  (* should return ["c", 3; "b", 2; "a", 1] *)

add "d" 4 foo;;

list_of_cache foo;;  (* should return ["b", 2; "a", 1; "d", 4] *)

find "c" foo;;       (* should raise Not_found *)

add "a" 5 foo;;

list_of_cache foo;;  (* should return ["b", 2; "d", 4; "a", 5] *)


*)
