(***************************************************************************

This module provides some kind of maps with efficient access to
elements. The size is bounded so when the cache is full, a free place
is obtained by removing the element with the oldest access time.

CiME Project - Dmons research team - LRI - Universit Paris XI

$Id: cache.mli,v 1.4 2002/07/01 12:01:36 contejea Exp $

***************************************************************************)


(*

This module provides some kind of maps with efficient access to
elements. The size is bounded so when the cache is full, a free place
is obtained by removing the element with the oldest access time.

*)

(*

\subsection{Type of caches, and basic access functions}

*)

type ('a,'b) cache;;


(*

[(create n)] returns a new cache of size [n]. [n] should be at least 1!

*)

val create : int -> ('a,'b) cache;;


(*

[(add x y c)] adds the pair [(x,y)] in cache [c]. If there is no room left
in [c], the earliest accessed pair of [c] is removed.

If there was already a pair [(x,z)] for some [z] in [c], it becomes
hidden. It will never accessible again since there is no remove
function on cache. It will eventually be removed after enough calls to
add.

*)


val add : 'a -> 'b -> ('a,'b) cache -> unit ;;


(*


[(find x c)] returns [y] such that [(x,y)] belongs to [c]. This pair
will be marked as the more recently accessed pair.

[find] raises exception [Not_found] if no such pair exists. 

*)


val find : 'a -> ('a,'b) cache -> 'b;;


(*


[int_int_find] is a specialized version of [find] whenever the type of its 
first argument is [int * int]. In this case, this equality test
should be more efficient.

*)

val int_int_find : int * int -> (int * int, 'b) cache -> 'b;;

(*

\subsection{Iterator}

[(iter f c)] applies function [f] to each pair [(x,y)] of the cache
c. the elements are processed from the newest accessed pair to the
oldest

*)


val iter : ('a -> 'b -> unit) -> ('a,'b) cache -> unit;;

(*

[function1 n f] is a function that calls [f] with caching of
arguments, in a cache of size [n]

*)

val function1 : int -> ('a -> 'b) -> ('a -> 'b) ;;


(*

\subsection{Statistics}

[(collision_rate c)] returns a real value [t] which estimate the
repartition of elements in the cache hash table. In fact, it is the
average time to access an element in the cache.

[(max_hash_length c)] is the maximal amount of time to access to an element. 

[(list_of_cache c)] returns the contents of the cache, useful for
 debugging purposes.

*)

val collision_rate : ('a,'b) cache -> float;;

val max_hash_length : ('a,'b) cache -> int;;

val list_of_cache : ('a,'b) cache -> ('a * 'b) list;;
