(***************************************************************************

General functions over graphs

CiME Project - Dmons research team - LRI - Universit Paris XI

$Id: labelled_graphs.ml,v 1.10 2001/02/13 14:33:49 marche Exp $

***************************************************************************)


type vertex = int;;

module VertexOrd = 
  struct
    type t = vertex
    let compare = (-)
  end;;


module VertexSet = Set.Make(VertexOrd);;

module VertexMap = Map.Make(VertexOrd);;

(*

  the label [first_unused_vertex] is a counter that permits to number
  each new vertex.

  [vertex_labels] is a map that associates a label to each node.

  [successors] is a map that associates to each vertex the set of its
  successors in the graph.

*)

type 'vertex_labels graph =
    {
      first_unused_vertex : vertex;
      vertex_labels : 'vertex_labels VertexMap.t;
      successors : VertexSet.t VertexMap.t;
    }
;;

let empty_graph = 
  {
    first_unused_vertex = 0;
    vertex_labels = VertexMap.empty;
    successors = VertexMap.empty;
  }
;;

let add_vertex g lab =
  let n = g.first_unused_vertex 
  in
    (n,
     {
       first_unused_vertex = succ n;
       vertex_labels = VertexMap.add n lab g.vertex_labels;
       successors = g.successors;
     })
;;

let add_arc g v1 v2 =
  try
    let v1_succ = VertexMap.find v1 g.successors
    in
      if VertexSet.mem v2 v1_succ
      then invalid_arg "add_arc : arc already exists"
      else
      	{
      	  first_unused_vertex = g.first_unused_vertex;
      	  vertex_labels = g.vertex_labels;
      	  successors = VertexMap.add 
			 v1 
			 (VertexSet.add v2 v1_succ)
			 g.successors;
	  (* note: [add] removes the old binding of v1 *)
      	}
  with
      Not_found ->
	{
      	  first_unused_vertex = g.first_unused_vertex;
      	  vertex_labels = g.vertex_labels;
      	  successors = VertexMap.add 
			 v1 
			 (VertexSet.add v2 VertexSet.empty)
			 g.successors;
      	}
;;


let get_vertex_label g v = 
  try
    VertexMap.find v g.vertex_labels
  with Not_found ->
    invalid_arg "label not found in get_vertex_label !";;

let get_successors g v = 
  try
    VertexMap.find v g.successors
  with Not_found ->
    VertexSet.empty
;;



(* 

   \subsection{Iterator}

*)

let fold_left_graph f a g =
  VertexMap.fold
    (fun v _ accu -> f accu v)
    g.vertex_labels
    a
;;


(*

  \subsection{Cycles}

*)

let vertex_has_a_1_cycle g v =
  VertexSet.mem v (get_successors g v);;

(*

  Warning : the following algorithm is certainly not optimal !

*)

let is_on_a_cycle v g = 

  let rec reachable_vertices v_set =
    
    if VertexSet.mem v v_set
    then true
    else
      let new_vertices =
	VertexSet.fold
	  (fun v e ->
	     VertexSet.union (get_successors g v) e)
	  v_set
	  v_set
	  
      in
	if VertexSet.cardinal new_vertices = VertexSet.cardinal v_set
	then false
	else reachable_vertices new_vertices

  in
    reachable_vertices (get_successors g v)
;;

(*
  
  Again !

*)

let vertices_on_cycle g =
  VertexMap.fold 
    (fun v l e ->
       if is_on_a_cycle v g
       then VertexSet.add v e
       else e)
    g.vertex_labels
    VertexSet.empty
;;

(*

  and again ! 

*)

let vertices_not_on_cycle g =
  VertexMap.fold 
    (fun v l e ->
       if is_on_a_cycle v g
       then e else VertexSet.add v e)
    g.vertex_labels
    VertexSet.empty
;;


(*

  \subsection{Remove functions}

*)

let remove_vertex_and_its_arcs g vert =
  { g with
      vertex_labels = VertexMap.remove vert g.vertex_labels;
      successors = VertexMap.fold
		     (fun v succ m ->
			VertexMap.add
			  v
			  (if VertexSet.mem vert succ 
			   then VertexSet.remove vert succ 
			   else succ)
			  m)
		     (VertexMap.remove vert g.successors)
		     VertexMap.empty;
  }


let remove_vertex_and_link_its_arcs g vert =
  let g_succ = get_successors g vert in
  let g_succ =
    if VertexSet.mem vert g_succ 
    then VertexSet.remove vert g_succ
    else g_succ
  in
  { g with
      vertex_labels = VertexMap.remove vert g.vertex_labels;
      successors = VertexMap.fold
		     (fun v succ m ->
			VertexMap.add
			  v
			  (if VertexSet.mem vert succ 
			   then 
			     VertexSet.union 
			       (VertexSet.remove vert succ)
			       g_succ
			   else succ)
			  m)
		     (VertexMap.remove vert g.successors)
		     VertexMap.empty;
  }



let get_subgraph g e =
  { g with
      vertex_labels = 
        VertexMap.fold
	  (fun v l accu -> 
	     if VertexSet.mem v e then VertexMap.add v l accu else accu)
	  g.vertex_labels
	  VertexMap.empty;
      successors =
        VertexMap.fold
	  (fun v succ m ->
	     if VertexSet.mem v e 
	     then 
	       VertexMap.add
		 v
		 (VertexSet.fold
		    (fun v accu ->
		       if VertexSet.mem v e
		       then VertexSet.add v accu
		       else accu)
		    succ
                    VertexSet.empty)  
		 m
             else
	       m)
	  g.successors
	  VertexMap.empty;
  }
;;      

(*

  \subsection{Printing}

*)


open Format;;

let print g print_label =
  VertexMap.iter
    (fun v succ ->
       let l = get_vertex_label g v in
       VertexSet.iter
	 (fun v' ->
	    print_label l;
	    print_string " ---> ";
	    print_label (get_vertex_label g v');
	    print_newline())
	 succ)
    g.successors
;;

		  
let compact_print g print_label =
  VertexMap.iter
    (fun v l ->
       print_int v;
       print_string ": ";
       print_label l;
       print_newline())
    g.vertex_labels;
  VertexMap.iter
    (fun v succ ->
       VertexSet.iter
	 (fun v' ->
	    print_int v;
	    print_string " -> ";
	    print_int v';
	    print_string " ,";
	    print_break 1 0)
	 succ)
    g.successors;
  print_newline();
;;

		  
let dot_print g print_label =
  printf "------cut here for the dot input file------@.";
  printf "digraph G {@.";
  VertexMap.iter
    (fun v l ->
       printf "n%d [shape=box, label = \"" v;
       print_label l;
       printf "\"];@.")
    g.vertex_labels;
  VertexMap.iter
    (fun v succ ->
       printf "n%d -> { " v;
       VertexSet.iter
	 (fun v' ->
	    printf "n%d ; " v')
	 succ;
       printf "};@."
    )
    g.successors;
  printf "}@.";
  printf "------cut here for the dot input file------@.";
  printf "------display using dot -Tps -o file.ps file.dot------@.";
;;

		  
(*

  \subsection{Cycles}

  The algorithm proceeds as follows : starting from any vertex, it
  performs a depth-first search in the graph, annotating each vertex
  with a number ([time]). During this search, it computes for each
  vertex [v] the minimum [mins.(v)] of its own number and the mins of
  its successors. At each step of this computation, [stack] records
  the vertices that have been followed, and if at any time we reach a
  vertex whose min is its own number, it means we have a strongly
  connected component formed by all vertices of the stack whose
  numbers are greater or equal to this min. In such a case, all these
  are removed from the stack, stored in a new component, are we put
  [mins.(v) = -1] to denote that this vertex has been definitely
  handled.

  At the end of this search some vertices has not been traversed, we
  start again a new search from any of these vertices.

*)


let split_cycles g =

  let mins = Array.create g.first_unused_vertex 0 
	       (* table of minimum son vnumbers *)
  and on_cycle = ref VertexSet.empty 
  and not_on_cycle = ref VertexSet.empty 
  and time = ref 1    (* counter for numbering the vertices *)

  in
    
  (*

    [follow_succ v mini] follow a branch to vertex v, computing
    recursively minimum sons values, returning the min of them and the
    minimum [mini] computed so far 

  *)
    
  let rec follow_succ v mini = 
    let m = mins.(v) in
      match m with
	  -1 -> mini   (* -1 means this vertex is already handled *)
	| 0 ->         (* 0 means this vertex has not yet been numbered *) 
	    let new_mini = search_cycles v in
	      min mini new_mini
	| _ -> min m mini

  (*

    [search_cycles v] performs a depth-first search of cycles
    starting from v, and returns the minimum son value

  *)

  and search_cycles v =
    mins.(v) <- !time;
    incr time;
    let succs = get_successors g v
    in
    let new_mini = VertexSet.fold follow_succ succs !time 
    in
      if new_mini <= mins.(v) (* means we have found a new cycle *)
      then on_cycle := VertexSet.add v !on_cycle
      else not_on_cycle := VertexSet.add v !not_on_cycle;
      min mins.(v) new_mini
	
  in

    (*

      perform the depth-first search starting form any vertex, as long as
      there are still non followed vertices
      
    *) 

    VertexMap.iter
      (fun v _  ->
	 if mins.(v) = 0
	 then let _ = search_cycles v in ())
      g.vertex_labels;

    (* and we are done ! *)

    !on_cycle,!not_on_cycle
    ;;

	    
(*

  \subsection{Strongly connected components}

  The algorithm proceeds as follows : starting from any vertex, it
  performs a depth-first search in the graph, annotating each vertex
  with a number ([time]). During this search, it computes for each
  vertex [v] the minimum [mins.(v)] of its own number and the mins of
  its successors. At each step of this computation, [stack] records
  the vertices that have been followed, and if at any time we reach a
  vertex whose min is its own number, it means we have a strongly
  connected component formed by all vertices of the stack whose
  numbers are greater or equal to this min. In such a case, all these
  are removed from the stack, stored in a new component, are we put
  [mins.(v) = -1] to denote that this vertex has been definitely
  handled.

  At the end of this search some vertices has not been traversed, we
  start again a new search from any of these vertices.

*)


let compute_strongly_connected_components even_if_no_cycles g =

  let mins = Array.create g.first_unused_vertex 0 
	       (* table of minimum son vnumbers *)
  and comps = ref []  (* records components computes so far *)
  and stack = ref []  (* traversing stack *)
  and time = ref 1    (* counter for numbering the vertices *)

  in
    
  (*
    
    [pop_stack min] pops from the stack each vertex that have a
    minimum sons values greater or equal to [min], and put them into
    a new component.
    
  *)
    
  let pop_stack mini = 
    let rec do_pop accu =
      match !stack with
	  [] -> accu
	| v::vl ->
	    if mins.(v) < mini
	    then accu
	    else
	      begin
(*
		print_int v;
*)
		mins.(v) <- -1;
		stack := vl;
		do_pop (VertexSet.add v accu) 
	      end
    in
(*
      print_string "pop stack : ";
*)
      let accu = do_pop VertexSet.empty
      in 
(*
	print_newline();
*)
	accu
	
	
  in

  (*

    [follow_succ v mini] follow a branch to vertex v, computing
    recursively minimum sons values, returning the min of them and the
    minimum [mini] computed so far 

  *)
    
  let rec follow_succ v mini = 
    let m = mins.(v) in
      match m with
	  -1 -> mini   (* -1 means this vertex is already in a component *)
	| 0 ->         (* 0 means this vertex has not yet been numbered *) 
	    let new_mini = compute_components v in
	      min mini new_mini
	| _ -> min m mini

  (*

    [compute_components v] performs a depth-first search of components
    starting from v, and returns the minimum son value

  *)

  and compute_components v =
    let old_mini = !time in
    mins.(v) <- old_mini;
    incr time;
    stack := v :: !stack;
    let succs = get_successors g v
    in
    let new_mini = VertexSet.fold follow_succ succs !time
    in
      if new_mini >= old_mini
	(* means we have found a new component *)
      then 
	begin
(*
	  print_string " old_mini = ";
	  print_int old_mini;
	  print_string " new_mini = ";
	  print_int new_mini;
*)
	  let p = pop_stack old_mini in
	    if new_mini = old_mini or even_if_no_cycles
	    then comps := p :: !comps;
	    old_mini
	end
      else new_mini
	
  in

    (*

      perform the depth-first search starting form any vertex, as long as
      there are still non followed vertices
      
    *) 

    VertexMap.iter
      (fun v _  ->
	 if mins.(v) = 0
	 then let _ = compute_components v in ())
      g.vertex_labels;

    (* and we are done ! *)

    !comps
    ;;



