(***************************************************************************

This module implements oriented graphs with labelled vertices (but no
labels on arcs), in a functional way.

CiME Project - Dmons research team - LRI - Universit Paris XI

$Id: labelled_graphs.mli,v 1.12 2001/02/13 14:33:49 marche Exp $

***************************************************************************)

(*

This module implements oriented graphs with labelled vertices (but no
labels on arcs), in a functional way.

*)


(*

\subsection{Vertices}

abstract type for vertices and related modules for sets and maps of
vertices

*)

type vertex;;

module VertexSet : Set.S with type elt=vertex;;
module VertexMap : Map.S with type key=vertex;;

(*

\subsection{Type for graphs}

abstract type for graphs, parameterized by the type of vertex labels

*)

type 'vertex_labels graph;;


(*

\subsection{Constructor functions for graphs} 

[empty_graph] is the graph with no vertex

[(add_vertex g l)] returns a pair [(v,g')] where [v] is the new vertex
having label [l] and [g'] is the graph obtained by adding [v] to [g]

%\ocwdefpretty{v1}{\nimportequoi}

[(add_arc g v1 v2)] returns the graph obtained by adding to [g] a new
arc from vertex [v1] to vertex [v2]. Raises Failure if the arc
already exists.

*)

val empty_graph : 'vertex_labels graph;;

val add_vertex : 
  'vertex_labels graph -> 'vertex_labels 
    -> vertex * 'vertex_labels graph;;

val add_arc : 
  'vertex_labels graph -> vertex -> vertex -> 'vertex_labels graph;;


(*

\subsection{Basic access functions}

[(get_vertex_label g v)] returns the label of the vertex [v] in graph [g]

[(get_successors g v)] returns the set of vertices reachables from [v]
in graph [g], i.e. the set of [v'] suvh that there is an arc from [v]
to [v'].

*)

val get_vertex_label : 'vertex_labels graph -> vertex -> 'vertex_labels;;

val get_successors : 'vertex_labels graph -> vertex -> VertexSet.t;;



(*

\subsection{Iterator}

[(fold_left_graph f a g)] returns [f (... (f (f a v1) v2) ...) vn] where
[v1],...,[vn] is the set of vertices of [g], in a non-specified order.

*)


val fold_left_graph : 
  ('a -> vertex -> 'a) -> 'a -> 'vertex_labels graph -> 'a
;;


(*

  \subsection{Extraction of subgraphs}

  \subsubsection{Removing one vertex}

  The next two functions provide two ways of removing a vertex from a
  graph. [(remove_vertex_and_its_arcs g v)] returns the graph obtained
  from [g] by removing the vertex [v] and all the arcs going to [v] or
  comming from [v]. [(remove_vertex_and_link_its_arcs g v)] returns
  the graph obtained from [g] by removing [g] and ``linking'' the arcs
  going to [v] with arcs coming from [v]: if (v1,v) and (v,v2) are
  arcs in [g] then (v1,v2) is an arc in the new graph returned.

*)


val remove_vertex_and_its_arcs : 
  'vertex_labels graph -> vertex -> 'vertex_labels graph;;

val remove_vertex_and_link_its_arcs : 
  'vertex_labels graph -> vertex -> 'vertex_labels graph;;


(*

  \subsubsection{Extraction of of a whole subgraph}

  [(get_subgraph g e)] returns the subgraph of [g] whose vertices are
  in [e], and arcs are the arcs of [g] that come from a vertex in [e]
  and go to a vertex in [e].

*)

val get_subgraph : 
  'vertex_labels graph -> VertexSet.t -> 'vertex_labels graph ;;

(*

  \subsection{Printing}

  [(print g f)] prints the graph [g], using [f] as printing function
  for labels. Warning: this function [f] must use Format library!

  [(compact_print g f)] does the same in a more compact format, using
  numbers to denote vertices.

  [(dot_print g f)] does the same in \textsl{dot} format, see
  \url{http://www.research.att.com/sw/tools/graphviz/}.

*)

val print : 'vertex_labels graph -> ('vertex_labels -> unit) -> unit;;

val compact_print : 'vertex_labels graph -> ('vertex_labels -> unit) -> unit;; 

val dot_print : 'vertex_labels graph -> ('vertex_labels -> unit) -> unit;; 

(*

  \subsection{Cycles, connected components and such}

  \subsubsection{Cycles}

  [(vertex_has_a_1_cycle g v)] returns true if there is an arc from
  [v] to [v] in [g]q.

  [(vertices_on_cycle g)] returns the set of vertices of [g] that
  belong to a cycle

  [(vertices_not_on_cycle g)] returns the set of vertices of [g] that
  do not belong to any cycle

  [(split_cycles g)] returns a pair [(a,b)] of sets of vertices, where
  [a] is the set of vertices that lay on a cycle of [g], [b] the set
  of vertices that do not belong to any cycle of [g]

*)

val vertex_has_a_1_cycle :
  'vertex_labels graph -> vertex -> bool;;

val split_cycles :
  'vertex_labels graph -> VertexSet.t * VertexSet.t ;;

(*

obsolete 

\begin{verbatim}
val vertices_on_cycle :
  'vertex_labels graph -> VertexSet.t;;

val vertices_not_on_cycle :
  'vertex_labels graph -> VertexSet.t;;
\end{verbatim}

*)

(* \subsubsection{Strongly connected components}

  [(compute_strongly_connected_components even_if_no_cycles g)]
  returns the set of strongly connected components of the directed
  graph [g]. These are defined as follows: let $G=(V,E)$ be a directed
  graph, where $V$ is the set of vertices and $E$ the arcs, then let
  $G'=(V,E')$ where $E'$ is defined by $v_1 \leftrightarrow_{E'} v_2$
  iff $v_1 \rightarrow^{\star}_{E} v_2$ and $v_2
  \rightarrow^{\star}_{E} v_1$. Then $\leftrightarrow_{E'}$ is an
  equivalence relation, and the strongly connected components of $G$
  are the equivalence classes of it.

   if [even_if_no_cycles] is set to false, then the components
   containing only one vertex which has no arc to itself are not
   returned. Example : for the graph $a\ra a, a\ra b, b\ra c,c\ra c$,
   the component [b] will not be returned.
   

*)

val compute_strongly_connected_components : 
  bool -> 'vertex_labels graph -> VertexSet.t list;; 





