(***************************************************************************

(This sentence has been added automatically, it should be replaced by a description of the module)

CiME Project - Dmons research team - LRI - Universit Paris XI

$Id: polynomials.ml,v 1.12 2000/03/26 17:35:51 marche Exp $

***************************************************************************)



module type RingType =
  sig
    type coef
    val zero : coef
    val one : coef
    val is_null : coef -> bool
    val add : coef -> coef -> coef
    val minus : coef -> coef
    val sub : coef -> coef -> coef
    val mult : coef -> coef -> coef
  end
;;


module type VarsType =
  sig
    type t
    val compare : t -> t -> int
  end
;;

module type PolynomialType =

  sig

(*

  types for coefficients, variable and polynomials

*)

    module Base_ring : RingType
    type coef = Base_ring.coef


    type variable 

    module Var_set : Ordered_sets.OrderedSet
    module Var_map : Ordered_maps.OrderedMap

    type monomial 

    val make_monomial : coef -> (variable,int) Var_map.t -> monomial 
    val monomial_coef : monomial -> coef
    val monomial_vars : monomial -> (variable,int) Var_map.t
    val monomial_degree : monomial -> int

    type poly

 (*

  zero is the polynomial 0, one is 1, [(cte a)] is the constant polynomial a, and [(var x)] is the polynomial consisting in the variable x

*)

    val zero : poly
    val one : poly
    val cte : coef -> poly
    val var : variable -> poly
    val poly_of_monomial : monomial -> poly

    val is_null : poly -> bool
    val constant_coef : poly -> coef
    val total_degree : poly -> int
    val partial_degree : variable -> poly -> int
    val set_of_vars : poly -> variable Var_set.t

    (* for debugging purposes *)

    val size : poly -> int

    val print_polynomial :
      (coef -> unit) -> (variable -> unit) -> poly -> unit
            
            
(*
  
  [(add p1 p2)] is p1+p2

  [(minus p)] is -p
  
  [(sub p1 p2)] is p1-p2

  [(mult p1 p2)] is p1*p2

  [(power p n)] returns p^n

*)

    val add : poly -> poly -> poly
    val minus : poly -> poly
    val sub : poly -> poly -> poly
    val mult : poly -> poly -> poly
    val power : poly -> int -> poly

  (*

      Polynomial evaluation.

      [(eval p m)] evaluates [p] with variable values given by [m],
      which is a map from variables to coefficients

    *)

    val eval : (variable -> coef) -> poly -> coef

(*

  [(substitute m p)] replaces in [p] each occurrence of variables by
  the corresponding polynomial in [m], which is a map from variables
  to polynomials

*)

    val substitute : (variable,poly) Var_map.t -> poly -> poly

(* 

  [(map f p)] applies [f] to each monomials of [p] and builds the sum
  of the results. WARNING : this fonction assumes that the resulting
  set of monomials does have any two monomials with a common power
  product. If not the case, please use fold and add instead.

*)
 
    val map : (monomial -> monomial) -> poly -> poly

(* 

  [(iter f p)] applies [f] to each monomials of [p]. 

*)
 
    val iter : (monomial -> unit) -> poly -> unit

(* 

  [(fold f p a)] computes [(f m1 ... (f mN a)...)] where mi is the
  ith monomial of p, given in the lexicographic order of power
  products.

*)
 
    val fold : (monomial -> 'a -> 'a) -> poly -> 'a -> 'a

(* 

  [(find f p)] search for a monomial m in p such that (f m) does not
  raise Not_found. raises Not_found if no such monomial exists.

*)
 
    val find : (monomial -> 'a) -> poly -> 'a

  end
;;


module Make 
  (Ring : RingType) 
  (Vars : VarsType) 
  (Var_set : Ordered_sets.OrderedSet with type 'a elt=Vars.t) 
  (Var_map : Ordered_maps.OrderedMap with type 'a key=Vars.t) =

  struct

    module Base_ring = Ring

    type coef = Ring.coef

    type variable = Vars.t

    type monomial =
  	{
	  coef : coef;
	  vars : (variable,int) Var_map.t
  	} 

    module Var_set = Var_set
    module Var_map = Var_map

    let make_monomial c m = { coef =c ; vars = m }

    let monomial_coef m = m.coef

    let monomial_vars m = m.vars

    let monomial_degree m =
      Var_map.fold (fun x exp n -> n+exp) m.vars 0

    type poly = monomial list

    let zero = ([] : poly) 

    let one = [ { coef = Ring.one; vars = Var_map.empty } ]

    let cte c = 
      if Ring.is_null c 
      then []
      else [ { coef = c; vars = Var_map.empty } ] 

    let var x = [ { coef = Ring.one; vars = Var_map.add x 1 Var_map.empty } ] 

    let poly_of_monomial x = [x]

    let is_null p = p=[]

    let number_of_variables m =
      Var_map.fold (fun x exp n -> n+1) m.vars 0

    let rec constant_coef = function
	[] -> Ring.zero 
      | m::p ->
	  if m.vars = Var_map.empty
	  then m.coef
	  else constant_coef p
	 
    let rec total_degree = function
    	[] -> 0
      | m::p -> max (monomial_degree m) (total_degree p)

    let monomial_partial_degree x m = 
      try
    	Var_map.find x m.vars
      with Not_found -> 0

    let rec partial_degree x = function
    	[] -> 0
      | m::p -> max (monomial_partial_degree x m) (partial_degree x p)

    let number_of_monomials p = List.length p

    let vars_of_monomial m =
      Var_map.fold
    	(fun x exp e -> Var_set.add x e)
    	m.vars
    	Var_set.empty

    let set_of_vars p =
      List.fold_left
    	(fun e m ->
	   Var_set.union (vars_of_monomial m) e)
    	Var_set.empty
    	p

    let size_of_monomial m =
      let n = 
	Var_map.fold
    	  (fun x exp n -> exp + n )
    	  m.vars
    	  (-1)
      in
	if n < 0 then 0 else n

    let size p =
      List.fold_left
    	(fun n m -> n + (size_of_monomial m))
    	0
    	p

    let rec compare_degrees v1 v2 = 
      try
    	let x1,e1 = Var_map.max v1
    	in
	  try
	    let x2,e2 = Var_map.max v2
	    in 
	      match Vars.compare x1 x2
	      with
		  0 -> 
		    if e1=e2
		    then compare_degrees 
		      (Var_map.remove x1 v1) (Var_map.remove x2 v2) 
		    else 
		      if e1<e2 then -1 else 1
		| x -> x	    
	  with Not_found -> 1
      with Not_found ->
    	try 
	  let x2,e2 = Var_map.max v2
	  in -1
    	with Not_found -> 0

    let compare_monomials m1 m2 = 
      compare_degrees m1.vars m2.vars
 
	
    let print_polynomial print_coef print_variable p =

      let print_monomial m =
	begin
	  if m.vars = Var_map.empty
	  then
	    print_coef m.coef
	  else
	    if not (Ring.is_null (Ring.sub m.coef Ring.one))
	    then
	      begin
		print_coef m.coef;
		Format.print_string "*";
	      end;
    	      let _ = Var_map.fold 
    		      (fun x exp sep -> 
			 (*
			   if exp<>0
			   then 
			   *)
			 begin
			   Format.print_string sep;
			   print_variable x;
			   if exp<>1 
			   then 
			     begin
			       Format.print_string "^";
			       Format.print_int exp
			     end;
			   "*"
			 end)
			m.vars
                      "" 
	      in ()
	end

      in
	
      let rec print_poly = function
	  [] -> Format.print_string "0"
  	| [m] -> print_monomial m
  	| m::l ->  
	    begin
	      print_monomial m;
	      Format.print_string " + ";
	      print_poly l
	    end
      in
	print_poly p
;;




(*

  (add p1 p2) returns the sum of p1 and p2, monomials are sorted in
  decreasing order for the lexicographic ordering of powers

*)

    let rec add = function
    	[] -> (function p2 -> p2)
      | (m1::r1) as p1 ->
	  function [] -> p1
	    | m2::r2 as p2 ->
	    	let c = compare_degrees m1.vars m2.vars 
	    	in 
		  if c < 0
		  then m2::(add p1 r2)
		  else
		    if c > 0
		    then  m1::(add r1 p2)
	    	    else
		      let a = Ring.add m1.coef m2.coef 
		      and r = (add r1 r2)
		      in
		    	if Ring.is_null a 
		    	then r
		    	else {coef=a;vars=m1.vars}::r
			  

(*

  (minus p) returns the opposite of p
  
*)


    let rec minus = function
    	[] -> []
      | m::p ->
	  {coef = Ring.minus m.coef ; vars = m.vars}::(minus p)


(*

  (sub p1 p2) returns p1-p2

*)

let rec sub = function
    [] -> minus
  | (m1::r1) as p1 -> function 
	[] -> p1
      | (m2::r2) as p2 ->
	  let c=compare_degrees m1.vars m2.vars 
	  in
	    if c<0
	    then
	      {coef=Ring.minus m2.coef;vars=m2.vars}::
	      (sub p1 r2)
	    else
	      if c>0
	      then
	     	m1::(sub r1 p2)
	      else
		let a = Ring.sub m1.coef m2.coef 
		and r = (sub r1 r2)
		in
		if Ring.is_null a
		then r
		else {coef=a;vars=m1.vars}::r


(*

  (power_product X_0^{a_0}..X_k^{a_k} X_0^{b_0}..X_k^{b_k})
 returns X_0^{a_0+b_0}..X_k^{a_k+b_k}

*)

let rec power_product_aux v1 v2 =
  match v1,v2 with
    [],v2 -> 
      List.fold_left 
	(fun m (x,d) -> Var_map.add x d m)
	Var_map.empty
	v2
  | v1,[] -> 
      List.fold_left 
	(fun m (x,d) -> Var_map.add x d m)
	Var_map.empty
	v1
  | (((x1,d1)::r1) as v1),(((x2,d2)::r2) as v2) ->
      let c = Vars.compare x1 x2
      in
      if c=0
      then Var_map.add x1 (d1+d2) (power_product_aux r1 r2)
      else
	if c<0
	then Var_map.add x1 d1 (power_product_aux r1 v2)
	else Var_map.add x2 d2 (power_product_aux v1 r2)


let power_product v1 v2 =
  let l1 = Var_map.elements_increasing_order v1
  and l2 = Var_map.elements_increasing_order v2
  in power_product_aux l1 l2




(*

  (mult_mono_poly m p) returns the product of monomial m by polynomial
  p

*)

let rec mult_mono_poly m1 = function
    [] -> []
  | m2::p2 ->
      { coef=Ring.mult m1.coef m2.coef;
	vars = power_product m1.vars m2.vars
      }::(mult_mono_poly m1 p2) 

(*

  (mult p1 p2) returns p1*p2

*)

let rec mult p1 = function
    [] -> []
  | m2::p2 ->
      add
	(mult_mono_poly m2 p1)
	(mult p1 p2)

(*
let mult p1 p2 =
   Format.print_string "Multiplication de ";
   Format.print_newline();
   print_polynomial 
   (fun _ -> Format.print_string "c")
   (fun _ -> Format.print_string "v")
   p1;
   Format.print_newline();
   Format.print_string "et";
   Format.print_newline();
   print_polynomial 
   (fun _ -> Format.print_string "c")
   (fun _ -> Format.print_string "v")
   p2;
   Format.print_newline();
   mult p1 p2
;;
*)

(*

  (power p n) returns p^n

*)

let rec power p = function	
    0 -> 
      if p=[]
	then failwith "attempt to compute 0^0 in Polynomial.power"
	else one
  | 1 -> p
  | n -> 
      let p2 = mult p p
      in
      let pn = power p2 (n/2) 
      in
      if n mod 2 =0
      then pn
      else mult p pn


let rec coef_power x = function
    0 -> Ring.one
  | 1 -> x
  | n ->
      let x2 = Ring.mult x x
      in
      let xn = coef_power x2 (n/2) 
      in  
      if n mod 2 =0
      then xn
      else Ring.mult x xn


let rec eval_monomial map m =
  Var_map.fold
    (fun x exp accu -> 
       try
	 let v = map x 
	 in
	   Ring.mult accu (coef_power v exp)
       with Not_found -> failwith "Value missing in Polynomials.eval")
  m.vars 
  m.coef


let rec eval map p =
  List.fold_right
    (fun m c -> Ring.add c (eval_monomial map m))
    p
    Ring.zero


(*

  (substitute list_p p) replaces in p each occurrence of variables by
  the corresponding polynomial in list_p, which is a association list
  of type (variable,polynomial).

*)



  let substitute list_p = 

    let exp_map = ref Var_map.empty

    in
    let get_power x n = 
      let powers = 
	try
	  Var_map.find x !exp_map
	with 
          Not_found -> Hashtbl.create 17
      and px = try Var_map.find x list_p with Not_found -> (var x)
      in
      let rec compute_power n =
	try 
	  Hashtbl.find powers n
	with 
	    Not_found ->
	      let result =
	      	if n=0 
		then one
		else
		    if n=1 
		    then px
		    else
		      let pns2 = compute_power (n/2)
		      in
		      let pn = mult pns2 pns2
		      in
			if n mod 2 =0
			then pn
			else mult px pn
	      in
		Hashtbl.add powers n result;result
      in
      let p = compute_power n
      in
	exp_map := Var_map.add x powers !exp_map;
	p

    in
	      
	      
    let substitute_in_monomial m = 
      let accu = ref (cte m.coef)
      in
        Var_map.iter
      	  (fun x exp -> 
	     accu := mult (get_power x exp) !accu)
      	m.vars;
        !accu
      	
	
	
    in 

    let substitute_list_p p =
      let accu = ref []
      in
        List.iter
          (function m -> accu := add (substitute_in_monomial m) !accu)
        p;
        !accu

   in
     function p -> 
       begin
(*
         Format.print_string "Substitution de ";
         Format.print_newline();
         print_polynomial 
           (fun _ -> Format.print_string "c")
           (fun _ -> Format.print_string "v")
           p;
         Format.print_newline();
         Format.print_string "avec";
         Format.print_newline();
         Var_map.iter
           (fun x p -> 
              Format.print_string "v -> ";
              print_polynomial 
                (fun _ -> Format.print_string "c")
                (fun _ -> Format.print_string "v")
                p;
              Format.print_newline())
           list_p;
         Format.print_newline();
*)
         substitute_list_p p
       end

  let map = List.map

  let iter = List.iter

  let rec fold f p a = match p with
      [] -> a
    | m::p -> fold f p (f m a)

  let rec find f = function
      [] -> raise Not_found
    | m::p ->
	try f m
	with Not_found -> find f p

  end




