(***************************************************************************

Module for polynomials over an arbitrary coefficient ring and an
arbitrary set of variables.

CiME Project - Dmons research team - LRI - Universit Paris XI

$Id: polynomials.mli,v 1.11 1999/09/23 15:35:06 marche Exp $

***************************************************************************)


(*

  This module allows to build polynomials over an arbitrary coefficient ring and an arbitrary set of variables. 

\subsection{Module for coefficient rings}

  Gives the type of coefficients, two constant 0 and 1, a test for
  nullity, and fonctions for addition, opposite, substraction and
  multiplication

*)

module type RingType =
  sig
    type coef
    val zero : coef
    val one : coef
    val is_null : coef -> bool
    val add : coef -> coef -> coef
    val minus : coef -> coef
    val sub : coef -> coef -> coef
    val mult : coef -> coef -> coef
  end
;;

(*

\subsection{Module for polynomials}

*)

module type PolynomialType =

  sig

    (* 

       \subsubsection{Coefficients and variables}

       Types for coefficients and variables, and modules for sets and
       maps of variables.
       
    *)

    module Base_ring : RingType

    type coef = Base_ring.coef

    type variable 
    module Var_set : Ordered_sets.OrderedSet 
    module Var_map : Ordered_maps.OrderedMap 

    (*

      \subsubsection{Monomials}
      
      [(make_monomial c v)] builds a monomial where [c] is the
      coefficient and v is a map giving to each variable its exponent
      (beware of giving only positive exponents and a non-zero
      coefficient) 
      
      [(monomial_coef m)] and [(monomials_vars)] are the
      corresponding acces functions.

      [(monomial_degree)] returns the degree of [m], that is the sum
      of its exponents.  

    *)

    type monomial 
    val make_monomial : coef -> (variable,int) Var_map.t -> monomial 
    val monomial_coef : monomial -> coef
    val monomial_vars : monomial -> (variable,int) Var_map.t
    val monomial_degree : monomial -> int

    (*

      \subsubsection{Type for polynomials, and basic polynomials}
      
      [zero] is the polynomial 0, [one] is 1

      [(cte a)] is the constant polynomial [a]

      [(var x)] is the polynomial consisting in the variable x

    *)

    type poly

    val zero : poly
    val one : poly
    val cte : coef -> poly
    val var : variable -> poly
    val poly_of_monomial : monomial -> poly

    (*

      \subsubsection{Basic functions over polynomials}

      [(is_null p)] is [true] if [p] is the polynomial 0

      [(constant_coef p)] is the constant coefficient of [p]

      [(total_degree p)] is the degree of [p] that is the maximum
      degree of its monomials

      [(partial_degre x p)] is the partial degree of [p] in variable
      [x], that is the maximum exponent of [x] in [p]

      [(set_of_vars p)] is the set of variables occurring in [p]

    *)

    val is_null : poly -> bool
    val constant_coef : poly -> coef
    val total_degree : poly -> int
    val partial_degree : variable -> poly -> int
    val set_of_vars : poly -> variable Var_set.t

    (* for debugging purposes *)

    val size : poly -> int

    (*

      \subsubsection{Printing function}

      [(print_polynomial fc fv p)] prints the polynomial [p] on
      standard output, where [fc] and [fv] are two printing functions
      for coefficients and variables. They must use the Format module
      for printing.

    *)

    val print_polynomial :
      (coef -> unit) -> (variable -> unit) -> poly -> unit
            
    (*

      \subsubsection{Polynomial operations}

      [(add p1 p2)] is $[p1]+[p2]$

      [(minus p)] is $-[p]$
  
      [(sub p1 p2)] is $[p1]-[p2]$

      [(mult p1 p2)] is $[p1]\times [p2]$

      [(power p n)] is $[p]^{[n]}$

    *)

    val add : poly -> poly -> poly
    val minus : poly -> poly
    val sub : poly -> poly -> poly
    val mult : poly -> poly -> poly
    val power : poly -> int -> poly

    (*

      \subsubsection{Polynomial evaluation}

      [(eval p m)] evaluates [p] with variable values given by [m],
      which is a map from variables to coefficients

    *)

    val eval : (variable -> coef) -> poly -> coef

    (*

      \subsubsection{Polynomial substitutions}

      [(substitute m p)] replaces in [p] each occurrence of variables
      by the corresponding polynomial in [m], which is a map from
      variables to polynomials

    *)

    val substitute : (variable,poly) Var_map.t -> poly -> poly

    (* 

       \subsubsection{Higher-order functions over polynomials}

       [(map f p)] applies [f] to each monomials of [p] and builds the
       sum of the results. WARNING : this fonction assumes that the
       resulting set of monomials does have any two monomials with a
       common power product. If not the case, please use fold and add
       instead.

       [(iter f p)] applies [f] to each monomials of [p]. 

       [(fold f p a)] computes [(f m1 ... (f mN a)...)] where [mi] is
       the ith monomial of [p], given in the lexicographic order of
       power products.

       [(find f p)] search for a monomial m in p such that (f m) does
       not raise [Not_found]. raises [Not_found] if no such monomial
       exists.

    *)
 
    val map : (monomial -> monomial) -> poly -> poly
    val iter : (monomial -> unit) -> poly -> unit
    val fold : (monomial -> 'a -> 'a) -> poly -> 'a -> 'a
    val find : (monomial -> 'a) -> poly -> 'a

  end
;;





(*

  \subsection{The polynomial functor}

  Functor for building polynomial modules from a coefficient ring, an
  ordered set of variables, and two modules for set and map of
  variables.

*)

module Make 
  (Ring : RingType) 
  (Vars : Ordered_types.OrderedType) 
  (Var_set : Ordered_sets.OrderedSet with type 'a elt=Vars.t) 
  (Var_map : Ordered_maps.OrderedMap with type 'a key=Vars.t) : 

  PolynomialType with 
                   module Base_ring = Ring
                 and 
		   type variable = Vars.t
		 and
  		   module Var_set = Var_set
		 and
  		   module Var_map = Var_map
;;
