(***************************************************************************

  This module provides parameterized functions to complete a rewriting
  system. They are intended to apply as well on terms and on words.

CiME Project - Dmons research team - LRI - Universit Paris XI

$Id: kb.ml,v 1.26 2002/07/19 15:52:19 contejea Exp $

***************************************************************************)

open Format
open Gen_terms
open Orderings_generalities
open Substitution

let coeff_size = ref 1
let coeff_age = ref 0

module type KB =
  sig
    type symbol
    type signature
    type variables
    type t
    type simple_rule
    exception KB_failed
    exception KB_aborted
    exception KB_conjecture_proved of (t * t) list
    val verbose : int ref
    val complete :
      signature -> variables -> t ordering -> 
	(t * t) list -> (t * t) list -> simple_rule list
    val prove_conj_by_comp : 
      signature -> variables -> t ordering -> 
	(t * t) list -> (t * t) list -> bool
  end



module Make (R : Abstract_rewriting.AbstractRewriting) =
struct
  

  type symbol = R.symbol

  type signature = R.signature
			     
  type variables = R.variables
		     
  type t = R.t
		     
  type simple_rule =  R.rule
    
  let verbose = ref 0
		  
  type pair_of_terms =
      { 
	lhs : R.t;
	rhs : R.t;
      }
      
  type equation =
      {
	e : pair_of_terms;
	birth : int;
	size : int;
	e_weight : int;
	is_oriented : bool option
      }
     
  type rule =
      {
	rule : R.rule;
	nb : int;
	r_weight : int;
      }
      
  module OrderedEquation =
    struct 
      type t = equation
      let compare e1 e2 = 
	let d_size = e1.size - e2.size 
	and d_birth = e1.birth - e2.birth
	and d_weight = e1.e_weight - e2.e_weight in
	  if (e1.birth = 0) && (e2.birth > 0)
	  then -1
	  else 
	    if (e2.birth = 0) && (e1.birth > 0)
	    then 1
	    else
	      if d_size <> 0
	      then 
		if d_weight <> 0
		then d_weight
		else d_birth
	      else
		if (R.equals e1.e.lhs e2.e.lhs) && (R.equals e1.e.rhs e2.e.rhs)
		then 0
		else 
		  if d_weight <> 0
		  then d_weight
		  else 
		    if d_birth <> 0
		    then d_birth
		    else 
		      let c_rhs = R.compare_t e1.e.rhs e2.e.rhs in
			if c_rhs <> 0
			then c_rhs
			else R.compare_t e1.e.lhs e2.e.lhs
    end

  module EquationSet = Ordered_sets.Make(OrderedEquation)

  type conjecture = Variables.var_id * (R.t * R.t) * (R.t * R.t) list

  type nar_type =
    | Nothing_to_do
    | Next_to_do of
	(unit -> conjecture * nar_type) *
	(simple_rule option -> simple_rule list -> unit -> nar_type) * 
	(unit -> conjecture) *
	(unit -> unit)

  module ConjectureSet = 
    Ordered_sets.Make(
      struct 
	type t = conjecture
	let compare = Pervasives.compare
      end)


  module NFTable =
    Hashtbl.Make(
      struct
	type t = R.t
	let equal = R.equals
	let hash = R.hash_t
      end)

  type completion_state =
      {
	eqs_to_process : equation EquationSet.t;
	unorientable_eqs : equation  EquationSet.t;
	oriented_eqs : equation EquationSet.t;
	rules : rule list;
	simple_rules : simple_rule list;
	non_oriented_rules : rule list;
	compiled_rules : R.compiled_rules;
	min_deduce_index : int;
	max_deduce_index : int;
	date : int;
	rule_nb : int;
	conjecture : (R.t * R.t) option;
	conjectures_to_narrow : nar_type;
	variables_of_conjectures : Variables.var_id list;
	solution_of_conjecture : (R.t * R.t) list option;
	nf_table : R.t NFTable.t;
	remaining_conj_to_process : (R.t * R.t) list
      }
      
  type choice =
    | ORIENT
    | SELECT_NEW_RULE
    | DEDUCE
    | NARROW
    | STOP
	
  exception Trivial
  exception KB_failed
  exception KB_aborted
  exception KB_conjecture_proved of (R.t * R.t) list

let print_equation sigma vars eq =
  open_box 0;
  R.print_t sigma vars eq.e.lhs;
  print_string " = ";
  R.print_t sigma vars eq.e.rhs;
  print_string " (birth = ";
  print_int eq.birth;
  print_string ", size = ";
  print_int eq.size;
  print_string ",weight = ";
  print_int eq.e_weight;
  print_string ")";
  print_newline();
  print_flush ()

let print_conjecture sigma vars (x,(s,t),lc) =
  open_box 0;
  print_string "shift var = ";
  print_string (Variables.default#string_of_var x);
  print_newline();
  R.print_t sigma vars s;
  print_string " = ";
  R.print_t sigma vars t;
  print_string " | {";
  List.iter 
    (function (l,r) ->
       R.print_t sigma vars l;
       print_string " = ";
       R.print_t sigma vars r;
       print_string "; ";
       print_newline())
    lc;
  print_string " }";
  print_newline();
  print_flush ()
    
let print_numbered_rule sigma vars nb simple_rule =
  printf "@[<hov>[%d]@ @[<hov>" nb;
  R.print_rewrite_rule sigma vars simple_rule;
  printf "@]@]"
    
let print_completion_state sigma vars comp_state =
  printf "-------------- step %d --------------@." comp_state.date;
  printf "min_deduce_index = %d, max_deduce_index = %d@." 
    comp_state.min_deduce_index comp_state.max_deduce_index;
  printf "Equations to process:@.";
  EquationSet.iter 
    (function eq -> print_equation sigma vars eq)
    comp_state.eqs_to_process;
  printf "Unorientable equations:@.";
  EquationSet.iter 
    (function eq -> print_equation sigma vars eq)
    comp_state.unorientable_eqs;
  printf "Oriented equations:@.";
  EquationSet.iter 
    (function eq -> print_equation sigma vars eq)
    comp_state.oriented_eqs;
  printf "Rules:@.";
  List.iter
    (function rule -> 
       print_numbered_rule sigma vars rule.nb rule.rule;
       print_newline())
      comp_state.rules;
  begin
    match
      comp_state.conjectures_to_narrow with
	| Nothing_to_do -> ()
	| Next_to_do (_,_,_,pr) -> pr()
  end;
  printf "-------------------------------------@."

let normalize nftbl sigma vars compiled_rules s =
  let nf_s = 
    R.normalize (NFTable.find nftbl) (NFTable.add nftbl) 
      sigma vars compiled_rules s in
    nf_s

let force_normalize nftbl sigma vars compiled_rules s =
  try
    let nf_s = NFTable.find nftbl s in
      if R.equals s nf_s
      then raise R.Irreducible
      else nf_s
  with Not_found ->
    let nf_s = 
      R.force_normalize (NFTable.find nftbl) (NFTable.add nftbl) 
	sigma vars compiled_rules s in
      NFTable.add nftbl s nf_s;
      nf_s

let normalize_pair nftbl sigma vars compiled_rules s t =
  try
    let s' = force_normalize nftbl sigma vars compiled_rules s in
      try
	let t' = force_normalize nftbl sigma vars compiled_rules t in
	  (s',t')
      with 
	  R.Irreducible -> (s',t)
  with 
      R.Irreducible ->
	try
	  let t' = force_normalize nftbl sigma vars compiled_rules t in
	    (s,t')
	with 
	    R.Irreducible -> raise R.Irreducible 
		
let simplify_pairs nftbl sigma vars compiled_rules list_of_pairs_of_terms =
  let rec rec_simplify = function
    | [] -> []
    |  (l,r)::list_of_p ->
	 if R.equals l r
	 then rec_simplify list_of_p
	 else
	   let l' = normalize nftbl sigma vars compiled_rules l
	   and r' = normalize nftbl sigma vars compiled_rules r in
	     if R.equals l' r'
	     then 
(*i
	     let _ = 
	       Format.print_string "***********simplify_pairs is called on***************";
	       Format.print_newline();
	       R.print_t sigma vars l;
	       Format.print_string " = ";
	       R.print_t sigma vars r;
	       Format.print_newline();
	       Format.print_string "the result is";
	       Format.print_newline();
	       Format.print_string "a collapsing pair";
	       Format.print_newline();
	       Format.print_flush() in
i*)
	       rec_simplify list_of_p
	     else 
(*i
	     let _ = 
	       Format.print_string "***********simplify_pairs is called on***************";
	       Format.print_newline();
	       R.print_t sigma vars l;
	       Format.print_string " = ";
	       R.print_t sigma vars r;
	       Format.print_newline();
	       Format.print_string "the result is";
	       Format.print_newline();
	       R.print_t sigma vars l';
	       Format.print_string " = ";
	       R.print_t sigma vars r';
	       Format.print_newline();
	       Format.print_flush() in
i*)
	       (l',r')::(rec_simplify list_of_p) in

	   let res = rec_simplify list_of_pairs_of_terms in
	   let _ =
	     if !verbose >= 5 then
	       begin
		 print_string "***********simplify_pairs is called on***************";
		 print_newline();
		 List.iter
		   (function (s,t) ->  
		      R.print_t sigma vars s;
		      print_string " = ";
		      R.print_t sigma vars t;
		      Format.print_newline();
		      Format.print_flush())
		   list_of_pairs_of_terms;
		 print_string "***********the result is***************";
		 print_newline();
		 List.iter
		   (function (s,t) ->  
		      R.print_t sigma vars s;
		      print_string " = ";
		      R.print_t sigma vars t;
		      Format.print_newline();
		      Format.print_flush())
		   res;
	       end in
	     res

      
let simplify nftbl sigma vars compiled_rules set_of_eqs =
  EquationSet.fold
    (fun eq equations ->
       try
	 if R.equals eq.e.lhs eq.e.rhs
	 then equations
	 else
	   let l',r' = 
	     normalize_pair nftbl sigma vars compiled_rules eq.e.lhs eq.e.rhs in
	     if R.equals l' r'
	     then equations
	     else 
	       let eq' =
		 {
		   eq with 
		     e =
		 {
		   lhs = l';
		   rhs = r';
		 }
       } in
		 EquationSet.add eq' equations
       with 
	   R.Irreducible -> 
	     EquationSet.add eq equations)
    set_of_eqs EquationSet.empty
      
let simplify_and_split nftbl sigma vars compiled_rules set_of_eqs =
  EquationSet.fold
    (fun eq (eqs, simpl_eqs as res) -> 
       if R.equals eq.e.lhs eq.e.rhs
       then res
       else
	 try
	   let l',r' = 
	     normalize_pair nftbl sigma vars compiled_rules eq.e.lhs eq.e.rhs in
	     if R.equals l' r'
	     then res
	     else 
	       let eq' =
		 {
		   eq with
		     e = {
		       lhs = l';
		       rhs = r';
		     }
         } in
		 (eqs, EquationSet.add eq' simpl_eqs)
	 with 
	     R.Irreducible -> ((EquationSet.add eq eqs),simpl_eqs))
    set_of_eqs (EquationSet.empty, EquationSet.empty) 

let simplify_and_split_pairs nftbl sigma vars compiled_rules list_of_pairs_of_terms =
  let (reversed_left_eqs,simplified_eqs) =
    List.fold_left
      (fun (eqs, simpl_eqs as res) (s,t as eq) -> 
	 try
	   let s',t' = 
	     normalize_pair nftbl sigma vars compiled_rules s t in
	     if R.equals s' t'
	     then res
	     else (eqs,(s',t')::simpl_eqs)
	 with 
	     R.Irreducible -> ((eq::eqs),simpl_eqs))
      ([],[]) list_of_pairs_of_terms in
    (List.rev reversed_left_eqs, simplified_eqs)

let collapse nftbl sigma vars date list_of_rules compiled_simple_rule =
  let (reversed_left_rules,collapsed_rules,collapsed_nbs) =
    List.fold_left
      (fun (rules,coll_rules,coll_nbs) rule ->
	 try
	   let l = R.lhs_of_rule rule.rule
	   and r = R.rhs_of_rule rule.rule in
	   let l' = force_normalize nftbl sigma vars compiled_simple_rule l in
	   let r' = normalize nftbl sigma vars compiled_simple_rule r in
	   if R.equals l' r'
	   then (rules, coll_rules, rule.nb::coll_nbs)
	   else 
	     let rule' = 
	       {
		 e = 
		  {
		    lhs = l';
		    rhs = r';
		  };
		 size = 0;
		 birth = date;
		 e_weight = 0;
		 is_oriented = None;
	       } in
	     (rules, (EquationSet.add rule' coll_rules), rule.nb::coll_nbs)
	 with 
	     R.Irreducible -> (rule::rules, coll_rules, coll_nbs))
      ([],EquationSet.empty,[]) list_of_rules in
  
  (List.rev reversed_left_rules,collapsed_rules,collapsed_nbs)
    
    
let compose nftbl sigma vars list_of_rules compiled_simple_rule =
  List.map
    (function rule ->
       let r = R.rhs_of_rule rule.rule in
	 try
	   let r' = force_normalize nftbl sigma vars compiled_simple_rule r in
	   let l = R.lhs_of_rule rule.rule in
	   let new_rule = 
	     {
	       rule with
		 rule = R.make_rule sigma l r'
	     } in
	   let _ =
	     if !verbose >= 1
	     then 
	       begin
		 print_string "Rule ";
		 print_numbered_rule sigma vars rule.nb rule.rule;
		 print_string " is composed into ";
		 print_numbered_rule sigma vars new_rule.nb new_rule.rule;
		 print_newline();
		 print_flush() 
	       end in
	     new_rule
		 
	 with 
	     R.Irreducible -> rule)
    list_of_rules

(*
let compute_weight date size birth = 
  2*(!coeff_size*size + ((!coeff_age*birth)/date))

let compute_rule_weight date size birth = 
  (!coeff_size*size + ((!coeff_age*birth)/date))
*)

let compute_weight date size birth = 
  2*(!coeff_size*size + ((!coeff_age*birth)/30))

let compute_rule_weight date size birth = 
  (!coeff_size*size + ((!coeff_age*birth)/30))

let add_eq date sigma eq set_of_eqs =
  let size = 
    if eq.size = 0
    then (R.size sigma eq.e.lhs) + (R.size sigma eq.e.rhs)
    else eq.size in
  let weight =
    if eq.e_weight = 0
    then compute_weight date size eq.birth
    else eq.e_weight in
  let new_eq =
    {
      eq with
	size = size;
	e_weight = weight
    } in
    EquationSet.add new_eq set_of_eqs
      


let developp_narrow_tree sigma vars o start =
  let closed = ref [] in
  let opened = ref [start] in
  let  sons = ref (fun n -> []) in
  let rec developp =
    (fun () ->
       match !opened with
	 | [] -> raise Exit
	 | n::l ->
	     closed := n :: !closed;
	     opened := 
	     l @ 
	     (List.filter 
		(fun x -> not ((List.mem x !opened) || (List.mem x !closed))) 
		(!sons n));
	     n,(Next_to_do (developp,update_sons,current_conjecture,print_conjectures)))
  and update_sons =
    (fun new_rule list_of_simple_rules () ->
       begin
	 match new_rule with
	   | None -> ()
	   | Some r ->
	       let compiled_rules_for_nar =
		 R.compile_for_unification sigma list_of_simple_rules 
	       and compiled_r = R.compile_for_unification sigma [r] in
	       sons := (fun c -> R.narrow sigma o compiled_rules_for_nar c);
	       let partial_sons =
		 (fun c ->  R.narrow sigma o compiled_r c) in
		 opened := 
		 List.fold_left
		   (fun acc x -> 
		      List.fold_left
			(fun acc' ps -> 
			   if (List.mem ps !opened) || (List.mem ps !closed)
			   then acc'
			   else ps :: acc')
			acc (partial_sons x))
		   !opened !closed 
       end;
       Next_to_do (developp,update_sons,current_conjecture,print_conjectures))
  and current_conjecture =
    fun () -> try List.hd !opened with Failure _ -> raise Not_found 
  and print_conjectures =
    fun () -> 
      Format.print_string "Already developped conjectures";
      Format.print_newline();
      List.iter (print_conjecture sigma vars) !closed;
      Format.print_string "Conjectures to be developped";
      Format.print_newline();
      List.iter (print_conjecture sigma vars) !opened;
  in
  Next_to_do (developp,update_sons,current_conjecture,print_conjectures)

let init_comp_state sigma vars o eq_set conjs = 
  if !verbose >= 1 then 
    printf "Initializing completion ...@.";
  let variables_of_conjectures =
    match conjs with
      | [] -> []
      | (s,t)::l -> (R.variables_of_t s) @ (R.variables_of_t t) in
  let closed_conjectures = (variables_of_conjectures = []) in
  {
    eqs_to_process =
     List.fold_left  
       (fun equations (l,r) ->
	  let s = (R.size sigma l) + (R.size sigma r) in
	  let eq = 
	    {
	      e = 
	       {
		 lhs = l;
		 rhs = r;
	       };
	      birth = 0;
	      size = s;
	      e_weight = compute_weight 1 s 0;
	      is_oriented = None;
	    } in
	    EquationSet.add eq equations)
       EquationSet.empty eq_set;
    unorientable_eqs = EquationSet.empty;
    oriented_eqs = EquationSet.empty;
    rules = [];
    simple_rules = [];
    non_oriented_rules = [];
    compiled_rules = R.compile sigma [];
    min_deduce_index = max_int;
    max_deduce_index = max_int;
    date = 0;
    rule_nb = 1;
    conjecture = 
     if closed_conjectures 
     then 
       (match conjs with 
	  | [] -> None 
	  | (s,t)::_ -> Some (s,t)) 
     else None;
    conjectures_to_narrow = 
     if closed_conjectures
     then Nothing_to_do
     else 
       (match conjs with 
	  | [] -> assert false 
	  | (s,t as c)::_ ->
	      developp_narrow_tree sigma vars o (R.init_narrow c));
    variables_of_conjectures = variables_of_conjectures;
    solution_of_conjecture = None;
    nf_table = NFTable.create 1997;
    remaining_conj_to_process = 
     match conjs with
       | [] -> []
       | _::l -> l
  }
      
let reset_comp_state sigma vars o comp_state =
  let variables_of_conjectures =
    match comp_state.remaining_conj_to_process with
      | [] -> []
      | (s,t)::l -> (R.variables_of_t s) @ (R.variables_of_t t) in
  let closed_conjectures = (variables_of_conjectures = []) in
    { 
      comp_state with
	conjecture = 
	if closed_conjectures 
	then 
	  (match comp_state.remaining_conj_to_process with
	     | [] -> None 
	     | (s,t)::_ -> Some (s,t)) 
	else None;
	conjectures_to_narrow = 
	if closed_conjectures
	then Nothing_to_do
	else 
	  (match comp_state.remaining_conj_to_process with
	     | [] -> assert false 
	     | (s,t as c)::_ ->
		 developp_narrow_tree sigma vars o (R.init_narrow c));
	variables_of_conjectures = variables_of_conjectures;
	remaining_conj_to_process = 
	match comp_state.remaining_conj_to_process with
	  | [] -> []
	  | _::l -> l
}

let rec get_nth_rule n = function
    [] -> raise Not_found
  | r::list_of_rules ->
      if r.nb = n
      then r
      else get_nth_rule n list_of_rules


let rec find_next_index_rule max_nb_rules rules n =
  if n <= max_nb_rules
  then 
    try
      let _ = get_nth_rule n rules in 
	n
    with 
	Not_found -> find_next_index_rule max_nb_rules rules (n+1)
  else raise Not_found 
     
let increment_min_deduce_index comp_state =
  let new_min_index =
    try 
      find_next_index_rule comp_state.rule_nb comp_state.rules 
	(comp_state.min_deduce_index + 1)
    with 
	Not_found -> max_int in
    {
      comp_state with
	min_deduce_index = new_min_index
    }
	    

let increment_max_deduce_index comp_state =
  try
    let new_max_index = 
      find_next_index_rule comp_state.rule_nb comp_state.rules 
	(comp_state.max_deduce_index + 1) in
      increment_min_deduce_index
	{
	  comp_state with
	    max_deduce_index = new_max_index;
	    min_deduce_index = 0
        }
  with 
      Not_found ->
	{
	  comp_state with
	    max_deduce_index = max_int
        }
      
let weight_orient comp_state =
  try
    let eq = EquationSet.min_elt comp_state.eqs_to_process in
      eq.e_weight
  with 
      Not_found -> max_int
	  
let weight_select_new_rule comp_state = 
  try
    let eq = EquationSet.min_elt comp_state.oriented_eqs in
      eq.e_weight
  with
      Not_found -> max_int
	  
let weight_deduce comp_state =
  try
    let min_rule = 
      get_nth_rule comp_state.min_deduce_index comp_state.rules
    and max_rule =
      get_nth_rule comp_state.max_deduce_index comp_state.rules in
      min_rule.r_weight + max_rule.r_weight
  with _ -> max_int
      
let weight_narrow sigma comp_state = 
  match comp_state.conjectures_to_narrow 
  with 
    | Nothing_to_do -> max_int
    | Next_to_do (_,_,current_conjecture,_) ->
	try 
	  let (_,(s,t),l) = current_conjecture () in
	    List.fold_left
	      (fun acc (u,v) -> acc + (R.size sigma u) + (R.size sigma v))
	      ((R.size sigma s) + (R.size sigma t))
	      l
	with Not_found -> max_int

let min4 n1 n2 n3 n4 =
  min (min (min n1 n2) n3) n4

let what_to_do sigma comp_state = 
  let wo = weight_orient comp_state
  and ws = weight_select_new_rule comp_state 
  and wd = weight_deduce comp_state 
  and wn = weight_narrow sigma comp_state in
  let m = min4 wo ws wd wn in
    if m = max_int
    then STOP
    else
      if wo = m
      then ORIENT
      else 
	if ws = m
	then SELECT_NEW_RULE
	else 
	  if wd = m
	  then 
	    if wd < max_int
	    then DEDUCE
	    else STOP
	  else
	    if wn = m
	    then NARROW
	    else STOP

exception Trivial_eq

let pre_orient nftbl sigma vars ordering comp_state = 
  let eq = EquationSet.min_elt comp_state.eqs_to_process in 
  let new_eqs_to_process = EquationSet.remove eq comp_state.eqs_to_process in
    try 
      let lhs,rhs =
	try 
	  let l,r = normalize_pair nftbl sigma vars comp_state.compiled_rules 
		      eq.e.lhs eq.e.rhs in 
	    if R.equals l r then raise Trivial_eq;
	    l,r
	with 
	    R.Irreducible -> 
	      if R.equals eq.e.lhs eq.e.rhs then raise Trivial_eq;
	      eq.e.lhs,eq.e.rhs in
      let e =
	match ordering lhs rhs with
	    Greater_than -> 
	      {
		lhs = lhs;
		rhs = rhs
	      }
	  | Less_than ->
	      {
		lhs = rhs;
		rhs = lhs
	      }
	  | _ -> raise Exit in

      let size = 
	if eq.size = 0
	then (R.size sigma lhs) + (R.size sigma rhs)
	else eq.size in
      let weight =
	if eq.e_weight = 0
	then compute_weight comp_state.date size eq.birth
	else eq.e_weight in
      let pre_rule =
	{
	  e = e; 
	  birth = eq.birth;
	  size = size;
	  e_weight = weight;
	  is_oriented = Some true;
	} in
      let new_oriented_eqs = 
	add_eq comp_state.date sigma pre_rule comp_state.oriented_eqs in
	{
	  comp_state with
	    eqs_to_process = new_eqs_to_process;
	    oriented_eqs = new_oriented_eqs;
	}
    with Trivial_eq -> 	
      if !verbose >= 5 then printf "Degenerated eq@.";
      { comp_state with eqs_to_process = new_eqs_to_process }


let orient nftbl sigma vars ordering comp_state = 
  try 
    pre_orient nftbl sigma vars ordering comp_state
  with 
      Exit -> 
	let eq = EquationSet.min_elt comp_state.eqs_to_process in 
	let new_eqs_to_process = 
	  EquationSet.remove eq comp_state.eqs_to_process in
	let new_unorientable_eqs = 
	  add_eq comp_state.date sigma eq comp_state.unorientable_eqs in
	  { 
	    comp_state with
	      eqs_to_process = new_eqs_to_process;
	      unorientable_eqs = new_unorientable_eqs
          }
      

let force_orient nftbl sigma vars ordering c comp_state = 
  try 
    pre_orient nftbl sigma vars ordering comp_state
  with 
      Exit ->
	let eq = EquationSet.min_elt comp_state.eqs_to_process in 
	let size = 
	  if eq.size = 0
	  then (R.size sigma eq.e.lhs) + (R.size sigma eq.e.rhs)
	  else eq.size in
	let weight =
	  if eq.e_weight = 0
	  then compute_weight comp_state.date size eq.birth
	  else eq.e_weight in
	let pre_rule1 =
	  Some ({
		  e = {lhs = eq.e.lhs; rhs = eq.e.rhs}; 
		  birth = eq.birth;
		  size = size;
		  e_weight = weight;
		  is_oriented = Some false
		})
	and 
	  pre_rule2 =
	  Some ({
		  e = {lhs = eq.e.rhs; rhs = eq.e.lhs}; 
		  birth = eq.birth;
		  size = size;
		  e_weight = weight;
		  is_oriented = Some false
		}) 
	and pre_rule3 = 
	  try 
	    let a = 
	      match c with 
		  Some cste -> cste
		| None -> R.a_variable_of_the_term eq.e.lhs in
	    let rhs_c = R.regular_pair sigma a eq.e.lhs eq.e.rhs in
	      if R.equals eq.e.lhs rhs_c
	      then None
	      else 
		Some ({
			e = {lhs = eq.e.lhs; rhs = rhs_c};
			birth = eq.birth;
			size = size;
			e_weight = weight;
			is_oriented =  
			 match ordering eq.e.lhs rhs_c with
			     Greater_than -> Some true
			   | _ -> Some false
		      })
	  with Exit -> None
	and pre_rule4 =
	  try 
	    let b = 
	      match c with 
		  Some cste -> cste
		| None -> R.a_variable_of_the_term eq.e.rhs in
	    let lhs_c = R.regular_pair sigma b eq.e.rhs eq.e.lhs in
	      if R.equals eq.e.rhs lhs_c
	      then None
	      else 
		Some ({
			e = {lhs = eq.e.rhs; rhs = lhs_c};
			birth = eq.birth;
			size = size;
			e_weight = weight;
			is_oriented =  
			 match ordering eq.e.rhs lhs_c with
			     Greater_than -> Some true
			   | _ -> Some false
		      })
	  with Exit -> None in
	  
	let list_of_pre_rules = [pre_rule1;pre_rule2;pre_rule3;pre_rule4] in
	let new_eqs_to_process = 
	  EquationSet.remove eq comp_state.eqs_to_process in
	let new_oriented_eqs = 
	  List.fold_left 
	    (fun res opt_pre_rule ->
		match opt_pre_rule with
		    None -> res
		  | Some pre_rule -> add_eq comp_state.date sigma pre_rule res)
	    comp_state.oriented_eqs list_of_pre_rules in
	  {
	    comp_state with
	      eqs_to_process = new_eqs_to_process;
	      oriented_eqs = new_oriented_eqs;
          }

let equation_extension o (e1,e2) (e'1,e'2) = 
  let o_e1_e'1 = o e1 e'1 
  and o_e1_e'2 = o e1 e'2
  and o_e2_e'1 = o e2 e'1
  and o_e2_e'2 = o e2 e'2 in
    match o_e1_e'1, o_e1_e'2, o_e2_e'1, o_e2_e'2 with
      | Equivalent, _, _, _ -> o_e2_e'2
      | _, Equivalent, _, _ -> o_e2_e'1
      | _, _, Equivalent, _ -> o_e1_e'2
      | _, _, _, Equivalent -> o_e1_e'1

      | Greater_than, Greater_than, _, _ -> Greater_than
      | _, _, Greater_than, Greater_than -> Greater_than
      | Greater_than, _, _, ( Greater_than
			    | Greater_or_equivalent) -> Greater_than
      | Greater_or_equivalent, _, _, Greater_than -> Greater_than
      | _, Greater_than, ( Greater_than
			 | Greater_or_equivalent), _ -> Greater_than
      | _, Greater_or_equivalent, Greater_than, _ -> Greater_than
      | _, Greater_or_equivalent, Greater_or_equivalent, _ -> Greater_or_equivalent
      | Greater_or_equivalent, _, _, Greater_or_equivalent -> Greater_or_equivalent


      | Less_than, _, Less_than, _ -> Less_than
      | _, Less_than, _, Less_than -> Less_than
      | Less_than, _, _, ( Less_than
			    | Less_or_equivalent) -> Less_than
      | _, Less_than, ( Less_than
			 | Less_or_equivalent), _ -> Less_than
      | Less_or_equivalent, _, _, Less_than -> Less_than
      | _, Less_or_equivalent, Less_than, _ -> Less_than
      | _, Less_or_equivalent, Less_or_equivalent, _ -> Less_or_equivalent
      | Less_or_equivalent, _, _, Less_or_equivalent -> Less_or_equivalent

      | _,_,_,_ -> Uncomparable

let rec is_a_max_conj ordering c = function
  | [] -> true
  | c'::l ->
      match equation_extension ordering c c' with
	| Greater_than 
	| Greater_or_equivalent 
	| Uncomparable -> is_a_max_conj ordering c l
	| _ -> false

let split_between_max_conj_and_others ordering l =
  let rec rec_split (c_max,c_others as res) = function
    | [] -> res
    | c :: l ->
	if (is_a_max_conj ordering c c_max) &&
	  (is_a_max_conj ordering c l) 
	then rec_split (c::c_max, c_others) l
	else rec_split (c_max, c::c_others) l in
    match l with
      | [_] -> (l,[])
      | _ -> rec_split ([],[]) l

let update_rules_for_select_new_rule sigma vars lhs rhs comp_state =
  let new_simple_rule = R.make_rule sigma lhs rhs in
  let new_rule =
    {
      rule = new_simple_rule;
      nb = comp_state.rule_nb;
      r_weight = 
       let size = 
	 (R.size sigma lhs) + 
	 (R.size sigma rhs) in
	 compute_rule_weight comp_state.date size comp_state.date
    } 
  and compiled_new_simple_rule = R.compile sigma [new_simple_rule] in

  let nft_new_simple_rule = NFTable.create 1997 in
    
  let new_rules, collapsed_rules, collapsed_numbers = 
    collapse nft_new_simple_rule sigma vars comp_state.date comp_state.rules 
      compiled_new_simple_rule in
  let composed_new_rules = 
    compose nft_new_simple_rule sigma vars new_rules compiled_new_simple_rule in
  let rules = new_rule :: composed_new_rules in
  let non_oriented_rules = 
    if (comp_state.non_oriented_rules = []) && 
      (R.is_oriented new_simple_rule)
    then []
    else 
      List.filter (function rule -> not (R.is_oriented rule.rule)) rules 
  and compiled_rules =
    R.compile sigma (List.map (function rule -> rule.rule) rules) in
  let simple_rules = 
    List.rev_map (function rule -> rule.rule) rules in
  let filter = EquationSet.filter 
		 (function e ->
		    let s = e.e.lhs and t = e.e.rhs in
		      not (List.exists
			     (function r ->
				R.is_encompassed_by sigma
				  (R.lhs_of_rule r.rule) 
				  (R.rhs_of_rule r.rule) s t)
			     non_oriented_rules)) in
    
  let red_collapsed_rules =
    if non_oriented_rules = []
    then simplify comp_state.nf_table sigma vars compiled_rules collapsed_rules
    else filter (simplify comp_state.nf_table sigma vars compiled_rules collapsed_rules) in

  let _ = 
    printf "@[<hov>New rule produced :@ ";
    print_numbered_rule sigma vars comp_state.rule_nb new_simple_rule;
    printf "@.@]";
    List.iter
      (function n ->
	 try
	   let r = get_nth_rule n comp_state.rules in
	     printf "@[<hov>Rule@ ";
	     print_numbered_rule sigma vars n r.rule;
	     printf "@ collapsed.@.@]"
	 with Not_found -> ())
      collapsed_numbers in
    
        new_simple_rule,
        compiled_new_simple_rule,
	rules,
	non_oriented_rules,
	simple_rules,
	compiled_rules,
	red_collapsed_rules



let select_new_rule sigma vars ordering comp_state = 
  NFTable.clear comp_state.nf_table;
  let new_rule_nb = comp_state.rule_nb + 1 in
  let selected_eq = EquationSet.min_elt comp_state.oriented_eqs in
  let oriented_eqs = EquationSet.remove selected_eq comp_state.oriented_eqs in
  
  let (new_simple_rule, compiled_new_simple_rule, rules, 
       non_oriented_rules, simple_rules, 
       compiled_rules,
       red_collapsed_rules)  =
    update_rules_for_select_new_rule sigma vars selected_eq.e.lhs selected_eq.e.rhs comp_state in

  let filter = EquationSet.filter 
		 (function e ->
		    let s = e.e.lhs and t = e.e.rhs in
		      not (List.exists
			     (function r ->
				R.is_encompassed_by sigma
				  (R.lhs_of_rule r.rule) 
				  (R.rhs_of_rule r.rule) s t)
			     non_oriented_rules)) in
    
  let new_eqs_to_process, simpl_eqs_to_process = 
    simplify_and_split comp_state.nf_table sigma vars compiled_new_simple_rule 
      comp_state.eqs_to_process
      (* comp_state.eqs_to_process,EquationSet.empty *)
  and new_unorientable_eqs, simpl_unorientable_eqs = 
    simplify_and_split comp_state.nf_table sigma vars compiled_new_simple_rule 
      comp_state.unorientable_eqs
  and new_oriented_eqs, simpl_oriented_eqs =
    simplify_and_split comp_state.nf_table sigma vars compiled_new_simple_rule 
      oriented_eqs in

  let red_eqs_to_process = 
    if non_oriented_rules = []
    then simplify comp_state.nf_table sigma vars compiled_rules simpl_eqs_to_process
    else filter (simplify comp_state.nf_table sigma vars compiled_rules simpl_eqs_to_process)
  and red_unorientable_eqs = 
    if non_oriented_rules = []
    then simplify comp_state.nf_table sigma vars compiled_rules simpl_unorientable_eqs 
    else filter (simplify comp_state.nf_table sigma vars compiled_rules simpl_unorientable_eqs)
  and red_oriented_eqs = 
    if non_oriented_rules = []
    then simplify comp_state.nf_table sigma vars compiled_rules simpl_oriented_eqs
    else filter (simplify comp_state.nf_table sigma vars compiled_rules simpl_oriented_eqs)
  and red_conjecture = 
    begin
      match comp_state.conjecture with 
	| None -> None
	| Some (s,t) as c ->
	    try
	      let s',t' = 
		normalize_pair (NFTable.create 1997) sigma vars compiled_new_simple_rule s t in
		if R.equals s' t'
		then None
		else 
		  let s'' = normalize comp_state.nf_table sigma vars compiled_rules s'
		  and t'' = normalize comp_state.nf_table sigma vars compiled_rules t' in
		    if R.equals s'' t''
		    then None
		    else Some(s'',t'')
	    with R.Irreducible -> c
     end in

    {
      comp_state with
	eqs_to_process = 
       EquationSet.fold (add_eq comp_state.date sigma) 
	 (EquationSet.fold (add_eq comp_state.date sigma)
	    (EquationSet.fold (add_eq comp_state.date sigma)
	       (EquationSet.fold (add_eq comp_state.date sigma)
		  (if non_oriented_rules = []
		   then new_eqs_to_process 
		   else filter new_eqs_to_process)
		  red_collapsed_rules)
	       red_eqs_to_process)
	    red_unorientable_eqs)
	 red_oriented_eqs;
	
	unorientable_eqs = new_unorientable_eqs;
	oriented_eqs = 
       if non_oriented_rules = [] 
       then new_oriented_eqs
       else filter new_oriented_eqs;
        rules = rules;
	non_oriented_rules = non_oriented_rules;
	simple_rules = simple_rules;
	compiled_rules = compiled_rules;
	max_deduce_index =
       if comp_state.min_deduce_index = max_int
       then comp_state.rule_nb
       else 
	 find_next_index_rule new_rule_nb rules comp_state.max_deduce_index;
	min_deduce_index =
       if comp_state.min_deduce_index = max_int
       then find_next_index_rule new_rule_nb rules 1
       else 
	 find_next_index_rule new_rule_nb rules comp_state.min_deduce_index;
	rule_nb = new_rule_nb;
	conjecture = 
       begin
	 let c = 
	   if non_oriented_rules = [] 
	   then red_conjecture
	   else 
	     match red_conjecture with
	       | None -> red_conjecture
	       | Some (s,t) ->
		   if (List.exists
			 (function r ->
			    R.is_encompassed_by sigma
			      (R.lhs_of_rule r.rule) 
			      (R.rhs_of_rule r.rule) s t)
			 non_oriented_rules)
		   then None
		   else red_conjecture in
	 let _ =
	   if c <> comp_state.conjecture
	   then 
	     begin
	       printf "@[<hov>The conjecture has been reduced.@ ";
	       print_newline();
	       printf "Conjecture is now:@ ";
	       (match c with
		  | None -> print_string "Trivial"
		  | Some (s,t) ->   
		      open_box 0;
		      R.print_t sigma vars s;
		      print_string " = ";
		      R.print_t sigma vars t;
		      print_newline();
		      print_flush ());
	       print_newline();
	       print_flush ();
	     end in
	   c
       end;
	conjectures_to_narrow =
       begin
	 match comp_state.conjectures_to_narrow with
	   | Nothing_to_do -> Nothing_to_do
	   | Next_to_do (_,f,_,_) -> 
	       f (Some new_simple_rule) simple_rules ()
       end;
}
    
let deduce sigma vars comp_state = 
  if comp_state.min_deduce_index > comp_state.max_deduce_index
  then increment_max_deduce_index comp_state
  else
    try
      let rule_max = 
	get_nth_rule comp_state.max_deduce_index comp_state.rules in
	try
	  let cp = 
	    if comp_state.min_deduce_index = comp_state.max_deduce_index
	    then 
	      begin
		let _ = 
		  if !verbose >= 2 
		  then printf "SelfDeduce(%d) -> " comp_state.max_deduce_index in
		  R.self_critical_pairs sigma rule_max.rule
	      end
	    else 
	      let rule_min = 
		get_nth_rule comp_state.min_deduce_index comp_state.rules in
	      let _ = 
		if !verbose >= 2 then 
		  printf "Deduce(%d,%d) -> " 
		    comp_state.max_deduce_index comp_state.min_deduce_index in
		R.critical_pairs sigma rule_min.rule rule_max.rule in
	  let _ =
	    if !verbose >= 2 then 
	      printf "%d critical_pairs@." (List.length cp);
	    if !verbose >= 5 then
	      List.iter
		(function (s,t) ->  
		   R.print_t sigma vars s;
		   print_string " = ";
		   R.print_t sigma vars t;
		   Format.print_newline();
		   Format.print_flush())
		cp in
	  let simplified_cp = 
	    R.canonize_pairs sigma
	      (simplify_pairs comp_state.nf_table sigma vars comp_state.compiled_rules cp) in 
	  let _ = 
	    if !verbose >= 2 then 
	      printf "%d simplified_critical_pairs@." (List.length simplified_cp);
	    if !verbose >= 5 then
	      List.iter
		(function (s,t) ->  
		   R.print_t sigma vars s;
		   print_string " = ";
		   R.print_t sigma vars t;
		   Format.print_newline();
		   Format.print_flush())
		simplified_cp in
	  let eqs_to_add =
	    List.fold_left
	      (fun equations (s,t) ->
		 let eq = 
		   {
		     e = {lhs = s; rhs = t };
		     size = 0;
		     birth = comp_state.date;
		     e_weight = 0;
		     is_oriented = None;
		   } in
		   EquationSet.add eq equations)
	      EquationSet.empty 
	      (if comp_state.non_oriented_rules = []
	       then simplified_cp 
	       else
		 List.filter 
		   (function (s,t) ->
		      not (List.exists
			     (function r ->
				R.is_encompassed_by sigma
				  (R.lhs_of_rule r.rule) 
				  (R.rhs_of_rule r.rule) s t)
			     comp_state.non_oriented_rules))
		   simplified_cp) in
	    
	  let new_eqs_to_process =
	    EquationSet.fold (add_eq comp_state.date sigma) 
	      eqs_to_add comp_state.eqs_to_process in
	    increment_min_deduce_index
	      {
		comp_state with
		  eqs_to_process = new_eqs_to_process
	}
	with Not_found -> increment_min_deduce_index comp_state
    with Not_found -> increment_max_deduce_index comp_state
	    	 
(*      
let narrow sigma vars comp_state = comp_state
*)
let complete sigma vars ordering eq_set conjs = 
  let comp_state = ref (init_comp_state sigma vars ordering eq_set conjs) in
    try
      while true do
	if !verbose >= 4 then
	  begin
	    print_completion_state sigma vars !comp_state;
	    print_string "Press Enter to continue";
	    print_flush();
	    ignore(read_line())
	  end;
	begin
	  match what_to_do sigma !comp_state with
	    | ORIENT -> 
		if !verbose >= 2 then printf "Orient@.";
		comp_state := orient !comp_state.nf_table sigma vars ordering !comp_state
	    | SELECT_NEW_RULE -> 
		if !verbose >= 2 then printf "Select new rule@.";
		comp_state := select_new_rule sigma vars ordering !comp_state
	    | DEDUCE -> 
		if !verbose >= 2 then printf "Deduction@.";
		comp_state := deduce sigma vars !comp_state 
	    | NARROW ->
		if !verbose >= 2 then printf "Narrowing@.";
		let ((_,(s,t),c),new_narrow) = 
		  (match !comp_state.conjectures_to_narrow with
		     | Nothing_to_do -> assert false
		     | Next_to_do (f,_,_,_) -> f) () in
		  begin
		    try
		      let sol = 
			R.solve_conj sigma ((s,t)::c) in
			comp_state := { !comp_state with
					  conjectures_to_narrow = Nothing_to_do;
					  solution_of_conjecture = Some sol }
		    with 
		      | R.Unsolvable -> 
			  comp_state := { !comp_state with conjectures_to_narrow = new_narrow }
		  end
	    | STOP -> raise Exit
	end;
	comp_state := { !comp_state with date = succ !comp_state.date }
      done;
      assert false
    with 
	Exit -> 
	  begin
	    if not (EquationSet.is_empty (!comp_state.unorientable_eqs))
	    then raise KB_failed
	    else 
	      List.map
		(function r -> r.rule)
		!comp_state.rules
	  end 
	  
let prove_conj_by_comp sigma vars ordering rule_set conj = 
  let comp_state = ref (init_comp_state sigma vars ordering rule_set conj) 
  and c = 
    try 
      Some (R.smallest_constant sigma ordering) 
    with Not_found -> None in
    try
      while true do
	try
	  if !verbose >= 4 then
	    begin
	      print_completion_state sigma vars !comp_state;
	      print_string "Press Enter to continue";
	      print_flush();
	      ignore(read_line())
	    end;
	  begin
	    match what_to_do sigma !comp_state with
	      | ORIENT -> 
		  if !verbose >= 2 then printf "Orient@.";
		  comp_state := force_orient !comp_state.nf_table sigma vars ordering c !comp_state
	      | SELECT_NEW_RULE -> 
		  if !verbose >= 2 then printf "Select new rule@.";
		  comp_state := select_new_rule sigma vars ordering !comp_state;
		  if !verbose >= 1
		  then 
		    begin
		      print_string "Current number of rules: ";
		      print_int (List.length !comp_state.rules);
		      print_newline();
		      print_flush()
		    end;
	      | DEDUCE -> 
		  if !verbose >= 2 then printf "Deduction@.";
		  comp_state := deduce sigma vars !comp_state 
	      | NARROW -> 
		  if !verbose >= 2 then printf "Narrowing@.";
		  let ((_,(s,t),c),new_narrow) = 
		    (match !comp_state.conjectures_to_narrow with
		       | Nothing_to_do -> assert false
		       | Next_to_do (f,_,_,_) -> f) () in
		    begin
		      try
			let sol = 
			  R.solve_conj sigma ((s,t)::c) in
			  raise (KB_conjecture_proved sol)
		      with 
			| R.Unsolvable -> 
			    comp_state := { !comp_state with conjectures_to_narrow = new_narrow }
		    end
	      | STOP -> raise Exit
	  end;
	  comp_state := { !comp_state with date = succ !comp_state.date };
	  if (!comp_state.conjecture = None) && 
	    (!comp_state.conjectures_to_narrow = Nothing_to_do) 
	  then 
	    raise (KB_conjecture_proved [])
	with 
	  | KB_conjecture_proved l -> 
	      begin
		if l = []
		then printf "The current conjecture is true and the solution is the identity\n"
		else
		  begin
		    printf "The current conjecture is true and the solution is\n";
		    R.print_sol sigma vars !comp_state.variables_of_conjectures l;
		    Format.print_newline()
		  end;
		if !comp_state.remaining_conj_to_process <> []
		then comp_state := reset_comp_state sigma vars ordering !comp_state
		else raise Exit
	      end
      done;
      assert false
    with
      |	Exit -> 
	  begin
	    if (!comp_state.conjecture <> None) || 
	      (!comp_state.conjectures_to_narrow <> Nothing_to_do) 
	    then
	      begin
		printf "Warning: some conjectures remain@.\n"; 
		false
	      end 
	    else 
	      begin
		printf "All conjectures have been proven@.\n";
		true
	      end
	  end 
end


open Abstract_rewriting;;

module StandardKBCompletion = 
  Make (MakeStandardRewriting (Confluence.User_symbol))

module OrderedKBCompletion 
  (TO : TermOrdering
   with type t =  User_signatures.symbol_id Gen_terms.term) =
  Make (MakeOrderedRewriting (Confluence.User_symbol) (TO))

module ACCompletion =
  Make (MakeACRewriting (Confluence.User_AC_symbol))

module OrderedACCompletion 
  (TO : TermOrdering
   with type t =  User_signatures.symbol_id Gen_terms.term) =
  Make (MakeOrderedACRewriting (Confluence.User_AC_symbol) (TO))

module String_symbol = struct type t = String_signatures.symbol_id end

module WordCompletion = 
  Make (MakeWordRewriting (String_symbol))

module PWordCompletion = Make (MakePWordRewriting)

(*i
Local Variables:
compile-command: "make -C .. opt"
End:
i*)

