open Unix
open String

let string_of_time i =
  let t = ref (int_of_float i) in
  let s = (!t) mod 60 in
  t := (!t)/60;
  let m = (!t) mod 60 in
  t := (!t)/60;
  let h = (!t) mod 24 in
  t := (!t)/24;
  (string_of_int h)^":"^(string_of_int m)^":"^(string_of_int s)

let now () = string_of_time (time ())

let is_digit c = c >= '0' && c <= '9'

let prefix s c = length s > length c && sub s 0 (length c) = c

exception IO_error of file_descr

let addr_of_host host = (gethostbyname host).h_addr_list.(0)

let addr_of_string s =
  if is_digit s.[0] then
    inet_addr_of_string s
  else
    addr_of_host s

let getaddr () = addr_of_host (gethostname ())

let stdin = Pervasives.stdin
and stdout = Pervasives.stdout

class server_socket port n = object (self)
  val fd = Array.create n stderr
  val tin = Array.create n stdin
  val tout = Array.create n stdout
  initializer
  let sock = socket PF_INET SOCK_STREAM 0 in
    bind sock (ADDR_INET (getaddr (), port));
    listen sock n;
    for i = 0 to pred n do
      fd.(i) <- fst (accept sock);
      tin.(i) <- in_channel_of_descr fd.(i);
      tout.(i) <- out_channel_of_descr fd.(i)
    done;
    close sock
  method write_line i s =
    output_string tout.(i) (s^"\n");
    flush tout.(i);
    print_string ("> "^s^"\n");
    flush stdout
  method read_line i =
    let s = (input_line tin.(i))^"\n" in
    print_string ("< "^s);
    flush stdout;
    s
  method read_until p i =
    let s = ref "" and c = ref "" in
      while not (prefix (!c) p) do
	c := self#read_line i;
	s := (!s)^(!c)
      done;
      !s
  method lexing i =
    let l = Lexing.from_channel tin.(i) in
      print_endline "<< ";
      flush stdout;
      l
  method close i =
    close_in tin.(i);
    close_out tout.(i);
    try
      close fd.(i)
    with Unix_error _ -> print_string "Socket already closed by client.";
    flush stdout;
  method close_all =
    for i = 0 to pred n do
      self#close i
    done
end

class client_socket port s = object (self)
  val c =
    let fd = socket PF_INET SOCK_STREAM 0 in
      begin try
	connect fd (ADDR_INET (addr_of_string s, port))
      with
	| Unix_error (_, "connect", _) -> failwith ("unable to connect host "^s^" port "^string_of_int port)
	| Not_found -> failwith ("unable to locate host "^s)
      end;
      (fd, (in_channel_of_descr fd, out_channel_of_descr fd))
  method write_line s =
    output_string (snd (snd c)) (s^"\n");
    flush (snd (snd c));
    print_string ("> "^s^"\n");
    flush stdout
  method read_line =
    let s = (input_line (fst (snd c)))^"\n" in
    print_string ("< "^s);
    flush stdout;
    s
  method read_until p =
    let s = ref "" and c = ref "" in
      while not (prefix (!c) p) do
	c := self#read_line;
	s := (!s)^(!c)
      done;
      !s
  method lexing =
    let l = Lexing.from_channel (fst (snd c)) in
      print_string "<< ";
      print_newline();
      flush stdout;
      l
  method close = close (fst c)
end
