(***************************************************************************

Solving finite domain integer constraints

CiME Project - Dmons research team - LRI - Universit Paris XI

$Id: finite_domains.ml,v 1.30 2002/04/12 15:33:58 marche Exp $

***************************************************************************)

(*
open Numbers;;
*)
open Format;;


let verbose = ref 0;;

(*

  the type of finite domain variables is simply int, which should
  always be non-negative

*)

type fd_var_id = int;;

let compare_fd_var_ids = (-);;

let new_fd_var_id =
  let n = ref (-1)
  in function () -> incr(n);!n;;

module Fd_var_ord =
  struct
    type t = fd_var_id
    let compare = compare_fd_var_ids
  end;;

module Fd_var_map = Ordered_maps.Make(Fd_var_ord);;
module Fd_var_set = Ordered_sets.Make(Fd_var_ord);;
      
let current_var_names = 
  ref (Fd_var_map.empty : (fd_var_id,string) Fd_var_map.t)
;;

exception Found of fd_var_id;;

let find_name name =
  try
    Fd_var_map.fold
      (fun id s () ->
	 if s=name
	 then raise (Found id))
      !current_var_names
      ();
    raise Not_found
  with
      Found x -> x
;;
      
let fd_var_id_of_string name =
  try
    find_name name
  with 
    Not_found -> 
      let x = new_fd_var_id ()
      in
	current_var_names := Fd_var_map.add x name !current_var_names;
	x
;;

let string_of_fd_var_id x = 
  try
    Fd_var_map.find x !current_var_names
  with
      Not_found -> 
	(* if Global_options.oz_output
	   then "A."^(string_of_int (x+1))
	   else
	*)
	"a"^(string_of_int x)
;;

let print_fd_var_id x =
  print_string (string_of_fd_var_id x)
;;



let print_fd_solution sol =
  Fd_var_map.iter
    (fun x n ->
       print_string (string_of_fd_var_id x);
       print_string " = ";
       print_string (Numbers.to_string n);
       print_string ";";
       force_newline())
    sol
;;


(*

  the domains operations. There are two version of division depending
  whether the result is rounded to the floor or the ceiling.

*)

type domain_operation =
    Add | Sub | Mult | Div_floor | Div_ceil;;

(*

  domain expressions are made from classical arithmetical operations
  and min(x) and max(x) for any domain variable x.

*)

type domain_expression =
    Const of Numbers.t
  | Min of fd_var_id
  | Max of fd_var_id
  | Oper of domain_expression * domain_operation * domain_expression
  | Sqrt_floor of domain_expression
  | Sqrt_ceil of domain_expression
  | Power of domain_expression * int
  | Root_floor of int * domain_expression
  | Root_ceil of int * domain_expression
;;


let string_of_op = function
    Add -> " + "
  | Sub -> " - "
  | Mult -> "*"
  | Div_floor -> "/_"
  | Div_ceil -> "/^"
;;

let rec print_domain_expression = function
    Const n -> print_string (Numbers.to_string n)
  | Min(x) -> 
      print_string "min(";
      (print_fd_var_id x);
      print_string ")"
  | Max(x) -> 
      print_string "max(";
      (print_fd_var_id x);
      print_string ")"
  | Oper(e1,op,e2) -> 
      print_string "(";
      print_domain_expression e1;
      print_string (string_of_op op);
      print_domain_expression e2;
      print_string ")"
  | Sqrt_floor(e) -> 
      print_string "sqrt_(";
      print_domain_expression e;
      print_string ")"
  | Sqrt_ceil(e) -> 
      print_string "sqrt^(";
      print_domain_expression e;
      print_string ")"
  | Power(e,n) -> 
      print_string "(";
      print_domain_expression e;
      print_string "^";
      print_int n;
      print_string ")"
  | Root_floor(n,e) -> 
      print_string "root_(";
      print_int n;
      print_string ",";
      print_domain_expression e;
      print_string ")"
  | Root_ceil(n,e) -> 
      print_string "root^(";
      print_int n;
      print_string ",";
      print_domain_expression e;
      print_string ")"
;;








let rec max_var_num_in_expression = function
    Const(n) -> 0
  | Min(x) -> x
  | Max(x) -> x
  | Oper(e1,op,e2) ->
      max (max_var_num_in_expression e1) (max_var_num_in_expression e2)
  | Sqrt_floor(e) -> max_var_num_in_expression e
  | Sqrt_ceil(e) -> max_var_num_in_expression e
  | Power(e,_) -> max_var_num_in_expression e
  | Root_floor(_,e) -> max_var_num_in_expression e
  | Root_ceil(_,e) -> max_var_num_in_expression e
;;

let rec set_of_vars_in_expression = function
    Const(n) -> Fd_var_set.empty
  | Min(x) -> Fd_var_set.add x Fd_var_set.empty
  | Max(x) -> Fd_var_set.add x Fd_var_set.empty
  | Oper(e1,op,e2) ->
      Fd_var_set.union 
	(set_of_vars_in_expression e1) (set_of_vars_in_expression e2)
  | Sqrt_floor(e) -> set_of_vars_in_expression e
  | Sqrt_ceil(e) -> set_of_vars_in_expression e
  | Power(e,_) -> set_of_vars_in_expression e
  | Root_floor(_,e) -> set_of_vars_in_expression e
  | Root_ceil(_,e) -> set_of_vars_in_expression e
;;

let rec add_nb_occ_in_expression e o = 
  match e with
      Const(n) -> o
    | Min(x) -> 
	Fd_var_map.add x (succ (try Fd_var_map.find x o with Not_found -> 0)) o
    | Max(x) -> 
	Fd_var_map.add x (succ (try Fd_var_map.find x o with Not_found -> 0)) o
    | Oper(e1,op,e2) ->
	add_nb_occ_in_expression e1
	  (add_nb_occ_in_expression e2 o)
    | Sqrt_floor(e) -> add_nb_occ_in_expression e o
    | Sqrt_ceil(e) -> add_nb_occ_in_expression e o
    | Power(e,_) -> add_nb_occ_in_expression e o
    | Root_floor(_,e) -> add_nb_occ_in_expression e o
    | Root_ceil(_,e) -> add_nb_occ_in_expression e o
;;

(*

  a finite domain constraint is always of the form x in min..max where
  min and max are two domain expressions

*)

type fd_constraint = 
  {
    var : fd_var_id;
    min : domain_expression;
    max : domain_expression;
  } 
;;

let print_fd_constraint c =
  (print_fd_var_id c.var);
  print_string " in [ ";
  print_domain_expression c.min;
  print_string " , ";
  print_domain_expression c.max;
  print_string " ] "
;;


let rec max_var_num_in_constraints = function
    [] -> 0
  | c::l ->
      max 
	(max c.var (max_var_num_in_expression c.min))
	(max (max_var_num_in_expression c.max) (max_var_num_in_constraints l))
;;

let add_nb_occ_constraint c o =
  add_nb_occ_in_expression c.max
    (add_nb_occ_in_expression c.min
       (Fd_var_map.add c.var 
	  (succ (try Fd_var_map.find c.var o with Not_found -> 0)) o))
;;

type store = 
  { 
    infinity : Numbers.t ;
    intervals : (fd_var_id,(Numbers.t * Numbers.t)) Fd_var_map.t ;
    occurrences_table : (fd_var_id,int) Fd_var_map.t ;
  } 
;;

let make_store bounds constraints =
  { 
    infinity = 
      Fd_var_map.fold 
	(fun var bound i -> Numbers.max i bound)
	bounds
	Numbers.zero;
    intervals =
      Fd_var_map.fold
	(fun var bound i -> Fd_var_map.add var (Numbers.zero,bound) i)
	bounds
	Fd_var_map.empty;
    occurrences_table =
      List.fold_right add_nb_occ_constraint constraints Fd_var_map.empty;
	
  } ;;

let extract_a_solution store =
  Fd_var_map.fold
    (fun var (m,_) sol -> Fd_var_map.add var m sol)
    store.intervals
    Fd_var_map.empty
;;


let print_store store =
  Fd_var_map.iter
    (fun var (mini,maxi) ->
       print_fd_var_id var;
       print_string " in [";
       print_string (Numbers.to_string mini);
       print_string ",";
       print_string (Numbers.to_string maxi);
       print_string "] "
    )
    store.intervals
;;


       

(*

  calculs du min et du max d'une domain-expression par rapport a un store 

*)

exception Undefined;;

let rec compute_min_in_store store = function
    Const(n) -> n
  | Min(x) -> fst (Fd_var_map.find x store.intervals)
  | Max(x) -> raise Undefined
  | Oper(e1,Add,e2) ->
      Numbers.add 
	(compute_min_in_store store e1) (compute_min_in_store store e2)
  | Oper(e1,Sub,e2) ->
      Numbers.max 
	Numbers.zero 
	(Numbers.sub 
	   (compute_min_in_store store e1) (compute_max_in_store store e2))
  | Oper(e1,Mult,e2) ->
      Numbers.mult
	(compute_min_in_store store e1) (compute_min_in_store store e2)
  | Oper(e1,Div_floor,e2) ->
      let v1 = compute_min_in_store store e1
      and v2 = compute_max_in_store store e2
      in 
      if Numbers.is_zero v2
      then raise Undefined
      else 
	Numbers.div_floor v1 v2
  | Oper(e1,Div_ceil,e2) ->
      let v1 = compute_min_in_store store e1
      and v2 = compute_max_in_store store e2
      in 
      if Numbers.is_zero v2
      then raise Undefined
      else Numbers.div_ceil v1 v2
  | Sqrt_floor(e) ->
      Numbers.sqrt_floor (compute_min_in_store store e)
  | Sqrt_ceil(e) ->
      Numbers.sqrt_ceil (compute_min_in_store store e)
  | Power(e,n) ->
      Numbers.power_int (compute_min_in_store store e) n
  | Root_floor(n,e) ->
      Numbers.root_floor n (compute_min_in_store store e)
  | Root_ceil(n,e) ->
      Numbers.root_ceil n (compute_min_in_store store e)

and compute_max_in_store store = function
    Const(n) -> n
  | Min(x) -> raise Undefined
  | Max(x) -> snd (Fd_var_map.find x store.intervals)
  | Oper(e1,Add,e2) ->
      Numbers.add 
	(compute_max_in_store store e1) (compute_max_in_store store e2)
  | Oper(e1,Sub,e2) ->
      Numbers.max
	Numbers.zero
      	(Numbers.sub 
	   (compute_max_in_store store e1) (compute_min_in_store store e2))
  | Oper(e1,Mult,e2) ->
      Numbers.mult
	(compute_max_in_store store e1) (compute_max_in_store store e2)
  | Oper(e1,Div_floor,e2) ->
      let v1 = compute_max_in_store store e1
      and v2 = compute_min_in_store store e2
      in
      if Numbers.is_zero v2
      then raise Undefined
      else Numbers.div_floor v1 v2
  | Oper(e1,Div_ceil,e2) ->
      let v1 = compute_max_in_store store e1
      and v2 = compute_min_in_store store e2
      in
      if Numbers.is_zero v2 
      then raise Undefined
      else Numbers.div_ceil v1 v2
  | Sqrt_floor(e) ->
      Numbers.sqrt_floor (compute_max_in_store store e)
  | Sqrt_ceil(e) ->
      Numbers.sqrt_ceil (compute_max_in_store store e)
  | Power(e,n) ->
      Numbers.power_int (compute_max_in_store store e) n
  | Root_floor(n,e) ->
      Numbers.root_floor n (compute_max_in_store store e)
  | Root_ceil(n,e) ->
      Numbers.root_ceil n (compute_max_in_store store e)
;;

let compute_max_in_store store e = 
try 
  compute_max_in_store store e 
with 
    Undefined -> store.infinity
;;

let compute_min_in_store store e = 
try 
  compute_min_in_store store e 
with 
    Undefined -> Numbers.zero
;;

(*

let compute_max_in_store = 
  Time_profiler.profile2 "compute_max_in_store" compute_max_in_store;;

let compute_min_in_store = 
  Time_profiler.profile2 "compute_min_in_store" compute_min_in_store;;

*)

(*

  graphs 

*)

(*
open Labelled_graphs;;

let make_dependency_graph constr = 
  let squelette = 
    List.fold_left 
      (fun g c -> 
	 let (_,g') = add_vertex g c
	 in g')
      empty_graph 
      constr
  in
  let sets_of_vars =
    fold_left_graph
      (fun m v -> 
	 let c = get_vertex_label squelette v
	 in 
	   VertexMap.add 
	     v 
	     (Fd_var_set.union 
		(set_of_vars_in_expression c.min)
		(set_of_vars_in_expression c.max)) 
	     m)
      VertexMap.empty
      squelette
  in
    fold_left_graph
      (fun g v1 ->
	 let c1 = get_vertex_label squelette v1 
	 in
	   fold_left_graph
	     (fun g v2 ->
		let c2 = get_vertex_label squelette v2 
	 	in
		  if Fd_var_set.mem c1.var (VertexMap.find v2 sets_of_vars)
		  then add_arc g v1 v2
		  else g)
	     g
	     squelette)
      squelette
      squelette	  
;;

let make_dependency_graph = 
  Time_profiler.profile1 "make_dependency_graph" make_dependency_graph ;;

*)

let make_dependency_graph constr = 
  let table = Array.of_list constr
  and dep = ref Fd_var_map.empty
  in
    for i=0 to pred(Array.length table) do
      let c = table.(i)
      in 
      let vars = 
	Fd_var_set.union 
	  (set_of_vars_in_expression c.min)
	  (set_of_vars_in_expression c.max)
      in
	Fd_var_set.iter
	  (fun v ->
	     dep := 
	       try
	         let current = Fd_var_map.find v !dep
		 in 
		   Fd_var_map.add v (Listutils.add i current) !dep
               with
		   Not_found -> Fd_var_map.add v [i] !dep)
	  vars
    done;
    (table,!dep)
;;




(*

  \subsection{Store saturation}

  (store_saturation graph store) returns the store saturated by
  deduction on the constraints given in the graph. The graph must be a
  dependency graph giving for each constraint c the list of
  constraints that depends on c.

  active constraints queue : to be explained

*)

type active_queue =
    {
      size : int;
      queue : int array;
      mutable head : int ;
      mutable tail : int ;
      presence : bool array ;
    }
;;

let create_full n =
  let q = Array.create (succ n) 0
  in
    for i=1 to (pred n) do
      q.(i) <- i
    done;
    {
      size = n;
      queue = q;
      head = 0;
      tail = n;
      presence = Array.create n true
    }
;;

let create_from_list n l =
  let q = Array.create (succ n) 0 
  and p = Array.create n false
  in
  let rec fill i l =
    match l with
	[] -> i
      | x::l ->
	  q.(i) <- x;
	  p.(x) <- true;
	  fill (succ i) l
  in
  let t = fill 0 l
  in
    {
      size = n;
      queue = q;
      head = 0;
      tail = t;
      presence = p;
    }
;;




let pop q =
  let h = q.head 
  in    
    if h = q.tail 
    then raise Not_found
    else
      let r = q.queue.(h)
      in
	q.presence.(r) <- false;
	q.head <- if h = q.size then 0 else succ h;
	r
;;

let push q n =
  if q.presence.(n)
  then ()
  else
    let t = q.tail
    in
      q.presence.(n) <- true;
      q.queue.(t) <- n;
      q.tail <- if t = q.size then 0 else succ t
;;

      

(*
let add_new_active_constraints = 
  Time_profiler.profile3
    "add_new_active_constraints" 
    add_new_active_constraints
;;
*)

let print_spaces n =
  for i=1 to n do
    print_string "   "
  done
;;

exception No_solution;;

let store_saturation graph depth store active_constraints = 
  
  let nb_iterations = ref 0
  in
  let rec saturation store constraints_queue = 
    
    try
      let v = pop constraints_queue
      in 
	(*
	  print_int v;
	  print_string " ";
	*)
	incr nb_iterations;
	let c = (fst graph).(v)
	in
	  if !verbose >= 2
	  then
	    begin
	      print_string "analyse de la contrainte\n";
	      print_fd_constraint c;
	      print_newline();
	    end;
      	    let mini = 
	      compute_min_in_store store c.min 
		
      	    and maxi = 
	      compute_max_in_store store c.max
      	    in
	    let (old_mini,old_maxi) = 
	      Fd_var_map.find c.var store.intervals
		
		
	    in
      	      if Numbers.le mini old_mini && Numbers.ge maxi old_maxi
      	      then
		saturation store constraints_queue
      	      else
  		(* new deduction obtained *)
		let mini = Numbers.max mini old_mini
		and maxi = Numbers.min maxi old_maxi
		in
	    	  if !verbose >= 2
	    	  then
		    begin
		      print_string "new deduction : ";
		      (print_fd_var_id c.var);
		      print_string " in ";
                      print_string (Numbers.to_string mini);
		      print_string "..";
		      print_string (Numbers.to_string maxi);
		      print_newline()
		    end;
	    	  if Numbers.gt mini maxi
	    	  then 
		    begin
		      if !verbose >= 1
		      then
			begin
			  print_spaces depth;
			  print_string "impossibilit, retour arrire";
			  print_newline()
			end;
	      	      raise No_solution
		    end
	    	  else 
		    begin
		      begin
			try
			  let l = Fd_var_map.find c.var (snd graph)
			  in
			    List.iter (push constraints_queue) l
			with Not_found -> ()
		      end;
              	      saturation 
	      		{store with
			   intervals = 
			   Fd_var_map.add c.var (mini,maxi) store.intervals
  		             (* note: [add] removes the old binding of c.var *)
			       
		        }
			constraints_queue
		    end

    with Not_found ->
      if !verbose >= 1
      then
	begin
	  print_string "(";
	  print_int !nb_iterations;
	  print_string " iterations)";	    
	  print_newline();
	end;
	(!nb_iterations > 0,store)
	

  in
    if !verbose >= 1
    then
      begin
	print_spaces depth;
	print_string "saturation...";
	print_newline()
      end;
    saturation store active_constraints



;;

(*i

let store_saturation = 
  Time_profiler.profile3 "store_saturation" store_saturation ;;

i*)

(* 

   \subsection{Heuristics for choosing the variable on which we
   perform reasoning by case}

  (finite_domain_solve_store graph store constraints) finds a solution
  to the constraints in the given store. raises No_solution if no
  solution exists.

  It is performed by first saturating the store with respect to the
  constraint, and then choosing a variable and non-deterministically
  choosing each possible values for this variable. The heuristics for
  choosing the variable is the smallest w.r.t the lex(min,max-min)

  Another good heuristics could be : divide interval by 2

*)

exception Exact;;

(*
let rec find_best_case_variable_method_1 store x best_var best_inter best_min =
  if x >= Array.length store.min_values
  then
    if gt_num best_inter store.infinity
    then raise Exact
    else best_var
  else
    let mini = store.min_values.(x)
    and maxi = store.max_values.(x)
    in
    if eq_num maxi mini
    then find_best_case_variable store (succ x) best_var best_inter best_min
    else
      let diff = sub_num maxi mini
      in 
      match compare_num mini best_min with
	-1 -> 
	  find_best_case_variable store (succ x) x diff mini
      |	0 -> 
	  if lt_num diff best_inter
	  then find_best_case_variable store (succ x) x diff mini
	  else find_best_case_variable store (succ x) best_var best_inter mini
      |	1 -> 
	  find_best_case_variable store (succ x) best_var best_inter best_min
      |	_ -> assert false
;;
*)

let find_best_case_variable_method_2 store =
  let (best_var,best_inter,best_min) =
    Fd_var_map.fold
      (fun x (mini,maxi) ((best_var,best_inter,best_min) as best) ->
	 if Numbers.eq maxi mini
	 then best
	 else
	   let diff = Numbers.sub maxi mini
	   in 
	     match Numbers.compare mini best_min with
		 -1 -> (x,diff,mini)
	       | 0 -> 
		   if Numbers.lt diff best_inter
		   then (x,diff,mini)
		   else best
	       | 1 -> best
	       | _ -> assert false)
      store.intervals
      (0,Numbers.succ store.infinity,Numbers.succ store.infinity)  
  in
    if Numbers.gt best_inter store.infinity
    then raise Exact
    else best_var   
;;

    
let find_best_case_variable_method_3 store =
  let (best_var,nb_occ,best_min) =
    Fd_var_map.fold
      (fun x (mini,maxi) ((best_var,nb_occ,best_min) as best) ->
	 if Numbers.eq maxi mini
	 then best
	 else
	   let nb_occ_x = 
	     try
	       Fd_var_map.find x store.occurrences_table
	     with Not_found -> assert false
	   in
	     match Numbers.compare mini best_min with
		 -1 -> (x,nb_occ_x,mini)
	       | 0 -> 
		   if nb_occ_x > nb_occ
		   then (x,nb_occ_x,mini)
		   else best
	       | 1 -> best
	       | _ -> assert false)
      store.intervals
      (-1,-1,Numbers.succ store.infinity)  
  in
    if best_var < 0
    then raise Exact
    else best_var   
;;

    
let find_best_case_variable_method_4 store =
  let (best_var,best_min,best_diff,best_nb_occ) =
    Fd_var_map.fold
      (fun x (mini,maxi) ((best_var,best_min,best_diff,best_nb_occ) as best) ->
	 if Numbers.eq maxi mini
	 then best
	 else
	   let diff = Numbers.sub maxi mini
	   and nb_occ_x = 
	     try
	       Fd_var_map.find x store.occurrences_table
	     with Not_found -> assert false
	   in
	     match Numbers.compare mini best_min with
		 -1 -> (x,mini,diff,nb_occ_x)
	       | 0 -> 
		   if Numbers.lt diff best_diff
		   then (x,mini,diff,nb_occ_x)
		   else
		     if nb_occ_x > best_nb_occ
		     then (x,mini,diff,nb_occ_x)
		     else best
	       | 1 -> best
	       | _ -> assert false)
      store.intervals
      (-1,Numbers.succ store.infinity,Numbers.succ store.infinity,0)  
  in
    if best_var < 0
    then raise Exact
    else best_var   
;;


(*

  Sort all variables in best order 

*)


let variables_best_order_method_4 store =
(*
  let default = (-1,Numbers.succ store.infinity,Numbers.succ store.infinity,0)  
  in
*)
  let l =
    Fd_var_map.fold
      (fun x (mini,maxi) accu ->
	 if Numbers.eq maxi mini
	 then accu
	 else
	   let diff = Numbers.sub maxi mini
	   and nb_occ_x = 
	     try
	       Fd_var_map.find x store.occurrences_table
	     with Not_found -> assert false
	   in
           (x,mini,diff,nb_occ_x)::accu)
      store.intervals
      []
  in
  Sort.list
    (fun (_,min1,diff1,nb_occ1) (_,min2,diff2,nb_occ2) ->
       let c = 
	 match Numbers.compare min1 min2 with
	     0 -> 
	       (match Numbers.compare diff1 diff2 with
		    0 -> nb_occ1 - nb_occ2
		  | x -> x)
	   | x -> x
       in c <=0)
    l
;;


(*i---------------------------------------------------------i*)
    
(*

\subsection{Solving in store}

\subsubsection{First method}

*)
    

let rec old_finite_domain_solve_store graph depth store active_constraints =

  (* heuristic 1 *)
  let rec iter_interval store x mini maxi =
    if Numbers.gt mini maxi
    then 
      begin
	if !verbose >= 1
	then
  	  begin
	    print_string "Plus de cas de l'intervalle, retour arriere";
	    print_newline();
	  end;
    	raise No_solution
      end
    else
      if !verbose >= 1
      then
  	begin
	  print_string "cas ";
          (print_fd_var_id x);
	  print_string " = ";
	  print_string (Numbers.to_string mini);
	  print_newline();
	end;
    
      let new_store = 
	{ store with
	  intervals = Fd_var_map.add x (mini,mini) store.intervals
	}
      in
      try
	old_finite_domain_solve_store 
	  graph 
	  (succ depth)
	  new_store 
	  (create_from_list 
	     active_constraints.size
	     (Fd_var_map.find x (snd graph)))

      with No_solution ->
	iter_interval store x (Numbers.succ mini) maxi

  (* heuristic 2 *)
  and split_interval store x mini maxi =
    let middle = Numbers.quo (Numbers.add mini maxi) Numbers.two
    in
    let new_store = 
      { store with
	intervals = Fd_var_map.add x (mini,middle) store.intervals;
      }
    in
      try 
      	old_finite_domain_solve_store 
	  graph 
	  (succ depth)
	  new_store 
	  (create_from_list 
	     active_constraints.size
	     (Fd_var_map.find x (snd graph)))
      with No_solution ->
      	let new_store = 
	  { store with
	    intervals = 
	     Fd_var_map.add x (Numbers.succ middle,maxi) store.intervals;
	  }
      	in
	  old_finite_domain_solve_store 
	    graph 
	    (succ depth)
	    new_store 
	    (create_from_list 
	       active_constraints.size
	       (Fd_var_map.find x (snd graph)))

  (* heuristic 3 *)
  and first_and_rest store x mini maxi =
    let new_store = 
      { store with
	intervals = 
	 Fd_var_map.add x (mini,mini) store.intervals;
      }
    in
    try
      if !verbose >= 1
      then
  	begin
	  print_spaces depth;
	  print_string "cas ";
          print_fd_var_id x;
	  print_string " = ";
	  print_string (Numbers.to_string mini);
	  print_newline()
	end;
      old_finite_domain_solve_store 
	graph 
	(succ depth)
	new_store 
	(create_from_list 
	   active_constraints.size
	   (try Fd_var_map.find x (snd graph) with Not_found -> []))

    with No_solution ->
      begin
	let new_store =
	  { store with
	    intervals = 
	     Fd_var_map.add x (Numbers.succ mini,maxi) store.intervals;
	  }
	in
	  if !verbose >= 1
	  then
  	    begin
	      print_spaces depth;
	      print_string "cas ";
              print_fd_var_id x;
	      print_string " > ";
	      print_string (Numbers.to_string mini);
	      print_newline()
	    end;
          old_finite_domain_solve_store 
	    graph 
	    (succ depth)
	    new_store 
	    (create_from_list 
	       active_constraints.size
	       (try Fd_var_map.find x (snd graph) with Not_found -> []))
      end
  
  in
    
  begin
    let 
      _,new_store = 
      store_saturation graph depth store active_constraints 
	
    in
    try
      let x = find_best_case_variable_method_4 new_store 
      in
      let (mini,maxi) = Fd_var_map.find x new_store.intervals
      in
	if !verbose >= 1
	then
	  begin
	    print_spaces depth;
	    print_string "par cas sur ";
	    print_fd_var_id x;
	    print_string " entre ";
	    print_string (Numbers.to_string mini);
	    print_string " et ";
	    print_string (Numbers.to_string maxi);
	    print_newline()
	  end;
	  
      (* heuristic 1 *)
(*
  iter_interval new_store x mini maxi;
*)
	
      (* heuristic 2 *)
(*
  split_interval new_store x mini maxi;
*)
      (* heuristic 3 *)
(**)
	first_and_rest new_store x mini maxi
(**)

    with Exact ->
      extract_a_solution new_store

  end
      
;;

(*

  \subsubsection{attempt for another method}

  searching the best case reasoning variable more clever ?

*)

let store_diameter s =
  Fd_var_map.fold
    (fun var (mini,maxi) accu -> Numbers.add accu (Numbers.sub maxi mini))
    s.intervals
    Numbers.zero
;;


let select_best_var table = 
  let (x,d) =
    Fd_var_map.fold 
      (fun var (s1,s2) ((best_var,best_diam) as best) ->
	 let d = Numbers.add (store_diameter s1) (store_diameter s2)
	 in 
	   match best_diam with
	       None -> (var,Some d)
	     | Some d' ->
		 if Numbers.lt d d' then (var,Some d) else best)
      table
      (0,None)
  in
    match d with
	None -> raise Exact
      | Some _ -> x
;;



exception Continue of store;;

let rec case_reasoning graph store =
  try
    let case_store_table = 
      Fd_var_map.fold 
	(fun x (mini,maxi) accu ->
	   if Numbers.eq mini maxi
	   then accu
	   else
	     try
	       let _,s1 = 
		 store_saturation graph 0
		   {store with 
		      intervals = Fd_var_map.add x (mini,mini) store.intervals}
		   (create_from_list 
		      (Array.length (fst graph))
		      (try Fd_var_map.find x (snd graph) with Not_found -> []))
	       in
		 try
		   let _,s2 = 
		     store_saturation graph 0
		       {store with 
			  intervals = Fd_var_map.add x (Numbers.succ mini,maxi) store.intervals} 
		       (create_from_list 
			  (Array.length (fst graph))
			  (try Fd_var_map.find x (snd graph) with Not_found -> []))
		   in
		     Fd_var_map.add x (s1,s2) accu
		 with
		     No_solution -> raise (Continue s1)
	     with
		 No_solution -> 
		   let _,s2 = 
		     store_saturation graph 0
		       {store with 
			  intervals = Fd_var_map.add x (Numbers.succ mini,maxi) store.intervals} 
		       (create_from_list 
			  (Array.length (fst graph))
			  (try Fd_var_map.find x (snd graph) with Not_found -> []))
		   in
		     raise (Continue s2)
	)
	store.intervals
	Fd_var_map.empty
    in
    let x = select_best_var case_store_table
    in
    let (s1,s2) = Fd_var_map.find x case_store_table
    in
      try
	case_reasoning graph s1 
      with 
	  No_solution -> 
	    case_reasoning graph s2
		  
    
  with
      Continue store -> case_reasoning graph store 
    | Exact -> extract_a_solution store
;;



(*

  \subsubsection{Solving method 3}

  Use a "credit" to bound the search.

*)

exception Credit_expired;;

let rec finite_domain_solve_credit_store_3 graph depth credit store =

  let first_and_rest credit store x mini maxi =

    try
      let new_store = 
	{ store with
	    intervals = Fd_var_map.add x (mini,mini) store.intervals;
        }
      in
      if !verbose >= 1
      then
  	begin
	  print_spaces depth;
	  print_string "cas ";
          print_fd_var_id x;
	  print_string " = ";
	  print_string (Numbers.to_string mini);
	  print_newline();
	end;
      let has_sat,satur_store =
	store_saturation graph depth new_store 
	  (create_from_list 
	     (Array.length (fst graph))
	     (try Fd_var_map.find x (snd graph) with Not_found -> []))
      in
      if !verbose >= 2
      then
	begin
	  print_spaces depth;
	  print_string "store = ";
	  print_store satur_store;
	  print_newline()
	end;
      finite_domain_solve_credit_store_3
	graph 
	(succ depth)
	(if true (* has_sat *) then credit else pred credit)
	satur_store
	
    with No_solution ->
      begin
	let new_store =
	  { store with
	      intervals = 
	        Fd_var_map.add x (Numbers.succ mini,maxi) store.intervals;
          } 
	in
	if !verbose >= 1
	then
  	  begin
	    print_spaces depth;
	    print_string "cas ";
            print_fd_var_id x;
	    print_string " > ";
	    print_string (Numbers.to_string mini);
	    print_newline();
	  end;
	let has_sat,satur_store =
	  store_saturation graph depth new_store 
	    (create_from_list 
	       (Array.length (fst graph))
	       (try Fd_var_map.find x (snd graph) with Not_found -> []))
	in
	if !verbose >= 2
	then
	  begin
	    print_spaces depth;
	    print_string "store = ";
	    print_store satur_store;
	    print_newline()
	  end;
	
	finite_domain_solve_credit_store_3
	  graph 
	  (succ depth)
	  (if false (* has_sat *) then credit else pred credit)
	  satur_store
      end
  
  in

  let rec case_reasoning new_store l credit =
    if credit <= 1 then raise Credit_expired;
    match l with
	[] -> raise Credit_expired
      | (x,_,_,_)::l ->
	  try
	    let (mini,maxi) = Fd_var_map.find x new_store.intervals
	    in
	    if !verbose >= 1
	    then
	      begin
		print_spaces depth;
		print_string "par cas sur ";
		print_fd_var_id x;
		print_string " entre ";
		print_string (Numbers.to_string mini);
		print_string " et ";
		print_string (Numbers.to_string maxi);
		print_string " (credit ";
		print_int credit;
		print_string ")";
		print_newline()
	      end;
	    
	    first_and_rest credit new_store x mini maxi
	  with
	  Credit_expired -> 
	    if !verbose >= 1
	    then
	      begin
		print_spaces depth;
		print_string "credit expir pour le cas sur ";
		print_fd_var_id x;
		print_newline()
	      end;
	    case_reasoning new_store l credit

  in
    
  begin
    if credit <= 0 then raise Credit_expired;

    let sorted_var_list = variables_best_order_method_4 store 
    in
    if sorted_var_list=[]
    then extract_a_solution store
    else case_reasoning store sorted_var_list credit

  end
 
;;

let finite_domain_solve_store_3 graph store =
  let rec increasing_credit credit =
    try
      let _,new_store = 
	store_saturation graph 0 store  
	  (create_full (Array.length (fst graph)))
      in
      finite_domain_solve_credit_store_3 graph 0 credit new_store 
    with Credit_expired ->
      increasing_credit (credit+credit)
  in 
  increasing_credit 2
;;
  
(*i---------------------------------------------------------i*)

(*

  \subsection{Main functions}

  Solving a constraint for given store and dependencies between
  variables and constraints.

  We call here the method currently in use.

*)


let new_finite_domain_solve_store graph store =
  let _,new_store = store_saturation graph 0 store 
		    (create_full (Array.length (fst graph)))
  in
    case_reasoning graph new_store  
;;

let finite_domain_solve_store graph store =
  old_finite_domain_solve_store 
    graph 0 store (create_full (Array.length (fst graph)))
;;

(*i
let rec finite_domain_solve_store graph store =
  finite_domain_solve_store_3 graph store
;;

i*)

(*i

let finite_domain_solve_store = 
  Time_profiler.profile2 
    "finite_domain_solve_store" (* new_ *) finite_domain_solve_store 
;;
i*)


(*

  \subsubsection{The main function}


  [(finite_domain_solve M C)] returns a solution of the constraints [C] in
  the domain $[0..M]$ for each variable. Raises exception [No_solution] if
  no solution exists.

  We must compute the initial store and the dependencies between
  variables and constraints.

*)

let finite_domain_solve bounds constraints =
  let store = make_store bounds constraints
  and graph = make_dependency_graph constraints
  in
(*i
    begin
      let c=open_out "graph"
      in 
	output_value c graph;
	close_out c
    end;
i*)
(*i
    begin
	Fd_var_map.iter
	(fun x l ->
	  print_fd_var_id x;
	  print_string " :";
	  List.iter 
	    (fun x -> print_string " ";print_int x)
	    l;
	  print_newline())
	(snd graph)
    end;
i*)
  finite_domain_solve_store graph store 

;;

(*i
let finite_domain_solve = 
    Time_profiler.profile2 "finite_domain_solve" finite_domain_solve 
  ;;
i*)


