(***************************************************************************

Finite domains

CiME Project - Dmons research team - LRI - Universit Paris XI

$Id: finite_domains.mli,v 1.17 2002/04/12 15:33:58 marche Exp $

***************************************************************************)


(*

This module defines finite domains constraints over integers, and allows to solve such constraints for given intervals to search for variables.


\subsection{Finite domain variables}

*)

val verbose : int ref;;

type fd_var_id;;

val string_of_fd_var_id : fd_var_id -> string;;

val print_fd_var_id : fd_var_id -> unit;;

val new_fd_var_id : unit -> fd_var_id;;

val fd_var_id_of_string : string -> fd_var_id;;

module Fd_var_ord :
  sig
    type t = fd_var_id
    val compare : fd_var_id -> fd_var_id -> int
  end;;


module Fd_var_map : Ordered_maps.OrderedMap with type 'a key = fd_var_id ;;

module Fd_var_set : Ordered_sets.OrderedSet with type 'a elt = fd_var_id ;;


(*

  \subsection{The domains expressions}


  Domain expressions are made from classical arithmetical operations,
  and $\min(x)$ and $\max(x)$ for any domain variable $x$.

  There are two version of division depending whether the result is
  rounded to the floor or the ceiling. The same for square root and
  arbitrary nth root.

*)

type domain_operation =
    Add | Sub | Mult | Div_floor | Div_ceil;;

type domain_expression =
    Const of Numbers.t
  | Min of fd_var_id
  | Max of fd_var_id
  | Oper of domain_expression * domain_operation * domain_expression
  | Sqrt_floor of domain_expression
  | Sqrt_ceil of domain_expression
  | Power of domain_expression * int
  | Root_floor of int * domain_expression
  | Root_ceil of int * domain_expression
;;
    
(*

  \subsection{Finite domain constraints}

  a finite domain constraint is always of the form $x \in [min..max]$
  where $min$ and $max$ are two domain expressions

*)

type fd_constraint = 
  {
    var : fd_var_id;
    min : domain_expression;
    max : domain_expression;
  } 
;;

val print_fd_constraint : fd_constraint -> unit;;

(*

  \subsection{Solving constraints}

  [(finite_domain_solve M C)] returns a solution of the constraints
  [C] in the domain [0..M(x)] for each variable [x]. Raises exception
  [Finite_domains.No_solution] if no solution exists.

*)

exception No_solution;;

val finite_domain_solve : 
  (fd_var_id,Numbers.t) Fd_var_map.t -> fd_constraint list 
    -> (fd_var_id,Numbers.t) Fd_var_map.t;;

val print_fd_solution : (fd_var_id,Numbers.t) Fd_var_map.t -> unit;;

