(***************************************************************************

Solving non linear Diophantine constraints

CiME Project - Dmons research team - LRI - Universit Paris XI

$Id: non_linear_solving.ml,v 1.31 2002/04/12 16:01:52 marche Exp $

***************************************************************************)

open Finite_domains;;
open Format;;

let verbose = ref 0;;


module NumRing =
  struct
    type coef = Numbers.t
    let zero = Numbers.zero
    let one = Numbers.one
    let is_null = Numbers.is_zero
    let add = Numbers.add
    let minus = Numbers.minus
    let sub = Numbers.sub
    let mult = Numbers.mult
  end
;;

module Fd_polynomials = 
  Polynomials.Make (NumRing) (Fd_var_ord) (Fd_var_set) (Fd_var_map);;

type constraint_op = Nul | Pos_or_nul ;;

let print_constraint_op = function
    Nul -> 
	(* if Global_options.oz_output
	   then print_string " =: 0"
	   else
	*)
      print_string " = 0"
  | Pos_or_nul  -> 
	(* if Global_options.oz_output
	   then print_string " >=: 0"
	   else
	*)
      print_string " >= 0"
;;

type non_linear_constraint = 
    {
      expr : Fd_polynomials.poly;
      cond : constraint_op;
    };;

let vars_of_constraint c = Fd_polynomials.set_of_vars c.expr;;

let vars_of_constraints c = 
    List.fold_left
      (fun e c ->
	 Fd_var_set.union (vars_of_constraint c) e)
      Fd_var_set.empty
      c

let size c = List.fold_left (fun n c -> n + 1 + Fd_polynomials.size c.expr) 0 c;;

let print_polynomial p =
  Fd_polynomials.print_polynomial 
      (fun x -> Format.print_string (Numbers.to_string x))
      (fun x -> Format.print_string (string_of_fd_var_id x))
      p
;;

let print_constraint c = 
  begin
    print_polynomial c.expr;
    print_constraint_op c.cond;
  end
;;

let print_non_linear_constraints c =
  print_string "{ ";
  open_hovbox 0;  
  List.iter 
    (fun c ->
       print_constraint c;
       print_string ";";
       force_newline())
    c;
  close_box ();
  print_string "}";
  print_break 1 0;
  print_string "(";
  print_int (List.length c);
  print_string " inequalitie(s) over ";
  print_int (Fd_var_set.cardinal (vars_of_constraints c));
  print_string " variable(s).)";
  force_newline()
;;



let rec from_abstract_expr e =
  match e with
    | Abstract_constraint.Var(x) -> 
	Fd_polynomials.var (Finite_domains.fd_var_id_of_string x) 
    | Abstract_constraint.Cte(n) -> 
	Fd_polynomials.cte n
    | Abstract_constraint.Plus(e1,e2) -> 
	Fd_polynomials.add (from_abstract_expr e1) (from_abstract_expr e2) 
    | Abstract_constraint.Mult(e1,e2) -> 
	Fd_polynomials.mult (from_abstract_expr e1) (from_abstract_expr e2) 
    | Abstract_constraint.Minus(x) -> 
	Fd_polynomials.minus (from_abstract_expr x) 
    | Abstract_constraint.Sub(e1,e2) -> 
	Fd_polynomials.add (from_abstract_expr e1)
	  (Fd_polynomials.minus (from_abstract_expr e2))
    | Abstract_constraint.Quotient(_) -> assert false
;;




let rec from_abstract_formula f = 
  match f with 
    | Abstract_constraint.True -> []
    | Abstract_constraint.Comp(e1,op,e2) ->
	let e1 = from_abstract_expr e1
	and e2 = from_abstract_expr e2
	in
	[
	  match op with
	    | "=" ->
		{ 
		  expr = Fd_polynomials.sub e1 e2; 
		  cond = Nul 
		}
	    | ">=" ->
		{ 
		  expr = Fd_polynomials.sub e1 e2; 
		  cond = Pos_or_nul 
		}
	    | "<=" ->
		{ 
		  expr = Fd_polynomials.sub e2 e1; 
		  cond = Pos_or_nul 
		}
	    | ">" ->
		{ 
		  expr = Fd_polynomials.add 
			   (Fd_polynomials.sub e1 e2)
			   (Fd_polynomials.cte Numbers.minus_one); 
		  cond = Pos_or_nul 
		} 
	    | "<" ->
		{ expr = Fd_polynomials.add 
			   (Fd_polynomials.sub e2 e1)
			   (Fd_polynomials.cte Numbers.minus_one); 
		  cond = Pos_or_nul 
		}
	    | "<>" -> invalid_arg "<> in from_abstract_formula"
	    | _ -> assert false
	]
    | Abstract_constraint.And(l) ->
	List.fold_left
	  (fun accu x -> (from_abstract_formula x)@accu)
	  [] l
    | _ -> invalid_arg "from_abstract_formula"
;;




let find_in_constraint cond c = 
  Fd_polynomials.find cond c.expr
;;

let replace_in_constraint cond c =
    {
      expr = Fd_polynomials.map cond c.expr;
      cond = c.cond;
    }
;;
      
      
let rec find_in_constraints cond = function
    [] -> raise Not_found
  | c::l ->
      try 
	find_in_constraint cond c 
      with Not_found ->
	find_in_constraints cond l
;;


let replace_in_constraints cond c = 
  List.map (replace_in_constraint cond) c
;;



(*


 trivial simplifications of non-linear constraints

*)

let common_coef_sign p =
  let s = ref 0
  in
    Fd_polynomials.iter
      (fun m ->
	 let c = Fd_polynomials.monomial_coef m
	 in
	   match !s with
	       0 -> s := Numbers.compare c Numbers.zero 
	     | 1 ->
		 if Numbers.lt c Numbers.zero
		 then raise Not_found
	     | -1 ->
		 if Numbers.gt c Numbers.zero
		 then raise Not_found
	     | _ -> assert false)
      p;
    !s
;;
   
	 
let each_monomial_nul p =
  if Numbers.is_zero (Fd_polynomials.constant_coef p)
  then
    Fd_polynomials.fold
      (fun m c ->
	 {expr =
	    Fd_var_map.fold
	      (fun x exp r -> Fd_polynomials.mult (Fd_polynomials.var x) r)
	      (Fd_polynomials.monomial_vars m)
	      (Fd_polynomials.cte Numbers.one);
	  cond=Nul}::c)
      p
      []
  else
    raise No_solution
;;

let non_linear_simplify c =
  try
    match common_coef_sign c.expr
    with
	0 -> []
      | 1 -> if c.cond=Nul then each_monomial_nul c.expr else []
      | -1 -> each_monomial_nul c.expr	    
      | _ -> assert false
  with
    Not_found -> [c]
;;

(*

  abstraction method 0 : translation without linearization

*)

let const_zero = Const Numbers.zero;;
let const_un = Const Numbers.one;;

let rec make_sum = function
    [] -> const_zero
  | [x] -> x
  | x::l ->
      if x=const_zero
      then make_sum l
      else
      	let r=make_sum l
      	in
	  if r=const_zero
	  then x
	  else Oper(x,Add,r)
;;



let rec make_prod = function
    [] -> const_un
  | [x] -> x
  | x::l ->
      if x=const_un
      then make_prod l
      else
	let r=make_prod l
	in 
	  if r=const_un
	  then x
	  else Oper(x,Mult,r)
;;

let sub e1 e2 =
  if e1 = const_zero
  then const_zero
  else
    if e2 = const_zero
    then e1
    else Oper(e1,Sub,e2)
;;

let div_ceil e1 e2 =
  if e2 = const_un
  then e1
  else Oper(e1,Div_ceil,e2)
;;

let div_floor e1 e2 =
  if e2 = const_un
  then e1
  else Oper(e1,Div_floor,e2)
;;


(*

  (split_poly m p) where p=m+\sum_{i\in I} a_i m_i returns a pair (A,B) where 
    A=\sum_{i\in I,a_i>0} a_i max(m_i) 
  and 
    B=\sum_{i\in I,a_i<0} (-a_i)min(m_i)  

*)

let prod_max m coef = 
  let r=
    Fd_var_map.fold
      (fun x exp p ->
	 if exp=1
	 then Max(x)::p
	 else Power(Max(x),exp)::p
      )
      m
      [Const(coef)]
  in make_prod r
;;

let prod_min m coef = 
  let r =
    Fd_var_map.fold
      (fun x exp p ->
	 if exp=1
	 then Min(x)::p
	 else Power(Min(x),exp)::p
      )
      m
      [Const(coef)]
  in make_prod r
;;

  
let split_poly m_exclude p = 
  let a,b =
    Fd_polynomials.fold
      (fun m ((posmax,negmin) as r) ->
	 if m=m_exclude
	 then r
	 else 
	   let coef = Fd_polynomials.monomial_coef m
	   and vars = Fd_polynomials.monomial_vars m
	   in
	     if Numbers.gt coef Numbers.zero
	     then 
	       (prod_max vars coef)::posmax,negmin
	     else 
	       posmax,(prod_min vars (Numbers.minus coef))::negmin
      )
      p
      ([],[])
  in (make_sum a,make_sum b)
;;

let split_poly_opposite m_exclude p = 
  let a,b =
    Fd_polynomials.fold
      (fun m ((posmin,negmax) as r) ->
	 if m=m_exclude
	 then r
	 else 
	   let coef = Fd_polynomials.monomial_coef m
	   and vars = Fd_polynomials.monomial_vars m
	   in
	     if Numbers.gt coef Numbers.zero
	     then 
	       (prod_min vars coef)::posmin,negmax
	     else 
	       posmin,(prod_max vars (Numbers.minus coef))::negmax
      )
      p
      ([],[])
  in (make_sum a,make_sum b)
;;

let finite_domain_constraint_of_monomial_positive bounds coef vars poly =
  Fd_var_map.fold 
    (fun x exp c ->
       let r=
	 Fd_var_map.fold
	   (fun x2 exp2 p ->
	      if x2=x
	      then p
	      else
	      	if exp2=1
	      	then Max(x2)::p
	      	else Power(Max(x2),exp2)::p
	   )
	   vars
	   [Const(coef)]
       in 
       let r' = div_ceil poly (make_prod r)
       in
       let r'' =
	 if exp=1
	 then 
	   r'
	 else
	   if exp=2
	   then Sqrt_ceil(r')
	   else Root_ceil(exp,r')
       in 
	 {
	   var = x;
	   min = r'';
	   max = Const(Fd_var_map.find x bounds);
	 }::c       
    )
    vars
    []
;;

let finite_domain_constraint_of_monomial_negative bounds coef vars poly =
  Fd_var_map.fold 
    (fun x exp c ->
       let r=
	 Fd_var_map.fold
	   (fun x2 exp2 p ->
	      if x2=x
	      then p
	      else
	      	if exp2=1
	      	then Min(x2)::p
	      	else Power(Min(x2),exp2)::p
	   )
	   vars
	   [Const(coef)]
       in 
       let r' = div_floor poly (make_prod r)
       in
       let r'' =
	 if exp=1
	 then 
	   r'
	 else
	   if exp=2
	   then Sqrt_floor(r')
	   else Root_floor(exp,r')
       in 
	 {
	   var = x;
	   min = const_zero;
	   max = r'';
	 }::c       
    )
    vars
    []
;;

let finite_domain_constraint_of_monomial_null 
  coef_was_pos bounds coef vars polymin polymax =
  Fd_var_map.fold 
    (fun x exp c ->
       let rmin,rmax =
	 Fd_var_map.fold
	   (fun x2 exp2 ((pmin,pmax) as p) ->
	      if x2=x
	      then p
	      else
	      	if exp2=1
	      	then Min(x2)::pmin,Max(x2)::pmax
	      	else Power(Min(x2),exp2)::pmin,Power(Max(x2),exp2)::pmax
	   )
	   vars
	   ([Const(coef)],[Const(coef)])
       in 
       let rmin' = div_ceil polymin (make_prod rmax)
       and rmax' = div_floor polymax (make_prod rmin)
       in
       let rmin'',rmax'' =
	 if exp=1
	 then 
	   rmin',rmax'
	 else
	   if exp=2
	   then 
	     if coef_was_pos 
	     then Sqrt_ceil(rmin'),Sqrt_floor(rmax')
	     else Sqrt_floor(rmin'),Sqrt_ceil(rmax')
	   else 
	     if coef_was_pos
	     then Root_ceil(exp,rmin'),Root_floor(exp,rmax')
	     else Root_floor(exp,rmin'),Root_ceil(exp,rmax')
       in 
	 (if coef_was_pos
	  then
	    {
	      var = x;
	      min = rmin'';
	      max = rmax'';
	    }
	  else
	    {
	      var = x;
	      min = rmax'';
	      max = rmin'';
	    })
	 ::c       
    )
    vars
    []
;;

    
let finite_domain_constraint_of_constraint_positive bounds poly =
  Fd_polynomials.fold
    (fun m c ->
       let coef = Fd_polynomials.monomial_coef m
       and vars = Fd_polynomials.monomial_vars m
       in
       let posmax,negmin = split_poly m poly
       in
       let new_c =
	 if Numbers.gt coef Numbers.zero
	 then
	   finite_domain_constraint_of_monomial_positive 
	     bounds coef vars (sub negmin posmax)
	 else
	   finite_domain_constraint_of_monomial_negative 
	     bounds (Numbers.minus coef) vars (sub posmax negmin)
       in 
	 new_c @ c
    )
    poly
    []
;;

let finite_domain_constraint_of_constraint_null bounds poly =
  Fd_polynomials.fold
    (fun m c ->
       let coef = Fd_polynomials.monomial_coef m
       and vars = Fd_polynomials.monomial_vars m
       in
       let posmax,negmin = split_poly m poly
       and posmin,negmax = split_poly_opposite m poly
       in
       let new_c =
	 if Numbers.gt coef Numbers.zero
	 then
	   finite_domain_constraint_of_monomial_null true
	     bounds coef vars (sub negmin posmax) (sub negmax posmin)
	 else
	   finite_domain_constraint_of_monomial_null false
	     bounds (Numbers.minus coef) vars (sub posmax negmin) (sub posmin negmax)
       in 
	 new_c @ c
    )
    poly
    []
;;

let finite_domain_constraint_of_constraint bounds c =
  if c.cond = Nul
  then
    finite_domain_constraint_of_constraint_null bounds c.expr
  else
    finite_domain_constraint_of_constraint_positive bounds c.expr
;;


let simple_translate_constraints bounds c = 
    print_string "size of constraint after (no) abstraction : ";
    print_int (size c);
    print_newline();
  let fd_constraints =
    Listutils.flat_map
      (finite_domain_constraint_of_constraint bounds)
      c
  in
    (fd_constraints,bounds)
;;

let simple_translate_constraints = 
  Time_profiler.profile2
    "simple_translate_constraints" simple_translate_constraints;;

(*

  Abstraction method 1 : linearization

*)


(*

 a simple constraint has the form 

 1) \sigma a_ix_i +c # 0 where # is either = or >=
 2) x^2 = y
 3) xy = z

*)

type simple_constraint =
    Sum of (fd_var_id * Numbers.t) list * Numbers.t * constraint_op
  | Square of fd_var_id * fd_var_id
  | Product of fd_var_id * fd_var_id *fd_var_id 
;;


(*

  square abstraction

*)


exception Found of Fd_polynomials.variable;;

let find_a_square_in_monomial m =
  try
    Fd_var_map.iter
      (fun x exp ->
      	if exp>=2
      	then raise (Found x))
      (Fd_polynomials.monomial_vars m);
    raise Not_found      
  with
    Found x -> x;;

(* 

remplace toute occurence de x^{2n} par z^n et x^{2n+1} par xz^{n}

*)

let replace_square_in_monomial x z m =
  try
    let exp = Fd_var_map.find x (Fd_polynomials.monomial_vars m)
    in
    if exp < 2
    then m
    else
      let m' =
	Fd_var_map.add z (exp/2) (Fd_polynomials.monomial_vars m)
      in
      let m'' =
	if exp mod 2 = 0
	then Fd_var_map.remove x m'
	else Fd_var_map.add x 1 m' (* note: [add] removes the old binding *)
      in
      Fd_polynomials.make_monomial (Fd_polynomials.monomial_coef m) m''

  with Not_found -> m;;



let square_abstraction constraints simple_constraints bounds = 
  let new_constraints = ref constraints
  and new_simple_constraints = ref simple_constraints
  and new_bounds = ref bounds 
  in
  try
    while true do
      let x = find_in_constraints find_a_square_in_monomial !new_constraints
      in 
      let z = new_fd_var_id()
      and x_bound = Fd_var_map.find x !new_bounds
      in
      new_constraints :=
	 replace_in_constraints (replace_square_in_monomial x z) !new_constraints; 
      new_simple_constraints := (Square (x,z))::!new_simple_constraints;
      new_bounds := Fd_var_map.add z (Numbers.mult x_bound x_bound) !new_bounds 
    done;
    !new_constraints,!new_simple_constraints,!new_bounds
    
  with Not_found ->
    !new_constraints,!new_simple_constraints,!new_bounds
;;



(*

  product abstraction

*)

exception Found of fd_var_id * fd_var_id;;

let find_a_product_in_monomial m =
  try
    let  _ = 
      Fd_var_map.fold
	(fun x exp prec ->
      	   if exp>=1
      	   then 
	     match prec with
		 [] -> [x]
	       | [y] -> raise (Found (x,y))
	       | _ -> failwith "bug dans find_a_product_in_monomial"
	   else prec)
	(Fd_polynomials.monomial_vars m)
	[]
    in
      raise Not_found      
  with
    Found (x,y) -> (x,y);;

(* 

remplace toute occurence de x^ny^m avec n>=1 et m>=1 par z

*)

let replace_product_in_monomial x y z m =
  try
    let expx = Fd_var_map.find x (Fd_polynomials.monomial_vars m)
    in
      if expx < 1
      then m
      else
      	let expy = Fd_var_map.find y (Fd_polynomials.monomial_vars m)
      	in
	  if expy < 1
	  then m
	  else
	    let m' =
	      (Fd_var_map.remove x 
		 (Fd_var_map.remove y (Fd_polynomials.monomial_vars m)))
	    in
	      Fd_polynomials.make_monomial 
		(Fd_polynomials.monomial_coef m)
		(Fd_var_map.add z 1 m')
	      
  with Not_found -> m;;
  


let product_abstraction constraints simple_constraints bounds = 
  let new_constraints = ref constraints
  and new_simple_constraints = ref simple_constraints
  and new_bounds = ref bounds 
  in
  try
    while true do    
      let (x,y) = find_in_constraints find_a_product_in_monomial !new_constraints
      in 
      let z = new_fd_var_id()
      and x_bound = Fd_var_map.find x !new_bounds
      and y_bound = Fd_var_map.find y !new_bounds
      in
      new_constraints :=
	 replace_in_constraints (replace_product_in_monomial x y z) !new_constraints;
      new_simple_constraints := (Product(x,y,z))::!new_simple_constraints;
      new_bounds := Fd_var_map.add z (Numbers.mult x_bound y_bound) !new_bounds
    done;
    !new_constraints,!new_simple_constraints,!new_bounds

  with Not_found ->
    !new_constraints,!new_simple_constraints,!new_bounds
;;

(*

Tests

*)


(*

let a = new_fd_var_id ();;
let b = new_fd_var_id ();;
let c = new_fd_var_id ();;

let p = 
  difference_polynomes
    (produit_polynomes 
       (poly_var b)
       (poly_var b))
    (produit_polynomes 
       (poly_var a)
       (poly_var c))
;;

string_of_polynome p;;

let ([c1],[c2]) = square_abstraction [Nul(p)] []
in
print_constraint c1;
print_newline();
print_constraint c2;
print_newline();;

let p2 = 
  somme_polynomes 
  (puissance_polynome (poly_var 0) 6)
  (puissance_polynome (poly_var 0) 7);;

string_of_polynome p2;;

let ([c1],[c2;c3]) = square_abstraction [Nul(p2)] [];;

print_constraint c1;
print_newline();
print_constraint c2;
print_newline();
print_constraint c3;
print_newline();;

let ([c4],[c5;c6]) = product_abstraction [c1] [];;


print_constraint c4;
print_newline();
print_constraint c5;
print_newline();
print_constraint c6;
print_newline();;


*)






(* 

   abstraire les expressions complexes : on applique les regles
   suivantes a P;\emptyset. 

   L'ordre de priorite des regles est :
   4 ,5
   1
   2
   3


   4) passage des contraintes simples (sommes)

      P\cup \{ax+by+cz # 0\};Q -> P ; Q\cup\{ax+by+cz # 0\}

      (x,y,z, variables distinctes, ou bien des constantes).

   5) passage des contraintes simples (produits)

      P\cup \{ax^2 +bz # 0\};Q -> P ; Q\cup\{ax^2 # -bz\}

      P\cup \{axy +bz # 0\};Q -> P ; Q\cup\{axy # -bz\}

      (x,y,z distincts)
      (changer signes si a<0)

   1) abstraction des carres

     P;Q -> P[x^2 -> z] ; Q \cup \{ x^2 = z \}

     si x^2 occurre dans P, z nouvelle variable
 
     noter que toute occurrence de x^3 devient x*z, x^4 devient z^2,
     de maniere generale x^{2n} devient z^n et x^{2n+1} devient x*z^n
 
   2) abstraction des produits 

     P;Q -> P[x*y -> z] ; Q \cup \{ x*y = z \}

     si x*y occurre dans P, z nouvelle variable.
     Efficacite : choisir le produit qui occure le plus souvent !

   3) abstraction des sommes d'au moins quatre termes

      P\cup \{ax+by+p+q # 0\};Q  -> P\cup\{z+p+q # 0\}; Q\cup {ax+by=z\} 

      ou # est =, >, <, >= ou <= ; x,y des variables ; p,q des
      polynomes et z nouvelle variable

*)



let rec linear_poly p = 
  Fd_polynomials.fold
    (fun m l ->
       if Fd_polynomials.monomial_degree m = 1
       then 
	 let (x,exp)= Fd_var_map.max (Fd_polynomials.monomial_vars m)
	 in (x,Fd_polynomials.monomial_coef m)::l
       else l)
    p
    []
;;

let variable_abstraction c bounds = 
  let (c1,sc1,b1) = square_abstraction c [] bounds
  in
  let (c2,sc2,b2) = product_abstraction c1 sc1 b1
  in
  let c3 =
    List.fold_left 
      (fun sc c -> 
	 Sum (linear_poly c.expr,Fd_polynomials.constant_coef c.expr,c.cond)
	 ::sc)
      sc2
      c2
  in (c3,b2)
;;

  
(*

  constraint de domaine fini associe a 
  
   ax + p = c 

  ou p = \Sigma a_ix_i 

*)


let add_const e a =
  if Numbers.is_zero a
  then e
  else Oper(e,Add,Const a)
;;


let mult_const e a =
  if Numbers.eq a Numbers.one
  then e
  else Oper(e,Mult,Const a)
;;

let div_ceil_const e a =
  if Numbers.eq a Numbers.one
  then e
  else Oper(e,Div_ceil,Const a)
;;

let div_floor_const e a =
  if Numbers.eq a Numbers.one
  then e
  else Oper(e,Div_floor,Const a)
;;



(*

  split_sum_min epsilon x = 
      (\Sigma_{b_i>0} b_i min(x_i),\Sigma_{b_i<0} b_i max(x_i)

     ou b_i = epsilon * a_i

  split_sum_max epsilon x = 
      (\Sigma_{b_i>0} b_i max(x_i),\Sigma_{b_i<0} b_i min(x_i)

     ou b_i = epsilon * a_i


*)

(*
let rec split_sum_min epsilon x = function
    [] -> [],[]
  | (y,a)::p ->
      let s1,s2 = split_sum_min epsilon x p
      in
      	if x=y
      	then s1,s2
      	else 
	  let b = mult_num epsilon a
	  in
	    if gt_num b Numbers.zero
	    then (mult_const (Min(y)) b)::s1,s2
	    else s1,(mult_const (Max(y)) (minus_num b))::s2
;;


let rec split_sum_max epsilon x = function
    [] -> [],[]
  | (y,a)::p ->
      let s1,s2 = split_sum_max epsilon x p
      in
      	if x=y
      	then s1,s2
      	else 
	  let b = mult_num epsilon a
	  in
	    if gt_num b Numbers.zero
	    then (mult_const (Max(y)) b)::s1,s2
	    else s1,(mult_const (Min(y)) (minus_num b))::s2
;;
*)

let rec split_sum epsilon x = function
    [] -> [],[],[],[]
  | (y,a)::p ->
      let s1,s2,s3,s4 = split_sum epsilon x p
      in
      	if x=y
      	then s1,s2,s3,s4
      	else 
	  let b = Numbers.mult epsilon a
	  in
	    if Numbers.gt b Numbers.zero
	    then (mult_const (Min(y)) b)::s1,s2, (mult_const (Max(y)) b)::s3,s4
	    else s1,(mult_const (Max(y)) (Numbers.minus b))::s2,s3,(mult_const (Min(y)) (Numbers.minus b))::s4 
;;





let finite_domain_constraint_of_sum a x p c =
  if Numbers.gt a Numbers.zero
  then
(*
    let s1,s2 = split_sum_min num_minus_one x p
    and s3,s4 = split_sum_max num_minus_one x p
*)
    let s1,s2,s3,s4 = split_sum Numbers.minus_one x p
    in
    {
      var = x;
      min = div_ceil_const
	      (sub (add_const (make_sum s1) (Numbers.minus c)) (make_sum s2))
	      a;
      max = div_floor_const
	      (sub (add_const (make_sum s3) (Numbers.minus c)) (make_sum s4))
	      a;
    }
  else
(*
    let s1,s2 = split_sum_min Numbers.one x p
    and s3,s4 = split_sum_max Numbers.one x p
*)
    let s1,s2,s3,s4 = split_sum Numbers.one x p
    in
    {
      var = x;
      min = div_ceil_const
	      (sub (add_const (make_sum s1) c) (make_sum s2))
	      (Numbers.minus a);
      max = div_floor_const
	      (sub (add_const (make_sum s3) c) (make_sum s4))
	      (Numbers.minus a);
    }
;;

let finite_domain_constraint_of_sum_positive bounds a x p c =
  if Numbers.gt a Numbers.zero
  then
(*
    let s1,s2 = split_sum_min num_minus_one x p
    and s3,s4 = split_sum_max num_minus_one x p
*)
    let s1,s2,s3,s4 = split_sum Numbers.minus_one x p
    in
    {
      var = x;
      min = div_ceil_const
	      (sub (add_const (make_sum s1) (Numbers.minus c)) (make_sum s2))
	      a;
      max = Const(Fd_var_map.find x bounds);
    }
  else
(*
    let s1,s2 = split_sum_min Numbers.one x p
    and s3,s4 = split_sum_max Numbers.one x p
*)
    let s1,s2,s3,s4 = split_sum Numbers.one x p
    in
    {
      var = x;
      min = Const(Numbers.zero);
      max = div_floor_const
	      (sub (add_const (make_sum s3) c) (make_sum s4))
	      (Numbers.minus a);
    }
;;

let finite_domain_constraint_of_simple_constraint bounds = function
    Sum(p,c,cond) ->
      begin
	match cond with 
	    Nul ->
	      List.map
		(function (x,a) -> 
		   finite_domain_constraint_of_sum a x p c)
		p		
	  | Pos_or_nul ->
	      List.map
		(function (x,a) -> 
		   finite_domain_constraint_of_sum_positive bounds a x p c) 
		p		
      end
  | Square(x,y) ->
      [
  	{ var = x;
	  min = Sqrt_ceil(Min(y));
	  max = Sqrt_floor(Max(y));
  	} ;
  	{ var = y;
	  min = Oper(Min(x),Mult,Min(x));
	  max = Oper(Max(x),Mult,Max(x));
       	}
      ]
  | Product(x,y,z) ->
      [
  	{ var = x;
	  min = Oper(Min(z),Div_ceil,Max(y));
	  max = Oper(Max(z),Div_floor,Min(y));
  	} ;
  	{ var = y;
	  min = Oper(Min(z),Div_ceil,Max(x));
	  max = Oper(Max(z),Div_floor,Min(x));
  	} ;
  	{ var = z;
	  min = Oper(Min(x),Mult,Min(y));
	  max = Oper(Max(x),Mult,Max(y));
       	}
      ]
;;

    
let translate_constraints bounds c = 
  let (simple_constraints,new_bounds) = variable_abstraction c bounds
  in
(*
    print_string "size of constraint after abstraction : ";
    print_int (size simple_constraints);
    print_newline();
*)

  let fd_constraints =
    Listutils.flat_map
      (finite_domain_constraint_of_simple_constraint new_bounds)
      simple_constraints
  in
    (fd_constraints,new_bounds)
;;

let translate_constraints = 
  Time_profiler.profile2 "translate_constraints" translate_constraints;;
  




(*

  abstraction method 2 : look for the most interesting product.

*)
  
(*
  
  sparse matrix : value is 0 by default
  
  access is done by hashing
  
*)

let matrix_size = 65500;;

let used_products_matrix = Array.create matrix_size [];;
  
let clear_matrix () =
  for i=0 to pred matrix_size do
    used_products_matrix.(i) <- []
  done
;;

let matrix_index x y = 
  ((Hashtbl.hash x) + 97 * (Hashtbl.hash y)) mod matrix_size
;;

let read_matrix x y =

  let rec do_read = function
      [] -> 0
    | (x1,y1,v1)::l -> if x1=x & y1=y then v1 else do_read l

  in
  let n = matrix_index x y
  in
    do_read used_products_matrix.(n)
;;

let write_matrix x y v = 
  
  let rec do_write = function
      [] ->   if v<>0 then [(x,y,v)] else []
    | ((x1,y1,_) as c)::l ->
	if x1=x & y1=y
	then if v<>0 then (x,y,v)::l else l
	else c::(do_write l)
  in
  let n = matrix_index x y
  in
      used_products_matrix.(n) <- do_write used_products_matrix.(n)
;;


exception No_change;;
  
let change_matrix x y v = 
  
  let rec do_change = function
      [] -> 
  	if v=0 then raise No_change else [(x,y,v)]
    | ((x1,y1,v1) as c)::l ->
	if x1=x & y1=y
	then
	  let v = v1+v
	  in
	    if v=v1 then raise No_change else (x,y,v)::l
	else
	  c::(do_change l)
  in
  let n = matrix_index x y
  in
  try
    used_products_matrix.(n) <- do_change used_products_matrix.(n)
  with No_change -> ()
;;

let update_row vars x expx =
  begin
  if expx>1
  then
    change_matrix x x (expx/2);
  Fd_var_map.iter
    (fun y expy ->
      if Fd_var_ord.compare x y < 0
      then
	change_matrix x y (min expx expy))
    vars
  end
;;

let compute_whole_table constraints =
  
  let find_products_in_monomial m =
    let vars = Fd_polynomials.monomial_vars m
    in
      Fd_var_map.iter
        (fun x expx -> update_row vars x expx)
        vars
  in
  begin
    clear_matrix();
    List.iter
      (fun c -> Fd_polynomials.iter find_products_in_monomial c.expr)
      constraints
  end
;;


let minimum_use_of_product = ref 1;;

let find_most_useful_product () =

  let max_found = ref (pred !minimum_use_of_product)
  and best_prod = ref None

  in

    for i=0 to pred matrix_size do
      List.iter 
        (function (x,y,n) ->
           if n > !max_found
	   then
	     begin
	       max_found := n;
               best_prod := Some (x,y)
	     end) 
    	used_products_matrix.(i)
    done;
    match !best_prod with 
      None -> raise Not_found
    | Some(x) -> x
;;



let write_matrix_symmetrically x y v =
  if Fd_var_ord.compare x y < 0
  then
    write_matrix x y v
  else
    write_matrix y x v
;;
		  
let change_matrix_symmetrically x y v =
  if Fd_var_ord.compare x y < 0
  then
    change_matrix x y v
  else
    change_matrix y x v
;;
		  
(* 

remplace toute occurence de x^{2n} par z^n et x^{2n+1} par xz^{n}

met a jour la table des produits

*)

let replace_square_in_monomial_and_update_table x z m =
  let vars = Fd_polynomials.monomial_vars m
  in
  try
    let expx = Fd_var_map.find x vars
    in
    if expx < 2
    then m
    else
      let expxs2 = expx/2
      in
      begin
(*
inutile : sera 0 a la fin 

  change_matrix x x -expxs2;
*)
      	let vars' =
	  if expx mod 2 = 0
	  then 
	    begin
	      Fd_var_map.iter
		(fun y expy ->
		  begin
		    change_matrix_symmetrically x y (-(min expx expy));
		    if Fd_var_ord.compare x y <> 0
		    then
		      change_matrix_symmetrically y z (min expxs2 expy)
		  end)
		vars;
	      Fd_var_map.remove x vars
	    end
	  else 
	    begin 
	      Fd_var_map.iter
		(fun y expy ->
		  begin
		    change_matrix_symmetrically x y (1-(min expx expy));
		    if Fd_var_ord.compare x y <> 0
		    then
		      change_matrix_symmetrically y z (min expxs2 expy)
		  end)
		vars;
	      change_matrix_symmetrically z x 1;
	      Fd_var_map.add x 1 vars (* note: [add] removes the old binding *)
	    end
      	in
	begin
	  change_matrix z z (expxs2/2);
      	  Fd_polynomials.make_monomial 
	    (Fd_polynomials.monomial_coef m)
	    (Fd_var_map.add z expxs2 vars')
	end
      end

  with Not_found -> m;;



(* 

replace toute occurence de (xy)^n avec n>=1 par z^n

met a jour la table des produits

*)

let replace_product_in_monomial_and_update_table x y z m =
  let vars = Fd_polynomials.monomial_vars m
  in
  try
    let expx = Fd_var_map.find x vars
    and expy = Fd_var_map.find y vars
    in
    let expmin = min expx expy
    in
      if expmin < 1
      then m
      else
	begin
	  let vars' =
	    if expx=expy
	    then
	      begin
	      	Fd_var_map.iter
		  (fun zz expzz ->
		    if Fd_var_ord.compare x zz <> 0 
			& Fd_var_ord.compare y zz <> 0 
		    then
		      begin
		      	change_matrix_symmetrically x zz (-(min expx expzz));
		      	change_matrix_symmetrically y zz (-(min expy expzz));
		      	change_matrix_symmetrically z zz (min expmin expzz)
		      end)
		  vars;	      
	  	change_matrix x x (-(expx/2));
	  	change_matrix y y (-(expy/2));
	      	Fd_var_map.remove y (Fd_var_map.remove x vars)
	      end
	    else
	      if expx<expy
	      then
	      	begin
	      	  Fd_var_map.iter
		    (fun zz expzz ->
		      if Fd_var_ord.compare x zz <> 0 
			  & Fd_var_ord.compare y zz <> 0 
		      then
		      	begin
		      	  change_matrix_symmetrically x zz (-(min expx expzz));
		      	  change_matrix_symmetrically y zz 
			    ((min (expy-expx) expzz)-(min expy expzz));
		      	  change_matrix_symmetrically z zz (min expmin expzz);
		      	end)
		  vars;	      		  
		  change_matrix x x (-(expx/2));
		  change_matrix y y ((expy-expx)/2-(expy/2));
	      	  Fd_var_map.add y (expy-expx) (Fd_var_map.remove x vars)
	      	end
	      else
	      	begin
	      	  Fd_var_map.iter
		    (fun zz expzz ->
		      if Fd_var_ord.compare x zz <> 0 
			  & Fd_var_ord.compare y zz <> 0 
		      then
		      	begin
		      	  change_matrix_symmetrically x zz 
			    ((min (expx-expy) expzz)-(min expx expzz));
		      	  change_matrix_symmetrically y zz (-(min expy expzz));
		      	  change_matrix_symmetrically z zz (min expmin expzz);
		      	end)
		  vars;	      		  		  
		  change_matrix x x ((expx-expy)/2-(expx/2));
		  change_matrix y y (-(expy/2));
		  Fd_var_map.add x (expx-expy) (Fd_var_map.remove y vars)
	      	end
	  in
	  begin
	    change_matrix z z (expmin/2);
	    Fd_polynomials.make_monomial 
	      (Fd_polynomials.monomial_coef m)
	      (Fd_var_map.add z expmin vars')
	  end
	end    
  with Not_found -> m;;
  


let smart_product_abstraction constraints simple_constraints bounds = 
  let new_constraints = ref constraints
  and new_simple_constraints = ref simple_constraints
  and new_bounds = ref bounds 
  in
  begin
    compute_whole_table !new_constraints;
    try
      while true do    
      	let (x,y) = find_most_useful_product ()
      	in 
      	let z = new_fd_var_id()
      	and x_bound = Fd_var_map.find x !new_bounds
      	and y_bound = Fd_var_map.find y !new_bounds
      	in
      	if Fd_var_ord.compare x y = 0
      	then 
	  begin
(*
	    Format.print_string "abstraction ";
	    print_fd_var_id z;
	    Format.print_string " = ";
	    print_fd_var_id x;
	    Format.print_string "^2 (";
	    Format.print_int (read_matrix x x);
	    Format.print_string " fois)";
	    Format.print_newline();
*)
	    write_matrix x x 0;
	    new_constraints :=
	       replace_in_constraints 
		 (replace_square_in_monomial_and_update_table x z) 
		 !new_constraints;
	    new_simple_constraints := Square(x,z)::!new_simple_constraints
	  end
      	else
	  begin
(*
	    Format.print_string "abstraction ";
	    print_fd_var_id z;
	    Format.print_string " = ";
	    print_fd_var_id x;
	    print_fd_var_id y;
	    Format.print_string " (";
	    Format.print_int (read_matrix x y);
	    Format.print_string " fois)";
	    Format.print_newline();
*)
	    write_matrix_symmetrically x y 0;
	    new_constraints :=
	       replace_in_constraints 
		 (replace_product_in_monomial_and_update_table x y z) 
		 !new_constraints;
	    new_simple_constraints := Product(x,y,z)::!new_simple_constraints;
(*
	    Format.print_int (read_matrix x y);
	    Format.print_string " fois";
	    Format.print_newline();
*)
	  end;
      	new_bounds := Fd_var_map.add z (Numbers.mult x_bound y_bound) !new_bounds
      done;
      !new_constraints,!new_simple_constraints,!new_bounds

    with Not_found ->
      !new_constraints,!new_simple_constraints,!new_bounds
  end
;;


let smart_translate_constraints bounds c = 
  let (constraints,simple_constraints,new_bounds) = 
    smart_product_abstraction c [] bounds 
  in
(*
    print_string "size of constraint after abstraction : ";
    print_int (size fd_constraints);
    print_newline();
*)

  let fd_constraints =
    (Listutils.flat_map
      (finite_domain_constraint_of_simple_constraint new_bounds)
      simple_constraints)
    @
    (Listutils.flat_map
      (finite_domain_constraint_of_constraint new_bounds)
      constraints)
  in
    (fd_constraints,new_bounds)
;;

let smart_translate_constraints = 
  Time_profiler.profile2 
    "smart_translate_constraints" smart_translate_constraints;;




(*

  (non_linear_solve M C) returns a solution of the constraints C in
  the domain [0..M] for each variable. Raises exception Not_found if
  no solution exists.

*)

let abstraction_method = ref 3;;

let map_length m =
  Fd_var_map.fold (fun _ _ x -> succ x) m 0
;;

let non_linear_solve bound constraints =
(**)
  let constraints = Listutils.flat_map non_linear_simplify constraints
  in

  if !verbose >= 3
  then
    begin
      print_string "Polynomial ordering constraints after simplification : ";
      print_newline();
      print_non_linear_constraints constraints;
      print_newline()
    end;
(**)
    
    let problem_vars = vars_of_constraints constraints
    in
    let bounds = 
      Fd_var_set.fold
	(fun x bounds -> Fd_var_map.add x bound bounds)
	problem_vars
	Fd_var_map.empty
    in
      
      if !verbose >= 2
      then
	
	begin
	  print_string "number of FD variables for the original problem : ";
	  print_int (map_length bounds);
	  print_newline();
	  print_newline()
	end;
      
      let max_degree =
	List.fold_right
	  (fun p n -> max (Fd_polynomials.total_degree p.expr) n)
	  constraints
	  0
      in
	if !verbose >= 2
	then
	  begin
	    print_string "maximum degree of monomials : ";
	    print_int max_degree;
	    print_newline();
	    print_newline()
	  end;
	
	
  let fd_constraints,bounds = 
    match !abstraction_method with
      0 -> simple_translate_constraints bounds constraints
    | 1 -> translate_constraints bounds constraints  
    | 2 -> 
	minimum_use_of_product:=1;
	smart_translate_constraints bounds constraints
    | 3 -> 
	minimum_use_of_product:=2;
	smart_translate_constraints bounds constraints
    | _ -> failwith "Unknown abstraction method in Non_linear_solving"
      
  in 
    if !verbose >= 2
    then
      begin
	List.iter 
	  (function c -> print_fd_constraint c;print_newline())
	  fd_constraints;
      end;
(*
    begin
      let c=open_out "fd_constraints"
      in 
	output_value c fd_constraints;
	close_out c
    end;
*)
    if !verbose >= 2
    then
      begin
	print_string "number of FD constraints : ";
	print_int (List.length fd_constraints);
	print_newline();
	print_string "number of FD variables in the FD constraints : ";
	print_int (map_length bounds);
	print_newline();
	print_newline()
      end;


      let fd_sol = 
	  finite_domain_solve bounds fd_constraints
    
      in
	Fd_var_map.fold
      	  (fun x n sol -> 
	     if Fd_var_set.mem x problem_vars
	     then Fd_var_map.add x n sol
	     else sol)
      	  fd_sol
	  Fd_var_map.empty
;;


let non_linear_solve = 
  Time_profiler.profile2 "non_linear_solve" non_linear_solve;;
