(***************************************************************************

(This sentence has been added automatically, it should be replaced by a description of the module)

CiME Project - Dmons research team - LRI - Universit Paris XI

$Id: bipartite_graph.ml,v 1.3 1998/05/19 11:55:27 contejea Exp $

***************************************************************************)

(*
  Algorithme de couplage maximum dans un graphe bipartite par 
  la methode primale-duale dite "hongroise".
  D'apres Chvatal, Linear Programming, p. 390.

  Le probleme est donne par un graphe bipartite, 
  - les noeuds source U = {ui} ont une capacite {ci}
  - les noeuds cible V = {vj} ont une demande {dj}
  - les arcs presents dans le graphe sont representes par une matrice 
  d'incidence gr (grij = k indique que l'arc ij a une capacite k, si
  grij = 0, l'arc ij n'est pas present dans le graphe).

  Il s'agit de trouver X = {xij} tel que
  - la capacite des noeuds source est completement utilisee
  \sum_j xij = ci \forall i
  - la demande des noeuds cible est satisfaite
  \sum_i xij = dj \forall j

  et bien sur, on n'utilise que les arcs permis.
*)

let auxiliary_graph nb_src nb_cbl nb_nd gr x y a_gr =
  for i = 0 to (pred nb_src) do
    for j = nb_src to (pred nb_nd) do
      a_gr.(i).(j) <- (-1);
      a_gr.(j).(i) <- (-1);
      if 0 < gr.(i).(j)  (* l'arc ij est present dans le graphe *)
      then
 	begin
	  if x.(i).(j) < gr.(i).(j) 
	  then a_gr.(i).(j) <- y.(i) - y.(j) + 1;
	  if 0 < x.(i).(j)
	  then a_gr.(j).(i) <- y.(j) - y.(i) - 1
	end
    done
  done
;;



let is_a_solution etat_nd =
  try 
    for i=0 to (pred (Array.length etat_nd)) do
      if (etat_nd.(i) <> 0)
      then raise Exit
    done;
    true
  with Exit -> false
;;



(*
  calcul de la distance minimale d par rapport aux noeuds mouilles,
  et des chemins associes, donnes par une fonction predecessur p.
*)

let array_exists f a =
  try
    for i=0 to (pred (Array.length a)) do
      if (f a.(i))
      then raise Exit
    done;
    false
  with Exit -> true
;;



let shortest_path_and_min_dist nb_nd etat_nd a_gr l d p =
  (* initialisation de la distance minimale courante *)
  (* et des labels permanents                        *)
  for i=0 to (pred nb_nd) do
    l.(i) <- 0;
    p.(i) <- -1;
    if (etat_nd.(i) > 0)
    then d.(i) <- 0
    else d.(i) <- max_int
  done;

  (* boucle sur les labels temporaires *)
  while (array_exists (function x -> x=0) l) do
  (* un label temporaire devient permanent, 
     on prend la plus petite distance *)
    let label = ref (-1) 
    and d_label = ref max_int 
    in
      for i=0 to (pred nb_nd) do
      	if (l.(i) = 0) & (d.(i) < !d_label)
      	then 
	  begin
	    label := i;
	    d_label := d.(i)
      	  end
      done;
    (* tous les noeuds restants sont a +oo, le graphe n'est plus 
       connecte , il faut tester que les noeuds non connectes sont
       equilibres. *)
      if !label = -1
      then 
      	begin
	  for i=0 to (pred nb_nd) do
	    if (l.(i) = 0)
	    then 
	      begin
	      	if etat_nd.(i) = 0
	      	then l.(i) <- 1
	      	else raise Exit
	      end
	  done
      	end
    (* sinon, on ajoute effectivement le label, 
       et on recalcule les distances *)
      else 
      	begin
	  l.(!label) <- 1;
    	  for i=0 to (pred nb_nd) do
	    if (l.(i) = 0) & (a_gr.(!label).(i) >= 0)
	    then 
 	      let di' = d.(!label) + a_gr.(!label).(i) in
      	    	if di' < d.(i)
	    	then 
		  begin
	    	    d.(i) <- di';
	    	    p.(i) <- !label
		  end
    	  done
      	end
  done
;;



let find_a_dry_node etat_nd = 
  let res = ref (-1) in
    begin
      try
      	for i=0 to (pred (Array.length etat_nd)) do
      	  if etat_nd.(i) < 0
      	  then 
      	    begin
	      res := i;
	      raise Exit
      	    end
      	done
      with Exit -> ()
    end;
    if !res = (-1) 
    then raise Exit (* on n'a pas trouve de noeud sec *)
    else !res
;;



let rec from_wet_to_dry_path node p etat_nd res =
  if etat_nd.(node) > 0
  then res
  else 
    let new_node = p.(node) in
      from_wet_to_dry_path new_node p etat_nd ((new_node,node)::res)
;;



let couplage_max src cbl gr = 
  (* recuperation d'informations *)
  let nb_src = Array.length src
  and nb_cbl = Array.length cbl 
  in
  let nb_nd = nb_src + nb_cbl 
  in

  (* initialisation *)
  let gr' = Array.make_matrix nb_nd nb_nd 0 
  and x = Array.make_matrix nb_nd nb_nd 0
  and y = Array.make nb_nd 0 
  and etat_nd = Array.make nb_nd 0 
  in
  (* on transforme le graphe en ne distingant plus le type des noeuds *)
    for i=0 to (pred nb_src) do
      for j=0 to (pred nb_cbl) do
      	gr'.(i).(nb_src + j) <- gr.(i).(j)
      done
    done;
  (* si l'etat d'un noeud est >0, ce noeud est humide, ie a des surplus
                              <0, ce noeud est sec, ie a un deficit
                              =0, ce noeud est equilibre *)
    for i=0 to (pred nb_src) do
      etat_nd.(i) <- src.(i)
    done;
    for j=0 to (pred nb_cbl) do
      etat_nd.(nb_src + j) <- -cbl.(j)
    done;

  (* allocation de place memoire *)
    let d = Array.make nb_nd 0                       (* distance minimale *)
    and p = Array.make nb_nd (-1)                    (* predecesseur      *)
    and l = Array.make nb_nd 0                       (* labels            *)
    and a_gr = Array.make_matrix nb_nd nb_nd (-1)    (* graphe auxiliaire *)
    in
  (* boucle *)
      while not (is_a_solution etat_nd)
      do  
    	auxiliary_graph nb_src nb_cbl nb_nd gr' x y a_gr;
    	shortest_path_and_min_dist nb_nd etat_nd a_gr l d p;
    	for i=0 to (pred nb_nd) do
	  y.(i) <- y.(i) + d.(i)
    	done;
    let dry_node = find_a_dry_node etat_nd 
    in
    let path = from_wet_to_dry_path dry_node p etat_nd [] 
    in
    let wet_node = fst(List.hd path) 
    in
    let t = min (-etat_nd.(dry_node)) etat_nd.(wet_node) 
    in
      List.iter 
      	(function (i,j) -> 
	   if i < j (* on a un noeud de type + *)
	   then x.(i).(j) <- x.(i).(j) + t
	   else x.(j).(i) <- x.(j).(i) - t)
      	path;
      etat_nd.(dry_node) <- etat_nd.(dry_node) + t;
      etat_nd.(wet_node) <- etat_nd.(wet_node) - t
      done;
      x;
      let res = Array.make_matrix nb_src nb_cbl 0 in
  	for i=0 to (pred nb_src) do
	  for j=0 to (pred nb_cbl) do
	    let v = x.(i).(nb_src + j) in
	      if v > 0
	      then 
		begin
(*
                  Printf.printf "%d" v;
*)
		  res.(i).(j) <- v
		end
(*
              else Printf.printf "%s" "-"
*)
	  done;
(*
          Printf.printf "%s" "\n"
*)
  	done;
  	res
;;



(*
let _ = 
  couplage_max 
    [|2; 3; 1; 2; 2|] 
    [|1; 5; 2; 2|] 
    [|[|1; 0; 0; 2|]; 
      [|0; 3; 0; 0|]; 
      [|1; 0; 1; 0|]; 
      [|0; 2; 0; 0|];
      [|1; 0; 1; 1|]|]
*)
