

open Signatures
open Variables
open Gen_terms
open Lazy_list
open Lazy_controle


let print_protected_term sigma vars t = 

  let rec print_term_rec is_deep = function
    | Term(Protected_var x,[]) -> 
	begin
      	  print_string "[";
	  print_string (vars#string_of_var x);
      	  print_string "]";
	end
    | Term((Original f),[]) ->
      	print_string (sigma#string_of_symbol f);
    | Term((Original f),t::l) ->
	Format.open_box 0;
	begin
	  match sigma#symbol_fix f with
	      Default | Prefix ->
	      	begin
      	      	  print_string (sigma#string_of_symbol f);
      		  print_string "(";
      		  print_term_rec false t.Hcons.node;
      		  print_end_term_list l
	      	end
	    | Postfix ->
	      	begin
      		  print_string "(";
      		  print_term_rec false t.Hcons.node;
      		  print_end_term_list l;
      	      	  print_string (sigma#string_of_symbol f);
	      	end
	    | Infix ->
	      	begin
		  if is_deep then print_string "(";
		  print_term_rec true t.Hcons.node;
      		  print_infix_list f l;
		  if is_deep then print_string ")"
	      	end
	end;
	Format.close_box();
    | Var(x) ->
	print_string (vars#string_of_var x)
    | _ -> assert false
  and print_end_term_list = function
      [] -> 
      	print_string ")"
    | t::l ->
      	print_string ",";
      	print_term_rec false t.Hcons.node;
      	print_end_term_list l
	  
  and print_infix_list f = function
      [] -> ()
    | t::l ->
	print_string " ";
      	print_string (sigma#string_of_symbol f);
	print_string " ";
      	print_term_rec true t.Hcons.node;
      	print_infix_list f l
	  
  in
    print_term_rec false t.Hcons.node

let debug_symbol sign f =
  let hash_f = Signatures.hash_symbol f in
    match f with
      | Original ff ->
	  print_string "hash_symbol(";
	  print_string (sign#string_of_symbol ff);
	  print_string ") = ";
	  print_int hash_f;
	  print_newline ();
	  Format.print_flush ()
      | Protected_var x ->
	  print_string "hash_symbol([";
	  print_string (Variables.default#string_of_var x);
	  print_string "]) = ";
	  print_int hash_f;
	  print_newline ();
	  Format.print_flush ()

let debug sign t =
(*  print_string "Building a term ; hkey = "; *)
  print_int t.Hcons.hkey;
  print_string " (hkey) ; ";
  print_int t.Hcons.tag;
  print_string " (tag) ; ";
  print_string " ; term = ";
  print_protected_term sign Variables.default t;
  print_newline();
  Format.print_flush ()


let make_var_term sign phct x =
  let x' = Gen_terms.make_var_term phct x in
(*i
  let _ = debug sign x' in
i*)
    x'
    
let make_cst_term sign phct c = 
  let c' = Gen_terms.make_term phct c [] in
(*i
  let _ = debug_symbol sign c; debug sign c' in
i*)
  c'


let make_term sign phct f l =
  match f with
    | Original _ ->
	let t' = Gen_terms.make_term phct f l in
(*i
	let _ = debug sign t' in
i*)
	  t'
    | Protected_var _ -> 
	let t' = Gen_terms.make_term phct f l in
(*i
	let _ = debug sign t' in
i*)
	  t'
	  

(*
  [protect_var_and_sort sign phct t] builds a copy of the term 
  [t] such that its variables are protected, that is are
  made constant since [t] is intended to be the subject in
  the matching. This copy is sorted in such a way that
  the arguments of a C/AC symbol are in increasing ordering.
*)

let rec protect_var_and_sort sign phct t = 
(*i
  let _ = Format.print_string "toto"; Format.print_newline (); Format.print_flush () in
i*)
  match t.Hcons.node with
    | Var(x) -> make_cst_term sign phct (Protected_var x)
    | Term(f,l) -> 
	let ls = List.map (protect_var_and_sort sign phct) l in
	  if sign#is_free f
	  then make_term sign phct (Original f) ls
	  else make_term sign phct (Original f) (List.sort compare_terms ls)


(* 
   [simple_copy_and_sort sign phct t] builds a copy of the term [t]
   which is sorted in such a way that the arguments of a C/AC symbol
   are in increasing ordering, the variable terms being at the end of
   the list of argument. [t] is intended to be the pattern in the 
   matching.
*)

let rec simple_copy_and_sort sign phct t = 
  match t.Hcons.node with
    | Var x -> make_var_term sign phct x
    | Term(f,l) -> 
	let ls = List.map (simple_copy_and_sort sign phct) l in
	  if sign#is_free f
	  then make_term sign phct (Original f) ls
	  else make_term sign phct (Original f) (List.sort compare_terms ls)


(*
  [unprotect hct t] builds the term corresponding with the
  term [t] on the initial signature.
*)

let rec unprotect hct t =
  match t.Hcons.node with
    | Var _ -> assert false
    | Term(Original f,l) -> 
	Gen_terms.make_term hct f (List.map (unprotect hct) l)
    | Term(Protected_var x,[]) -> Gen_terms.make_var_term hct x
    | _ -> assert false


exception No_match;;

type 'symbol problem =
    {
      (*i
	clef : int list;
      i*)
      unsolved_part : ('symbol protect_var_sig term * 'symbol protect_var_sig term) list;
      solved_part : 'symbol protect_var_sig Substitution.substitution;
    }





let print_pb sign pb =
  (*i
  Printf.printf "clef = ";
  List.iter (function i -> Printf.printf " %d;" i) pb.clef;
  Printf.printf "\n";
  i*)
  Printf.printf "solved equations =\n";
  VarMap.iter 
    (fun x t -> 
       print_string (default#string_of_var x);
       Format.print_flush ();
       Printf.printf " = ";
       print_protected_term sign default t;
       Format.print_flush ();
       Printf.printf "\n")
    pb.solved_part;
  Printf.printf "unsolved equations =\n";
  List.iter
    (fun (s,t) ->
       print_protected_term sign default s;
       Format.print_flush ();
       Printf.printf " = ";
       print_protected_term sign default t;
       Format.print_flush ();
       Printf.printf "\n")
    pb.unsolved_part


let rec remove a = function
    [] -> raise Not_found
  | b::l -> if a == b then l else b::(remove a l)

let one_level_sort_term sign phct t =
  match t.Hcons.node with 
    | Term ((Original f as original_f),l) -> 
	if sign#is_free f
	then t
	else make_term sign phct original_f (List.sort compare_terms l)
    | _ -> t

let make_protected_sorted_term sign phct ff l =
  match ff with
    | (Original f) ->
	if sign#is_free f
	then make_term sign phct ff l
	else make_term sign phct ff (List.sort compare_terms l)
    | _ -> make_term sign phct ff l

(*

  [substitute sign phct subst t] builds the term [t] instantiated
  by the substitution [subst]. [t] and [subst] are build over the
  protected version of the signature [sign], hence the substitute
  generic function does not the job.

*)
let rec substitute sign phct subst t = 
  match t.Hcons.node with
    | Var id -> 
	begin
	  try 
	    Variables.VarMap.find id subst
	  with Not_found -> t
	end
    | Term ((Protected_var _),[]) -> t
    | Term ((Original f as original_f),l) -> 
	if sign#is_ac f 
	then
	  let list_of_substituted_subterms = 
	    List.sort 
	      compare_terms
	      (List.fold_left
		 (fun acc x -> 
		    let x_subst = substitute sign phct subst x in
		      begin
			match x_subst.Hcons.node with 
			    Term (s,l2) -> 
			      if s = original_f then l2 @ acc else x_subst :: acc
			  | Var _ -> x_subst::acc
		      end)
		 [] l) in
	    make_term sign phct original_f list_of_substituted_subterms

	else make_term sign phct original_f (List.map (substitute sign phct subst) l)
    | _ -> assert false
;;

let apply_subst_to_match_eq sign phct subst (s,t) =
  (one_level_sort_term sign phct 
     (substitute sign phct subst s), t)

let apply_subst_to_match_eqs sign phct subst = 
  List.map (apply_subst_to_match_eq sign phct subst)

(*

  [ac_elementary_solve tt1,tt2] performs an elementary solving
  step on an equation [tt1,tt2] such that [tt1] and [tt2] are
  terms which have the same AC top function symbol.

*)

let ac_elementary_solve sign phct pb (tt1,tt2) list_of_equations solved_part =
  match tt1.Hcons.node,tt2.Hcons.node with
    | Term (f,[]), Term (_,l') ->
	let _ = assert (l'=[]) in
	let new_pb = 
	  {
	    (*i
	      clef = pb.clef @ [1];
	    i*)
	    unsolved_part = list_of_equations;
		    solved_part = solved_part;
	  } in
	  Cons {head = new_pb; tail = Val Nil}
    | Term (f,(s :: list_of_terms as l)), Term (_,l') ->
	begin
	  match s.Hcons.node with 
	    | Term (g,_) ->
		begin
		  (*i
		    let cmpt = ref 0 in
		  i*)
		  Lazy_list.map
		    (function t ->
		       match t.Hcons.node with
			 | Term (h,_) ->
			     if g = h
			     then 
			       let s' = 
				 match list_of_terms with
				   | [a_single_term] -> a_single_term
				   | _ -> make_term sign phct f list_of_terms
			       and l'_without_t = remove t l' in
			       let t' = 
				 match l'_without_t with
				   | [a_single_term] -> a_single_term
				   | _ -> make_term sign phct f l'_without_t in  
				 {
				   (*i
				     clef = (let _ = cmpt := !cmpt + 1 in
				     pb.clef @ [!cmpt]);
				     i*)
				   unsolved_part = 
				    (s,t)::(s',t')::list_of_equations;
				   solved_part = solved_part;
				 }
			     else raise Exit
			 | Var _ -> assert false) 
		    l' 
		end
	    | Var _ -> 
		begin
		  (*i
		    let cmpt = ref 0 in
		  i*)
		  match l' with 
		    | t :: l'' ->
			Lazy_list.map2_without_repetition
			  (function v ->
			     match v.Hcons.node with
			       | Var x ->
				   let y = fresh_var_for_unif () in
				   let y_plus_t = 
				     make_protected_sorted_term sign phct f [(make_var_term sign phct y);t] in
				   let x_mapsto_t =
				     VarMap.add x t VarMap.empty 
				   and x_mapsto_y_plus_t =
				     VarMap.add x y_plus_t VarMap.empty in
				   let l_without_v = remove v l in
				   let t_f_l'' = 
				     match l'' with
				       | [a_single_term] -> a_single_term
				       | _ -> make_term sign phct f l'' in
				   let new_pb =
				     {
				       (*i
					 clef = (let _ = cmpt := !cmpt + 1 in
					 pb.clef @ [!cmpt]);
				       i*)
				       unsolved_part =
					begin
					  let new_t1 =
					    match l_without_v with
					      | [a_single_term] -> a_single_term
					      | _ ->  make_term sign phct f l_without_v in
					  let new_list_of_equations =
					    (new_t1, t_f_l'') :: list_of_equations in
					    apply_subst_to_match_eqs sign phct x_mapsto_t 
					      new_list_of_equations
					end;
				       solved_part = 
					VarMap.add x t
					  (VarMap.map 
					     (substitute sign phct x_mapsto_t) 
					     solved_part)
				     }
				   and  new_pb' =
				     {
				       (*i
					 clef = (let _ = cmpt := !cmpt + 1 in
					 pb.clef @ [!cmpt]);
				       i*)
				       unsolved_part =
					begin
					  let new_list_of_equations =
					    (make_protected_sorted_term sign phct f 
					       ((make_var_term sign phct y) :: (remove v l)),
					     t_f_l'') :: list_of_equations in
					    apply_subst_to_match_eqs sign phct x_mapsto_y_plus_t 
					      new_list_of_equations
					end;
				       solved_part = 
					VarMap.add x y_plus_t
					  (VarMap.map (substitute sign phct x_mapsto_y_plus_t) 
					     solved_part)
				     } in
				     (new_pb, new_pb')
				     
			       | Term _ -> assert false)
			  l 
		    | [] -> Nil 
		end
	end
    | _,_ -> assert false

(*

  [solve sign phct pb] solve the first unsolved equation of
  the problem [pb], at the top level, and returns the obtained 
  problem.

*)

let solve sign phct pb =
  (*i
  let _ = 
    Format.print_string "solve est appele sur le probleme";
    Format.print_newline();
    print_pb sign pb
  in
  i*)
  match pb.unsolved_part with
    | (tt1,tt2 as eq) :: list_of_equations ->
	begin
	  match tt1.Hcons.node,tt2.Hcons.node with
	    | Var x,_ ->
		let x_mapsto_tt2 = VarMap.add x tt2 VarMap.empty in
		let new_pb =
		  {
		    (*i
		      clef = pb.clef @ [1];
		    i*)
		    unsolved_part = 
		     apply_subst_to_match_eqs sign phct x_mapsto_tt2 list_of_equations;
		    solved_part = 
		     VarMap.add x tt2 
		       (VarMap.map 
			  (substitute sign phct x_mapsto_tt2) 
			  pb.solved_part)
		  } in
		  Cons {head = new_pb; tail = Val Nil}
		    
	    | (Term ((Protected_var x as protected_x),[]) ,Term (f',l')) ->
		if protected_x <> f'
		then raise No_solution
		else 
		  let _ = assert (l' = []) in
		  let new_pb = 
		    {
		      (*i
		      clef = pb.clef @ [1];
		      i*)
		      unsolved_part = list_of_equations;
		      solved_part = pb.solved_part;
		    } in
		    Cons {head = new_pb; tail = Val Nil}

	    | (Term ((Original f as original_f),l), Term (f',l')) ->
		if original_f <> f'
		then raise No_solution
		else 
		  if sign#is_free f
		  then 
		    let new_pb =
		      try
			{
			  (*i
			    clef = pb.clef @ [1];
			  i*)
			  unsolved_part = 
			   List.fold_left2
			     (fun current_list_of_eqs s t -> 
				(s,t)::current_list_of_eqs)
			     list_of_equations l l';
			  solved_part = pb.solved_part;
			} 
		      with 
			  Invalid_argument "List.fold_left2" -> raise No_solution in
		      Cons {head = new_pb; tail = Val Nil}
		  else 
		    if sign#is_ac f
		    then 
		      if (List.length l) > (List.length l')
		      then raise No_solution
		      else ac_elementary_solve sign phct pb eq list_of_equations pb.solved_part
		    else (*i sign#is_commutative f i*)
		      begin
			match l,l' with
			    [s1;s2], [t1;t2] ->
			      let new_pb1 =
				{
				  (*i
				    clef = pb.clef @ [1];
				  i*)
				  unsolved_part = (s1,t1)::(s2,t2)::list_of_equations;
				  solved_part = pb.solved_part;
				}
			      and new_pb2 =
				{
				  (*i
				    clef = pb.clef @ [2];
				  i*)
				  unsolved_part = (s1,t2)::(s2,t1)::list_of_equations;
				  solved_part = pb.solved_part;
				} in 
				Cons {head = new_pb1; 
				      tail = Val (Cons {head = new_pb2; tail = Val Nil})}
				  
			  | _,_ -> assert false
		      end
		      
	    | Term _, Var _  -> assert false
	    | Term (Protected_var _, _::_),_-> assert false
	end
    | [] -> raise Not_appliable

	      

let matching sign hct phct t1 t2 = 
(*i
  let _ = 
    begin
      Format.print_string "matching sur ";
      Gen_terms.print_term sign default t1;
      Format.print_string " = ";
      Gen_terms.print_term sign default t2;
      Format.print_newline();
      Format.print_flush () 
    end in
i*)    
    if (Gen_terms.size sign t1) > (Gen_terms.size sign t2) 
    then raise No_match
    else
      begin
	let t1' = simple_copy_and_sort sign phct t1
	and t2' = protect_var_and_sort sign phct t2 in
	let vars_of_t1 = (set_of_variables t1) in
	let _ = init_for_unif vars_of_t1 in
	let init_pb = 
	  {
	    (*i
	      clef = [];
	      i*)
	    unsolved_part = [t1',t2'];
	    solved_part = VarMap.empty
	  } in
	let list_of_solutions =
	  repeat (solve sign phct) init_pb in
	  try 
	    let solved_pb = Lazy_list.hdl list_of_solutions in
	    let sol = Variables.VarMap.map (unprotect hct) solved_pb.solved_part in
	      (*i
	    let _ = 
	      begin
		if sol = Variables.VarMap.empty
		then 
		  begin
		    Format.print_string "{}";
		    Format.print_newline()
		  end
		else Substitution.print sign default sol;
		Format.print_flush ()
	      end in
	      i*)
	      sol
	  with
	      Failure "hdl" -> 
		begin 
		  (*i
		  Format.print_string "no solution";
		  Format.print_newline();
		  Format.print_flush ();
		  i*)
		  raise No_match 
		end
      end

let all_matchs sign hct phct vars t1 t2 =
    let t1' = simple_copy_and_sort sign phct t1
    and t2' = protect_var_and_sort sign phct t2 in
    let vars_of_t1 = (set_of_variables t1') in
    let _ = init_for_unif vars_of_t1 in
    let init_pb = 
      {
	(*i
	clef = [];
	i*)
	unsolved_part = [t1',t2'];
	solved_part = VarMap.empty
      } in
    let list_of_solutions =
      repeat (solve sign phct) init_pb in
    let rec print_solved_pbs n = function
	Nil -> 
	  print_int n;
	  print_string " solutions";
	  print_newline()
      | Cons {head = solved_pb; tail = Val l} ->
	  let sol = Variables.VarMap.map (unprotect hct) solved_pb.solved_part in
(*
	    print_string "-------- solution ";
	    print_int n;
	    print_string " --------";
	    print_newline ();
	    Substitution.print sign vars sol;
*)
	    print_solved_pbs (n+1) l
      | Cons ({head = solved_pb; tail = Freeze th} as cell) ->
	  let sol = Variables.VarMap.map (unprotect hct) solved_pb.solved_part in
(*
	    print_string "-------- solution ";
	    print_int n;
	    print_string " --------";
	    print_newline ();
	    Substitution.print sign vars sol;
*)
	    let tail = th () in
	      cell.tail <- Val tail;
	      print_solved_pbs (n+1) tail in
      print_solved_pbs 0 list_of_solutions



