(***************************************************************************

  Several types of equations for completion (interface).

CiME Project - Dmons research team - LRI - Universit Paris XI

$Id: equations_as_obj.mli,v 1.2 2001/04/20 13:42:21 marche Exp $

***************************************************************************)

(*

Obsolete header:

  CiME Project - Dmons research team - LRI - Universit Paris XI

$Id: equations_as_obj.mli,v 1.2 2001/04/20 13:42:21 marche Exp $

***************************************************************************)

open Orderings;;
open Content;;
open Terms;;
open Dnet;;
open Equations;;

class pair_of_terms : (Terms.term) -> (Terms.term) ->
  object
    val mutable term1 : Terms.term
    val mutable term2 : Terms.term
    method first_term : Terms.term
    method second_term : Terms.term
    method set_first_term : Terms.term -> unit
    method set_second_term : Terms.term -> unit
    method print : unit
  end;;

val print_list_of_pairs_of_terms : 
  < print : 'a; .. > list -> unit
;;

(*
  EQUATION is the type of the equations used by the completion.
*)

module type EQUATION =
  sig
    type t 
      (* The type of equations. An equation possesses
	 - a first term
	 - a second term
	 - 3 different types of interest
	 - an age
	 - an orientation
	 - a field norm_wrt_rule. If this field is equal to n, this
	 means that the equation is in a normal form wrt to the rules
	 in the completion whose number is less than or equal to n. *)

    val interest1_of : t -> i1 
      (* the global size of equations, both sides are considered *)

    val interest2_of : t -> i2
      (* the size of equations as ordered pairs, only the lhs is considered *)

    val interest3_of : t -> i3
      (* the interest of equations for proving the conjectures *)

    val set_interest1 : i1 -> t -> unit
    val set_interest2 : i2 -> t -> unit
    val set_interest3 : i3 -> t -> unit
      (* set_interest{1,2,3} [interest] [equation] gives the value [interest]
	 to the field interest{1,2,3} of the equation [equation]. *)

    val compare_equation : t ordering
      (* ordering on equations, based on their interest *)

    val first_term : t -> term
      (* yields the first term equations *)

    val second_term : t -> term
      (* yields the second term of equations *)

    val set_first_term : term -> t -> unit
      (* set_first_term [term] [equation] gives the value [term] to 
	 the field first term of the equation [equation] (in-place 
	 modification). *)

    val set_second_term : term -> t -> unit
      (* set_second_term [term] [equation] gives the value [term] to 
	 the field second term of the equation [equation] (in-place 
	 modification). *)

    val make_equation : term -> term -> t
      (* make_equation [term1] [term2] builds the equation of which
	 the first term is [term1] and the second term is [term2]. 
	 The other fields are not specified. *)

    val make_aged_equation : int -> term -> term -> t
      (* make_aged_equation [age] [term1] [term2] builds the equation 
	 of which the first term is [term1] and the second term is 
	 [term2] and the age is [age]. The other fields are not specified. *)

    val is_oriented : t -> bool
      (* is_oriented [equation] is true if the equation can be and has 
	 actually been oriented. An equation can be oriented if its first
         and second terms are comparable wrt the ordering used by the 
         completion, and in this case, the first term is intented to be
	 the greatest one. *)

    val set_orientation : bool -> t -> unit
      (* set_orientation [orientation] [equation] gives the value 
	 [orientation] to the field orientation of the equation
	 [equation]. This should be done according to the above
	 specification *)

    val reset_orientation : t -> unit
      (* reset_orientation [equation] gives the default value, that
	 is false, to the field orientation of the equation [equation]. *)

    val get_age : t -> int
      (* get_age [equation] yields the value of the field age of the
	 equation [equation]. *)

    val set_age : int -> t -> unit
      (* set_age [age] [equation] gives the value [age] to the
	 field age of the equation [equation]. *)

    val get_norm_wrt_rule : t -> int
      (* get_norm_wrt_rule [equation] yields the value of the
	 field norm_wrt_rule of the equation [equation]. *)

    val set_norm_wrt_rule : int -> t -> unit
      (* set_norm_wrt_rule [rule_number] [equation] gives the value
	 [rule_number] to the field norm_wrt_rule of the equation
	 [equation]. *)

    val reset_interest : t -> unit
      (* reset_interest [equation] gives the default values to the
	 fields interest{1,2,3} of the equation [equation]. *)

    val print_interest : i1 -> i2 -> i3 -> unit
      (* print_interest [interest1] [interest2] [interest3] prints
	 the three interest values [interest1], [interest2] 
	 and [interest3]. *)

    val print_equation : t -> unit
      (* print_equation [equation] prints the equation [equation]. *)

    val print_list_of_equations : t list -> unit
      (* print_list_of_equations [list_of equations] prints
	 the list of equations [list_of_equations]. *)

    val is_a_trivial_pair_of_terms : 
      int * regle_etendue dnet -> term * term -> bool
      (* is_a_trivial_pair_of_terms [completion_rules] [term1] [term2]
	 check if the terms [term1] and [term2] are equal wrt the
	 rules [completion_rules] generated by the completion. *)
  end;;

module type EQUATION_SET =
sig
  type equation
      (* type equation, intended as the element type of sets *)
    
  type eq_set
      (* type of sets of equations *)
    
  exception No_more_equation
    
  val make_empty_equation_set : unit -> eq_set
      (* builds an empty equation set. *)
      
  val next_equation : eq_set -> equation
      (* next_equation [equation_set] yields the minimal (wrt interest) 
	 equation of the set of equations [equation_set], and removes it 
	 from [equation_set], as pop does for stacks. When [equation_set]
	 is empty, the exception No_more_equation is raised. *)

  val insert_equation : equation -> eq_set -> unit
      (* insert_equation [equation] [equation_set] inserts in place 
	 the equation [equation] in the set of equations [equation_set]. *)
      
  val carefully_insert_equation : equation -> eq_set -> unit
      (* carefully_insert_equation [equation] [equation_set] inserts in place 
	 the equation [equation] in the set of equations [equation_set]. 
	 This is done carefully, if the equation to be inserted is redundant,
	 nothing is done, whereas if some equations in the set of equations
	 are redundant wrt the new equation, they are removed. The tests 
	 rather costly. *)
      
  val is_empty : eq_set -> bool
      (* is_empty [equation_set] checks if the set of equations 
	 [equation_set] is empty, i.e. does not contains any equation. *)
      
  val card : eq_set -> int
      (* card [equation_set] gives the number of equations contained in
	 the set of equations [equation_set]. *)

  val min_interest1 : eq_set -> i1
  val min_interest2 : eq_set -> i2
  val min_interest3 : eq_set -> i3
      (* min_interest{1,2,3} gives the value of the field interest{1,2,3}
	 of the minimal equation in the set of equations [equation_set]. *)

  val set_min_interest1 : i1 -> eq_set -> unit
  val set_min_interest2 : i2 -> eq_set -> unit
  val set_min_interest3 : i3 -> eq_set -> unit
      
  val print_min_interest : eq_set -> unit
      (* print_min_interest [equation_set] prints the 3 interests of
	 the minimal equation of the set of equations {equation_set]. *)
      
  val print_equation_set : eq_set -> unit
      (* print_equation_set [equation_set] prints the set of 
	 equations [equation_set]. *)
      
  val fold_left : ('a -> equation -> 'a) -> 'a -> eq_set -> 'a
      (* fold_left f a0 [equation_set] returns
	 f(...(f a0 [equation0])...[equationk]) if the set of 
	 equations [equation_set] contains exactly the equations
	 [equation0],...,[equationk]. The order in which the 
	 equation arguments are given is not specified. *)	 
      
  val fold_right : (equation -> 'a -> 'a) -> eq_set -> 'a -> 'a
      (* fold_left f [equation_set] a0 returns
	 f([equationk]...(f [equation0] a0 ))))) if the set of 
	 equations [equation_set] contains exactly the equations
	 [equation0],...,[equationk]. The order in which the 
	 equation arguments are given is not specified. *)	 
      
  val iter : (equation -> unit) -> eq_set -> unit
      (* iter f [equation_set] aplies iteratively f to each
	 equation of the set of equations [equation_set]. *)
      
  val find : (equation -> bool) -> eq_set -> equation
      (* find [predicat] [equation_set] returns an equation
	 of the set of equations [equation_set] which satisfies
	 the predicat [predicat]. If no equation fulfills this
	 requirement, the exception Not_found is raised. *)
      
  val set_to_list : eq_set -> equation list
      (* set_to_list [equation_set] returns a list containing
	 the elements of the set of equations [equation_set]. *)
end




module Equation1 : EQUATION;;
module Equation_set1 : 
  (EQUATION_SET with type equation = Equation1.t);;

(*
module Equation2 : EQUATION;;
module Equation_set2 : 
  (EQUATION_SET with type equation = Equation2.t);;
*)

module Equation_with_goal : EQUATION;;
module Equation_with_goal_set :
  (EQUATION_SET with type equation = Equation_with_goal.t);;


module Conjectures : EQUATION;;
module Conjectures_set :
  (EQUATION_SET with type equation = Conjectures.t);;


module Rules_set :
  (EQUATION_SET with type equation = regle);;




