(***************************************************************************

Generic Normalization

CiME Project - Dmons research team - LRI - Universit Paris XI

$Id: generic_normalization.ml,v 1.1 2000/05/16 17:36:42 monate Exp $

***************************************************************************)


open Symbols;;
open Positions;;
open Terms;;
open Args;;
open Messages;;
open Format;;
open Errors;;

exception Irreducible

(*
  innermost_normalize red t retourne la forme normale de t vis-a-vis 
  de la relation de reduction red appliquee avec la strategie innermost, et 
  echoue s'il faut plus d'etapes qu'autorisees par les options courantes.

  red sera en general une fonction qui effectue un pas de S-reecriture ou
  de R/S-reecriture en tete.

*)

(*
  innermost_normalized red t retourne la forme normale de t vis-a-vis 
  de la relation de reduction red, et echoue s'il faut plus d'etapes 
  qu'autorisees par les options courantes.
*)

let increase_red_number bound = 
  bound := succ (!bound);
  if !bound>=options_courantes.normalization_steps_bound
  then 
    begin
      message ("There have been more than "^(string_of_int !bound)^" consecutive rewrite steps.");
      message "Please check that the rewrite system is terminating,";
      message "or use -rew-bound option.";
      abort()
    end
;;

let rec innermost_normalize_avec_test bound red = function
    VAR x as v -> v
  | TERM (f,l) -> 
      let l'= List.map (innermost_normalize_avec_test bound red)l in
      let t' = 
      	if (is_ac f) 
      	then 
      	  let l'' = flatten_term_list f l' 
      	  in TERM(f,l'')
      	else 
      	  TERM(f,l') 
      in
      	try 
	  let t'' = red t' 
	  in
            increase_red_number bound;
            innermost_normalize_avec_test bound red t''
      	with Irreducible -> t'
;;

let innermost_normalize red t = 
  let bound = ref 0 in
    innermost_normalize_avec_test bound red t
;;

(*
  outermost_normalized red t retourne la forme normale de t vis-a-vis 
  de la relation de reduction red, et echoue s'il faut plus d'etapes 
  qu'autorisees par les options courantes.
*)

let rec outermost_normalize_pos bound red t p q =  
  match t with
      VAR _ -> t
    | _  -> 
      	begin
	  try 
	    let s = red (sous_terme_a_la_position t q) in
              increase_red_number bound;
              let t' = replace t s q in
              	outermost_normalize_path bound red t' q Top
          with Irreducible -> 
	    begin
	      try let q' = next_pos_l t q in
              	outermost_normalize_pos bound red t p q'
	      with No_more_pos -> t
	    end
      	end

and outermost_normalize_path bound red t p q = 
  match t with
      VAR _ -> t
    | _  -> 
    	begin
	  try 
	    let s = red (sous_terme_a_la_position t q) in
	      increase_red_number bound;
	      let t' = replace t s q in
             	outermost_normalize_path bound red t' q Top
	  with Irreducible -> 
	    begin
	      try let q' = allonge p q in
              	outermost_normalize_path bound red t p q'
              with No_more_pos -> 
		try let q''= next_pos_l t q in
		  outermost_normalize_pos bound red t p q''
            	with No_more_pos -> t
	    end
	end
;;

let outermost_normalize red t = 
  let bound = ref 0 in
    outermost_normalize_pos bound red t Top Top
;;



(*


Attempt of an more efficient innermost normalization

we use an auxiliary function 

[substitute_and_innermost_normalize : 
  (term -> term * substitution) -> substitution -> term ]

[(substitute_and_innermost_normalize red sigma t)] returns the
innermost normal form of [(t sigma)] for the reduction relation [red]
where we assume that [sigma] is already in normal form

for a term [t], [(red t)] must return a pair [(u,tau)] such that [t]
rewrites to [(u tau)] at root, or raise [Irreducible] if [t] is in
normal form

*)



let rec substitute_and_innermost_normalize bound red sigma t = 

  let normalize_at_top t = 
      try 
	let (t',sigma') = red t in
          increase_red_number bound;
          substitute_and_innermost_normalize bound red sigma' t'
      with Irreducible -> t
      
  in

  let normalize_subst f sigma = 
    VarMap.fold
    (fun x t s ->
	 let t' =
	   try
	     if f = top_symbol_id t
	     then normalize_at_top t	     
	     else t
	   with
	       Invalid_argument _ -> t
	 in
	   VarMap.add x t' s)
      sigma
      VarMap.empty

  in 
    match t with

    	VAR x as v -> 
	  begin
	    try 
	      VarMap.find x sigma	            	
      	    with Not_found -> v
	  end
	  
      | TERM (f,l) -> 
	  let l'= 
	    List.map (substitute_and_innermost_normalize bound red sigma) l 
	  in
	  let t' = 
	    if is_ac f
            then TERM(f,flatten_term_list f l')
	    else TERM(f,l') 
	  in
	    try 
	      let (t'',sigma') = red t' in
          	increase_red_number bound;
	  	let sigma'' =
		  if is_ac f
		  then normalize_subst f sigma'
		  else sigma'
		in
		  substitute_and_innermost_normalize bound red sigma'' t''

	    with Irreducible -> t'
;;

let efficient_innermost_normalize_avec_test = 
  substitute_and_innermost_normalize
;;

let efficient_innermost_normalize red t = 
  let bound = ref 0 in
  substitute_and_innermost_normalize bound red VarMap.empty t
;;
