(***************************************************************************

Th\'eorie canonique pour la r\'e\'ecriture normalis\'ee : corps

CiME Project - Dmons research team - LRI - Universit Paris XI

$Id: norm_theory.ml,v 1.1 2000/05/16 17:36:46 monate Exp $

***************************************************************************)


open Format;;
open Symbols;;
open Terms;;
open Equations;;
open Dnet;;
open Positions;;


(*=========================================

La th\'eorie canonique modulo laquelle on travaille est m\'emoris\'ee
dans une variable globale.

=========================================*)

type theorie_connue =
  No_theory 
| A of symbol_id 
| AU of symbol_id * symbol_id
| ACU of symbol_id * symbol_id
| ACI of symbol_id
| ACUI of symbol_id * symbol_id
| ACUN of symbol_id * symbol_id * int
| G of symbol_id * symbol_id * symbol_id
| AG of symbol_id * symbol_id * symbol_id
| CR of symbol_id * symbol_id * symbol_id * symbol_id * symbol_id
| BR of symbol_id * symbol_id * symbol_id * symbol_id 
| FP of int * symbol_id * symbol_id * symbol_id * symbol_id * symbol_id
| Int of symbol_id * symbol_id * symbol_id * symbol_id * symbol_id * symbol_id * symbol_id
| Rat of symbol_id * symbol_id * symbol_id * symbol_id * symbol_id * symbol_id * symbol_id * symbol_id
| USER of regle list
;;

type description_theorie_canonique =
  {
    mutable theories : theorie_connue list;
    mutable theory_assoc_list : (symbol_id,theorie_connue) SymbolMap.t;
    mutable regles : regle list;
    mutable regles_compilees : int * regle_etendue dnet
  }
;;


let theorie_canonique_courante = 
  {
    theories=[];
    theory_assoc_list=SymbolMap.empty;
    regles=[];
    regles_compilees= (max_int, empty_dnet())
  }
;;


let clear_current_theory () =
  
  theorie_canonique_courante.theories <- [];
  theorie_canonique_courante.theory_assoc_list <- 
     SymbolMap.empty;
  theorie_canonique_courante.regles <- [];
  theorie_canonique_courante.regles_compilees <- (max_int, empty_dnet())
;;


let string_of_theorie = function
    No_theory -> "No theory"
  | ACU(n,m)  -> "ACU(" ^ (string_of_symbol_id n) ^
                 "," ^ (string_of_symbol_id m) ^ ")"
  | AU(n,m)   -> "AU(" ^ (string_of_symbol_id n) ^
                 "," ^ (string_of_symbol_id m) ^ ")"
  | A(n)      -> "A(" ^ (string_of_symbol_id n) ^ ")"
  | ACI(n)    -> "ACI(" ^ (string_of_symbol_id n) ^ ")"
  | ACUI(n,m) -> "ACUI(" ^ (string_of_symbol_id n) ^
                  "," ^ (string_of_symbol_id m) ^ ")"
  | ACUN(n,m,k) -> "AG(" ^ (string_of_symbol_id n) ^
                  "," ^ (string_of_symbol_id m) ^
                  "," ^ (string_of_int k) ^ ")"
  | G(n,m,k)  -> "G(" ^ (string_of_symbol_id n) ^
                  "," ^ (string_of_symbol_id m) ^
                  "," ^ (string_of_symbol_id k) ^ ")"
  | AG(n,m,k) -> "AG(" ^ (string_of_symbol_id n) ^
                  "," ^ (string_of_symbol_id m) ^
                  "," ^ (string_of_symbol_id k) ^ ")"
  | CR(n,m,k,j,i) -> "CR(" ^ (string_of_symbol_id n) ^
                  "," ^ (string_of_symbol_id m) ^
                  "," ^ (string_of_symbol_id k) ^ 
		  "," ^ (string_of_symbol_id j) ^
                  "," ^ (string_of_symbol_id i) ^ ")"
  | BR(n,m,k,j) -> "BR(" ^ (string_of_symbol_id n) ^
                  "," ^ (string_of_symbol_id m) ^
                  "," ^ (string_of_symbol_id k) ^ 
		  "," ^ (string_of_symbol_id j) ^ ")"
  | FP(p,n,m,k,j,i) -> "FP(" ^ (string_of_int p) ^
                  "," ^ (string_of_symbol_id n) ^
                  "," ^ (string_of_symbol_id m) ^
                  "," ^ (string_of_symbol_id k) ^ 
		  "," ^ (string_of_symbol_id j) ^
                  "," ^ (string_of_symbol_id i) ^ ")"
  | Int(plus_z,coef,n,m,k,j,i) -> 
      "Int(" ^ (string_of_symbol_id plus_z) ^
      "," ^ (string_of_symbol_id coef) ^
      "," ^ (string_of_symbol_id n) ^
      "," ^ (string_of_symbol_id m) ^
      "," ^ (string_of_symbol_id k) ^ 
      "," ^ (string_of_symbol_id j) ^
      "," ^ (string_of_symbol_id i) ^ ")"
  | Rat(un_z,coef,irred,n,m,k,j,i) -> 
      "Rat(" ^ (string_of_symbol_id un_z) ^
      "," ^ (string_of_symbol_id coef) ^
      "," ^ (string_of_symbol_id irred) ^ 
      "," ^ (string_of_symbol_id n) ^
      "," ^ (string_of_symbol_id m) ^
      "," ^ (string_of_symbol_id k) ^ 
      "," ^ (string_of_symbol_id j) ^
      "," ^ (string_of_symbol_id i) ^ ")"
  | USER r_list -> 
      let rule = List.hd r_list 
      in
        "USER(" ^ (string_of_term rule.l) ^ 
        "->" ^ (string_of_term rule.r) ^
      	(List.fold_left
           (fun str r ->
                "; " ^ (string_of_term rule.l) ^ 
                "->" ^ (string_of_term rule.r) ^ str) 
           ")" (List.tl r_list))
;;

let print_theorie th = 
  open_hvbox 0;
  begin
  match th with
    []   -> print_string "Empty"
  | r::l -> print_string "{";
            print_string (string_of_theorie r);
  let rec print_fin_de_liste_de_theories = function
     []     -> print_string "}"
   | (r::l) -> print_string ",";
               (fun(x,y)->print_break x y) (1,0);
               print_string (string_of_theorie r);
               print_fin_de_liste_de_theories l in 
    print_fin_de_liste_de_theories l
  end;
  close_box ()
;;

let norm_theory_of f =

let rec find_theory = function
    [] -> 
      theorie_canonique_courante.theory_assoc_list <-
      SymbolMap.add f No_theory theorie_canonique_courante.theory_assoc_list;
      No_theory
  | th::thl -> 
      let th' = 
	match th with
  	    No_theory -> th
	  | A(mult) -> 
	      if f=mult then th else find_theory thl
	  | AU(mult,un) -> 
	      if f=mult or f=un then th else find_theory thl
	  | ACU(plus,zero) -> 
	      if f=plus or f=zero then th else find_theory thl
	  | ACI(plus) -> 
	      if f=plus then th else find_theory thl
	  | ACUI(plus,zero) -> 
	      if f=plus or f=zero then th else find_theory thl
	  | ACUN(plus,zero,_) -> 
	      if f=plus or f=zero then th else find_theory thl
	  | G(mult,un,inv) -> 
	      if f=mult or f=un or f=inv then th else find_theory thl
	  | AG(plus,zero,moins) -> 
	      if f=plus or f=zero or f=moins then th else find_theory thl
	  | CR(plus,zero,moins,mult,un) -> 
	      if f=plus or f=zero or f=moins or f=mult or f=un 
	      then th 
	      else find_theory thl
	  | BR(plus,zero,mult,un) -> 
	      if f=plus or f=zero or f=mult or f=un 
	      then th 
	      else find_theory thl
	  | FP(p,plus,zero,moins,mult,un) -> 
	      if f=plus or f=zero or f=moins or f=mult or f=un 
	      then th 
	      else find_theory thl
	  | Int(plus_z,coef,plus,zero,moins,mult,un) -> 
	      if f=plus_z or f=coef or f=plus or f=zero or f=moins 
							     or f=mult or f=un 
	      then th 
	      else find_theory thl
	  | Rat(un_z,coef,irred,plus,zero,moins,mult,un) -> 
	      if f=un_z or f=coef or f=irred 
				       or f=plus or f=zero 
						      or f=moins or f=mult 
								      or f=un 
	      then th 
	      else find_theory thl
	  | USER _ -> 
	      find_theory thl
      in
	theorie_canonique_courante.theory_assoc_list <-
  	SymbolMap.add f th' theorie_canonique_courante.theory_assoc_list;
	th' 

  in

  try 
    SymbolMap.find f theorie_canonique_courante.theory_assoc_list
  with 
      Not_found -> find_theory theorie_canonique_courante.theories

;;


let associativite f =
    (* (x.y).z -> x.(y.z) *)
    let x = fresh_new_var () in
    let y = next_new_var x in
    let z = next_new_var y in
    (make_basic_rule (TERM(f,[TERM(f,[(VAR x);(VAR y)]);(VAR z)]),
                      TERM(f,[(VAR x);TERM(f,[(VAR y);(VAR z)])])))
;;

let element_neutre_ac f e =
    (* x+0 -> x *)
    let x = fresh_new_var () in
    (make_basic_rule (TERM(f,[(VAR x);TERM(e,[])]),(VAR x)))
;;

let element_neutre_a_droite f e =
    (* x.e -> x *)
    let x = fresh_new_var () in
    (make_basic_rule (TERM(f,[(VAR x);TERM(e,[])]),(VAR x)))
;;

let element_neutre_a_gauche f e =
    (* e.x -> x *)
    let x = fresh_new_var () in
    (make_basic_rule (TERM(f,[TERM(e,[]);(VAR x)]),(VAR x)))
;;

let idempotence f =
    (* x+x -> x *)
    let x = fresh_new_var () in
    (make_basic_rule (TERM(f,[(VAR x);(VAR x)]),(VAR x)))
;;

let nilpotence f e n =
    (* nx -> 0 *)
    let x = fresh_new_var () in
    (make_basic_rule (TERM(f,duplique n (VAR x)),TERM(e,[])))
;;

let inverse_a_droite f e i =
    (* x.I(x) -> e *)
    let x = fresh_new_var () in
    (make_basic_rule (TERM(f,[(VAR x);TERM(i,[(VAR x)])]),
                      TERM(e,[])))
;;

let inverse_ac f e i =
    (* x+(-x) -> 0 *)
    let x = fresh_new_var () in
    (make_basic_rule (TERM(f,[(VAR x);TERM(i,[(VAR x)])]),
                      TERM(e,[])))
;;

let inverse_a_gauche f e i =
    (* I(x).x -> e *)
    let x = fresh_new_var () in
    (make_basic_rule (TERM(f,[TERM(i,[(VAR x)]);(VAR x)]),
                      TERM(e,[])))
;;

let inverse_a_droite_etendue f e i =
    (* x.(I(x).y) -> y *)
    let x = fresh_new_var () in
    let y = next_new_var x in
    (make_basic_rule (TERM(f,[(VAR x);TERM(f,[TERM(i,[(VAR x)]);(VAR y)])]),
                      (VAR y)))
;;

let inverse_a_gauche_etendue f e i =
    (* I(x).(x.y) -> y *)
    let x = fresh_new_var () in
    let y = next_new_var x in
    (make_basic_rule (TERM(f,[TERM(i,[(VAR x)]);TERM(f,[(VAR x);(VAR y)])]),
                      (VAR y)))
;;

let inverse_d_un_produit f i = 
    (* I(x.y) -> I(y).I(x) *)
    let x = fresh_new_var () in
    let y = next_new_var x in
    (make_basic_rule (TERM(i,[TERM(f,[(VAR x);(VAR y)])]),
                      TERM(f,[TERM(i,[(VAR y)]);TERM(i,[(VAR x)])])))
;;

let inverse_d_une_somme f i = 
    (* I(x+y) -> I(x)+I(x) *)
    let x = fresh_new_var () in
    let y = next_new_var x in
    (make_basic_rule (TERM(i,[TERM(f,[(VAR x);(VAR y)])]),
                      TERM(f,[TERM(i,[(VAR x)]);TERM(i,[(VAR y)])])))
;;

let inverse_du_neutre e i = 
    (* I(e) -> e *)
    (make_basic_rule (TERM(i,[TERM(e,[])]),
                      TERM(e,[])))
;;

let inverse_de_l_inverse i = 
    (* I(I(x)) -> x *)
    let x = fresh_new_var () in
    (make_basic_rule (TERM(i,[TERM(i,[(VAR x)])]),
                      (VAR x)))
;;

let distributivite_ac_ac f g =
    (* x*(y+z) -> (x*y)+(x*z) *)
    let x = fresh_new_var () in
    let y = next_new_var x in
    let z = next_new_var y in
    (make_basic_rule (TERM(f,[(VAR x);TERM(g,[(VAR y);(VAR z)])]),
                      TERM(g,[TERM(f,[(VAR x);(VAR y)]);TERM(f,[(VAR x);(VAR z)])])))
;;

let absorbant f e =
    (* x+0 -> 0 *)
    let x = fresh_new_var () in
    (make_basic_rule (TERM(f,[(VAR x);TERM(e,[])]),TERM(e,[])))
;;

let produit_d_un_oppose f i =
    (* x*(-y) -> -(x*y) *)
    let x = fresh_new_var () in
    let y = next_new_var x in
    (make_basic_rule (TERM(f,[(VAR x);TERM(i,[(VAR y)])]),
                      TERM(i,[TERM(f,[(VAR x);(VAR y)])])))
;;

(*=====================================================

Construction du systeme canonique associe a une theorie
predefinie

======================================================*)


let systeme_canonique_associe_a = function
    No_theory -> []
  | ACU(plus,zero)  -> [(element_neutre_ac plus zero)]

  | AU(mult,one)   -> [(associativite mult);
                       (element_neutre_a_droite mult one);
                       (element_neutre_a_gauche mult one)]

  | A(mult)      -> [(associativite mult)]

  | ACI(plus)    -> [(idempotence plus)]

  | ACUI(plus,zero) -> [(element_neutre_ac plus zero);
                        (idempotence plus)]

  | ACUN(plus,zero,n) -> [(element_neutre_ac plus zero);
                        (nilpotence plus zero n)]

  | G(mult,one,inv)  -> [(associativite mult);
                         (element_neutre_a_droite mult one);
                         (element_neutre_a_gauche mult one);
                         (inverse_a_droite mult one inv);
                         (inverse_a_gauche mult one inv);
                         (inverse_a_droite_etendue mult one inv);
                         (inverse_a_gauche_etendue mult one inv);
                         (inverse_d_un_produit mult inv);
                         (inverse_du_neutre one inv);
                         (inverse_de_l_inverse inv)]

  | AG(plus,zero,opp) -> [(element_neutre_ac plus zero);
                          (inverse_ac plus zero opp);
                          (inverse_du_neutre zero opp);
                          (inverse_d_une_somme plus opp);
                          (inverse_de_l_inverse opp)]

  | CR(plus,zero,opp,mult,one) -> [(element_neutre_ac plus zero);
                                   (inverse_ac plus zero opp);
                                   (inverse_du_neutre zero opp);
                                   (inverse_d_une_somme plus opp);
                                   (inverse_de_l_inverse opp);
		                   (element_neutre_ac mult one);
		                   (distributivite_ac_ac mult plus);
		                   (absorbant mult zero);
		                   (produit_d_un_oppose mult opp)]

  | BR(plus,zero,mult,one) -> [(element_neutre_ac plus zero);
                               (element_neutre_ac mult one);
                               (absorbant mult zero);
                               (let x = fresh_new_var () in
                                 (make_basic_rule 
                                             (TERM(plus,[(VAR x);(VAR x)]),
                                              TERM(zero,[]))));
                               (let x = fresh_new_var () in
                                 (make_basic_rule 
                                             (TERM(mult,[(VAR x);(VAR x)]),
                                              (VAR x))));
                               (distributivite_ac_ac mult plus)]

  | FP(p,plus,zero,opp,mult,one) -> [(element_neutre_ac plus zero);
                                     (element_neutre_ac mult one);
	  	                     (distributivite_ac_ac mult plus);
		                     (absorbant mult zero);
      	       	       	             (* -x -> (p-1)x *)
      	       	       	             (let x = fresh_new_var () in
                                     (make_basic_rule (TERM(opp,[(VAR x)]),
                                                       TERM(plus,duplique (p-1) (VAR x)))));
                                     (* px -> 0 *)
      	       	       	             (let x = fresh_new_var () in
                                     (make_basic_rule (TERM(plus,duplique p (VAR x)),
                                                       TERM(zero,[]))))]

  | Int(plus_z,coef,plus,zero,moins,mult,one) -> 
      [(element_neutre_ac plus zero);
       (inverse_ac plus zero moins);
       (inverse_du_neutre zero moins);
       (inverse_d_une_somme plus moins);
       (inverse_de_l_inverse moins);
       (element_neutre_ac mult one);
       (distributivite_ac_ac mult plus);
       (absorbant mult zero);
       (produit_d_un_oppose mult moins)]
      	
  | Rat(un_z,coef,irred,plus,zero,moins,mult,one) -> 
      [(element_neutre_ac plus zero);
       (inverse_ac plus zero moins);
       (inverse_du_neutre zero moins);
       (inverse_d_une_somme plus moins);
       (inverse_de_l_inverse moins);
       (element_neutre_ac mult one);
       (distributivite_ac_ac mult plus);
       (absorbant mult zero);
       (produit_d_un_oppose mult moins)]
      	
  | USER r_list -> r_list
;;


let construit_theorie_canonique = 
  let ajoute_theorie t = 
    theorie_canonique_courante.theories <- 
    t::theorie_canonique_courante.theories;
    theorie_canonique_courante.regles <- 
    (systeme_canonique_associe_a t) @ theorie_canonique_courante.regles
  in List.iter ajoute_theorie;;


let compile_theorie_courante () =
  theorie_canonique_courante.regles_compilees <- 
  List.fold_right
    (fun r d -> add_rule_to_dnet LEFT r d)
    theorie_canonique_courante.regles
    theorie_canonique_courante.regles_compilees;;

