(*

  Compilation of string rewriting system into an automaton and computations 
  of its normal forms.

*)


module CharSet = 
  Ordered_sets.MakePoly
    (struct type 'a t = 'a  let compare = compare end);;


(*

  \subsection{The type of automata}

  Automata considered here are deterministic.

  The states of these automata are always represented as natural
  numbers, the initial state always being 0.

  An automaton is then made of a transition table containing at [i*j] the 
  number of transitions from [i] to [j]. 
*)
 

type 'a full_automaton =
    {
      full_auto_trans : int array array;
      full_auto_accept : bool array;
      full_auto_init : bool array;
    }

(* Raised when one tries to count the number of words in an infinite 
   language.*)

exception Infinite_language

let count_words {full_auto_accept = a ; 
		 full_auto_trans = t; 
		 full_auto_init = initials} =
  let n = Array.length t in
  let reaches_final = Array.create n false in
  let reaches_final_on_cycle = Array.create n false in
  let is_open = Array.create n false in
  let rec traverse i =
    if is_open.(i) or reaches_final_on_cycle.(i) then begin
      reaches_final_on_cycle.(i) <- true
    end else begin
      reaches_final.(i) <- true;
      is_open.(i) <- true;
      for j = 0 to n-1 do if t.(j).(i) > 0 then traverse j done;
      is_open.(i) <- false
    end
  in
    for i = 0 to n-1 do if a.(i) then traverse i done;
  let count = Array.create n None in
  let rec count_from i = match count.(i) with
    | Some k ->
	k
    | None -> 
	let k = ref 0 in
	for j = 0 to n-1 do
	  if t.(i).(j) > 0 && reaches_final.(j) then begin
	    if reaches_final_on_cycle.(j) then raise Infinite_language;
	    k := !k + t.(i).(j) + count_from j
	  end
	done;
	count.(i) <- Some !k;
	!k
  in
  let total = ref 1 in
  for i = 0 to n-1 do if initials.(i) then total := !total + count_from i done;
  !total


let rec get_all_strict_sub w = 
  if Array.length w <= 1 then CharSet.empty else
    begin
      let result = ref (CharSet.empty) in
       for i = 1 to (Array.length w - 1) do
	 result := CharSet.add (Array.sub w i (Array.length w - i)) !result
       done;
       !result
    end


let does_overlap l r n = 
  assert (n<= Array.length r);
  if n > Array.length l then false else 
    let first_l = Array.length l - n in
    let first_r = 0 in
      try 
	for i = 0 to n-1 do
	  if l.(first_l+i)<>r.(first_r+i) then raise Exit
	done;
	true
      with Exit -> false

exception Found of int	  

let longest_right_overlap left_array right = 
try 
  for n = Array.length right downto 1 do 
    Array.iteri (fun i l -> if i = 0 then () else
		   if does_overlap l right n then raise (Found i))
      left_array
  done;
  assert false
with 
    Found n -> n

let count_nf sigma word_list compiled_srs =
  let word_array = List.map Array.of_list word_list in
  let sigma_list = CharSet.elements sigma in
  let states_from_constructors = CharSet.fold
				   (fun x acc -> CharSet.add [|x|] acc)
				   sigma CharSet.empty
  in
  let states_from_sub_terms = 
    List.fold_left 
      (fun acc x -> CharSet.union acc (get_all_strict_sub x))
      states_from_constructors word_array
  in
  let state_number = 1 + CharSet.cardinal states_from_sub_terms in
  let _ = Format.printf "Number of states of the normal forms automaton: %d\n" state_number in
  let states_array = Array.create state_number [||] in
  let counter = ref 1 in
  let _ = CharSet.iter 
	    (fun x -> states_array.(!counter)<- x; incr counter) 
	    states_from_sub_terms 
  in
  let automaton = 
    { full_auto_trans = Array.make_matrix state_number state_number 0;
      full_auto_accept =  Array.init state_number (fun x -> true) ;
      full_auto_init =  Array.init state_number (function x -> x=0) ;
    }
  in
    for i = 0 to state_number-1 do
      List.iter 
	(function letter -> 
	   let new_word = Array.append [|letter|] states_array.(i) in
	     if String_rewriting.is_nf_compiled compiled_srs 
	       (Array.to_list new_word)
	     then 
	       ( 
		 let best = longest_right_overlap states_array new_word in
		   automaton.full_auto_trans.(i).(best) <- 
		   automaton.full_auto_trans.(i).(best) + 1
	       )
	)
	sigma_list
    done;
    count_words automaton
