(*


\subsection{regexp datatype and simplifying constructors}

*)


module CharSet = 
  Inttagset.MakePoly
    (struct type 'a t = 'a  let tag = Hashtbl.hash end);;

  
type 'a regexp =
  | Empty
  | Epsilon
  | Char of 'a CharSet.elt
  | Charset of 'a CharSet.t      (*r cardinal at least 2 *) 
  | Star of 'a regexp
  | Alt of 'a regexp_set
  | Seq of 'a regexp * 'a regexp
and 'a regexp_set = 'a regexp list (*r naive implementation of sets *)
;;


let add e l =
  if List.mem e l then l else e::l
;;

let rec union l1 l2 = 
  match l1 with
    | [] -> l2
    | e::l -> if List.mem e l2 then union l l2 else e::union l l2 
;;

let empty = Empty;;

let epsilon = Epsilon;;

let char c = Char c;;

let star e = 
  match e with
    | Empty | Epsilon -> Epsilon
    | Star _ -> e
    | _ -> Star e


let alt e1 e2 =
  match e1,e2 with
    | Empty,_ -> e2
    | _,Empty -> e1
    | Alt(l1),Alt(l2) -> Alt(union l1 l2)
    | Alt(l1),_ -> Alt(add e2 l1)
    | _,Alt(l2) -> Alt(add e1 l2)
    | _ -> if e1=e2 then e1 else Alt([e1;e2])
;;

let seq e1 e2 =
  match e1,e2 with
    | Empty,_ -> Empty
    | _,Empty -> Empty
    | Epsilon,_ -> e2
    | _,Epsilon -> e1
    | _ -> Seq(e1,e2)

(*

\subsection{extended regexp}

*)

let some e = (seq e (star e));;

let opt e = (alt e epsilon);;


(*

  \subsection{Regexp match by run-time interpretation of regexp}

*)

let rec nullable r =
  match r with 
    | Empty -> false
    | Epsilon ->  true
    | Char _ -> false
    | Charset _ -> false
    | Star e -> true
    | Alt(l) -> List.exists nullable l
    | Seq(e1,e2) -> nullable e1 && nullable e2


 let rec residual r c =
  match r with
    | Empty -> Empty
    | Epsilon ->  Empty
    | Char a -> if a=c then Epsilon else Empty
    | Charset s -> if CharSet.mem c s then Epsilon else Empty
    | Star e -> seq (residual e c) r
    | Alt(l) -> 
	List.fold_right
	  (fun e accu -> alt (residual e c) accu)
	  l
	  Empty
    | Seq(e1,e2) -> 
	if nullable(e1) 
	then alt (seq (residual e1 c) e2) (residual e2 c) 
	else seq (residual e1 c) e2

(*

  [firstchars r] returns the set of characters that may start a word
  in the language of [r]

*)

 let rec firstchars r =
   match r with
     | Empty -> CharSet.empty
     | Epsilon ->  CharSet.empty
     | Char a -> CharSet.singleton a
     | Charset s -> s
     | Star e -> firstchars e
     | Alt(l) -> 
	 List.fold_right
	   (fun e accu ->
	      CharSet.union (firstchars e) accu)
	   l
	   CharSet.empty
     | Seq(e1,e2) -> 
	 if nullable e1
	 then CharSet.union (firstchars e1) (firstchars e2)
	 else firstchars e1

