(***************************************************************************

  This module provides usual narrowing functions on a term wrt. a set of
  rules.  The signature may or may not contain only free function symbols.

  CiME Project - Dmons research team - LRI - Universit Paris XI
  
  $Id: innermost_narrowing.ml,v 1.2 2002/07/16 16:03:47 contejea Exp $

***************************************************************************)

open Gen_terms
open Rewrite_rules
open Orderings_generalities

type 'symbol constraints = ('symbol term * 'symbol term) list


let print_conjecture sigma vars (x,(s,t),lc) =
  Format.print_string "shift var = ";
  Format.print_string (Variables.default#string_of_var x);
  Format.print_newline();
  Gen_terms.print_term sigma vars s;
  Format.print_string " = ";
  Gen_terms.print_term sigma vars t;
  Format.print_string " | {";
  List.iter 
    (function (l,r) ->
       Gen_terms.print_term sigma vars l;
       Format.print_string " = ";
       Gen_terms.print_term sigma vars r;
       Format.print_string "; ";
       Format.print_newline())
    lc;
  Format.print_string " }";
  Format.print_newline();
  Format.print_flush() 

let my_print_nar sigma vars nar =
  List.iter (print_conjecture sigma vars ) nar;;

let print_nar_term sigma vars (x,t,lc) =
  Format.print_string "shift var = ";
  Format.print_string (Variables.default#string_of_var x);
  Format.print_newline();
  Gen_terms.print_term sigma vars t;
  Format.print_string " | {";
  List.iter 
    (function (l,r) ->
       Gen_terms.print_term sigma vars l;
       Format.print_string " = ";
       Gen_terms.print_term sigma vars r;
       Format.print_string "; ";
       Format.print_newline())
    lc;
  Format.print_string " }";
  Format.print_newline();
  Format.print_flush() 

let my_print_nar1 sigma vars nar =
  List.iter (print_nar_term sigma vars ) nar;;

exception Incompatible

let merge_standard_constr sign hct c1 c2 =  
  try 
    List.sort compare
      (Unification.free_solve_constraints sign hct (List.rev_append c1 c2))
  with 
      Controle.No_solution -> raise Incompatible

let merge_ac_constr sign hct c1 c2 =  
  let c = List.sort compare (List.rev_append c1 c2) in
    if Unification.has_an_ac_solution sign hct c
    then c
    else raise Incompatible

let empty_constr () = []

(*

  [narrow_at_top sign hct ruleset shift t l] returns the terms obtained by
  narrowing the term [t] at the top position by all the rules of the
  set of rules [ruleset] which unify with [t].  The signature [sign] over
  which the term [t] and the set of rules [ruleset] are built may
  contain only free function symbols.

*)

let narrow_at_top sign hct ruleset shift t l =
  let rec rec_narrow partial_res = function
    | [] -> partial_res
    | r::s ->
	try
	  let l' = 
	    List.sort compare 
	      (Unification.free_solve_constraints sign hct 
		 (((shift_term hct shift r.lhs),t)::l)) in
	  let new_shift = 
	    Variables.shift_variable shift 
	      (Variables.var_outside_set 
		 (Variables.VarSet.union 
		    (set_of_variables r.lhs)
		    (set_of_variables r.rhs))) in
	    rec_narrow 
	      ((new_shift,(shift_term hct shift r.rhs),l') :: partial_res) 
	      s
	with
	    Controle.No_solution -> rec_narrow partial_res s in
    rec_narrow [] ruleset

(*

  [ac_narrow_at_top sign hct ruleset t l] returns the terms obtained by
  narrowing the term [t] at the top position by all the rules of the
  set of rules [ruleset] which unify with [t].  The signature [sign] over
  which the term [t] and the set of rules [ruleset] are built may
  contain free, commutative and AC function symbols.

*)

let ac_narrow_at_top sign hct ruleset shift t l =
  let rec rec_ac_narrow partial_res = function
    | [] -> partial_res
    | r::s ->
	let l' = 
	  List.sort compare (((shift_term hct shift r.lhs),t)::l) in
	  if Unification.has_an_ac_solution sign hct l' 
	  then 
	    let new_shift = 
	      Variables.shift_variable shift 
		(Variables.var_outside_set 
		   (Variables.VarSet.union 
		      (set_of_variables r.lhs)
		      (set_of_variables r.rhs))) in
	      rec_ac_narrow 
		((new_shift,(shift_term hct shift r.rhs),l') :: partial_res) s
	  else rec_ac_narrow partial_res s in
    rec_ac_narrow [] ruleset

let special_cartesian_product merge_const sign hct list_of_sets = 
  let rec rec_scp partial_res = function
    | [] -> 
	List.fold_left 
	  (fun res (shift,lt,lc) -> (shift,List.rev lt,lc) :: res)
	  [] partial_res
    | l::ll ->
	let partial_res' = 
	  List.fold_left
	    (fun acc (shift,(term : 'a term),constraints as e) -> 
	       let res_e = 
		 List.fold_left
		   (fun acc' (sh,lt,lc) -> 
		      try
			  ((Variables.max_variable shift sh),
			   (term::lt),
			   (merge_const sign hct constraints lc)) ::
			  acc'
		      with Incompatible -> acc')
		   [] partial_res in
		 res_e @ acc)
	    [] l in
	  rec_scp partial_res' ll in
    rec_scp 
      [(Variables.min_var,[],empty_constr())] 
      list_of_sets
	  
let special_cartesian_product2 f l constraints =
  match l with
    | [] -> []
    | c::l' -> 
	let rec rec_scp partial_res prev a after = 
	  match after with 
	    | [] -> 
		List.fold_left
		  (fun acc (shift,f_a,constr) -> 
		     (shift,(List.rev_append prev [f_a]),constr) :: acc)
		  partial_res (f a constraints)
	    | b::after' ->
		let partial_res' =
		  List.fold_left
		    (fun acc (shift,f_a,constr) -> 
		       (shift,(List.rev_append prev (f_a :: after)),constr) :: acc)
		    partial_res (f a constraints) in
		  rec_scp partial_res' (a::prev) b after' in
    rec_scp [] [] c l'


let rec innermost_narrow_a_term sign hct red merge_const (shift,s,lc as s_lc) =
(*i
  let _ = 
    Format.print_string "innermost_narrow_a_term called on ";
    Format.print_newline ();
    print_nar_term sign Variables.default (shift,s,lc) in
  let res = 
i*)
  match s.Hcons.node with
    | Var _ -> [s_lc]
    | Term (f,[]) ->
	let after_a_step_at_the_top = red shift s lc in
	  s_lc :: 
	    (List.fold_left 
	       (fun acc x ->
		  (innermost_narrow_a_term sign hct red merge_const x) @ acc) 
	       [] after_a_step_at_the_top)
    | Term (f,l) ->
	let list_of_results_on_args =
	  List.map 
	    (function t -> innermost_narrow_a_term sign hct red merge_const (shift,t,lc))
	    l in
	let set_of_list_of_shift_args_and_constr = 
	  special_cartesian_product merge_const sign hct list_of_results_on_args in
	let after_a_step_at_the_top = 
	  List.fold_left
	    (fun acc (shift',l',lc') -> 
	       let t' = make_term hct f l' in
		 if s.Hcons.tag = t'.Hcons.tag
		 then (red shift' t' lc') @ acc
		 else 
		   (shift',t',lc') :: ((red shift' t' lc') @ acc))
	    [] set_of_list_of_shift_args_and_constr in
	  s_lc  ::
	    (List.fold_left 
	       (fun acc x ->
		  (innermost_narrow_a_term sign hct red merge_const x) @ acc)
	       [] after_a_step_at_the_top) 
(*i
  in
  let _ = 
    Format.print_string "result = ";
    my_print_nar1 sign Variables.default res in
    res
i*)

let rec one_step_narrow_at_all_positions sign hct red unif shift s constraints =
  match s.Hcons.node with
    | Var _ -> []
    | Term (f,l) -> 
	let set_of_list_of_args =
	  special_cartesian_product2
	    (fun t constr -> one_step_narrow_at_all_positions sign hct red unif shift t constr) 
	    l
	    constraints in
	let not_at_the_top = 
	  List.fold_left
	    (fun acc (shift',l',constr') -> 
	       (shift',(make_term hct f l'),constr') :: acc)
	    [] set_of_list_of_args in
	  try
	    let at_the_top = red shift s constraints in
	      at_the_top @ not_at_the_top
	  with
	      Not_found -> not_at_the_top
(*
  [innermost_narrow_a_pair ...] is intended NOT to return the original 
  problem it was called on.
*)


let innermost_narrow_a_pair sign hct order red merge_const (shift,(s,t),l) =
(*i
  let _ = 
    Format.print_string "innermost_narrow_a_pair called on ";
    Format.print_newline ();
    print_conjecture sign Variables.default (shift,(s,t),l) in
  let res = 
i*)
  match s.Hcons.node,t.Hcons.node with
    | Var _, Var _ -> []
    | Var _, Term _ ->
	List.rev_map
	  (fun (shift',t',l') -> (shift',(s,t'),l'))
	  (one_step_narrow_at_all_positions sign hct red merge_const shift t l)
    | Term _, Var _ ->
	List.rev_map
	  (fun (shift',s',l') -> (shift',(s',t),l'))
	  (one_step_narrow_at_all_positions sign hct red merge_const shift s l)
    | Term _, Term _ ->
	match order s t with
	  | Greater_than ->
	      List.rev_map
		(fun (shift',s',l') -> (shift',(s',t),l'))
		(one_step_narrow_at_all_positions sign hct red merge_const shift s l)
	  | Less_than ->
	      List.rev_map
		(fun (shift',t',l') -> (shift',(s,t'),l'))
		(one_step_narrow_at_all_positions sign hct red merge_const shift t l)
	  | _ ->
	      List.rev_append
		(List.rev_map
		   (fun (shift',s',l') -> (shift',(s',t),l'))
		   (one_step_narrow_at_all_positions sign hct red merge_const shift s l))
		(List.rev_map
		   (fun (shift',t',l') -> (shift',(s,t'),l'))
		   (one_step_narrow_at_all_positions sign hct red merge_const shift t l))
(*i
  in
  let _ = 
    Format.print_string "result = ";
    my_print_nar sign Variables.default res in
    res
i*)

let init (s,t as s_t) =
    ((Variables.max_var_of_set
	(Variables.VarSet.union (Gen_terms.set_of_variables s)
	   (Gen_terms.set_of_variables t))),
     s_t,
     empty_constr())

let one_step_standard sign hct order dnet_for_unif x =
  innermost_narrow_a_pair sign hct order 
    (fun shift t l -> 
       let ruleset = Naive_dnet.discriminate (snd dnet_for_unif) t 
	 (*(fst dnet_for_unif)*) in
	 narrow_at_top sign hct ruleset shift t l) 
    merge_standard_constr x
   
let one_step_ac sign hct order dnet_for_unif x =
  innermost_narrow_a_pair sign hct order 
    (fun shift t l -> 
       let ruleset = Naive_dnet.discriminate (snd dnet_for_unif) t 
	 (* (fst dnet_for_unif) *) in
	 ac_narrow_at_top sign hct ruleset shift t l)
    merge_ac_constr x
