(***************************************************************************

  This module provides usual narrowing functions on a term wrt. a set of
  rules.  The signature may contain only free function symbols.

  CiME Project - Dmons research team - LRI - Universit Paris XI
  
  $Id: standard_innermost_narrowing.ml,v 1.2 2002/07/11 13:25:23 pguillon Exp $

***************************************************************************)

open Gen_terms
open Rewrite_rules

(*

  [narrow_at_top_by_a_rule sign hct rule (t,l)] returns a pair
  [(rule.rhs, l')] obtained by narrowing by the rule [rule], at the
  top position the term [t], under the unifiability constraints [l],
  ([l] being a list of equations between terms) [l'] is the new
  constraints obtained by the combination [l] and the constraint
  [t=rule.lhs] If [l'] is not satisfiable, the exception Not_found is
  raised. The signature [sign] over which the term [t] and the rule
  [rule] are built may contain only free function symbols.

*)

let narrow_at_top_by_a_rule sign hct rule (t,l) =
  try
    let l' = Unification.free_solve_constraints sign hct ((rule.lhs,t)::l) in
      (rule.rhs,l')
  with
     Controle.No_solution -> raise Not_found


(*

  [narrow_at_top sign hct ruleset t] returns the term obtained by
  rewriting the term [t] at the top position by the first rule of the
  set of rules [ruleset] which matches [t]. When there is not such a
  rule in [ruleset], the exception Not_found is raised. The signature
  [sign] over which the term [t] and the set of rules [ruleset] are
  built may contain only free function symbols.

*)

let rec narrow_at_top sign hct ruleset (t,l as t_l) = 
  match ruleset with
    | [] -> raise Not_found
    | r::s ->
	try
	  let l' = Unification.free_solve_constraints sign hct ((r.lhs,t)::l) in
	    (r.rhs,l')
	with
	    Controle.No_solution -> narrow_at_top sign hct s t_l

let special_cartesian_product list_of_lists = 
  let rec rec_scp partial_res = function
    | [] -> List.rev_map (function (lt,lc) -> (List.rev lt,lc)) partial_res
    | l::ll ->
	let partial_res' = 
	  List.fold_left 
	    (fun acc (term,constraints as e) -> 
	       let res_e = 
		 List.map 
		   (function (lt,lc) -> (term::lt),(constraints@lc)) 
		   partial_res in
		 res_e @ acc)
	    [] l in
	  rec_scp partial_res' ll in
    rec_scp [[],[]] list_of_lists
	  

let rec innermost_narrow_a_term sign hct vars red (s,lc as s_lc) =
  match s.Hcons.node with
    | Var _ -> [s_lc]
    | Term (f,l) ->
	let list_of_results_on_args =
	  List.map 
	    (function t -> innermost_narrow_a_term sign hct vars red (t,lc))
	    l in
	let list_of_list_of_args = special_cartesian_product list_of_results_on_args in
	let after_a_step_at_the_top = 
	  List.fold_left
	    (fun acc (l',lc') -> 
	       let t' = make_term hct f l' in
		 try 
		   (red t' lc') :: acc
		 with Not_found -> 
		   if s.Hcons.tag = t'.Hcons.tag
		   then acc
		   else 
		     try 
		       let lc'' = Unification.free_solve_constraints sign hct lc' in
			 (t',lc'') :: acc
		     with 
			 Controle.No_solution -> acc)
	    [] list_of_list_of_args in
	  s_lc :: 
	  (Listutils.flat_map (innermost_narrow_a_term sign hct vars red) after_a_step_at_the_top)


(*
let rec innermost_narrow_a_pair sign hct vars order red ((s,t as s_t),l as s_t__l) =
  match s.Hcons.node,t.Hcons.node with
    | Var _, Var _ -> []
    | Var _, Term(f,l) ->
	
    | Term _, Var _ ->
    | Greater_than ->
	
    | Less_than ->
    | _ ->
*)
