(***************************************************************************

  AC-compatible Recursive Path Ordering [Rubio99rta]

CiME Project - Dmons research team - LRI - Universit Paris XI

$Id: acrpo.ml,v 1.6 2002/07/11 13:26:07 pguillon Exp $

***************************************************************************)

open Orderings_generalities
open Gen_terms
open Rpo
open Variables
open Format

let number_of_cache_add = ref 0
let number_of_cache_find = ref 0
let number_of_failed_cache_find = ref 0

let acrpo_statistics () = 
  (!number_of_cache_add,!number_of_cache_find,!number_of_failed_cache_find)


(*

  [build_embterm f left t right] returns the term [f(rev
  left,t,right)] with taking care of flattening [t] if needed.

  CAUTION: [f] is assumed to be an AC symbol.

*)

let build_embterm sign ht f left t right =
  match t.Hcons.node with
    | Term(g,l) ->
	if g==f then make_sorted_term sign ht f (List.rev_append left (l@right))
	else make_sorted_term sign ht f (List.rev_append left (t::right))
    | _ -> make_sorted_term sign ht f (List.rev_append left (t::right))
;;


(*

  [(emb_no_big prec f l)] is the multiset of terms embedded in [f(l)] through
  an argument headed by a symbol not bigger than [f].

  CAUTION: [f] is assumed to be an AC symbol.

  \[ 
  EmbNoBig(f(s_1,\ldots,s_n)) =
  \{f(s_1,\ldots,tf_f(v_j),\ldots,s_n) \mid s_i = h(v_1,\ldots,v_r), h
  \not> f, j\in\{1,\ldots,r\} \}
  \]

*)


let emb_no_big sign h prec f st =
  let rec traverse_subterms left right =
    match right with
      | [] -> []
      | s::rest ->
	  match s.Hcons.node with
	    | Term(g,vl) ->
		if prec g f = Greater_than 
		then traverse_subterms (s::left) rest
		else 
		  List.fold_left
		    (fun acc v ->(build_embterm sign h f left v rest)::acc)
		    (traverse_subterms (s::left) rest)
		    vl
		    
	    | _ -> traverse_subterms (s::left) rest
  in	
  traverse_subterms [] st
;;

let no_small_head prec f st =
  List.fold_left
    (fun acc s ->
       match s.Hcons.node with
	 | Var _ -> s::acc
	 | Term(h,_) ->
	     if prec f h = Greater_than
	     then acc
	     else s::acc)
    []
    st
;;

let big_head prec f st =
  List.fold_left
    (fun acc s ->
       match s.Hcons.node with
	 | Var _ -> acc
	 | Term(h,_) ->
	     if prec h f = Greater_than
	     then s::acc
	     else acc)
    [] st
;;

let vars_count l =
  List.fold_left
    (fun (v,n) t ->
       match t.Hcons.node with
	 | Var x -> 
	     let new_v =
	       try
		 let k = VarMap.find x v in
		 (VarMap.add x (succ k) v)
	       with
		   Not_found -> VarMap.add x 1 v
	     in 
	     (new_v,succ n)
	 | Term _ -> (v,succ n))
    (VarMap.empty,0)
    l
;;

let print_vars_count my_vars (v,n) =
  printf "%d" n;
  VarMap.iter
    (fun x n -> printf "+%d%s" n (my_vars#string_of_var x))
    v
;;

let compare_var_count (v1,n1) (v2,n2) =
  let b = 
    VarMap.fold
      (fun x n2 acc ->
	 try
	   let n1 = VarMap.find x v1 in
	   if n1 >= n2 then acc else false
	 with
	     Not_found -> false)
      v2
      true
  in
  if b 
  then 
    if n1 > n2
    then Greater_than
    else 
      if n1==n2 
      then Greater_or_equivalent
      else Uncomparable
  else Uncomparable
    
;;

let acrpo sign ht (*i vars i*) prec status =
  let acrpo_cache = Cache.create 1031 in
  let get_in_cache t1 t2 =
    incr number_of_cache_find;
    Cache.int_int_find (t1.Hcons.tag,t2.Hcons.tag) acrpo_cache 

  and return_via_cache t1 t2 result =
    incr number_of_cache_add;
    Cache.add (t1.Hcons.tag,t2.Hcons.tag) result acrpo_cache ;
    result
  in

  let rec acrpo_rec t1 t2 = 
    match t1.Hcons.node,t2.Hcons.node with
	Term(f1,l1), Term(f2,l2) -> 
	  begin
	    try 
	      get_in_cache t1 t2		
	    with Not_found ->	      
	      incr number_of_failed_cache_find;
	      if exists_greater_or_equal acrpo_rec l1 t2
	      then return_via_cache t1 t2 Greater_than
	      else 
		if exists_greater_or_equal acrpo_rec l2 t1
		then return_via_cache t1 t2 Less_than
		else 
		  begin
                    match prec f1 f2 with
			Uncomparable          -> 
			return_via_cache t1 t2 Uncomparable
                    | Greater_than -> 
			if greater_than_all acrpo_rec t1 l2
			then return_via_cache t1 t2 Greater_than
			else return_via_cache t1 t2 Uncomparable
                    | Less_than -> 
			if greater_than_all acrpo_rec t2 l1
			then return_via_cache t1 t2 Less_than
			else return_via_cache t1 t2 Uncomparable
                    | Equivalent            ->
			  if sign#is_ac f1 
			  then
			    (* The important part ! *)
			    let embnobig1 = emb_no_big sign ht prec f1 l1
			    and embnobig2 = emb_no_big sign ht prec f2 l2
			    in
			    if exists_greater_or_equal acrpo_rec embnobig1 t2
			    then return_via_cache t1 t2 Greater_than
			    else
			      if exists_greater_or_equal acrpo_rec embnobig2 t1
			      then return_via_cache t1 t2 Less_than
			      else
				acrpo_main_case t1 t2 f1 f2 l1 l2 embnobig1 embnobig2
			  else
			    begin
			      match (status f1,status f2) with
				  (Multiset,Multiset)   -> 
   	       	       		    return_via_cache t1 t2 
				      (multiset_extension acrpo_rec l1 l2)
				| (Lr_lexico,Lr_lexico) ->
  	       	       		    return_via_cache t1 t2 
				      (acrpo_lexico t1 t2 l1 l2)
				| (Lr_lexico,Rl_lexico) ->
  	       	       		    return_via_cache t1 t2 
				      (acrpo_lexico t1 t2 l1 (List.rev l2))
				| (Rl_lexico,Lr_lexico) ->
  	       	       		    return_via_cache t1 t2 
				      (acrpo_lexico t1 t2 (List.rev l1) l2)
				| (Rl_lexico,Rl_lexico) ->
   	       	       		    return_via_cache t1 t2 
				      (acrpo_lexico t1 t2 (List.rev l1) 
					 (List.rev l2))
				| _ -> assert false		
			    end		  
                    | _ -> assert false
                end
	end
	
    | (Var x,Term(_,l)) -> 
	if var_occurs_in_term_list x l
	then Less_than
	else Uncomparable
    | (Term(_,l),Var x) -> 
	if var_occurs_in_term_list x l
	then Greater_than
	else Uncomparable
    | (Var x1,Var x2) -> 
     if x1==x2
     then Equivalent
     else Uncomparable
	  
  and acrpo_lexico s1 s2 l1 l2 =
    match lexicographic_extension acrpo_rec l1 l2 with
	Equivalent            -> 
          Equivalent
      | Greater_than -> 
          if greater_than_all acrpo_rec s1 l2
          then Greater_than
          else Uncomparable
      | Less_than -> 
          if greater_than_all acrpo_rec s2 l1
          then Less_than
          else Uncomparable
      | Uncomparable          ->
          Uncomparable
      | _ -> assert false (* path ordering never returns >= or <= *)
	    


  and acrpo_main_case t1 t2 f1 f2 l1 l2 embnobig1 embnobig2 =
    let nosmallhead1 = no_small_head prec f1 l1
    and nosmallhead2 = no_small_head prec f2 l2
    in
    let c= f_multiset_extension f1 f2 nosmallhead1 nosmallhead2 in
(*i
    printf "compare nosmallhead : %s@." (string_of_comparison_result c);
i*)
    match c with
      | Greater_than -> 
          if last_conditions_ok t1 t2 f1 f2 l1 l2 embnobig2
	  then Greater_than
	  else Uncomparable
      | Less_than -> 
          if last_conditions_ok t2 t1 f2 f1 l2 l1 embnobig1
	  then Less_than
	  else Uncomparable
      | Uncomparable          ->
          Uncomparable
      |	Equivalent            -> 
          if last_conditions_ok t1 t2 f1 f2 l1 l2 embnobig2
          then 
	    if last_conditions_ok t2 t1 f2 f1 l2 l1 embnobig1
	    then Equivalent
	    else Greater_than
          else
	    if last_conditions_ok t2 t1 f2 f1 l2 l1 embnobig1
	    then Less_than
	    else Uncomparable
      | _ -> assert false (* path ordering never returns >= or <= *)

      
    and last_conditions_ok t1 t2 f1 f2 l1 l2 embnobig2 =
      if greater_than_all acrpo_rec t1 embnobig2
      then 
	if (multiset_extension acrpo_rec 
	      (big_head prec f1 l1) (big_head prec f2 l2)) = Greater_than 
	then true
	else
	  let var_count1 = vars_count l1
	  and var_count2 = vars_count l2
	  in
(*i
	  printf "debug_var_count : ";
	  print_vars_count vars var_count1;
	  printf " and ";
	  print_vars_count vars var_count2;
	  printf " : %s@." 
	    (string_of_comparison_result 
	       (compare_var_count var_count1 var_count2));
i*)
	  match compare_var_count var_count1 var_count2 with
	      Greater_than -> true
	    | Greater_or_equivalent | Equivalent ->
		begin
		  match multiset_extension acrpo_rec l1 l2 
		  with 
		      Greater_than | Greater_or_equivalent | Equivalent -> true
		    | _ -> false
		end
	    | _ -> false
      else false

    and f_acrpo f s t =
      match acrpo_rec s t with
	| Greater_than ->
	    begin
	      match s.Hcons.node with
		| Term(h,_) ->
		    if prec h f = Greater_than 
		    then Greater_than
		    else 
		      begin 
			match t.Hcons.node with 
			  | Var _ -> Uncomparable
			  | Term(g,_) -> 
			      begin
				match prec h g with
				  | Greater_than 
				  | Greater_or_equivalent
				  | Equivalent -> Greater_than
				  | _ -> Uncomparable
			      end
		      end
		| _ -> assert false 
	    end
	| _ ->	Uncomparable (* other cases are not needed below *)    


    and f_multiset_extension f1 f2 l1 l2 =
      match remove_equivalent_elements acrpo_rec l1 l2 with
	  ([],[]) -> Equivalent
	| (l1,l2) -> 
	    if forall_exists_greater (f_acrpo f1) l1 l2 
            then Greater_than
            else 
	      if forall_exists_greater (f_acrpo f2) l2 l1
              then Less_than
              else Uncomparable

  in
    
    acrpo_rec





(*
let check_acrpo_conditions n prec = ()
*)



