(***************************************************************************

Polynomial term orderings : body

CiME Project - Dmons research team - LRI - Universit Paris XI

$Id: poly_interp.ml,v 1.12 2002/07/11 13:26:07 pguillon Exp $

***************************************************************************)

open Signatures;;
open Variables;;
open Gen_terms;;
open Orderings_generalities;;
open Polynomials;;
(*i
open Unix;;
open Format;;
i*)
open Num;;



(*

  [(check_poly_interp f p)] checks whether the polynomial [p] is a
  valid interpretation for the symbol [f]. Raise a [Semantical_error]
  exception if not.

  Verifications are :

    . the dimension of [p] must be the arity of [f]

    . [p] must be positive or null

    . [p] must be increasing in each variable, that is each derivatives
      must be positive

    . if [f] is commutative, [p] has to be symmetric

    . if [f] is AC, [p] has to be of the form $axy + b(x+y) + c$ where
    $b^2=b+ac$

*)



(*

  Polynomials over coefficients num of variables indexed by integers

*)

module NumRing =
  struct
    type coef = num
    let zero = Int 0
    let one = Int 1
    let is_null n = eq_num n zero
    let add = add_num
    let minus = minus_num
    let sub = sub_num
    let mult = mult_num
  end
;;

module IntOrd =
  struct
    type t = int
    let compare = (-)
  end
;;

module IntSet = Ordered_sets.Make (IntOrd);;

module IntMap = Ordered_maps.Make (IntOrd);;

module IntPolynomials =
  Polynomials.Make (NumRing) (IntOrd) (IntSet) (IntMap)


let current_signature = ref (User_signatures.from_string "");;

let current_poly_vars = ref [];;

let print_int_polynomial p =
  IntPolynomials.print_polynomial
    (fun x -> Format.print_string (string_of_num x))
    (fun n -> Format.print_string ("X"^(string_of_int n)))
    p
;;

(* 

   [(build_var_interp_of_term t)] builds a one-to-one map from the
   (term) variables of [t] to integers

   [(complete_var_interp_of_term old_interp t)] builds a one-to-one
   map from the (term) variables of [t] to integers variables by
   completing the old interpretations [old_interp]

*)



let rec complete_var_interp_of_term old_interp next_new_var t =
  match t.Hcons.node with
    Var(x)    -> 
      begin 
	try
	  VarMap.find x old_interp;
	  old_interp,next_new_var
      	with
	    Not_found -> 
	      VarMap.add x next_new_var old_interp,succ next_new_var
      end

  | Term(f,l) -> 
      List.fold_left 
	(fun (interp,next) t -> 
	   complete_var_interp_of_term interp next t)
	(old_interp,next_new_var)
	l
;;

let build_var_interp_of_term t = 
  complete_var_interp_of_term VarMap.empty 0 t;;


(* 

   Term interpretation.

   [(interpret symb_interp var_interp t)] returns the polynomial
   interpretation of term [t] with respect to the polynomial
   interpretation of function symbols given by [symb_interp] and the
   interpretation of variables given by [var_interp]

*)

let interpret sigma symb_interp var_interp t =

  let rec substituer_ac_list f l =
    IntPolynomials.substitute
      (IntMap.add 0 (substituer_term (List.hd l))
	 (IntMap.add 1 
	    (if (List.length l)>2
	     then 
	       (substituer_ac_list f (List.tl l))
	     else
	       (substituer_term (List.hd (List.tl l))))
	    IntMap.empty))
      (symb_interp f)
      
  and substituer_term u = 
    match u.Hcons.node with
      | Var(x) -> IntPolynomials.var (VarMap.find x var_interp)
      | Term(f,l) -> 
      	  if (sigma#is_ac f)
          then substituer_ac_list f l
	  else 
	    IntPolynomials.substitute
      	      (let (map,_) =
		 (List.fold_left
		    (fun (map,n) t ->
		       (IntMap.add n (substituer_term t) map,(succ n)))
		    (IntMap.empty,0)
		    l)
	       in map)
      	      (symb_interp f)

  in substituer_term t 
;;




(*

  ordering on polynomials

*)

let rec poly_sign_pos p = 
  let result = ref Greater_than
  in
    begin 
      try
      	IntPolynomials.iter
    	  (function m ->
	     if lt_num (IntPolynomials.monomial_coef m) NumRing.zero
	     then 
	       begin
		 result := Uncomparable;
		 raise Exit
	       end)
	  p;
      with Exit -> ();
    end;
    !result
;;

let rec poly_sign_neg p = 
  let result = ref Less_than
  in
    begin
      try
	IntPolynomials.iter
    	  (function m ->
	     if gt_num (IntPolynomials.monomial_coef m) NumRing.zero
	     then 
	       begin
		 result := Uncomparable;
		 raise Exit
	       end)
	  p;
      with Exit -> ();
    end;
    !result
;;

let rec poly_sign_pos_or_null p = 
let result = ref Greater_or_equivalent
  in
    begin
      try
	IntPolynomials.iter
    	  (function m ->
	     if lt_num (IntPolynomials.monomial_coef m) NumRing.zero
	     then 
	       begin
		 result := Uncomparable;
		 raise Exit
	       end)
	  p;
      with Exit -> ();
    end;
    !result
;;

let rec poly_sign_neg_or_null p = 
  let result = ref Less_or_equivalent
  in
    begin
      try
	IntPolynomials.iter
    	  (function m ->
	     if gt_num (IntPolynomials.monomial_coef m) NumRing.zero
	     then 
	       begin
		 result := Uncomparable;
		 raise Exit
	       end)
	  p;
      with Exit -> ();
    end;
    !result
;;

let rec poly_sign_null p =
  let result = ref Equivalent 
  in
    begin
      try
	IntPolynomials.iter
    	  (function m ->
	     match compare_num (IntPolynomials.monomial_coef m) NumRing.zero
	     with
		 -1 -> 
		   begin
		     result := poly_sign_neg_or_null p;
		     raise Exit
		   end
	       |  0 -> ()
	       |  1 -> 
		    begin
		      result := poly_sign_pos_or_null p;
		      raise Exit
		    end
	       |  _ -> failwith "bug in poly_sign_null")
	  p
      with Exit -> ();
    end;
    !result
;;
      	
let poly_sign p = 
  if IntPolynomials.is_null p
  then Equivalent 
  else
    let a = IntPolynomials.constant_coef p
    in
      match compare_num a NumRing.zero
      with 
	  -1 -> poly_sign_neg p
      	|  0 -> poly_sign_null p
      	|  1 -> poly_sign_pos p
      	|  _ -> failwith "bug in poly_sign"
;;
    
let poly_compare p1 p2 = 
  let p = IntPolynomials.sub p1 p2
  in 
(*
  print_int_polynomial p;
  force_newline();
*)
  poly_sign p
;;



(* 

   ordering by polynomial term interpretations 

*)

let user_time_poly = ref 0.0;;
let system_time_poly = ref 0.0;;


let poly_ord sigma symb_interp t1 t2 =
(*
  let times_before_poly = Unix.times() in
*)
  let var_interp1,next1 = build_var_interp_of_term t1 in
  let var_interp2,next2 = complete_var_interp_of_term var_interp1 next1 t2 in 
  let p1 = interpret sigma symb_interp var_interp2 t1
  and p2 = interpret sigma symb_interp var_interp2 t2 in 
(*
    if Args.options_courantes.Args.niveau_de_verbosite >=2
    then 
      begin
	force_newline();
	print_string "p1 = ";
	print_int_polynomial p1;
	force_newline();
	print_string "p2 = ";
	print_int_polynomial p2;
	force_newline();
      end;
*)
    let result = poly_compare p1 p2 in
(*
    let times_after_poly = times() in	
      user_time_poly := 
      !user_time_poly 
      +. ( times_after_poly.tms_utime -. times_before_poly.tms_utime);
      system_time_poly := 
      !system_time_poly 
      +. ( times_after_poly.tms_stime -. times_before_poly.tms_stime);
*)
      result


let poly_times ()=
  !user_time_poly,!system_time_poly
;;





(*

  (mu_translate n mu P) retourne la translation de P(X_1,...,X_n) par
  mu c-a-d P(X_1+mu,..,X_n+mu)-mu

*)


let mu_translate mu pol = 

  let num_mu = Int mu 
  in
  let mu_poly_list =
    IntSet.fold
      (fun x m ->
	 IntMap.add 
	   x 
	   (IntPolynomials.add 
	      (IntPolynomials.var x) (IntPolynomials.cte num_mu))
	   m)
      (IntPolynomials.set_of_vars pol)
      IntMap.empty
      
  in 
    IntPolynomials.add
      (IntPolynomials.substitute mu_poly_list pol)
      (IntPolynomials.cte (Int (-mu)))
;;


(*

  (check_poly_interp f p) checks whether the polynomial p is a valid
  interpretation for the symbol f. Raise a semantical_error exception
  if not.

  Verifications are :

    . the dimension of p must be the arity of f

    . p must be positive or null

    . p must be increasing in each variable, that is each derivatives
      must be positive

    . if f is commutative, p has to be symmetric

    . if f is AC, f has to be of the form axy + b(x+y) + c where b^2=b+ac 

*)


let ac_poly p = 
  if IntSet.cardinal (IntPolynomials.set_of_vars p)<>2
  then false
  else
    if IntPolynomials.partial_degree 0 p > 1
	or IntPolynomials.partial_degree 1 p > 1
    then false
    else       
      let a = ref NumRing.zero
      and b1 = ref NumRing.zero
      and b2 = ref NumRing.zero
      and c = ref NumRing.zero
      in
    	IntPolynomials.iter
	  (fun m ->
	     let vars = IntPolynomials.monomial_vars m
	     and coef = IntPolynomials.monomial_coef m
	     in
	       match ((try IntMap.find 0 vars with Not_found -> 0),
		      (try IntMap.find 1 vars with Not_found -> 0))
		 with
		     0,0 -> c := coef
		   | 1,0 -> b1 := coef
		   | 0,1 -> b2 := coef
		   | 1,1 -> a := coef
		   | _   -> failwith "bug in Poly_interp.ac_poly")
	  p;
      	(eq_num !b1 !b2) & 
      	(eq_num (mult_num !b1 !b1) (add_num !b1 (mult_num !a !c)))
;;



let check_poly_interp sigma f p = 

(* test non valide : certaines variables peuvent ne pas apparaitre 
  if arity f <> IntSet.cardinal (IntPolynomials.set_of_vars p)
  then semantical_error 
      "A polynomial interpretation must have as many variables as the arity of the symbol";
*)

  if not (is_greater_or_equal (poly_compare p IntPolynomials.zero)) 
  then failwith 
    "A polynomial interpretation must be into [mu,+oo[";
  if (sigma#is_ac f)
  then 
    if not (ac_poly p)
    then failwith 
      "This is not an AC-compatible polynomial representation" 
;;      
