(***************************************************************************

Polynomial term orderings: specification

CiME Project - Dmons research team - LRI - Universit Paris XI

$Id: poly_interp.mli,v 1.8 2002/05/27 12:07:25 contejea Exp $

***************************************************************************)

open Orderings_generalities;;
open Signatures;;
open Variables;;
open Gen_terms;;


(*

  Polynomials used in polynomial orderings are multivariate
  polynomials with arbitrary size integers as coefficients

*)

module NumRing : Polynomials.RingType with type coef = Num.num;;

module IntPolynomials : 
  Polynomials.PolynomialType
with
  module Base_ring = NumRing
and
  type variable = int;;


val current_signature : User_signatures.user_signature ref;;

(*
  
  [current_signature] is a reference onto the signature used for 
  parsing the polynomial interpretations.

*)

val current_poly_vars : string list ref;;

(*

  [current_poly_vars] is a reference onto the list of variables
  occurring in a polynomial interpretation (used in parsing).

*)


val print_int_polynomial : IntPolynomials.poly -> unit;;

(* 

   [(build_var_interp_of_term t)] builds a one-to-one map from the
   (term) variables of [t] to integers

   [(complete_var_interp_of_term old_interp t)] builds a one-to-one
   map from the (term) variables of [t] to integers variables by
   completing the old interpretations [old_interp]

*)



val complete_var_interp_of_term : 
  (var_id,int) VarMap.t -> int -> 'symbol term -> (var_id,int) VarMap.t * int;;

val build_var_interp_of_term :  'symbol term -> (var_id,int) VarMap.t * int;;

   
(*

  [(check_poly_interp sigma f p)] checks whether the polynomial [p] is
  a valid interpretation for the symbol [f]. Raise a [Failure]
  exception if not.

  Verifications are :
  \begin{itemize}
  \item the dimension of [p] must be the arity of [f] ;
  \item [p] must be positive or null ;
  \item [p] must be increasing in each variable, that is each 
      derivatives must be positive ;
  \item if [f] is commutative, [p] has to be symmetric ;
  \item if [f] is AC, [f] has to be of the form 
    $axy + b(x+y) + c$ where $b^2=b+ac$.
  \end{itemize}

*)


val check_poly_interp : 
  'symbol #signature -> 'symbol -> IntPolynomials.poly -> unit;;


(*

  [(mu_translate mu P)] returns the translation of $P(X_1,...,X_n)$ by
  $\mu$ that is $P(X_1+\mu,..,X_n+\mu)-\mu$.

*)

val mu_translate : int -> IntPolynomials.poly -> IntPolynomials.poly ;;


(*

  [(poly_ord sigma P)] is the ordering on [sigma]-term generated by [P],
  that is:
  \[
  s > t \mbox{ iff } I(s) > I(t)
  \]
  where
  \[
  I(f(t_1,...,t_n)) = P(f)(I(t_1),...,I(t_n))
  \]

*)

val poly_ord : 
  'symbol #signature 
  -> ('symbol -> IntPolynomials.poly) -> 'symbol term ordering;;

(*

[(poly_times ())] returns the pair (user time, system time) spent in
poly, in seconds

*)

val poly_times :
  unit
  -> float * float
;;

