(***************************************************************************

Dependency pairs criteria, description and specification

CiME Project - Dmons research team - LRI - Universit Paris XI

$Id: dependency_pairs_criteria.ml,v 1.15 2002/07/11 13:26:15 pguillon Exp $

***************************************************************************)

open Signatures;;
open Gen_terms;;
open Rewrite_rules;;
open Termination_constraints;;

open Format;;

(*

  [(simple_dependency_pair_criterion R)] returns the set of ordering
  existence constraints given by the simple dependency pairs
  termination criterion on [R], which is the following :

  \begin{enumerate} 

  \item let $D$ be the set of \emph{defined} symbols of [R], that is
    the set of symbols occuring at root of a lefthand side of a rule
    of [R] ;

  \item a \emph{dependency pair} is any pair of terms $(s,t)$ such
    that $s$ is the lefthand side of a rule $s\rightarrow r$ in [R]
    and $t$ is any subterm of $r$ whose root symbol is defined.

  \item a condition for [R] terminating is that there is a weakly
    monotonic ordering $\geq$ such that $l\geq r$ for all rules of [R]
    and $s>t$ for all dependency pair of [R] ;

  \item $\geq$ must be AC-compatible.

  \end{enumerate} 

*)

let compute_defined_symbols r = 
  List.fold_left
    (fun set rule -> SymbolSet.add (root_symbol rule.lhs) set)
    SymbolSet.empty
    r
;;    
    
	
let rec defined_subterms defined_symbols t = 
  match t.Hcons.node with
    | Var _ -> []
    | Term(f,l) ->
	let others = 
	  Listutils.flat_map (defined_subterms defined_symbols) l
	in
      	  if SymbolSet.mem f defined_symbols
      	  then t::others
	  else others


let dependency_pairs_of_rule defined_symbols l r =
  let all_defined_subterms = 
    defined_subterms defined_symbols r
  in
    List.map (fun t -> (l,t)) all_defined_subterms
;;


let compute_dependency_pairs defined_symbols r =
  Listutils.flat_map
    (function rule ->
       (dependency_pairs_of_rule defined_symbols rule.lhs rule.rhs)
       @(match rule.ext with
	     Some(l,r) -> dependency_pairs_of_rule defined_symbols l r
	   | None -> []))
    r
;;



let all_rules_weakly_decreasing r =
  List.fold_right
    (fun rule accu ->
       AtomicTerminationConstraintSet.add
	 {
	   left = rule.lhs;
	   right = rule.rhs;
	   constraint_op = Greater_or_equal
	 }
	 accu)
    r
    AtomicTerminationConstraintSet.empty
;;

let all_dp_strictly_decreasing dp =
  List.fold_right
    (fun (l,r) accu ->
       AtomicTerminationConstraintSet.add
	 {
	   left = l;
	   right = r;
	   constraint_op = Greater
	 }
	 accu)    
    dp
    AtomicTerminationConstraintSet.empty
;;


let simple_dependency_pair_criterion sigma r =
  let defined_symbols = compute_defined_symbols r 
  in

  let dp = compute_dependency_pairs defined_symbols r
  in
    AtomicTerminationConstraintSet.union
      (all_rules_weakly_decreasing r)
      (all_dp_strictly_decreasing dp)
;;
