(***************************************************************************

Parametric Polynomials: coefficient are themselves polynomials over
finite domain variables.

CiME Project - Dmons research team - LRI - Universit Paris XI

$Id: generic_polynomials.ml,v 1.18 2002/07/11 13:26:15 pguillon Exp $

***************************************************************************)

open Signatures;;
open Variables;;
open Gen_terms;;
open Polynomials;;
open Format;;
open Num;;

open Finite_domains;;
open Non_linear_solving;;

open Termination_constraints;;

type polyinterptype = Linear | Simple | Simple_Mixed;;

let use_degree_two_interp = ref false;;
let use_linear_interp = ref false;;

(*

  Polynomials over 
    coefficients ring : polynomials of finite domain variables
    variables : first order variables

*)

module IntOrd =
  struct
    type t = int
    let compare = (-)
  end
;;

module IntSet = Ordered_sets.Make (IntOrd);;

module IntMap = Ordered_maps.Make (IntOrd);;

module Fd_polynomials_ring =
  struct
    type coef = Fd_polynomials.poly
    let zero = Fd_polynomials.zero
    let one = Fd_polynomials.one
    let is_null = Fd_polynomials.is_null
    let add = Fd_polynomials.add
    let minus = Fd_polynomials.minus
    let sub = Fd_polynomials.sub
    let mult = Fd_polynomials.mult
  end
;;

module GenericPolynomials =
  Polynomials.Make (Fd_polynomials_ring) (IntOrd) (IntSet) (IntMap)

let print_generic_polynomial p =
  GenericPolynomials.print_polynomial
    Non_linear_solving.print_polynomial
    (fun n -> Format.print_string ("X"^(string_of_int n)))
    p
;;

(* 

   [(build_var_interp_of_term t)] builds a map from the (term)
   variables of [t] to polynomial variables 

   [(complete_var_interp_of_term old_interp t)] builds a map from the
   (term) variables of [t] to polynomial variables by completing the
   old interpretations [old_interp]

*)



let rec complete_var_interp_of_term old_interp next_new_var tt =
  match tt.Hcons.node with 
    | Var(x)    -> 
	begin 
	  try
	    VarMap.find x old_interp;
	    old_interp,next_new_var
      	  with
	      Not_found -> 
		VarMap.add x next_new_var old_interp,succ next_new_var
	end
	
    | Term(f,l) -> 
	List.fold_left 
	  (fun (interp,next) t -> 
	     complete_var_interp_of_term interp next t)
	  (old_interp,next_new_var)
	  l
;;

let build_var_interp_of_term t = 
  complete_var_interp_of_term VarMap.empty 0 t;;


(* 

   Term interpretation.

   [(interpret sigma symb_interp var_interp t)] returns the polynomial
   interpretation of term [t] over the signature [sigma] with respect
   to the polynomial interpretation of function symbols given by
   [symb_interp] and the interpretation of variables given by
   [var_interp]

*)


let substitute_verbose map p =
  begin
(*
    Format.print_string "Substitution de ";
    Format.print_newline();
    print_generic_polynomial p;
    Format.print_newline();
    Format.print_string "avec";
    Format.print_newline();
    IntMap.iter
      (fun n p -> 
        Format.print_string ("X"^(string_of_int n));
        Format.print_string " -> ";
        print_generic_polynomial p;
        Format.print_newline())
      map;
    Format.print_newline();
*)
    GenericPolynomials.substitute map p
  end
;;


let interpret sigma symb_interp var_interp t =

  let rec substituer_term tt =
    match tt.Hcons.node with

      | Var(x)    -> 
	  GenericPolynomials.var (VarMap.find x var_interp)
	    
      | Term(f,l) -> 
      	  if (sigma#is_ac f)
          then 
  (*
	  let accu = ref (substituer_term (List.hd l))
	  and interp_f = symb_interp f
	  in
	  List.iter
	    (function t ->
	      let interp_t = substituer_term t
	      in
	      let map = IntMap.add 0 !accu (IntMap.add 1 interp_t IntMap.empty)
    	      in
      	      accu := substitute_verbose map interp_f)
	    (List.tl l);
	  !accu
*)

	    let interp_f = symb_interp f
	    in
	    let rec build_poly = function
	      | [] -> 
		  failwith "bug in Generic_polynomials.build_poly (AC symbol with no argument)"
	      | [t] -> substituer_term t
	      | t::l ->
	    	  let interp_t = substituer_term t
	    	  and interp_l = build_poly l
	    	  in
	    	  let map = 
		    IntMap.add 0 interp_t
		      (IntMap.add 1 interp_l IntMap.empty)
    	    	  in
      	    	    substitute_verbose map interp_f
		  in
		    build_poly l
		      
	  else 
	    
      	    let (map,_) =
	      List.fold_left
		(fun (map,n) t ->
		   (IntMap.add n (substituer_term t) map,(succ n)))
		(IntMap.empty,0)
		l
	    in
	      
	      substitute_verbose map (symb_interp f)
		
  in substituer_term t 
;;


let rec print_var_list_end n m =
  if n=m
  then print_string ")"
  else
    begin
      print_string ",X";
      print_int n;
      print_var_list_end (succ n) m
    end;;
	
let print_interp 
  (sigma : 'a #Signatures.signature) 
  (interp :('a,GenericPolynomials.poly) Signatures.SymbolMap.t ) =

  open_hovbox 0;
  SymbolMap.iter
    (fun f p ->
       print_string "[";
       print_string (sigma#string_of_symbol f);
       print_string "]";
       let a = sigma#arity f
       in
	 if a >= 1
	 then 
	   begin
	     print_string "(X0";
	     print_var_list_end 1 a
	   end;
	 print_string " = ";
	 print_generic_polynomial p;
	 print_string ";";
	 force_newline())
    interp;
  close_box()
;;




(*

  [(specialize_polynomial p s)] returns the polynomial obtained by
  instantiation of finite domain variables in coefficients of [p] by
  their value given by the map [s]

*)

let specialize_polynomial pol s =
  GenericPolynomials.fold
  (fun m p ->
     let coef = GenericPolynomials.monomial_coef m
     and vars = GenericPolynomials.monomial_vars m
     in
     let new_coef = 
       Fd_polynomials.eval 
       (fun a -> try Fd_polynomials.Var_map.find a s 
	         with Not_found -> Fd_polynomials.Base_ring.zero)
       coef 
     in
       if Fd_polynomials.Base_ring.is_null new_coef 
       then p
       else 
	 GenericPolynomials.add 
	   (GenericPolynomials.poly_of_monomial 
	      (GenericPolynomials.make_monomial 
		 (Fd_polynomials.cte new_coef) vars))
	   p
  )
  pol
  GenericPolynomials.zero
;;









let rec linear_interp ar =
  if ar=0
  then 
    let coef = Fd_polynomials.var (Finite_domains.new_fd_var_id ())
    in
      (GenericPolynomials.cte coef,[])
  else
    let p1,constr_mon1 = linear_interp (pred ar)
    in
    let coef = Fd_polynomials.var (Finite_domains.new_fd_var_id ())
    in
    let p = GenericPolynomials.add
	      (GenericPolynomials.mult 
		 (GenericPolynomials.cte coef) 
		 (GenericPolynomials.var (pred ar)))
	      p1
    and constr_monotonic = 
      {
	expr = Fd_polynomials.sub coef Fd_polynomials.one;  
	cond = Pos_or_nul
      }
    in p,constr_monotonic::constr_mon1
	
;;


let rec interp_of_degree_one ar =
  if ar=0
  then 
    (GenericPolynomials.cte 
       (Fd_polynomials.var (Finite_domains.new_fd_var_id ())),[])
  else
    let p1,constr_mon1 = interp_of_degree_one (pred ar)
    and p2,constr_mon2 = interp_of_degree_one (pred ar)
    in
    let p = GenericPolynomials.add
	      (GenericPolynomials.mult 
		 p1 (GenericPolynomials.var (pred ar)))
	      p2
    and constr_monotonic = 
      {
	expr = Fd_polynomials.sub
		 (GenericPolynomials.constant_coef p1)
		 Fd_polynomials.one;  
	cond = Pos_or_nul
      }
    in p,constr_monotonic::constr_mon2
	
;;


(*

  [(compute_simple_mixed n)] returns a simple-mixed polynomial of
  arity [n] and new generic coefficients

*)

  let compute_simple_mixed n =
    
      match n with
	  0 -> 
	    let p = GenericPolynomials.cte 
		      (Fd_polynomials.var (Finite_domains.new_fd_var_id ()))
	    in (p,[])
      	| 1 ->
	    if !use_degree_two_interp
	    then
	  (* interpretation of degree 2 for unary symbols *)
	    let a = Fd_polynomials.var (Finite_domains.new_fd_var_id ())
	    and b = Fd_polynomials.var (Finite_domains.new_fd_var_id ())
	    and c = Fd_polynomials.var (Finite_domains.new_fd_var_id ())
	    and x = GenericPolynomials.var 0
	    in
	    let p = GenericPolynomials.add
	      	      (GenericPolynomials.mult
			 (GenericPolynomials.cte a)
			 (GenericPolynomials.power x 2))
	      	      (GenericPolynomials.add
			 (GenericPolynomials.mult (GenericPolynomials.cte b) x)
			 (GenericPolynomials.cte c))
	    and constr_monotonic = 
	      {
	    	expr = Fd_polynomials.sub b Fd_polynomials.one;  
	    	cond = Pos_or_nul
	      }
    	    in (p,[constr_monotonic])
	    else
	      interp_of_degree_one 1
	| _ -> 
	    if !use_linear_interp
	    then linear_interp n
	    else interp_of_degree_one n
;;


(*

  [(compute_commutative_simple_mixed ())] returns a commutative
  polynomial with new generic coefficients

*)

  let compute_commutative_simple_mixed () =
    
    let a = Fd_polynomials.var (Finite_domains.new_fd_var_id ())
    and b = Fd_polynomials.var (Finite_domains.new_fd_var_id ())
    and c = Fd_polynomials.var (Finite_domains.new_fd_var_id ())
    and x = GenericPolynomials.var 0
    and y = GenericPolynomials.var 1
    in
    let p = GenericPolynomials.add
	      (GenericPolynomials.mult
		 (GenericPolynomials.cte a)
		 (GenericPolynomials.mult x y))
	      (GenericPolynomials.add
		 (GenericPolynomials.mult (GenericPolynomials.cte b) x)
		 (GenericPolynomials.add
		    (GenericPolynomials.mult (GenericPolynomials.cte b) y)
		    (GenericPolynomials.cte c)))	
    and constr_monotonic = 
      {
	expr = Fd_polynomials.sub b Fd_polynomials.one;		 
	cond = Pos_or_nul
      }
    in p,constr_monotonic

;;


(*

  [(compute_ac_simple_mixed ())] returns a pair [p,c] where [p] is a
  commutative polynomial and [c] is the diophantine constraint
  required by the Ben Cherifa-Lescanne criterion (that is
  p(x,y)=axy+b(x+y)+c where b^2=b+ac)

*)

  let compute_ac_simple_mixed () =
    
    let a = Fd_polynomials.var (Finite_domains.new_fd_var_id ())
    and b = Fd_polynomials.var (Finite_domains.new_fd_var_id ())
    and c = Fd_polynomials.var (Finite_domains.new_fd_var_id ())
    and x = GenericPolynomials.var 0
    and y = GenericPolynomials.var 1
    in
    let p = GenericPolynomials.add
	      (GenericPolynomials.mult
		 (GenericPolynomials.cte a)
		 (GenericPolynomials.mult x y))
	      (GenericPolynomials.add
		 (GenericPolynomials.mult (GenericPolynomials.cte b) x)
		 (GenericPolynomials.add
		    (GenericPolynomials.mult (GenericPolynomials.cte b) y)
		    (GenericPolynomials.cte c)))	
	 
    and constr_ac = 
      {
	expr = Fd_polynomials.sub
		 (Fd_polynomials.add (Fd_polynomials.mult a c) b)
		 (Fd_polynomials.power b 2);
	cond = Nul
      }

    and constr_monotonic = 
      {
	expr = Fd_polynomials.sub b Fd_polynomials.one;		 
	cond = Pos_or_nul
      }

    in p,constr_ac,constr_monotonic
;;

(*

  [(build_dioph_constraints p cond)] returns the list of diophantine
  constraints sufficient for [p] satisfying [cond], where [cond]
  either =0 or >=0.

  This sufficient condition is simply that the same condition is
  satisfied for each coefficients.

*)

let build_dioph_constraints p cond =
  GenericPolynomials.fold 
    (fun m l -> 
       { expr = GenericPolynomials.monomial_coef m ;
	 cond = cond ;
       }::l)
    p
    []
;;





(* 

   [(build_var_interp_of_term t)] builds a one-to-one map from the
   (term) variables of [t] to integers

   [(complete_var_interp_of_term old_interp t)] builds a one-to-one
   map from the (term) variables of [t] to integers variables by
   completing the old interpretations [old_interp]

*)



let rec complete_var_interp_of_term old_interp next_new_var tt =
  match tt.Hcons.node with
    | Var(x)    -> 
	begin 
	  try
	    VarMap.find x old_interp;
	    old_interp,next_new_var
      	  with
	      Not_found -> 
		VarMap.add x next_new_var old_interp,succ next_new_var
	end
	
    | Term(f,l) -> 
	List.fold_left 
	  (fun (interp,next) t -> 
	     complete_var_interp_of_term interp next t)
	  (old_interp,next_new_var)
	  l
;;

let build_var_interp_of_term t = 
  complete_var_interp_of_term VarMap.empty 0 t;;


(*

  [(build_decrease_constraints sigma interp c)] builds the list of
  constraints obtained by interpreting constraints [c] by the
  polynomial interpretation [interp]

*)


let build_decrease_constraints sigma interp constr =
  
  let finterp f = SymbolMap.find f interp
  in
    AtomicTerminationConstraintSet.fold
      (fun c accu ->
	 let var_interp1,next1 = 
	   build_var_interp_of_term c.left
	 in
	 let var_interp2,next2 = 
	   complete_var_interp_of_term var_interp1 next1 c.right
	 in 
	 let p = GenericPolynomials.sub
		   (interpret sigma finterp var_interp2 c.left)
		   (interpret sigma finterp var_interp2 c.right)
	 in
	 let cres =
	   match c.constraint_op with
	       Equal -> build_dioph_constraints p Nul
	     | Greater -> 
		 build_dioph_constraints 
		   (GenericPolynomials.sub p GenericPolynomials.one) 
		   Pos_or_nul
	     | Greater_or_equal -> build_dioph_constraints p Pos_or_nul
	 in cres@accu)
      constr
      []
;;


(*

  [(polynomial_constraints_of_termination_constraints c)]
  returns a pair (interp,d) where interp is a set of generic
  polynomial interpretations of symbols in [c] and [d] is the list of
  diophantine constraints over variables of [interp] implied by [c]

*)


let polynomial_constraints_of_termination_constraints 
  (sigma : 'a #Signatures.signature) 
  (initial_interp : ('a, GenericPolynomials.poly) Signatures.SymbolMap.t )
  constraints =

  let all_symbols = set_of_symbols constraints
  in
  let simple_mixed_interp,ac_constraints,monotonicity_constraints =
    SymbolSet.fold
      (fun f (interp,ac,monotonic) -> 
	 try
	   let _ = SymbolMap.find f initial_interp
	   in (interp,ac,monotonic)
	 with Not_found ->
	   if sigma#is_ac f
	   then
	     let p,c_ac,c_monotonic = compute_ac_simple_mixed ()
	     in (SymbolMap.add f p interp,c_ac::ac,c_monotonic::monotonic)
	   else
	     if sigma#is_commutative f
	     then
	       let p,c_monotonic = compute_commutative_simple_mixed ()
	       in (SymbolMap.add f p interp,ac,c_monotonic::monotonic)
	     else
	       let p,c_monotonic = compute_simple_mixed (sigma#arity f)
	       in (SymbolMap.add f p interp,ac,c_monotonic@monotonic))
      all_symbols
      (initial_interp,[],[])
  in
    (simple_mixed_interp,
     ac_constraints,
     monotonicity_constraints,
     build_decrease_constraints sigma simple_mixed_interp constraints)
;;    
    



let build_new_interp interp sol =
  SymbolMap.fold
  (fun f p i ->
     SymbolMap.add 
       f (specialize_polynomial p sol) i)
  interp
  SymbolMap.empty
;;



let poly_solve_termination_constraint bound sigma c =
  let interp,constr_ac,constr_mon,constr_dp = 
    polynomial_constraints_of_termination_constraints
      sigma SymbolMap.empty c
  in
  let poly_constr = constr_ac@constr_mon@constr_dp
  in
  try
    let sol =
      Non_linear_solving.non_linear_solve bound
	poly_constr
    in
    let new_interp = build_new_interp interp sol
    in
    print_interp sigma new_interp      
  with 
      Finite_domains.No_solution -> 
	printf "No polynomial interpretation found in [0;%s]@." 
	  (Numbers.to_string bound)
;;


	








(*

  ordering on polynomials

*)

(*

let rec poly_sign_pos p = 
  let result = ref Greater
  in
    begin 
      try
      	GenericPolynomials.iter
    	  (function m ->
	     if lt_num (GenericPolynomials.monomial_coef m) NumRing.zero
	     then 
	       begin
		 result := Uncomparable;
		 raise Exit
	       end)
	  p;
      with Exit -> ();
    end;
    !result
;;

let rec poly_sign_neg p = 
  let result = ref Less_than
  in
    begin
      try
	GenericPolynomials.iter
    	  (function m ->
	     if gt_num (GenericPolynomials.monomial_coef m) NumRing.zero
	     then 
	       begin
		 result := Uncomparable;
		 raise Exit
	       end)
	  p;
      with Exit -> ();
    end;
    !result
;;

let rec poly_sign_pos_or_null p = 
let result = ref Greater_or_equivalent
  in
    begin
      try
	GenericPolynomials.iter
    	  (function m ->
	     if lt_num (GenericPolynomials.monomial_coef m) NumRing.zero
	     then 
	       begin
		 result := Uncomparable;
		 raise Exit
	       end)
	  p;
      with Exit -> ();
    end;
    !result
;;

let rec poly_sign_neg_or_null p = 
  let result = ref Less_or_equivalent
  in
    begin
      try
	GenericPolynomials.iter
    	  (function m ->
	     if gt_num (GenericPolynomials.monomial_coef m) NumRing.zero
	     then 
	       begin
		 result := Uncomparable;
		 raise Exit
	       end)
	  p;
      with Exit -> ();
    end;
    !result
;;

let rec poly_sign_null p =
  let result = ref Equivalent 
  in
    begin
      try
	GenericPolynomials.iter
    	  (function m ->
	     match compare_num (GenericPolynomials.monomial_coef m) NumRing.zero
	     with
		 -1 -> 
		   begin
		     result := poly_sign_neg_or_null p;
		     raise Exit
		   end
	       |  0 -> ()
	       |  1 -> 
		    begin
		      result := poly_sign_pos_or_null p;
		      raise Exit
		    end
	       |  _ -> failwith "bug in poly_sign_null")
	  p
      with Exit -> ();
    end;
    !result
;;
      	
let poly_sign p = 
  if GenericPolynomials.is_null p
  then Equivalent 
  else
    let a = GenericPolynomials.constant_coef p
    in
      match compare_num a NumRing.zero
      with 
	  -1 -> poly_sign_neg p
      	|  0 -> poly_sign_null p
      	|  1 -> poly_sign_pos p
      	|  _ -> failwith "bug in poly_sign"
;;
    
let poly_compare p1 p2 = 
  let p = GenericPolynomials.sub p1 p2
  in 
(*
  print_int_polynomial p;
  force_newline();
*)
  poly_sign p
;;

*)


(* 

   ordering by polynomial term interpretations 

*)

open Unix;;

let user_time_poly = ref 0.0;;
let system_time_poly = ref 0.0;;


let poly_ord sigma symb_interp (t1, t2) =
  let times_before_poly = times()
  in
  let var_interp1,next1 = build_var_interp_of_term t1
  in
  let var_interp2,next2 = complete_var_interp_of_term var_interp1 next1 t2
  in 
  let p1 = interpret sigma symb_interp var_interp2 t1
  and p2 = interpret sigma symb_interp var_interp2 t2
  in 
(*
  if Args.options_courantes.Args.niveau_de_verbosite >=2
  then 
    begin
      force_newline();
      print_string "p1 = ";
      print_generic_polynomial p1;
      force_newline();
      print_string "p2 = ";
      print_generic_polynomial p2;
      force_newline();
    end;
*)
  let result = () (* poly_compare p1 p2 *)
  in
  let times_after_poly = times()
  in	
  user_time_poly := 
    !user_time_poly 
    +. ( times_after_poly.tms_utime -. times_before_poly.tms_utime);
  system_time_poly := 
    !system_time_poly 
    +. ( times_after_poly.tms_stime -. times_before_poly.tms_stime);
  result
;;

let poly_times ()=
  !user_time_poly,!system_time_poly
;;






(*

  (check_poly_interp f p) checks whether the polynomial p is a valid
  interpretation for the symbol f. Raise a semantical_error exception
  if not.

  Verifications are :

    . the dimension of p must be the arity of f

    . p must be positive or null

    . p must be increasing in each variable, that is each derivatives
      must be positive

    . if f is commutative, p has to be symmetric

    . if f is AC, f has to be of the form axy + b(x+y) + c where b^2=b+ac 

*)

(*
let ac_poly p = 
  if IntSet.cardinal (GenericPolynomials.set_of_vars p)<>2
  then false
  else
    if GenericPolynomials.partial_degree 0 p > 1
	or GenericPolynomials.partial_degree 1 p > 1
    then false
    else       
      let a = ref NumRing.zero
      and b1 = ref NumRing.zero
      and b2 = ref NumRing.zero
      and c = ref NumRing.zero
      in
    	GenericPolynomials.iter
	  (fun m ->
	     let vars = GenericPolynomials.monomial_vars m
	     and coef = GenericPolynomials.monomial_coef m
	     in
	       match ((try IntMap.find 0 vars with Not_found -> 0),
		      (try IntMap.find 1 vars with Not_found -> 0))
		 with
		     0,0 -> c := coef
		   | 1,0 -> b1 := coef
		   | 0,1 -> b2 := coef
		   | 1,1 -> a := coef
		   | _   -> failwith "bug in Poly_interp.ac_poly")
	  p;
      	(eq_num !b1 !b2) & 
      	(eq_num (mult_num !b1 !b1) (add_num !b1 (mult_num !a !c)))
;;

*)


(*

let check_poly_interp f p = 

  if arity f <> IntSet.cardinal (GenericPolynomials.set_of_vars p)
  then semantical_error 
      "A polynomial interpretation must have as many variables as the arity of the symbol";

  if (match (poly_compare p GenericPolynomials.zero) with
	  Equivalent | Greater | Greater_or_equivalent -> false
  	|  _ -> true)
  then semantical_error 
    "A polynomial interpretation must be into [mu,+oo[";
  if (is_ac f)
  then 
    if not (ac_poly p)
    then semantical_error 
      "This is not an AC-compatible polynomial representation" 
;;      
*)
