(***************************************************************************

A module implementing graphs with a strongly connected parts operation.

CiME Project - Dmons research team - LRI - Universit Paris XI

$Id: graph.ml,v 1.2 2002/06/25 08:32:44 urbain Exp $

***************************************************************************)

(*
module type OrderedType =
  sig
    type t
    val compare: t -> t -> int
  end
;;
*)

module type G =
  sig
    type 'a elt
          (* The type of the set elements. *)
    type 'a t
          (* The abstract type of graphs. *)
    
    exception NodeNotInGraph
    exception NodeAlreadyInGraph

    val empty : 'a t

    val add_node : 'a t -> 'a elt -> 'a t
    val is_node : 'a t -> 'a elt -> bool

    val add_edge : 'a t -> 'a elt -> 'a elt -> 'a t
    val is_edge : 'a t -> 'a elt -> 'a elt -> bool

    val del_node : 'a t -> 'a elt -> 'a t
    val del_edge : 'a t -> 'a elt -> 'a elt -> 'a t

    val node_list : 'a t -> 'a elt list
    val edge_list : 'a t -> ('a elt * 'a elt) list
    val neighbours_list : 'a t -> 'a elt -> 'a elt list

    val iter_node : 'a t -> ('a elt -> unit) -> unit
    val iter_edge : 'a t -> ('a elt -> 'a elt -> unit) -> unit

    val fold : ('a elt -> 'b -> 'b) -> 'b -> 'a t -> 'b

    val transpose : 'a t ->  'a t
    val scc :  'a t -> 'a elt list list
(*i    val scc_set :  'a t -> 'a EltSet.t list i*)
    val scc_clean : 'a t -> 'a elt list list
(*i    val scc_clean_set :'a t -> 'a EltSet.t list i*)
    val build_subgraph : 'a t -> 'a elt list -> 'a t
  end

module Make(Ord: Ordered_types.OrderedPolyType) =
  struct
    type 'a elt = 'a Ord.t

    module EltMap = Ordered_maps.MakePoly(Ord)

    module EltSet = Ordered_sets.MakePoly(Ord)

    type  'a t = ('a, 'a EltSet.t) EltMap.t 

    exception NodeAlreadyInGraph
    exception NodeNotInGraph

    let empty = EltMap.empty

    let add_node g n =
      if EltMap.mem n g then
	raise NodeAlreadyInGraph
      else
	EltMap.add n EltSet.empty g

    let is_node g n =
      EltMap.mem n g

    let add_edge g x y =
      if is_node g x & is_node g y then
	let vx = EltMap.find x g in
	  EltMap.add x (EltSet.add y vx) g
      else
	raise NodeNotInGraph

    let is_edge g x y =
      EltSet.mem y (EltMap.find x g)

    let del_node g x =
      EltMap.map (fun v -> EltSet.remove x v) (EltMap.remove x g)

    let del_edge g x y =
      EltMap.add x (EltSet.remove y (EltMap.find x g)) g

    let node_list g =
      EltMap.fold (fun x _ l -> x::l) g []

    let edge_list g =
      EltMap.fold (fun x vx l -> EltSet.fold (fun y l -> (x,y)::l) vx l) g []

    let neighbours_list g x =
      EltSet.fold (fun y l -> y::l) (EltMap.find x g) []
	
    let iter_node g f =
      EltMap.iter (fun n _ -> f n) g

    let iter_edge g f =
      EltMap.iter (fun x -> EltSet.iter (f x)) g

    let fold f v0 g =
      EltMap.fold (fun x _ v -> f x v) g v0

    let transpose graph = 
      EltMap.fold 
	(fun k d g -> EltSet.fold (fun elem g -> add_edge g elem k) d g) 
	graph 
	(fold (fun e g -> add_node g e) empty graph)


(* Depth first search in graphs *)

(* [dfs_visit g u (d,t)] performs a depth first search in graph [g]
starting with vertex [u] with [d] as set of non-white nodes and [t] as
timestamp list for completed nodes. It returns a couple [(dv,ts)] with
the new set [dv] of visited nodes and updated time stamps [ts]. *)

    let rec dfs_visit g  u ((dejavu, timestamp) as cpl) =
      if EltSet.mem u dejavu 
      then cpl
      else 
	let (a,b) = EltSet.fold 
		      (dfs_visit g) 
		      (EltMap.find u g) 
		      (EltSet.add u dejavu, timestamp) 
	in (a, u::b)
;;
    let rec dfs_visit_set g  u ((dejavu, timestamp) as cpl) =
      if EltSet.mem u dejavu 
      then cpl
      else 
	let (a,b) = EltSet.fold 
		      (dfs_visit_set g) 
		      (EltMap.find u g) 
		      (EltSet.add u dejavu, timestamp) 
	in (a, EltSet.add u b)
;;
    
    
(* [dfs_time g] returns a sequence [a_1,...a_n] of blackened nodes
such that [a_n] was the first black node of graph [g] in the performed
DFS. *)

    let dfs_time g = 
      snd
	(EltMap.fold 
	   (fun k _ acc -> dfs_visit g k acc)
	   g
	   (EltSet.empty, []))
;;



(* [dfs_forest g [n_1,...]] returns a sequence [t_1,...t_n] of
predecessor trees computed while performing DFS in graph [g]. Nodes
are considered in order: [n_1] then [n_2], etc.*)

    let dfs_forest g l = 
      snd
	(List.fold_left
	   (fun (dv,forest) k -> 
	      let (new_dv,new_tree) = dfs_visit g k (dv,[])
	      in (new_dv, 
		  (match new_tree with 
		       [] -> forest
		     | a::l -> new_tree::forest))) 
	   (EltSet.empty, [])
	   l)
;;

    let dfs_forest_set g l = 
      snd
	(List.fold_left
	   (fun (dv,forest) k -> 
	      let (new_dv,new_tree) = dfs_visit_set g k (dv,EltSet.empty)
	      in (new_dv, 
		  (if new_tree =  EltSet.empty then forest
		   else new_tree::forest))) 
	   (EltSet.empty, [])
	   l)
;;

(* [scc g] returns a list of the strongly connected components of graph [g]
as lists of vertices. We use Tarjan 1972. *)

let scc g = dfs_forest (transpose g) (dfs_time g)
;;

(* [scc_set g] returns a list of the strongly connected components of graph [g]
as sets of vertices. We use Tarjan 1972. *)

let scc_set g = dfs_forest_set (transpose g) (dfs_time g)
;;

(* [scc_clean g] returns a list of the strongly connected components
 of graph [g] which are not reduced to a single node with no edge, as
 lists of vertices. *)

let scc_clean g = List.fold_left 
		    (fun acc c ->
		     ( match c with 
			  [a] -> 
			    if List.mem a (neighbours_list g a) 
			    then c::acc
			    else acc
			| _ -> c::acc))
		    []
		    (scc g)

(* [build_subgraph g l] returns the subgraph of [g] restricted to
   vertices in list [l]. *)    

let build_subgraph g l = 
  List.fold_left
    (fun acc vfrom -> 
       let acc2 = 
	 if is_node acc vfrom then acc 
	 else (add_node acc vfrom) in	     
	 List.fold_left (fun acc2 vgoal -> 
			   if List.mem vgoal l then 
			     try add_edge acc2 vfrom vgoal
			     with NodeNotInGraph -> 
			       add_edge (add_node acc2 vgoal) vfrom vgoal
			   else acc2)
	   acc2
	   (neighbours_list g vfrom)
    )
    empty
    l
;;

let build_subgraph_from_set g s = 
  EltSet.fold
    (fun vfrom acc -> 
       let acc2 = 
	 if is_node acc vfrom then acc 
	 else (add_node acc vfrom) in	     
	 List.fold_left (fun acc2 vgoal -> 
			   if EltSet.mem vgoal s then 
			     try add_edge acc2 vfrom vgoal
			     with NodeNotInGraph -> 
			       add_edge (add_node acc2 vgoal) vfrom vgoal
			   else acc2)
	   acc2
	   (neighbours_list g vfrom)
    )
    s
    empty
;;



(* [scc_graph g] returns a list of the strongly connected components
    of graph [g] as subgraphs of [g]. We use Tarjan 1972. *)

let scc_graph g = List.map (build_subgraph g) (scc g)


(* [scc_graph_clean g] returns a list of the strongly connected components
    of graph [g] which are not reduced to a single node with no edge,
    as subgraphs of [g]. *) 

let scc_graph_clean g = 
  List.fold_left
    (fun acc n_list -> let g = (build_subgraph g n_list) in 
       match node_list g with 
	 | [one] -> if edge_list g == [] then acc else g::acc 
	 | _ -> g::acc) 
    [] 
    (scc g)

end
