(***************************************************************************

Splitting htrs in minimal modules

CiME Project - Dmons research team - LRI - Universit Paris XI

$Id: minimal_split.ml,v 1.11 2002/07/11 13:26:15 pguillon Exp $

***************************************************************************)

open Signatures;;
open Variables;;
open Gen_terms;;
open User_terms;;
open Rewrite_rules;;
open Format;;
open Hierarchical_trs;;
open Graph;;

open Signature_syntax;; (* Part II *)


module Symgraph = Graph.Make(SymbolOrd)
;;

let doo_bee graph root set =
  Signatures.SymbolSet.fold
    (fun elem acc -> 
       Symgraph.add_edge 
	 (if Symgraph.is_node acc elem 
	  then acc 
	  else Symgraph.add_node acc elem)
	 root elem)
    set
    graph
;;

let g_from_rule graph rule =
  let lhs = rule.lhs in
  let root = Gen_terms.root_symbol lhs in
    doo_bee 
      (doo_bee    
	 (if Symgraph.is_node graph root
	  then graph
	  else Symgraph.add_node graph root)
	 root 
	 (Gen_terms.set_of_symbols rule.lhs))
      root 
      (Gen_terms.set_of_symbols rule.rhs)
;;      


(* [graph_from_a_list l] returns a dependency graph of symbols of
    rules in [l] *)

let graph_from_a_list rule_list =
  List.fold_left 
    g_from_rule
    Symgraph.empty
    rule_list
;;


(* [packs_from_a_list l] returns a list of \emph{minimal} packs of
    symbols of rules in [l] according to the hierarchical dependencies. *) 

let packs_from_a_list rule_list =
  Symgraph.scc (graph_from_a_list rule_list)
;;


(* Printing packs. *)

let print_pack sgn pack =
  List.iter 
    (fun x -> 
       begin 
	 print_string ((sgn#string_of_symbol x)^" ; ")
       end ) 
    pack
;;


(* [rules_from_root sym l] returns a pair [(l1,l2)] of lists of rules
   such that [l1] contains rules of [l] the root symbol of which is
   [sym] and [l2] being the list of other rules of [l]. *)

let rules_from_root sym rule_list =
  List.fold_left
    (fun (rules_of_sym, rules_left) elem ->
       if Gen_terms.root_symbol elem.lhs = sym
       then (elem::rules_of_sym, rules_left)
       else (rules_of_sym, elem::rules_left)
    )
    ([],[])
    rule_list
;;

let rules_from_roots_in_packs pack rule_list =
  List.fold_left
    (fun (rules,l2) elem -> 
       let (used,left) = rules_from_root elem l2
       in
	 (used@rules,left))
    ([],rule_list)
    pack
;;


(* [packs_and_rules_from_packs rl pl] returs a list of pairs
   [(p_i,r_i); \ldots] where [p_i] is a pack from [pl] and [r_i] a
   list of the corresponding rules in [rl].*)

let packs_and_rules_from_packs rule_list pack_list =
  fst (
    List.fold_left
	 (fun (packs_and_rules,rules) elem ->
	    let (rules_from_elem,left) = rules_from_roots_in_packs elem rules
	    in
	      (((elem,rules_from_elem)::packs_and_rules),left))
	 ([],rule_list)
	 pack_list)
;;




(* This graph is presently implemented with neighbours lists. *)

(* [contains sym pack_list] returns a pack of [pack_list] the
   first part of which contains [sym].  *)

(* We can safely suppose that packs are disjoint. Moreover each symbol
   is in a pack: no exception is to be thrown. *)
let contains sym pack_list =
  List.find (fun (sl,l) -> List.mem sym sl) pack_list
;;


(* [depends sym sym_graph pack_list] returns a list of packs from
   [pack_list] of which symbol [sym] depends according to hierarchy
   graph of symbols [sym_graph]. *)

let depends orig_acc sym sym_graph pack_list =
  List.fold_left
    (fun acc elem -> let new_pack =  contains elem pack_list in
       if List.mem new_pack acc then acc else new_pack::acc)
    orig_acc
    (Symgraph.neighbours_list sym_graph sym)
;;

let neighbours_of_a_pr sym_graph pr_list (sl,_) =
  List.fold_left
    (fun acc elem ->
       depends acc elem sym_graph pr_list
    )
    []
    sl
;;


(* Elimination of reflexivity. *)
let clean_neighbours n_list (a1,_) = 
  List.fold_left
    (fun acc ((a2,_) as elem) -> if a1 = a2 then acc else elem::acc)
    []
    n_list
;;


let pack_graph sym_graph pr_list =
  List.fold_left
    (fun acc pr ->
       (pr,(neighbours_of_a_pr sym_graph pr_list pr))::acc)
    []
    pr_list
;;


(*II*)
(* Here starts the actual building of modules from packs. *)

let string_of_term sgn (*vars*) tt = 

  let rec string_of_term_rec is_deep t =
    match t.Hcons.node with
      | Term(f,[]) -> sgn#string_of_symbol f
      | Term(f,t::l) ->
	(
	  match sgn#symbol_fix f with
	      Default | Prefix ->
      	      	  (sgn#string_of_symbol f)^"("^
      		  (string_of_term_rec false t)^
      		  (string_of_end_term_list l)
	    | Postfix ->
      		  "("^
      		  (string_of_term_rec false t)^
      		  (string_of_end_term_list l)^
      	      	  (sgn#string_of_symbol f)
	    | Infix ->
		  (if is_deep then "(" else "")^
		  (string_of_term_rec true t)^
      		  (string_of_infix_list f l)^
		  (if is_deep then ")" else "")
	)
    | Var(x) ->
	((*vars*)default#string_of_var x)
	  
  and string_of_end_term_list = function
      [] -> 
      	")"
    | t::l ->
      	","^
      	(string_of_term_rec false t)^
      	(string_of_end_term_list l)
	  
  and string_of_infix_list f = function
      [] -> ""
    | t::l ->
	" "^
      	(sgn#string_of_symbol f)^
	" "^
      	(string_of_term_rec true t)^
      	(string_of_infix_list f l)
	  
  in
    string_of_term_rec false tt
;;

let string_of_arity  = function
    Prefix -> "prefix" 
  | Infix -> "infix"
  | Postfix -> "postfix"
  | Default -> ""
;;
(*i Should use pattern matching on symbol theory rather than if then if.
   Moreover testing if sgn is free may result in increasing perfomances. i*)
let string_of_pack sgn p =
  List.fold_left
    (fun acc sym -> 
       (sgn#string_of_symbol sym)^" : "^
       (if sgn#is_free(sym) 
	then
       (string_of_arity (sgn#symbol_fix(sym)))
	  ^" "^(string_of_int(sgn#arity(sym)))
	else (
	  if sgn#is_ac(sym) 
	  then (string_of_arity (sgn#symbol_fix(sym)))^" AC" 
	  else failwith 
	    "only free and AC symbols are suitable for modular proofs so far" 
	)
       )
       ^" ; "^acc)
    ""
    p
;;

let string_of_rewrite_rule sgn (*vars*) r =
  (string_of_term sgn (*vars*) r.lhs)^
  " -> "^
  (string_of_term sgn (*vars*) r.rhs)^"; "
;;

let string_of_rewrite_rule_set sgn (*vars*) rset =
  List.fold_left
    (fun acc r -> (string_of_rewrite_rule sgn (*vars*) r)^acc) 
    ""
    rset
;;

module OrdSymList = 
struct
  type 'a t =  'a SymbolOrd.t list
  let compare = compare
end
;;
  
module PackMap = Ordered_maps.MakePoly(OrdSymList)
;;


(*i SI ON VEUT CALCULER LES MODULES D'UNE LISTE DE PACKS, IL FAUT
   ENCORE LE CODER DE MANIRE  NE PASSER QU'UN PR_GRAPH NON
   REFLEXIF. PLUS EFFICACE SI KEY = SYMBOLE PLUTT QUE LISTE ? i*)

(*i RCURSIF TERMINAL AVEC UN TRY SOME ? i*)


(* [module_from_pr sym_graph pr_list sgn vars deja_vu pr] returns a pair
   [(m,mmap)] consisting of a module [m] together with a map [mmap] from 
   packs to their corresponding module. *)

let rec module_from_pr sym_graph pr_list sgn (*vars*) deja_vu ((p,r)as pr) =
  try ((PackMap.find p deja_vu), deja_vu)
  with Not_found ->
    let pack_neigh = (clean_neighbours 
			(neighbours_of_a_pr sym_graph pr_list pr) pr)
    in let (mod_neigh,new_dv) = (****)
	List.fold_left
	  (fun (acc,dv) ((l,r) as elem) ->
	     try ((PackMap.find l dv)::acc, dv)
	     with Not_found -> 
	       let (new_mod,ndv) = (****)
		 module_from_pr sym_graph pr_list sgn (*vars*) dv elem
	       in
		 (new_mod::acc, PackMap.add l new_mod ndv)) (****)
	  ([],deja_vu)
	  pack_neigh
       and mod_sgn = User_signatures.from_string (string_of_pack sgn p)  
       and mod_rules = string_of_rewrite_rule_set sgn (*vars*) r
       in
	 ((extend mod_neigh 
	     mod_sgn Variables.default
	     (*vars*) 
	     mod_rules)
	    ,new_dv)(****)
;;

(* [mod_map_from_list sym_graph sgn pr_list] returns a map from
   packs to their corresponding modules. *)

let mod_map_from_list sym_graph sgn (*vars*) pr_list =
    List.fold_left
      (fun acc elem ->
	 snd (module_from_pr 
		sym_graph pr_list sgn (*vars*) 
		acc
		elem))
      PackMap.empty
      pr_list
;;

let minimal_split sgn (*vars*) my_htrs =  
  let my_all = my_htrs.htrs_all_rules 
(*i and _ = print_rewrite_rule_set my_htrs.htrs_sig Variables.default 
(my_htrs.htrs_all_rules) i*)
  in
  let g_htrs = graph_from_a_list my_all
  and pre_mod_htrs = 
    packs_and_rules_from_packs my_all (packs_from_a_list my_all) 
  in
(*i
    List.iter (fun (a,b) -> print_pack my_htrs.htrs_sig a ; print_newline()) pre_mod_htrs ;
i*)
(*i
  let result =
i*)
    (fst(List.fold_left
	   (fun (acc,dv) elem ->
	      let (a1,b1) = (module_from_pr 
			       g_htrs pre_mod_htrs my_htrs.htrs_sig (*vars*) 
			       dv 
			       elem)
	      in
		(a1::acc,b1))
	   ([],PackMap.empty)
	   pre_mod_htrs))
(*i
  in print_int (List.length result) ; print_newline () ; result
i*)
;;



