(***************************************************************************

Definition of termination constraints

CiME Project - Dmons research team - LRI - Universit Paris XI

$Id: termination_constraints.ml,v 1.3 2001/02/13 14:33:52 marche Exp $

***************************************************************************)


(*i*)
open Signatures;;
open Gen_terms;;
open Format;;
(*i*)

(*

  An atomic termination constraint is either $s=t$, $s>t$ or $s>=t$
  for two terms $s$ and $t$

*)

type constraint_op_type = Equal | Greater | Greater_or_equal;;

type 'symbol atomic_termination_constraint =
    {
      left : 'symbol term;
      right : 'symbol term;
      constraint_op : constraint_op_type;
    }
;;

(*

  a termination constraint is a disjunction of conjonction of atomic
  constraints.

*)

module TermConstrOrd = 
  struct
    type 'a t = 'a atomic_termination_constraint
    let compare = Pervasives.compare
  end;;


module AtomicTerminationConstraintSet = Ordered_sets.MakePoly(TermConstrOrd);;

type 'symbol conj_termination_constraints = 
    'symbol AtomicTerminationConstraintSet.t;;

module ConjTermConstrOrd = 
  struct
    type 'a t = 'a conj_termination_constraints
    let compare = Pervasives.compare
  end;;


module ConjTerminationConstraintSet = Ordered_sets.MakePoly(ConjTermConstrOrd);;


type 'symbol termination_constraints =
    'symbol ConjTerminationConstraintSet.t;;


let print_atomic_termination_constraint sigma vars c =
  print_term sigma vars c.left;
  print_string (match c.constraint_op with
		    Greater -> " > "
		  | Equal -> " = "
		  | Greater_or_equal -> " >= ");
  print_term sigma vars c.right
;;

let print_conj_termination_constraints sigma vars constr =
  print_string "{ ";
  open_hovbox 0;
  AtomicTerminationConstraintSet.iter
    (fun c -> 
       print_atomic_termination_constraint sigma vars c;
       print_string " ;";
       print_newline())
    constr;
  close_box ();
  print_string "}";
  print_newline();
;;


let print_termination_constraints sigma vars constr =
  ConjTerminationConstraintSet.iter
    (fun c ->
       print_conj_termination_constraints sigma vars c;
       print_string "or";
       print_newline();
    )
    constr
;;


(*

  [(set_of_symbols c)] returns the set of symbols occuring in the list
  of constraints [c]

*)

let set_of_symbols constr = 
  AtomicTerminationConstraintSet.fold
    (fun c s ->
       SymbolSet.union 
	 (Gen_terms.set_of_symbols c.left)
	 (SymbolSet.union 
	    (Gen_terms.set_of_symbols c.right)
	    s))
    constr
    SymbolSet.empty
;;


(*

  \subsection{Checking constraints}

*)


let check_conj_constraints sigma vars constr order =
  AtomicTerminationConstraintSet.iter
    (fun c -> 
       print_atomic_termination_constraint sigma vars c;
       let comp = order c.left c.right in
       let test = 
	 match c.constraint_op with
	     Equal -> comp = Orderings_generalities.Equivalent
	   | Greater -> comp =  Orderings_generalities.Greater_than
	   | Greater_or_equal -> 
	       comp = Orderings_generalities.Greater_than or 
	       comp =  Orderings_generalities.Greater_or_equivalent
       in
       printf " : %b@." test)
    constr;
