(***************************************************************************

Main.ml for testing termination criteria

CiME Project - Dmons research team - LRI - Universit Paris XI

$Id: termination_expert.ml,v 1.29 2002/07/01 12:01:36 contejea Exp $

***************************************************************************)


open Format;;
open Arg;;

let standard_criterion = ref true;;
let dp_criterion_uses_markings = ref true;;
let dp_criterion_uses_graph = ref true;;
let enable_split = ref false;;
let bound = ref 6;;
let verbose = ref 0;;
let user_dhct = Gen_terms.HashTerm.create 1997

open Signatures;;
open User_signatures;;
open Rewrite_rules;;
open User_terms;;
open Variables;;

open Dependency_pairs_criteria;;
open Marked_dp_criteria;;

open Graph_criteria;;

open Hierarchical_signatures;;
open Hierarchical_trs;;
open Relative_dp;;
open Minimal_split;;


open Non_linear_solving;;

open Format;;

(* DEBUG MARKED *)
(*
let display_marked sigma vars t =
  let a = List.hd 
	    (List.tl
	       (List.tl
		  (List.tl 
		     (List.tl(Marked_dp_criteria.split sigma vars t)))))
  in
  print_string "##############################\n";
  Gen_terms.print_term sigma vars a;
  print_string "\n##############################\n"
;;
*)

let display_pairs sigma vars pairs =
  begin
    let _ = 
      List.fold_right 
	(fun (s,t) n ->
	   print_int n;
	   print_string ": < ";
	   Gen_terms.print_term sigma vars s;
	   print_string " , ";
	   Gen_terms.print_term sigma vars t;
	   print_string " >";
	   print_newline();
	   succ n
	)
	pairs 1
    in ()
  end
;;


let display_graph sigma vars g =
(*
  if !verbose >= 2
  then
*)
    begin
      Labelled_graphs.dot_print g 
	(fun (s,t) -> 
	   Gen_terms.print_term sigma vars s;
	   print_string " ; ";
	   Gen_terms.print_term sigma vars t
	);
      print_newline();
    end
;;

let display_termination_constraints sigma vars c =
  if !verbose >= 1
  then
    begin
      print_string "Termination constraints:";
      print_newline();
      Termination_constraints.print_termination_constraints 
    	sigma vars c;
      print_string "(";
      print_int (Termination_constraints.ConjTerminationConstraintSet.cardinal c);
      print_string " termination constraints)";
      print_newline();
      print_newline();
    end
;;

let display_conj_termination_constraints sigma vars c =
  if !verbose >= 1
  then
    begin
      print_string "Termination constraints:";
      print_newline();
      Termination_constraints.print_conj_termination_constraints 
    	sigma vars c;
      print_string "(";
      print_int (Termination_constraints.AtomicTerminationConstraintSet.cardinal c);
      print_string " termination constraints)";
      print_newline();
      print_newline();
    end
;;

let display_interp sigma i =
  if !verbose >= 2
  then
    begin
      print_string "Generic Polynomial Interpretation:";
      print_newline();
      Generic_polynomials.print_interp sigma i;
      print_newline();
      print_newline()
    end
;;

let display_polynomial_constraints c =
  if !verbose >= 3
  then
    begin
      print_string "Polynomial ordering constraints:";
      print_newline();
      print_non_linear_constraints c;
      print_newline()
    end
  else
    if !verbose >= 2
    then
      begin
	print_int (List.length c);
      	print_string " polynomial ordering constraints";
      	print_newline();
      end

;;

let display_solution sol =
  if !verbose >= 2
  then
    begin
      print_string "Solution:";
      print_newline();
      Finite_domains.print_fd_solution sol;
      print_newline()
    end
;;



(*

  \subsection{displaying pairs and graph}

*)

let show_pairs sigma hct marked_hct vars trs =
  (* displaying pairs *)
  if !dp_criterion_uses_markings
  then
    let sigma_dupl = signature_with_marks sigma 
    and def =  Dependency_pairs_criteria.compute_defined_symbols trs in
    let pairs = 
      Marked_dp_criteria.compute_dependency_pairs_with_markings 
	sigma hct marked_hct def trs 
    in 
      if !dp_criterion_uses_graph
      then
	let def_dupl =
	  SymbolSet.fold 
	    (fun f s ->
	       SymbolSet.add (Original f) s)
	    def
	    SymbolSet.empty
	in
	let g = compute_dependency_graph sigma_dupl marked_hct def_dupl pairs in
	  display_graph sigma_dupl vars g
      else
	display_pairs sigma_dupl vars pairs
  else
    let def =  Dependency_pairs_criteria.compute_defined_symbols trs in
    let pairs = Dependency_pairs_criteria.compute_dependency_pairs def trs in
      if !dp_criterion_uses_graph
      then
	let g = compute_dependency_graph sigma hct def pairs in
	  display_graph sigma vars g
      else
	display_pairs sigma vars pairs
      
    

;;


(*

  \subsection{Computing termination constraint}

*)
(*i
let compute_modular_termination_constraints_without_marks htrs =
  if !standard_criterion 
  then
    failwith "compute_modular_termination_constraints_without_marks"
  else
    if !dp_criterion_uses_graph
    then
      failwith "not implemented"
    else
      if !dp_criterion_uses_markings
      then
	failwith "compute_modular_termination_constraints_without_marks"
      else
	relative_dp_criterion htrs
;;

let compute_modular_termination_constraints_with_marks htrs =
  if !standard_criterion 
  then
    failwith "compute_modular_termination_constraints_without_marks"
  else
    if !dp_criterion_uses_graph
    then
      failwith "not implemented"
    else
      if !dp_criterion_uses_markings
      then
	marked_relative_dp_criterion htrs
      else
	failwith "compute_modular_termination_constraints_without_marks"
;;
i*)

let compute_termination_constraints_without_marks sigma vars r =
  if !standard_criterion
  then
    Basic_criterion.all_decreasing_constraints r
  else 
    if !dp_criterion_uses_markings
    then
      failwith "compute_termination_constraints_without_marks"
    else
      if !dp_criterion_uses_graph
      then
	failwith "not implemented"
	  (*i 
      	  simple_dependency_graph_criterion sigma r
	  i*)
      else
	simple_dependency_pair_criterion sigma r
;;

let compute_termination_constraints_with_marks sigma sigma_marked vars hct hct_marked r =
  if !standard_criterion
  then
    failwith "compute_termination_constraints_with_marks"
  else 
    if !dp_criterion_uses_markings
    then
      if !dp_criterion_uses_graph
      then
	failwith "not implemented"
	  (*i 
      	  simple_dependency_graph_criterion sigma r
	  i*)
      else
	dependency_pair_criterion_with_markings sigma hct hct_marked r
    else
    failwith "compute_termination_constraints_with_marks"
;;


(*

  \subsection{The termination expert}

*)
exception Proof_found;;

let find_termination_proof sigma hct marked_hct vars r =

  if !standard_criterion
  then
    
    let c = Basic_criterion.all_decreasing_constraints r
    in
    display_conj_termination_constraints sigma vars c;
    let interp,constr_ac,constr_mon,constr_dp = 
      Generic_polynomials.polynomial_constraints_of_termination_constraints
	sigma SymbolMap.empty c
    in
    display_interp sigma interp;
    let poly_constr = constr_ac@constr_mon@constr_dp
    in
    display_polynomial_constraints poly_constr;
    let sol =
      Non_linear_solving.non_linear_solve (Numbers.of_int !bound)
	poly_constr
    in
    display_solution sol;
    let new_interp = Generic_polynomials.build_new_interp interp sol
    in
    Generic_polynomials.print_interp sigma new_interp      
      
      
  else
    begin
      if !verbose >= 1
      then
	show_pairs sigma hct marked_hct vars r;
      if !dp_criterion_uses_markings
      then
	let sigma_dupl = signature_with_marks sigma
	in
	if !dp_criterion_uses_graph
	then
      	  let comps = 
	    dependency_graph_criterion_with_markings 
	      sigma hct marked_hct r
	  in
	  (*
	    display_graph sigma_dupl vars g;
	  *)
	  print_string "checking each of the ";
	  print_int (List.length comps);
	  print_string " strongly connected components :"; 
	  print_newline();
	  let _ = List.fold_right
		    (fun c count-> 
		       try
			 print_string "checking component ";
			 print_int count;
			 print_string " (disjunction of ";
			 print_int (Termination_constraints.ConjTerminationConstraintSet.cardinal c);
			 print_string " constraints)";
			 print_newline();
			 display_termination_constraints sigma_dupl vars c;
			 Termination_constraints.ConjTerminationConstraintSet.iter
			   (fun c ->
			      display_conj_termination_constraints sigma_dupl vars c;
			      let interp,constr_ac,constr_mon,constr_dp = 
				Generic_polynomials.polynomial_constraints_of_termination_constraints
				  sigma_dupl SymbolMap.empty c
			      in
      			      display_interp sigma_dupl interp;
			      let poly_constr = constr_ac@constr_dp
			      in
			      display_polynomial_constraints poly_constr;
			      try
      				let sol =
				  Non_linear_solving.non_linear_solve (Numbers.of_int !bound)
	    			    poly_constr
      				in
				display_solution sol;
				let new_interp = Generic_polynomials.build_new_interp interp sol
				in
	      			Generic_polynomials.print_interp 
				  sigma_dupl new_interp;
				print_newline();
				raise Proof_found
				  
			      with Finite_domains.No_solution -> 
				print_string 
				  "No termination proof found yet, trying next in disjunction...";
				print_newline())
			   c;
			 raise Finite_domains.No_solution
		       with Proof_found ->
			 succ count 
		    ) 
		    comps 1
		    
	  in
	  print_string "Termination proof found.";
	  print_newline()
	    
	else
	  let c = 
      	    dependency_pair_criterion_with_markings sigma hct marked_hct r
	  in
	  display_conj_termination_constraints sigma_dupl vars c;
	  let interp,constr_ac,constr_mon,constr_dp = 
	    Generic_polynomials.polynomial_constraints_of_termination_constraints
	      sigma_dupl SymbolMap.empty c
	  in
      	  display_interp sigma_dupl interp;
	  let poly_constr = constr_ac@constr_dp
	  in
	  display_polynomial_constraints poly_constr;
	  (*
	    begin
	    let c=open_out "poly_constr"
	    in 
	    output_value c poly_constr;
	    close_out c
	    end;
	  *)
      	  let sol =
	    Non_linear_solving.non_linear_solve (Numbers.of_int !bound)
	      poly_constr
      	  in
	  display_solution sol;
	  let new_interp = Generic_polynomials.build_new_interp interp sol
	  in
	  Generic_polynomials.print_interp sigma_dupl new_interp      
      else
	if !dp_criterion_uses_graph
	then
      	  let c = simple_dependency_graph_criterion sigma hct r
	  in 
	  (*
	    display_graph sigma vars g;
	  *)
	  print_string "checking each of the ";
	  print_int (List.length c);
	  print_string " strongly connected components :"; 
	  print_newline();
	  let _ = List.fold_right
		    (fun c count -> 
		     try
		       print_string "checking component ";
		       print_int count;
		       print_string " (disjunction of ";
		       print_int (Termination_constraints.ConjTerminationConstraintSet.cardinal c);
		       print_string " constraints)";
		       print_newline();
		       display_termination_constraints sigma vars c;
  		       Termination_constraints.ConjTerminationConstraintSet.iter
			 (fun c ->
			    let interp,constr_ac,constr_mon,constr_dp = 
			      Generic_polynomials.
				polynomial_constraints_of_termination_constraints
				sigma SymbolMap.empty c
			    in
			    display_interp sigma interp;
			    let poly_constr = constr_ac@constr_dp
			    in
			    display_polynomial_constraints poly_constr;
			    try
      			      let sol =
				Non_linear_solving.non_linear_solve 
				  (Numbers.of_int !bound) 
				  poly_constr
      			      in
			      display_solution sol;
			      let new_interp = Generic_polynomials.build_new_interp interp sol
			      in
	      		      Generic_polynomials.print_interp sigma new_interp;
			      print_newline();
			      raise Proof_found
			    with
				Finite_domains.No_solution -> 
				  print_string 
				    "No termination proof found yet, trying next in disjunction...";
				  print_newline())
			 c;
		       raise Finite_domains.No_solution
		     with Proof_found -> succ count)
		  c 1
	in
	print_string "Termination proof found.";
	print_newline()
	  
      else
	let c =
      	  simple_dependency_pair_criterion sigma r
	in
	display_conj_termination_constraints sigma vars c;
  	let interp,constr_ac,constr_mon,constr_dp = 
	  Generic_polynomials.polynomial_constraints_of_termination_constraints
	    sigma SymbolMap.empty c
	in
	display_interp sigma interp;
	let poly_constr = constr_ac@constr_dp
	in
	display_polynomial_constraints poly_constr;
      	let sol =
	  Non_linear_solving.non_linear_solve 
	    (Numbers.of_int !bound) 
	    poly_constr
      	in
	display_solution sol;
	let new_interp = Generic_polynomials.build_new_interp interp sol
	in
	Generic_polynomials.print_interp sigma new_interp      
  end
;;


(***********************************************************************)
let  solve_htrs vars hr =
  if !dp_criterion_uses_markings
  then
    let sigma_dupl = Marked_dp_criteria.signature_with_marks (hr.htrs_sig)
    in
      if !dp_criterion_uses_graph
      then
      	let comps = modular_marked_graph_criterion hr sigma_dupl
	in
	print_string "checking each of the ";
	print_int (List.length comps);
	print_string " strongly connected components :"; 
	print_newline();
	let _ = List.fold_right
		  (fun c count-> 
		     try
		       print_string "checking component ";
		       print_int count;
		       print_string " (disjunction of ";
		       print_int (Termination_constraints.ConjTerminationConstraintSet.cardinal c);
		       print_string " constraints)";
		       print_newline();
		       display_termination_constraints sigma_dupl vars c;
		       Termination_constraints.ConjTerminationConstraintSet.iter
			 (fun c ->
			    display_conj_termination_constraints sigma_dupl vars c;
			    let interp,constr_ac,constr_mon,constr_dp = 
			      Generic_polynomials.polynomial_constraints_of_termination_constraints
				sigma_dupl SymbolMap.empty c
			    in
      			    display_interp sigma_dupl interp;
			    let poly_constr = constr_ac@constr_dp
			    in
			    display_polynomial_constraints poly_constr;
			    try
      			      let sol =
				Non_linear_solving.non_linear_solve (Numbers.of_int !bound)
	    			  poly_constr
      			      in
				display_solution sol;

			      let new_interp = Generic_polynomials.build_new_interp interp sol
			      in
	      		      Generic_polynomials.print_interp 
				sigma_dupl new_interp;
			      print_newline();
			      raise Proof_found
				
			    with Finite_domains.No_solution -> 
			      print_string 
				"No termination proof found yet, trying next in disjunction...";
			      print_newline())
			 c;
		       raise Finite_domains.No_solution
		     with Proof_found ->
		       succ count 
		  ) 
		  comps 1
		  
	in
	  hr.proved_c_e_terminating := true ;
	  print_string "Termination proof found.";
	  print_newline()
	    
	    
	    
      else
	let c =
	  marked_relative_dp_criterion hr sigma_dupl
	in
	  (**)display_conj_termination_constraints sigma_dupl vars c;(**)
	  let interp,constr_ac,constr_mon,constr_dp = 
	    Generic_polynomials.polynomial_constraints_of_termination_constraints
	      sigma_dupl SymbolMap.empty c
	  in
	    (* display_interp sigma interp;*)
	  let poly_constr = constr_ac@constr_dp
	  in
	    (*display_polynomial_constraints poly_constr;*)
	  let sol =
	    Non_linear_solving.non_linear_solve 
	      (Numbers.of_int !bound) 
	      poly_constr
	  in
	    display_solution sol; hr.proved_c_e_terminating := true ;
	    let new_interp = Generic_polynomials.build_new_interp interp sol
	    in
	      Generic_polynomials.print_interp sigma_dupl new_interp      
  else
    let sigma = (hr.htrs_sig)
    in      
      if !dp_criterion_uses_graph
      then
	
      	let c = modular_simple_graph_criterion hr
	in 
	  (*
	    display_graph sigma vars g;
	  *)
	  print_string "checking each of the ";
	  print_int (List.length c);
	  print_string " strongly connected components :"; 
	  print_newline();
	  let _ = List.fold_right
		    (fun c count -> 
		       try
			 print_string "checking component ";
			 print_int count;
			 print_string " (disjunction of ";
			 print_int (Termination_constraints.ConjTerminationConstraintSet.cardinal c);
			 print_string " constraints)";
			 print_newline();
			 display_termination_constraints sigma vars c;
  			 Termination_constraints.ConjTerminationConstraintSet.iter
			   (fun c ->
			      let interp,constr_ac,constr_mon,constr_dp = 
				Generic_polynomials.
				  polynomial_constraints_of_termination_constraints
				  sigma SymbolMap.empty c
			      in
				display_interp sigma interp;
				let poly_constr = constr_ac@constr_dp
				in
				  display_polynomial_constraints poly_constr;
				  try
      				    let sol =
				      Non_linear_solving.non_linear_solve 
					(Numbers.of_int !bound) 
					poly_constr
      				    in
				      display_solution sol;

				      let new_interp = 
					Generic_polynomials.build_new_interp interp sol
				      in
	      				Generic_polynomials.print_interp 
					  sigma new_interp;
					print_newline();
					raise Proof_found
				  with
				      Finite_domains.No_solution -> 
					print_string 
					  "No termination proof found yet, trying next in disjunction...";
					print_newline())
			   c;
			 raise Finite_domains.No_solution
		       with Proof_found -> succ count)
		    c 1
	  in
	    hr.proved_c_e_terminating := true ;
	    print_string "Termination proof found.";
	    print_newline()

      else
	let c =
	  relative_dp_criterion hr
	in
	  (**)display_conj_termination_constraints sigma vars c;(**)
	  let interp,constr_ac,constr_mon,constr_dp = 
	    Generic_polynomials.polynomial_constraints_of_termination_constraints
	      sigma SymbolMap.empty c
	  in
	    (* display_interp sigma interp;*)
	  let poly_constr = constr_ac@constr_dp
	  in
	    (*display_polynomial_constraints poly_constr;*)
	  let sol =
	    Non_linear_solving.non_linear_solve 
	      (Numbers.of_int !bound) 
	      poly_constr
	  in
	    display_solution sol; hr.proved_c_e_terminating := true ;
	    let new_interp = Generic_polynomials.build_new_interp interp sol
	    in
	      Generic_polynomials.print_interp sigma new_interp      
;;
    
let exec_param (interptype,i) = 
  begin
(*    Generic_polynomials.use_linear_interp := b ; *)
    bound := i;
    match interptype with
      |	Generic_polynomials.Linear ->
	  begin 
	    print_string "##  Linear\n##  Bound is : ";
	    Generic_polynomials.use_linear_interp := true
	  end
      | Generic_polynomials.Simple ->
	  begin 
	    print_string "##  Simple\n##  Bound is : ";
	    Generic_polynomials.use_linear_interp := false
	  end
      | Generic_polynomials.Simple_Mixed ->
	  begin 
	    print_string "##  Simple-mixed\n##  Bound is : ";
	    Generic_polynomials.use_linear_interp := false;
	    Generic_polynomials.use_degree_two_interp := true
	  end;
    print_int i; print_newline(); flush stdout
  end

let rec htrs_solve_with_param my_vars my_htrs param_list =
  match param_list with
    | [] -> solve_htrs my_vars my_htrs
    | p::p_list -> 
	begin exec_param p ;
	try solve_htrs my_vars my_htrs 
	with Finite_domains.No_solution ->  
	  begin print_string "### No termination proof found yet, trying with next parameters...";
	    print_newline(); flush stdout;
	    htrs_solve_with_param my_vars my_htrs p_list
	  end
	end

let rec solve_for_module my_vars my_htrs param_list =
    if !enable_split 
    then 
      List.iter
	(fun m -> 
	   begin
	     print_string "Checking module:";
	     print_newline();
	     print_rewrite_rule_set m.htrs_sig my_vars m.htrs_new_rules;
	     print_newline();
	     print_newline();
	     htrs_solve_with_param my_vars m param_list;
		   print_newline() 
	   end)
	(minimal_split my_htrs.htrs_sig (*i my_vars i*) my_htrs)
    else
    let _  =
    List.iter
      (fun m -> 
	 begin 
	   if !(m.proved_c_e_terminating) 
	   then 
	     begin
	       print_string "CE-termination of";print_newline();
	       print_rewrite_rule_set m.htrs_sig my_vars m.htrs_all_rules;
	       print_newline();print_string "is already proven.";
	       print_newline()
	     end
	   else 
	     begin 
	       solve_for_module my_vars m param_list 
	     end ; 
	   print_newline() end)
     (my_htrs.htrs_imported)
    in
      htrs_solve_with_param my_vars my_htrs param_list 
(*i 03/06
      solve_htrs my_vars my_htrs
i*)
;;


(* [modular\_expert X m plist] search a termination proof for
   hierarchy [m] with variables [X] trying successively the
   (termination expert) parameters of list [plist] *)

let modular_expert vars m plist =
  try
    print_string "Entering the termination expert for modules. Verbose level = ";
    print_int !verbose;
    print_newline();
   (*i debug marked i*) 
   (*i display_marked sigma vars (List.hd r).lhs; i*)
    solve_for_module vars m plist
  with 
      Finite_domains.No_solution -> 
	print_string "No polynomial interpretation found in [0;";
	print_int !bound;
	print_string "]";
	print_newline()
    | Proof_found -> 
	print_string "Termination proof found.";
	print_newline()
;;

(*i*********************************************************************i*)

let expert sigma vars r =
  let hct = Gen_terms.HashTerm.create 1997
  and marked_hct = Gen_terms.HashTerm.create 1997 in
  try
    print_string "Entering the termination expert. Verbose level = ";
    print_int !verbose;
    print_newline();
   (*i debug marked i*) 
   (*i display_marked sigma vars (List.hd r).lhs; i*)
    find_termination_proof sigma hct marked_hct vars r
  with 
      Finite_domains.No_solution -> 
	print_string "No polynomial interpretation found in [0;";
	print_int !bound;
	print_string "]";
	print_newline()
    | Proof_found -> 
	print_string "Termination proof found.";
	print_newline()
;;








