
open Gen_terms;;

type comparison_op = Eq | Ne | Le | Lt | Ge | Gt;;
    
type 'symbol formula =
  | True
  | False
  | Or of 'symbol formula list
  | And of 'symbol formula list
  | Not of 'symbol formula
  | Atom of 
      'symbol term * comparison_op * 'symbol term 
;;

let conj f1 f2 = 
  match (f1,f2) with
    | (True,_) -> f2
    | (_,True) -> f1
    | (False,_) -> False
    | (_,False) -> False
    | (And(l1),And(l2)) -> And(l1@l2)
    | (And(l1),_) -> And(f2::l1)
    | (_,And(l2)) -> And(f1::l2)
    | (_,_) -> And([f1;f2])
;;

let disj f1 f2 = 
  match (f1,f2) with
    | (True,_) -> True
    | (_,True) -> True
    | (False,_) -> f2
    | (_,False) -> f1
    | (Or(l1),Or(l2)) -> Or(l1@l2)
    | (Or(l1),_) -> Or(f2::l1)
    | (_,Or(l2)) -> Or(f1::l2)
    | (_,_) -> Or([f1;f2])
;;

let rec conj_all l = 
  match l with
    | [] -> True
    | [f] -> f
    | f::l -> conj f (conj_all l)
;;

let rec disj_all l = 
  match l with
    | [] -> False
    | [f] -> f
    | f::l -> disj f (disj_all l)
;;


let rec make_formula_from_string_formula sigma h f =
  match f with 
    | True -> True
    | False -> False
    | Or(l) -> 
	Or(List.rev_map (make_formula_from_string_formula sigma h) l)
    | And(l) -> 
	And(List.rev_map (make_formula_from_string_formula sigma h) l)
    | Not(f) -> 
	Not(make_formula_from_string_formula sigma h f)
    | Atom(t1,op,t2) -> 
	Atom(Gen_terms.make_term_from_string_term sigma h t1,
	    op,
	    Gen_terms.make_term_from_string_term sigma h t2)
;;


open Format;;

let string_of_comparison_op o = 
  match o with
    | Eq -> "="
    | Ne -> "<>"
    | Le -> "<="
    | Lt -> "<"
    | Ge -> ">="
    | Gt -> ">"
;;

let rec print_formula sign vars f =
  match f with 
    | True | And([]) -> printf "true"
    | False | Or([]) -> printf "false"
    | Or(f::l) -> 
	printf "@[(";
	print_formula sign vars f;
	List.iter 
	  (fun f -> printf "@ \\/@ "; print_formula sign vars f)
	  l;
	printf ")@]"
    | And(f::l) -> 
	printf "@[(";
	print_formula sign vars f;
	List.iter 
	  (fun f -> printf "@ /\\@ "; print_formula sign vars f)
	  l;
	printf ")@]"
    | Not(f) -> 
	printf "@[(";
	printf "~@ ";
	print_formula sign vars f;
	printf ")@]"
    | Atom(t1,op,t2) -> 
	printf "@[";
	Gen_terms.print_term sign vars t1;
	printf "@ %s@ " (string_of_comparison_op op);
	Gen_terms.print_term sign vars t2;
	printf "@]"
;;
