(********************************************************************

This is the main module of CiME 2.
It starts the main interaction loop, after having processed the files
  given on the command line.


CiME Project - Dmons research team - LRI - Universit Paris XI

$Id: main.ml,v 1.51 2002/07/16 09:29:36 contejea Exp $

********************************************************************)

open Format
open Abstract_syntax
open Eval
open Values
open Unix

let show_time = ref false

let prompt = "CiME> "


(*
  
  [(time f x)] is [(f x)] but also show the execution time

*)

let time f x =
  let time_before = Unix.times ()
  in
  let show_time () =
    let time_after = Unix.times ()
    in
      printf "Execution time: %f sec@."
	(time_after.Unix.tms_utime -. time_before.Unix.tms_utime)
(*
      Time_profiler.print_profile()
*)      	
  in
    try
      let res = f x
      in show_time();res
    with e -> show_time();raise e      




let global_env = ref (Predef.initial_env ());;

let global_env_type = ref (Predef.initial_env_type ());;

let do_eval env e =
  if !show_time
  then time (eval env) e
  else eval env e
;;

let rec all_different = function
    [] -> true
  | x::l -> if List.mem x l then false else all_different l
;;

let make_rec_fun env args body = 
  let rec make_body = function
      [] -> body
    | y::l -> Fun(y,make_body l)

  in
    match args with
	x::l -> Fun_value(x,env,make_body l)
      | [] -> assert false 
;;

let create_env_type f args =
  let result_type = Typing.Type_variable (ref None)
  in
  let rec make_env = function
      [] -> [],result_type
    | x::l ->
	let env,t = make_env l
	and alpha = Typing.Type_variable (ref None)
	in 
	  ((x,Typing.Non_generalized(alpha))::env,
	   Typing.Dep_arrow(ref Typing.Anonymous, alpha, fun _ -> t))
	
  in
  let (env,t) = make_env args
  in
    ((f,Typing.Non_generalized(t))::env,t,result_type)
;;


let set_verbose_level n =
  Non_linear_solving.verbose := n;
  Termination_expert.verbose := n;
  Srs_completion.verbose := n;
  Unification.verbose := n;
  Kb.StandardKBCompletion.verbose := n;
  Kb.ACCompletion.verbose := n;
  printf "Verbose level is now %d@." n
;;


let name_value name = function
    Sig_value(sigma) -> sigma.sig_name <- name
  | Var_value(vars) -> vars.vars_name <- name
  | _ -> ();;


let rec execute c = 
  match c with
      Def(x,e) ->
	let t = Typing.type_expr !global_env_type !global_env_type e
	in 
 	let v = do_eval !global_env e
	in
	  name_value x v;
	  global_env := (x,{value=v}) :: !global_env;
	  global_env_type := (x,Typing.Non_generalized t) :: !global_env_type;
	  printf "%s : " x;
	  Typing.print_type t;
	  printf " = ";
	  print_value v
	    
    | Deffun(f,args,e) ->
	if all_different (f::args)
	then
	  let new_env_type, function_type, result_type =
	    create_env_type f args
	  in
	  let t = 
	    Typing.type_expr 
	      !global_env_type (new_env_type @ !global_env_type) e
	  in
	  if Typing.unify_types result_type t
	  then
	    begin
	      global_env_type := 
	      (f,Typing.generalize function_type)::!global_env_type;
	      let new_cell = {value=Unit_value}
	      in
	      global_env := (f,new_cell) :: !global_env;
	      new_cell.value <- make_rec_fun !global_env args e ;
	      printf "@[%s :@ " f;
	      Typing.print_type function_type;
	      printf " =@ ";
	      print_value new_cell.value;
	      printf "@]";
	    end
	  else
	    raise 
	      (Typing.Type_error "bad function result type")
	else
	  raise 
	    (Typing.Type_error "function name and arguments must all be different")
	    
    | Eval(e) ->
	let t = Typing.type_expr !global_env_type !global_env_type e
	in 
	let this_eval () = 
	  let v = do_eval !global_env e in
	  print_string "- : ";
	  Typing.print_type t;
	  print_string " = ";
	  print_value v;
	  print_flush ()
	in
	this_eval ()
    | Directive(s,t) ->
	match s with 
	  | "mem" | "memory" -> 
	      Format.printf "Size of allocated OCaml heap is %f Ko." 
		(Profiling_tools.memory_kb ())
	  | "load" -> load_file t
	  | "smp" -> 
	      begin
		try 
		  match t with
		      "" -> Parallel.M.smp := true
		    | n -> 
			Parallel.M.smp := true ;
			Parallel.M.max_process := int_of_string n
		with _ -> 
		  Format.printf "Bad argument to #smp directive@."
	      end;
	      Format.printf 
		"Smp is now on %d processes@." 
		!Parallel.M.max_process
	  | "quit" -> raise Exit
	  | "time" ->
	      begin
		match t with
		    "" -> show_time := not !show_time
		  | "on" -> show_time := true
		  | "off" -> show_time := false
		  | _ -> Format.printf "Bad argument to #time directive@."
	      end;
	      Format.printf "time is now o%s@."
		(if !show_time then "n" else "ff")
	  | "verbose" ->
	      begin
		try
		  let n = 
		    if t = "" then 1 else int_of_string t
		  in
		  set_verbose_level n
		with
		    Failure _ -> Format.printf 
			"Bad argument to #verbose directive@."
	      end;
	  | "quiet" -> set_verbose_level 0
	  | "completion_coeff_size" ->
	      begin
		try
		  let n = 
		    if t = "" then 1 else int_of_string t
		  in
		    Kb.coeff_size := n
		with
		    Failure _ -> print_string 
			("bag argument to #completion_coeff_size directive\n")
	      end;
	  | "completion_coeff_age" ->
	      begin
		try
		  let n = 
		    if t = "" then 0 else int_of_string t
		  in
		    Kb.coeff_age := n
		with
		    Failure _ -> print_string 
			("bag argument to #completion_coeff_birth directive\n")
	      end;
	  | "stats" ->
	      begin
		printf "***** Hashtable for user terms **********************************";
		print_newline();
		Gen_terms.HashTerm.stat User_terms.user_hct;
 		printf "***** Hashtable for protected user terms (matching) *************";
		print_newline();
		Gen_terms.HashTerm.stat Confluence.user_phct;
		print_newline();
		print_flush () 
	      end; 
	  | "help" ->
	      begin
		if t="" then
		  begin
		    printf 
		      "@[Type '#help \"category\"' with 'category' being one of :@.";
		    List.iter 
		      (fun s -> printf "%s@ " s)
		      (Predef.get_predef_category_list ());
		    printf "@]"
		  end
		else
		  begin
		    try
		      let l = Predef.get_predef_command_list t in
		      printf "@[Type '#help \"command\"' with 'command' being one of :@.";
		      List.iter 
			(fun s -> 
			   let (_,h) = Predef.get_help_of_command s in
			   if h<>"" 
			   then printf "%s@ " s)
			l;
		      printf "@]"
		    with Not_found ->
		      let (cat,help) = Predef.get_help_of_command t
		      and ty = 
			Typing.type_expr 
			  !global_env_type !global_env_type (Var t)
		      in
		      printf "%s (category \"%s\"): " t cat;
		      Typing.print_type ty;
		      printf "@.@[%s@]" help
		  end
	      end
	  | "extern" -> ignore (Sys.command t)
	  | _ -> 
	      printf "@[Unknown directive '%s'. Use one of:@." s;
	      List.iter (fun s -> printf "#%s@ " s) 
		["load";"mem";"quit";"time";"verbose";"quiet";"help";"extern";"smp"];
	      printf "@]"
and
  input_loop show_prompt channel =
  try
    let b = Lexing.from_channel channel
    in
    while true do
      if show_prompt 
      then
	begin
	  print_string prompt;
	  print_flush ();
	end;
      begin
	try
	  let c = Toplevel_parser.command Toplevel_lexer.token b
	  in	
	  execute c
	with

	    Parsing.Parse_error ->
	      printf "Syntax error"	
	  | Toplevel_lexer.Lexical_error s ->
	      printf "Lexical error: %s" s
	  | Eval_error s ->
	      printf "Evaluation error: %s" s
	  | Hierarchical_trs.Overriden s ->
	      printf "Semantical error: %s" s
	  | Typing.Type_error s ->
	      printf "Typing error: %s" s
	  | Sys.Break ->
	      printf "Command aborted by user"
	  | Invalid_argument s -> 
	      printf "Invalid argument: %s" s
(*	  | Out_of_memory -> 
	      printf "Out of memory, command aborted"
	  | Stack_overflow ->
	      printf "Stack overflow, command aborted"
 *)     end;
      printf "@."
    done
  with
      End_of_file -> ()

and load_file name =
  try
    let ch = open_in name
    in input_loop false ch
  with
      Sys_error s ->
	Format.printf "load failed: %s@."s

(*i Uncomment to test Parallel computations.
let add_10 x = x+10
let get_integer =
  let i = ref 0 in
  function () -> if !i>=100 then None else  
    Some (incr i;!i)
let l = Parallel.M.compute add_10 get_integer
i*)

let main () = 
  try
(*i Probably not a good idea...
    Sys.set_signal Sys.sigsegv
      (Sys.Signal_handle 
	 (function _ -> 
	    printf "Unrecoverable stack overflow : please report\n"; 
	    print_newline (); exit (-1)));
i*)
    Sys.catch_break true;
    set_ellipsis_text "...";
    set_max_boxes 20;
    printf "Welcome to CiME version %s - Built on %s@.@." 
      Version.version Version.date;
    for i = 1 to pred(Array.length Sys.argv) do
      load_file Sys.argv.(i)
    done;
    input_loop true Pervasives.stdin
  with 
    | Assert_failure (file,c1,c2) -> 
	printf "@.@.Assert failed, file \"%s\", chars %d-%d@."
	file c1 c2;

	printf "Please report to Claude.Marche@@lri.fr@.";
    | Exit ->
	printf"@.Quitting.@."
