(***************************************************************************

CiME Project - Dmons research team - LRI - Universit Paris XI

$Id: predef.ml,v 1.90 2002/07/19 15:52:19 contejea Exp $

***************************************************************************)


open Format;;
open Abstract_syntax;;
open Values;;
open Typing;;


type termination_parameter = 
    {
      linear : bool ;
      bound : int
    }

(* 

\subsection{Functions on diophantine constraints}

*)

let dioph_table =
  [
    "presburger_satisfiable",
    ("presburger_satisfiable f tells if f is satisfiable using Presburger's quantifier elimination method. f has to be linear.",
     1,
     Non_generalized(arrow_type dioph_constraint_type unit_type),
     function
	 [  Dioph_constraint c ] ->
	   begin
	     let _,c = Linear_constraints.from_abstract_formula [] c in
	     let b = Presburger.is_satisfiable c in
	     printf 
	       "This formula %s satisfiable.@."
	       (if b then "IS" else "is NOT");
	     Unit_value
	   end
       | _ -> raise (Eval_error "wrong arguments for presburger_satisfiable"))

    ;
    
    "presburger_valid",
    ("presburger_valid f tells if f is valid using Presburger's quantifier elimination method. f has to be linear.",
     1,
     Non_generalized(arrow_type dioph_constraint_type unit_type),
     function
	 [  Dioph_constraint c ] ->
	   begin
	     let _,c = Linear_constraints.from_abstract_formula [] c in
	     let b = Presburger.is_valid c in
	     printf 
	       "This formula %s valid.@."
	       (if b then "IS" else "is NOT");
	     Unit_value
	   end
       | _ -> raise (Eval_error "wrong arguments for presburger_valid"))

    ;
    
    "dioph_solve",
    ("dioph_solve B C solves Diophantine constraints C. It prints the first solution found with values between 0 and B.",
     2,
     Non_generalized(arrow_type int_type
		       (arrow_type dioph_constraint_type unit_type)),
     function
	 [ Int_value(n) ; Dioph_constraint c ] ->
	   begin
	     try	       
	       let c = Non_linear_solving.from_abstract_formula c in
      	       let sol = Non_linear_solving.non_linear_solve n c
      	       in
      		 print_string "Solution : \n";
		 Finite_domains.Fd_var_map.iter
		   (fun x n ->
		      print_string (Finite_domains.string_of_fd_var_id x);
		      print_string " = ";
		      print_string (Numbers.to_string n);
		      print_newline()) sol;
		 Unit_value
      	     with
		 Finite_domains.No_solution -> 
		   print_string 
		     ("No solution in [0,..,"^(Numbers.to_string n)^"]");
		   print_newline();
		   Unit_value
	   end
       | _ -> raise (Eval_error "wrong arguments for dioph_solve"))
    ;

    "dioph_constraint",
    ("Builds a Diophantine constraint from a given string.",
     1,
     Non_generalized(arrow_type string_type dioph_constraint_type),
     function 
	 [ String_value(s) ] ->
	   begin
	     try
	       Dioph_constraint(Poly_syntax.constraint_of_string s)
	     with
		 Poly_syntax.Syntax_error msg ->
		   raise (Eval_error "Syntax error in Diophantine constraint")
	   end
       | _ -> raise (Eval_error "wrong argument for dioph_constraint"))
    
  ]
;;

(*

\subsection{Function on words}

*)

let words_table = 
  [
    "word_signature",
    ("Builds a word signature from a given string.",
     1,
     Non_generalized(arrow_type string_type word_signature_type),
     function 
	 [String_value chaine] ->
	   begin
	     (*
	       try
	     *)
	     Word_sig_value(
	       {
		 word_sig_name = "?";
		 word_sig_val = 
		  String_signatures.from_string chaine;
	       }
	     ) 
	       (*
		 with Signature_syntax.Syntax_error s ->
		 raise (Eval_error ("Syntax error in signature ("^s^")"))
	       *)
	   end
       | _ -> raise (Eval_error "wrong arguments for word_signature"))
    ;

    "word",
    ("(word S s) builds the word on signature S from the given string.",
     2,
     Non_generalized
       (Dep_arrow(
	  ref Depend_var, word_signature_type,
	  fun sigma -> 
	    arrow_type string_type (word_type sigma))),
       function 
	   [Word_sig_value sigma;String_value chaine] ->
	     begin
	       try
		 Word_value(sigma,
			    User_words.word_of_string 
			      sigma.word_sig_val chaine)
	       with 
		   Word_syntax.Syntax_error s ->
		     raise (Eval_error ("Syntax error in word ("^s^")"))
	     end
	 | _ -> raise (Eval_error "wrong arguments for word"))
    ;
    
    "word_precedence",
    ("(word_precedence S s) builds the precedence on signature S given by string s",
     2,
     Non_generalized
       (Dep_arrow
	  (ref Depend_var, word_signature_type,
	   fun sigma -> 
	     arrow_type string_type (word_precedence_type sigma))),
       function 
	   [Word_sig_value sigma;String_value chaine] ->
	     begin
	       try
		 Word_precedence_value(sigma,
				       User_words.word_prec_of_string 
					 sigma.word_sig_val chaine)
	       with 
		   Word_syntax.Syntax_error s ->
		     raise (Eval_error ("Syntax error in word_precedence ("^s^")"))
		 | Finite_orderings.Incompatible ->
		     raise (Eval_error "Incompatible precedence")

	     end
	 | _ -> raise (Eval_error "wrong arguments for word_precedence"))
    ;

    "multi_lex",
    ("(multi_lex p) is the ordering on words that compares words w.r.t. the multiset of their letters w.r.t the precedence p, and when equal, lexicographically (left to right) w.r.t the precedence p.", 
     1,
     Generalized
       (fun sigma ->
	  Non_generalized
	    (arrow_type 
	       (word_precedence_type sigma) (word_ordering_type sigma))),
       function 
	   [Word_precedence_value(sigma,prec)] ->
	     Word_ordering_value(sigma,
				 Word_orderings.multi_lex prec)
	 | _ -> raise (Eval_error "wrong arguments for multi_lex"))
    ;
    
    "length_lex",
    ("(length_lex p) is the ordering on words that compares words w.r.t. their length, and when equal length, lexicographically (left to right) w.r.t the precedence p.",
     1,
     Generalized
       (fun sigma ->
	  Non_generalized
	    (arrow_type 
	       (word_precedence_type sigma) (word_ordering_type sigma))),
       function 
	   [Word_precedence_value(sigma,prec)] ->
	     Word_ordering_value(sigma,
				 Word_orderings.length_lex prec)
	 | _ -> raise (Eval_error "wrong arguments for length_lex"))
    ;
    
    "word_compare",
    ("(word_compare o w1 w2) compares the words w1 and w2 w.r.t. the word ordering o.",
     3,
     Generalized
       (fun sigma ->
	  Non_generalized
	    (arrow_type 
	       (word_ordering_type sigma)
	       (arrow_type 
		  (word_type sigma) 
		  (arrow_type (word_type sigma) string_type)))),
       function 
	   [Word_ordering_value(sigma,order); 
	    Word_value(_,w1); Word_value(_,w2)] ->
	     String_value
	       (Orderings_generalities.string_of_comparison_result
		  (order w1 w2))
	 | _ -> raise (Eval_error "wrong arguments for word_compare"))
    ;

    "SRS",
    ("(SRS S s) builds the SRS on signature S defined by string s.",
     2,
     Non_generalized
       (Dep_arrow
	  (ref Depend_var, 
	   word_signature_type, 
	   fun sigma -> 
	     arrow_type string_type (srs_type sigma))),
       function 
	   [Word_sig_value sigma; String_value chaine] ->
	     begin
	       try
		 let srs = 
		   User_words.srs_of_string sigma.word_sig_val chaine 
		 in
		   SRS_value(sigma,srs, lazy (String_rewriting.compile_srs srs))
	       with Word_syntax.Syntax_error s ->
		 raise (Eval_error ("Syntax error in SRS ("^s^")"))
	     end
	 | _ -> raise (Eval_error "wrong arguments for SRS"))
    ;

    "word_normalize",
    ("word_normalize R w computes the normal of word w w.r.t. SRS R, with the rightmost strategy. Beware that this function will loop if w has no normal form. You may interrupt computation with Ctrl-C.",
     2,
     Generalized
       (fun sigma -> 
	  Non_generalized
	    (arrow_type 
  	       (srs_type sigma)
	       (arrow_type (word_type sigma) (word_type sigma)))),
       function
	   [ SRS_value(_,srs,compiled_srs) ; Word_value(sigma,w) ] ->
	     let w' = String_rewriting.compiled_normalize (Lazy.force compiled_srs) w
	     in Word_value(sigma,w')
	 | _ -> raise (Eval_error "wrong arguments for word_normalize"))
    ;
    "word_counter",
    ("word_counter R computes the number of normal forms of the SRS R if it is finite or -1.",
     1,
       Generalized
	 (fun sigma -> 
	    Non_generalized 
	      (arrow_type 
  		 (srs_type sigma)
		 int_type)),
       function
	   [ SRS_value(_,srs,_) ] ->
	     (try 
	       Int_value(Numbers.of_int (Regexp_stub.count srs))
	     with
		 _ ->  Int_value(Numbers.minus_one))
	 | _ -> raise (Eval_error "wrong arguments for word_normalize"))
    ;
    "word_confluence",
    ("word_confluence R checks if R is localy confluent.",
     1,
       Generalized
	 (fun sigma -> 
	    Non_generalized 
	      (arrow_type 
  		 (srs_type sigma)
		 bool_type)),
       function
	   [ SRS_value(_,srs,_) ] ->
	     Bool_value(Local_confluence.is_localy_confluent srs)
	     
	 | _ -> raise (Eval_error "wrong arguments for word_confluence"))
    ;


    "word_completion",
    ("Performs the KB completion of the given SRS. You may interrupt computation with Ctrl-C (optimized version).",
     2,
      Generalized
	(fun sigma -> 
	   Non_generalized
	     (arrow_type 
		(word_ordering_type sigma)
		(arrow_type (srs_type sigma) (srs_type sigma)))),
	function
	    [ Word_ordering_value(sigma,order) ; 
	      SRS_value(_,srs,compiled_srs) ] ->
	      let completed_srs = 
		Srs_completion.complete_srs_std_strategy 
		  sigma.word_sig_val 
		  order
		  srs
	      in 
	      SRS_value(sigma,completed_srs,lazy (String_rewriting.compile_srs srs))
       | _ -> raise (Eval_error "wrong arguments for word_completion"))

    ;
    
    "new_word_completion",
    ("Performs the KB completion of the given SRS. You may interrupt computation with Ctrl-C (generic version).",
     2,
      Generalized
	(fun sigma -> 
	   Non_generalized
	     (arrow_type 
		(word_ordering_type sigma)
		(arrow_type (srs_type sigma) (srs_type sigma)))),
	function
	    [ Word_ordering_value(sigma,order) ; 
	      SRS_value(_,srs,compiled_srs) ] ->
	      let completed_srs = 
		Kb.WordCompletion.complete 
		  (sigma.word_sig_val :> 'symbol String_signatures.word_signature)
		  ()
		  order 
		  srs 
		  []
	      in 
	      SRS_value(sigma,completed_srs,
			lazy (String_rewriting.compile_srs srs))
       | _ -> raise (Eval_error "wrong arguments for word_completion"))
    
  ]
;;


(*

\subsection{Function on parameterized words}

*)

let parameterized_words_table = 
  [
    "parameters",
    ("Builds a set of parameters from a given string.",
     1,
     Non_generalized(arrow_type string_type parameter_set_type),
     function 
	 [String_value chaine] ->
	   Parameters_value
	     {parameters_name = "?";
	      parameters_val = 
		User_parameterized_signatures.parameters_from_string chaine;
	     }
       | _ -> raise (Eval_error "wrong arguments for word_signature"))
    ;

    "pword_signature",
    ("Builds a word parameterized signature from a given string.",
     2,
     Non_generalized(arrow_type parameter_set_type 
		       (arrow_type string_type pword_signature_type)),

     function
	 [Parameters_value v ; String_value chaine] ->
	   begin
	     try
	       Pword_sig_value(
		 {
		   pword_sig_name = "?";
		   pword_sig_val = 
		    User_parameterized_signatures.from_string 
		      v.parameters_val
		      chaine;
		 }
	       )
	     with Parameterized_signatures_syntax.Syntax_error s ->
	       raise (Eval_error 
			("Syntax error in parameterized signature ("^s^")"))
	   end
       | _ -> raise (Eval_error "wrong arguments for pword_signature"))
    ;

    "pword",
    ("(pword S s) builds the parameterized word on signature S from the given string.",
     2,
     Non_generalized
       (Dep_arrow(
	  ref Depend_var, pword_signature_type,
	  fun sigma -> 
	    arrow_type string_type (pword_type sigma))),
       function 
	   [Pword_sig_value sigma;String_value chaine] ->
	     begin
	       try
		 Pword_value(sigma,
			     User_parameterized_words.from_string 
			       sigma.pword_sig_val chaine)
	       with 
		   Parameterized_signatures_syntax.Syntax_error s ->
		     raise (Eval_error ("Syntax error in parameterized word ("^s^")"))
	     end
	 | _ -> raise (Eval_error "wrong arguments for parameterized word"))
    ;


    "psrs",
    ("(psrs S s) builds the parameterized string rewriting system on signature S from the given string.",
     2,
     Non_generalized
       (Dep_arrow(
	  ref Depend_var, pword_signature_type,
	  fun sigma -> 
	    arrow_type string_type (psrs_type sigma))),
       function 
	   [Pword_sig_value sigma;String_value chaine] ->
	     begin
	       try
		 Psrs_value(sigma,
			     User_parameterized_rewriting.from_string 
			       sigma.pword_sig_val chaine)
	       with 
		   Parameterized_signatures_syntax.Syntax_error s ->
		     raise (Eval_error ("Syntax error in parameterized srs ("^s^")"))
	     end
	 | _ -> raise (Eval_error "wrong arguments for parameterized srs"))
    ;

    "psim",
    ("(psim w1 w2) builds the parameterized similitude formula of w1 and w2.",
     2,
     Generalized
       (fun sigma ->
	  Non_generalized
	  (arrow_type 
	     (pword_type sigma) 
	     (arrow_type (pword_type sigma) dioph_constraint_type))) ,
    function 
      |	[Pword_value(sigma,w1) ; Pword_value(_,w2)] ->
	  Linear_constraint (Parameterized_similitude.compute_similitude w1 w2)
      | _ -> raise (Eval_error "wrong arguments for parameterized srs"))
    ;
    "ptest_sim",
    ("(psim w1 w2) tests if w1 and w2 are localy similar.",
     2,
     (Generalized
       (fun sigma ->
	  Non_generalized
	  (arrow_type 
	     (pword_type sigma) 
	     (arrow_type (pword_type sigma) bool_type)))),
     function 
       | [Pword_value(sigma,w1) ; Pword_value(_,w2)] ->
	   Bool_value (Parameterized_similitude.test_similitude w1 w2)
       | _ -> raise (Eval_error "wrong arguments for parameterized srs"))
    ;
	  
    "preducible",
    ("(preducible r w) tests if w is reducible by r.",
     2,
     (Generalized
	(fun sigma -> 
	   Non_generalized
	   (arrow_type 
	      (psrs_type sigma)
	      (arrow_type (pword_type sigma) bool_type)))),
     function 
       | [ Psrs_value(_,r) ; Pword_value(_,w) ] ->
	   Bool_value 
	   (Parameterized_rewriting.is_reducible r w
	   )
       | _ -> raise (Eval_error "wrong arguments for preducible"))
    ;

    "pnormalize",
    ("(pnormalize r w) computes the normal form of w by r. It may not terminate if the system is non terminating.",
     2,
     (Generalized
	(fun sigma -> 
	   Non_generalized
	   (arrow_type 
	      (psrs_type sigma)
	      (arrow_type (pword_type sigma) (pword_type sigma))))),
     function 
       | [ Psrs_value(sigma,r) ; Pword_value(_,w) ] ->
	   Pword_value
	   (sigma,
	    (Parameterized_rewriting.normalize r w
	    ))
       | _ -> raise (Eval_error "wrong arguments for preducible"))
    ;

    "pconfluent",
    ("(pconfluent r) tests if r is locally confluent.",
     1,
     (Generalized
	(fun sigma ->
           Non_generalized
	     (arrow_type 
		(psrs_type sigma) bool_type))),
     function 
       | [ Psrs_value(sigma,r) ] ->
	   Bool_value 
	   (Parameterized_rewriting.is_localy_confluent r)
       | _ -> raise (Eval_error "wrong arguments for pconfluent"))
    ;
    
    "pconfluent_ext",
    ("(pconfluent r1 r2) tests if r1 is locally confluent using r1 union r2 to normalize.",
     2,
     (Generalized
	(fun sigma ->
           Non_generalized
	   ((arrow_type 
		  (psrs_type sigma)
		  (arrow_type (psrs_type sigma) bool_type))))),
     function 
       | [ Psrs_value(sigma,r1) ; Psrs_value(_,r2) ] ->
	   Bool_value 
	   (Parameterized_rewriting.is_localy_confluent_extended
	      r1
	      r2
	   )
       | _ -> raise (Eval_error "wrong arguments for pconfluent"))
    ;
    
    "psubst",
    ("Builds a substitution from parameters to integers.",
     1, 
     Non_generalized(arrow_type string_type psubst_type),
     function
       |[String_value c] ->  
	   (try
	      Psubst_value {psubst_name = "?";
			    psubst_val = 
			      failwith "not yet implemented"
			   (* Parameterized_signatures.Psubst.from_string c *) 
			   }
	    with
		_ -> raise (Eval_error "syntax error in psubst."))
       | _ -> raise (Eval_error "wrong arguments for psubst"))
    
    ;
    
    "pinst_sig",
    ("Instantiate a parameterized signature onto a string signature.",
     2,
     Non_generalized(arrow_type pword_signature_type 
		       (arrow_type psubst_type word_signature_type)),
     function 
       | [Pword_sig_value sigma ; Psubst_value subst] -> 
	   Word_sig_value{
	     word_sig_name = "?";
	     word_sig_val = 
	      (User_parameterized_signatures.instantiate_signature 
		 sigma.pword_sig_val 
		 subst.psubst_val)}
       | _ -> raise (Eval_error "wrong arguments for pinst_sig"))
    ;
    "pword_completion",
    ("Performs the KB completion of the given parameterized SRS. You may interrupt computation with Ctrl-C (generic version).",
     1,
      Generalized
	(fun sigma -> 
	   Non_generalized
	   (arrow_type (psrs_type sigma) (psrs_type sigma))),
	function
	    [ Psrs_value(sigma,r) ] ->
	      let completed_srs = 
		Kb.PWordCompletion.complete 
		  sigma.pword_sig_val
		  ()
		  User_parameterized_words.interactive_ordering
		  (Obj.magic r)
		  []
	      in 
	      Psrs_value(sigma,completed_srs)
       | _ -> raise (Eval_error "wrong arguments for pword_completion"))
       
  ]
;;



(* 

\subsection{Functions on first-order terms}

*)


let terms_table =
  [
    "signature",
    ("Builds a first-order signature from a given string.",
     1,
     Non_generalized(arrow_type string_type signature_type),
     function 
	 [String_value chaine] ->
	   begin
	     try
	       Sig_value(
		 {
		   sig_name = "?";
		   sig_val = 
		    User_signatures.from_string chaine;
		 }
	       ) 
	     with Signature_syntax.Syntax_error s ->
	       raise (Eval_error ("Syntax error in signature ("^s^")"))
	   end
       | _ -> raise (Eval_error "wrong arguments for signature"))
    ;

    "theory",
    ("(theory S s) builds the first-order (equational) theory on the signature S defined by the string s.",
     2,
     Non_generalized
       (Dep_arrow
	  (ref Depend_var, 
	   signature_type,
	   fun sigma -> 
	     arrow_type string_type (theo_type sigma))),
     function 
	 [Sig_value sigma; String_value chaine] ->
	   begin
	     try
	       Theo_value(sigma, User_theory.from_string sigma.sig_val chaine)
	     with Theory.Syntax_error s ->
	       raise (Eval_error ("Syntax error in theory ("^s^")"))
	   end
       | _ -> raise (Eval_error "wrong arguments for theory"))
    ;
(*i
    "position",
    ("(position p) builds a position from the string p.",
     1,
     Non_generalized (arrow_type string_type position_type),
     function
	 [String_value chaine] ->
	   Position_value (User_terms.position_of_string chaine)
       | _ -> raise (Eval_error "wrong arguments for position"))
    ;
i*)

    "vars",
    ("Builds a set of first-order variables from the given string.",
     1,
     Non_generalized(arrow_type string_type variable_set_type),
     function 
	 [String_value chaine] ->
	   begin
	     try
	       Var_value(
		 {
		   vars_name = "?";
                   vars_val = Variables.from_string chaine;
		 }
	       ) 
	     with Variables.Syntax_error s ->
	       raise (Eval_error ("Syntax error in variables declaration ("^s^")"))
	   end
       | _ -> raise (Eval_error "wrong arguments for vars"))
    ;
 
    "term",
    ("(term S X s) builds the term on signature S defined by string s.",
     3,
     Non_generalized
       (Dep_arrow
	  (ref Depend_var, 
	   signature_type,
	   fun sigma -> 
	     Dep_arrow
	       (ref Depend_var, 
		variable_set_type,
		fun vars ->
		  arrow_type string_type (term_type sigma vars)))),
       function 
	   [Sig_value sigma;Var_value vars;String_value chaine] ->
	     begin
	       try
		 Term_value(sigma,vars,
			    User_terms.term_of_string 
			      sigma.sig_val User_terms.user_hct vars.vars_val chaine)
	       with 
		   Term_syntax.Syntax_error s ->
		     raise (Eval_error ("Syntax error in term ("^s^")"))
		 | Gen_terms.Bad_arity op ->
		     raise (Eval_error ("Bad arity for " ^ op))
	     end
	 | _ -> raise (Eval_error "wrong arguments for term"))
    ;

    "equations",
    ("(equations F X s) builds from the string s a list of equalities between terms (built over the signature F and the variables X). This set of equalities is usually considered as an equational specification to be completed.",
     3,
     Non_generalized
       (Dep_arrow
	  (ref Depend_var, 
	   signature_type, 
	   fun sigma -> 
	     Dep_arrow
	       (ref Depend_var, 
		variable_set_type, 
		fun vars ->
		  arrow_type string_type (equations_type sigma vars)))),
       function 
	   [Sig_value sigma;Var_value vars;String_value chaine] ->
	     begin
	       try
		 let l = 
		   User_terms.equation_set_of_string 
		     sigma.sig_val User_terms.user_hct vars.vars_val chaine
		 in
		 Equations_value(sigma,vars,l)
	       with Term_syntax.Syntax_error s ->
		 raise (Eval_error ("Syntax error in equations ("^s^")"))
	     end
	 | _ -> raise (Eval_error "wrong arguments for equations"))
    ;

     
    
    "conjecture",
    ("(conjecture F X s) builds from the string s an equality between terms (built over the signature F and the variables X). This equality is usually considered as a conjecture which is intended to be proven.",
     3,
     Non_generalized
       (Dep_arrow
	  (ref Depend_var, 
	   signature_type, 
	   fun sigma -> 
	     Dep_arrow
	       (ref Depend_var, 
		variable_set_type, 
		fun vars ->
		  arrow_type string_type (conjecture_type sigma vars)))),
       function 
	   [Sig_value sigma;Var_value vars;String_value chaine] ->
	     begin
	       try
		 let e = 
		   User_terms.equation_of_string 
		     sigma.sig_val User_terms.user_hct vars.vars_val chaine
		 in
		 Conjecture_value(sigma,vars,e)
	       with Term_syntax.Syntax_error s ->
		 raise (Eval_error ("Syntax error in conjecture ("^s^")"))
	     end
	 | _ -> raise (Eval_error "wrong arguments for conjecture"))
    ;

     
    
(*i
    "hterm",
    ("(hterm M X s) builds the term of module M on variables X defined by string s.", 
     3,
     Generalized
       (fun sigma ->
	  Non_generalized
	    (arrow_type htrs_type 
	       (Dep_arrow
		  (ref Depend_var, variable_set_type,
		   (fun vars ->
		      (arrow_type string_type (term_type sigma vars))))))),
       function 
	   [HTRS_value m ; Var_value vars ; String_value chaine] ->
	     begin
	       let hsig_val = m.Hierarchical_trs.htrs_sig in
	       let sigma = { hsig_name = "" ; hsig_val = hsig_val } in
	       try
		 Term_value(sigma,vars,
			    User_terms.term_of_string 
			      sig_val vars.vars_val chaine)
	       with 
		   Term_syntax.Syntax_error s ->
		     raise (Eval_error ("Syntax error in term ("^s^")"))
		 | Gen_terms.Bad_arity op ->
		     raise (Eval_error ("Bad arity for " ^ op))
	     end
	 | _ -> raise (Eval_error "wrong arguments for hterm"))
    ;
i*) 
    "precedence",
    ("(precedence S s) builds the precedence on signature S given by string s",
     2,
     Non_generalized
       (Dep_arrow
	  (ref Depend_var, signature_type,
	   fun sigma -> 
	     arrow_type string_type (precedence_type sigma))),
       function 
	   [Sig_value sigma;String_value chaine] ->
	     begin
	       try
		 Precedence_value(sigma,
				  User_terms.prec_of_string 
				    sigma.sig_val chaine)
	       with 
		   Term_syntax.Syntax_error s ->
		     raise (Eval_error ("Syntax error in precedence ("^s^")"))
		 | Finite_orderings.Incompatible ->
		     raise (Eval_error "Incompatible precedence")
	     end
	 | _ -> raise (Eval_error "wrong arguments for precedence"))
    ;

    "status",
    ("(status S s) builds the status function on signature S given by string s",
     2,
     Non_generalized
       (Dep_arrow
	  (ref Depend_var, signature_type,
	   fun sigma -> 
	     arrow_type string_type (status_type sigma))),
       function 
	   [Sig_value sigma;String_value chaine] ->
	     begin
	       try
		 Status_value(sigma,
			      User_terms.status_of_string 
				    sigma.sig_val chaine)
	       with 
		   Term_syntax.Syntax_error s ->
		     raise (Eval_error ("Syntax error in status ("^s^")"))
		 | User_terms.Invalid_status s ->
		     raise (Eval_error ("Incompatible status for symbol "^s))
	     end
	 | _ -> raise (Eval_error "wrong arguments for status"))
    ;

    "MPO",
    ("(MPO p) is the multiset path ordering on terms defined the precedence p.",
     1,
     Generalized
       (fun sigma ->
	  Non_generalized
	    (arrow_type 
	       (precedence_type sigma) (term_ordering_type sigma))),
       function 
	   [Precedence_value(sigma,prec)] ->
	     Term_ordering_value(sigma,
				 Rpo.rpo prec (fun _ -> Rpo.Multiset))
	 | _ -> raise (Eval_error "wrong arguments for MPO"))
    ;
    
    "RPO",
    ("(RPO s p) is the usual recursive path ordering on terms defined by the status s and the precedence p.",
     2,
     Generalized
       (fun sigma ->
	  Non_generalized
	    (arrow_type 
	       (status_type sigma) 
	       (arrow_type 
		  (precedence_type sigma) (term_ordering_type sigma)))),
       function 
	   [Status_value(sigma,status); Precedence_value(_,prec)] ->
	     Term_ordering_value(sigma,
				 Rpo.rpo prec status)
	 | _ -> raise (Eval_error "wrong arguments for RPO"))
    ;

    "ACRPO",
    ("(ACRPO s p) is the AC path ordering on terms defined by the status s and the precedence p.",
     2,
     Generalized
       (fun sigma ->
	  Non_generalized
	    (arrow_type 
	       (status_type sigma) 
	       (arrow_type 
		  (precedence_type sigma) (term_ordering_type sigma)))),
       function 
	   [Status_value(sigma,status); Precedence_value(_,prec)] ->
	     Term_ordering_value(sigma,
				 Acrpo.acrpo sigma.sig_val User_terms.user_hct prec status)
	 | _ -> raise (Eval_error "wrong arguments for ACRPO"))
    ;

    "POLY",
    ("(POLY S s) is the polynomial ordering on terms built over the signature S defined the string s.",
     2,
     Non_generalized
       (Dep_arrow
	  (ref Depend_var,
	   signature_type,
	   fun sigma -> 
	     arrow_type string_type (term_ordering_type sigma))),
    function 
	[Sig_value sigma;String_value chaine] ->
	  begin
	    try
	      Term_ordering_value(sigma, User_poly_ordering.of_string 
				    sigma.sig_val chaine)
	    with 
		Poly_interp_lexer.Syntax_error s ->
		  raise (Eval_error s)
	      | User_poly_ordering.Syntax_error s ->
		  raise (Eval_error s)
	      | Failure "lexing: empty token" -> 
		  raise (Eval_error "Problems in POLY")
	  end
      | _ -> raise (Eval_error "wrong arguments for POLY"))
    ;

    "term_compare",
    ("(term_compare o t1 t2) compares the terms t1 and t2 w.r.t. the term ordering o.",
     3,
     Generalized
       (fun sigma ->
	  Generalized
	    (fun vars ->
	       Non_generalized
		 (arrow_type 
		    (term_ordering_type sigma)
		    (arrow_type 
		       (term_type sigma vars) 
		       (arrow_type (term_type sigma vars) string_type))))),
       function 
	   [Term_ordering_value(sigma,order); 
	    Term_value(_,_,t1); Term_value(_,_,t2)] ->
	     String_value
	       (Orderings_generalities.string_of_comparison_result
		  (order t1 t2))
	 | _ -> raise (Eval_error "wrong arguments for term_compare"))
    ;

    "TRS",
    ("(TRS S s) builds the TRS on signature S defined by string s.",
     3,
     Non_generalized
       (Dep_arrow
	  (ref Depend_var, 
	   signature_type, 
	   fun sigma -> 
	     Dep_arrow
	       (ref Depend_var, 
		variable_set_type, 
		fun vars ->
		  arrow_type string_type (trs_type sigma vars)))),
       function 
	   [Sig_value sigma;Var_value vars;String_value chaine] ->
	     begin
	       try
		 let l = 
		   User_terms.rule_set_of_string 
		     sigma.sig_val User_terms.user_hct vars.vars_val chaine
		 in
		 TRS_value(sigma,vars,l,Naive_dnet.compile sigma.sig_val l)
	       with Term_syntax.Syntax_error s ->
		 raise (Eval_error ("Syntax error in TRS ("^s^")"))
	     end
	 | _ -> raise (Eval_error "wrong arguments for TRS"))
    ;
    
    "HTRS",
    ("(HTRS H S V s) extends the set H of HTRSs by adding signature S and rules defined by string s on variables V.",
     4,
     Non_generalized
       (arrow_type (set_type htrs_type)
	  (Dep_arrow
	     (ref Depend_var, 
	      signature_type, 
	      fun sigma -> 
		Dep_arrow
		  (ref Depend_var, 
		   variable_set_type, 
		   fun vars ->
		     arrow_type string_type htrs_type)))),
       function 
	   [Set_value set; Sig_value sigma; Var_value vars;
	    String_value chaine] ->
	     begin
	       try
		 let imported_htrs = 
		   List.map 
		     (function 
			| HTRS_value(h) -> h
			| _ ->
			    raise (Eval_error "wrong arguments for HTRS"))
		     set
		 in
		 let htrs = 
		   Hierarchical_trs.extend 
		     imported_htrs sigma.sig_val vars.vars_val chaine 
		 in
		   HTRS_value(htrs)
	       with Term_syntax.Syntax_error s ->
		 raise (Eval_error ("Syntax error in HTRS ("^s^")"))
	     end
	 | _ -> raise (Eval_error "wrong arguments for HTRS"))
    ;
    

    "normalize",
    ("normalize R t computes the normal of term t w.r.t. TRS R, with the innermost strategy. Beware that this function will loop if t has no normal form. You may interrupt computation with Ctrl-C.",
     2,
     Generalized
       (fun sigma -> 
	  Generalized
	    (fun vars -> 
	       Non_generalized
		 (arrow_type 
  	            (trs_type sigma vars)
	            (arrow_type (term_type sigma vars) 
		       (term_type sigma vars))))),
       function
	   [ TRS_value(_,_,rules,d) ; Term_value(sigma,vars,t) ] ->
	     let t = 
	       Innermost.innermost_normalize 
		 (fun _ -> raise Not_found) (fun _ _ -> ()) 
		 sigma.sig_val User_terms.user_hct
		 (Rewriting.compiled_ac_rewrite_at_top 
		    sigma.sig_val User_terms.user_hct Confluence.user_phct d) t
	     in Term_value(sigma,vars,t)
	 | _ -> raise (Eval_error "wrong arguments for normalize"))
    ;
    "traced_normalize",
    ("normalize R t computes the normal of term t w.r.t. TRS R, with the innermost strategy. Beware that this function will loop if t has no normal form. You may interrupt computation with Ctrl-C. (prints out the trace)",
     2,
     Generalized
       (fun sigma -> 
	  Generalized
	    (fun vars -> 
	       Non_generalized
		 (arrow_type 
  	            (trs_type sigma vars)
	            (arrow_type (term_type sigma vars) 
		       (term_type sigma vars))))),
       function
	   [ TRS_value(_, _, rules, d) ; Term_value(sigma, vars, t) ] ->
	     let (t, tr) = 
	       Traced_rewriting.traced_innermost_normalize 
		 User_terms.user_hct
		 (Traced_rewriting.new_compiled_rewrite_at_top d) t
	     in
	       tr#print (sigma.sig_val :> User_signatures.symbol_id Signatures.signature) vars.vars_val;
	       (Term_value(sigma, vars, t))
	 | _ -> raise (Eval_error "wrong arguments for traced_normalize"))
    ;
    ]
;;


(* 

\subsection{Functions on termination}

*)

let termination_table =
  [
    "termcrit",
    ("termcrit \"c\" selects the termination criterion c, either standard, minimal, dp, marks, nomarks, graph or nograph. See manual for details.",
     1,
     Non_generalized(arrow_type string_type unit_type),
     function
	 [ String_value(t) ] ->
	   begin
	     match t with
		 "minimal" ->
(*		   Termination_expert.standard_criterion := false;*)
		   Termination_expert.enable_split := true;
		   print_string "Termination now uses minimal decomposition";
	       | "dp" -> 
		   Termination_expert.standard_criterion := false;
		   print_string "Termination now uses dependency pair criterion";
	       | "standard" -> 
		   Termination_expert.standard_criterion := true;
		   Termination_expert.enable_split := false;
		   print_string "Termination now uses the standard criterion";
	       | "marks" -> 
		   Termination_expert.dp_criterion_uses_markings := true;
		   print_string "Dependency pair criterion now uses marks";
	       | "nomarks" -> 
		   Termination_expert.dp_criterion_uses_markings := false;
		   print_string "Dependency pair criterion now do not use marks";
	       | "graph" -> 
		   Termination_expert.dp_criterion_uses_graph := true;
		   print_string "Dependency pair criterion now uses graph";
	       | "scparts" -> 
		   Graph_criteria.uses_all_strongly_connected_parts := true;
		   print_string "Graph criterion now uses strongly connected parts";
	       | "noscparts" -> 
		   Graph_criteria.uses_all_strongly_connected_parts := false;
		   print_string "Graph criterion now do not use strongly connected parts";	       
	       | "nograph" -> 
		   Termination_expert.dp_criterion_uses_graph := false;
		   print_string "Dependency pair criterion now do not use graph";
	       | _ -> 
		   print_string "Unknown termination criterion option: should be one of dp, standard, marks, nomarks, graph, nograph, scparts, noscparts, minimal"
	   end;
	   print_newline();
	   Unit_value
       | _ -> raise (Eval_error "wrong argument for termcrit"))
    ;    

    "polyinterpkind",
    ("polyinterpkind \"c\" selects the kind c of polynomial interpretations, either linear, simple, or simple-mixed. See manual for details.",
     1,
     Non_generalized(arrow_type string_type unit_type),
     function
	 [ String_value(t) ] ->
	   begin
	     match t with
		 "linear" -> 
		   Generic_polynomials.use_linear_interp := true;
		   Generic_polynomials.use_degree_two_interp := false;
		   print_string "Now searching only linear polynomial interpretations";
	       | "simple" -> 
		   Generic_polynomials.use_linear_interp := false;
		   Generic_polynomials.use_degree_two_interp := false;
		   print_string "Now searching simple polynomial interpretations";
	       | "simple-mixed" -> 
		   Generic_polynomials.use_linear_interp := false;
		   Generic_polynomials.use_degree_two_interp := true;
		   print_string "Now searching simple-mixed polynomial interpretations";
	       | _ -> 
		   print_string "Unknown polynomial interpretation kind: should be one of linear, simple, or simple-mixed"
	   end;
	   print_newline();
	   Unit_value
       | _ -> raise (Eval_error "wrong argument for polyinterpkind"))
    ;    

    "termpolybound",
    ("Selects the search bound on coefficients of polynomial interpretations.",
     1,
     Non_generalized(arrow_type int_type unit_type),
     function
	 [ Int_value(n) ] ->
	   begin
	     try
	       let b = Numbers.to_int n
	       in
		 Termination_expert.bound := b;
		 print_string ("Search bound of polynomial coefficients is now "^(string_of_int b))
		   
	     with Failure _ ->
	       print_string "bound is too large"
	   end;
	   print_newline();
	   Unit_value
       | _ -> raise (Eval_error "wrong arguments for termpolybound"))
    ;
    "printstate",
    ("(printstate \"c\") prints the global state of CiME component c, i.e. \"expert\" for termination expert (presently only one available).",
     1,
     Non_generalized(arrow_type string_type unit_type),
     function
	 [String_value(t)] -> 
      	   begin
	     match t with
		 "expert" -> 
		   begin
		     print_string "The termination expert now uses:";
		     print_newline();
		     if !Termination_expert.standard_criterion 
		     then (print_string "  Standard criterion."; 
			   print_newline())
		     else (print_string
			     ((if !Termination_expert.dp_criterion_uses_markings 
			       then "  Marked" 
			       else "  Unmarked")^
			      " Dependency Pairs criterion");
			   print_newline();
			   print_string ("  with"^
					 (if !Termination_expert.dp_criterion_uses_graph 
					  then " " 
					  else "out ")^
					 "Dependency Graphs.");
			   if !Graph_criteria.uses_all_strongly_connected_parts
			   then (print_newline(); 
				 print_string "  Graphs are split up into strongly connected parts");
			   if !Termination_expert.enable_split 
			   then (print_newline(); 
				 print_string "  Minimal Decomposition is enabled.");
			   print_newline());		
		     print_string ("Bound of polynomial coefficient is "^
				   (string_of_int !Termination_expert.bound));
		     print_newline();
		     print_string ("Searched Interpretations are "^
				   (if !Generic_polynomials.use_linear_interp 
				    then "Linear."
				    else 
				      if !Generic_polynomials.use_degree_two_interp 
				      then "Simple-Mixed."
				      else "Simple.")
				  );
		     print_newline()
		   end;
		   Unit_value
	       | _ -> raise (Eval_error "Unknown CiME component, it should be one of: expert.")
	   end
       | _ -> raise (Eval_error "wrong arguments for printstate"))
    ;
    "assume_ce",
    ("(assume_ce H) enable the termination analysis to assume that htrs H is CE-terminating.",
     1,
     Non_generalized(arrow_type htrs_type unit_type),
     function 
	 [HTRS_value(h)] ->
	  
	     Hierarchical_trs.assume_ce h;
	     print_string "Assuming CE-termination";
	     Unit_value
	       
       | _ -> raise (Eval_error "wrong arguments for assume_ce"))
    ;

    "clear_ce",
    ("(clear_ce H) remove assumptions over CE-termination of htrs H.",
     1,
     Non_generalized(arrow_type htrs_type unit_type),
     function 
	 [HTRS_value h] ->
	   
	   Hierarchical_trs.clear_ce h;
	   print_string "No more assumption over CE-termination";
	   print_newline();
	   Unit_value
	     
       | _ -> raise (Eval_error "wrong arguments for clear_ce"))
    ;

    "check_ce",
    ("(check_ce H) check assumptions over CE-termination of htrs H.",
     1,
     Non_generalized
       (arrow_type htrs_type unit_type),
       function 
	   [HTRS_value h] ->

	     if !(h.Hierarchical_trs.proved_c_e_terminating)
	     then 
	       begin
		 print_string "Proven CE-terminating";
		 print_newline()
	       end
	     else 
	       begin
		 print_string "Not yet proven CE-terminating";
		 print_newline()
	       end;
	     Unit_value
	       
	 | _ -> raise (Eval_error "wrong arguments for clear_ce"))
    ;

    "h_termination",
    ("Tries to prove termination of a HTRS.",
     1,
     Generalized(
       fun sigma ->
	 Generalized(
	   fun vars ->
	     Non_generalized(
	       (arrow_type (htrs_type) (* sigma vars)*) unit_type)))),
     function
	 [ HTRS_value(r) ] ->
	     Termination_expert.modular_expert Variables.default r [];
	   Unit_value
	     
       | _ -> raise (Eval_error "wrong arguments for h_termination"))
    ;
(**)
    "h_termination_with",
    ("Tries to prove termination of a HTRS with tuning of the expert.",
     2,
     Generalized(
       fun sigma ->
	 Generalized(
	   fun vars ->
	     Non_generalized(
	       (arrow_type (set_type (pair_type string_type int_type))
		  (arrow_type (htrs_type) unit_type))))),
     function
  [Set_value plist_cime ; HTRS_value(r) ] ->
  let plist = List.map 
               (function 
		    (Prod_value [String_value s; Int_value i]) -> 
		      let bound =  Numbers.to_int i in
			(match s with
			   | "linear"  -> 
			       Generic_polynomials.Linear, bound 
			   | "simple" -> 
			       Generic_polynomials.Simple, bound
			   | "simple-mixed" -> 
			       Generic_polynomials.Simple_Mixed, bound 
			   | _ -> raise (Eval_error "Unknown polynomial interpretation kind: should be one of linear, simple, or simple-mixed" ))
       | _ -> assert false) 
    plist_cime in
    Termination_expert.modular_expert Variables.default r plist;
    Unit_value	     
  | _ -> raise (Eval_error "wrong arguments for h_termination_with"))
    ;
(**)
    "termination",
    ("Tries to prove termination of a TRS.",
     1,
     Generalized(
       fun sigma ->
	 Generalized(
	   fun vars ->
	     Non_generalized(
	       (arrow_type (trs_type sigma vars) unit_type)))),
     function
	 [ TRS_value(sigma,vars,r,_) ] ->
	   Termination_expert.expert sigma.sig_val vars.vars_val r;
	   Unit_value
	     
       | _ -> raise (Eval_error "wrong arguments for termination"))
    ;

    "dependency_pairs",
    ("Prints the dependency pairs of a TRS.",
     1,
     Generalized
       (fun sigma ->
	  Generalized
	    (fun vars ->
	       Non_generalized
		 (arrow_type (trs_type sigma vars) unit_type))),
       function
	   [ TRS_value(sigma,vars,r,_) ] ->
	     Termination_expert.show_pairs sigma.sig_val 
	       User_terms.user_hct Termination_expert.user_dhct vars.vars_val r;
	     Unit_value
	       
	 | _ -> raise (Eval_error "wrong arguments for dependency_pairs"))
    ;

    "order_constraint",
    ("(order_constraint F X s) builds the order constraint syntactically from string s.", 
     3,
     Non_generalized
       (Dep_arrow
	  (ref Depend_var, 
	   signature_type,
	   fun sigma -> 
	     Dep_arrow
	       (ref Depend_var, 
		variable_set_type,
		fun vars ->
		  arrow_type string_type (order_constraint_type sigma vars)))),
       function 
	   [Sig_value sigma;Var_value vars;String_value chaine] ->
	     begin
	       try
		 Order_constraint(sigma,vars,
			    User_terms.order_constraint_of_string 
			      sigma.sig_val User_terms.user_hct vars.vars_val chaine)
	       with 
		   Term_syntax.Syntax_error s ->
		     raise (Eval_error ("Syntax error in term ("^s^")"))
		 | Gen_terms.Bad_arity op ->
		     raise (Eval_error ("Bad arity for " ^ op))
	     end
	 | _ -> raise (Eval_error "wrong arguments for term"))
    ;

    "check_order_constraint",
    ("(check_order_constraint o f) checks whether o satisfies f.", 
     2,
     Generalized
       (fun sigma ->
	  Generalized
	    (fun vars ->
	       Non_generalized
		 (arrow_type (term_ordering_type sigma) (arrow_type (order_constraint_type sigma vars) bool_type)))),
       function
	   [ Term_ordering_value(_,o) ; Order_constraint(_,_,c) ] ->
	     let b = Check_order_constraints.check_order_constraint o c in
	     Bool_value(b)
	       
	 | _ -> raise (Eval_error "wrong arguments for check_order_constraint"))
    ;


    "all_decrease_termination_constraints",
    ("builds the termination constraint obtained by asking each rule of a TRS strictly decreasing.",
     1,
     Generalized
       (fun sigma ->
	  Generalized
	    (fun vars ->
	       Non_generalized
		 (arrow_type (trs_type sigma vars) (termination_constraint_type sigma vars)))),
       function
	   [ TRS_value(sigma,vars,r,_) ] ->
	     let c = 
	       Basic_criterion.all_decreasing_constraints 
		 r
	     in
	     Termination_constraint(sigma,vars,c)
	       
	 | _ -> raise (Eval_error "wrong arguments for all_decrease_termination_constraints"))
    ;

    

    "termination_constraint_without_marks",
    ("computes the constraint for proving termination of a TRS, with respect to the current termination criterion.",
     1,
     Generalized
       (fun sigma ->
	  Generalized
	    (fun vars ->
	       Non_generalized
		 (arrow_type (trs_type sigma vars) (termination_constraint_type sigma vars)))),
       function
	   [ TRS_value(sigma,vars,r,_) ] ->
	     let c = 
	       Termination_expert.compute_termination_constraints_without_marks
		 sigma.sig_val vars.vars_val r
	     in
	     Termination_constraint(sigma,vars,c)
	       
	 | _ -> raise (Eval_error "wrong arguments for termination_constraint_without_marks"))
    ;

(*

    "termination_constraint_with_marks",
    ("computes the constraint for proving termination of a TRS, with respect to the current termination criterion.",
     1,
     Generalized
       (fun sigma ->
	  Generalized
	    (fun vars ->
	       Non_generalized
		 (arrow_type (trs_type sigma vars) (termination_constraint_type sigma vars)))),
       function
	   [ TRS_value(sigma,vars,r) ] ->
	     let sigma' = Marked_dp_criteria.signature_with_marks sigma.sig_val
	     and c = 
	       Termination_expert.compute_termination_constraints_with_marks
		 sigma.sig_val vars.vars_val r
	     in
	     Termination_constraint(
	       { sig_name = "'" ^ sigma.sig_name ^ "`";
		 sig_val = sigma' },vars,c)
	       
	 | _ -> raise (Eval_error "wrong arguments for termination_constraint_without_marks"))
    ;

*)
    "check_termination_constraint",
    ("checks whether a termination constraint is satisfied by an ordering.",
     2,
     Generalized
       (fun sigma ->
	  Generalized
	    (fun vars ->
	       Non_generalized
		 (arrow_type (termination_constraint_type sigma vars) 
		    (arrow_type (term_ordering_type sigma) unit_type)))),
       function
	   [ Termination_constraint(sigma,vars,constr) ;
	     Term_ordering_value(_,order) ] ->
	     let c = 
	       Termination_constraints.check_conj_constraints
		 sigma.sig_val vars.vars_val constr order
	     in
	     Unit_value
	       
	 | _ -> raise (Eval_error "wrong arguments for check_termination_constraints"))
    ;
    "poly_solve_termination_constraint",
    ("tries to find a polynomial ordering satisfying a termination constraint.",
     2,
     Generalized
       (fun sigma ->
	  Generalized
	    (fun vars ->
	       Non_generalized
		 (arrow_type int_type 
		    (arrow_type (termination_constraint_type sigma vars) 
		       unit_type)))),
       function
	   [ Int_value(n) ; 
	     Termination_constraint(sigma,vars,constr) ] ->
	     let c = 
	       Generic_polynomials.poly_solve_termination_constraint
		 n sigma.sig_val constr 
	     in
	     Unit_value
	       
	 | _ -> raise (Eval_error "wrong arguments for poly_solve_termination_constraint"))
    ;
  ]
;;


(* 

\subsection{Functions on matching and unification}

*)

let unification_table =
  [
    "ac_match",
    ("ac_match pat subj matches subj against pat, modulo AC, and displays the result",
     2,
     Generalized(
       fun sigma ->
	 Generalized(
	   fun vars ->
	     Non_generalized(
	       (arrow_type 
		  (term_type sigma vars)
		  (arrow_type (term_type sigma vars) unit_type))))),
     function
	 [ Term_value(sigma,vars,t1); Term_value(_,_,t2)] ->
	   begin
	     try
	       let subst = 
		 Matching.matching sigma.sig_val User_terms.user_hct 
		   Confluence.user_phct t1 t2 in
	       Substitution.print sigma.sig_val vars.vars_val subst
	     with
		 Matching.No_match -> 
		   printf "No match@."
	   end;
	   Unit_value
	     
       | _ -> raise (Eval_error "wrong arguments for ac_unify"))
    ;
    "all_ac_matchs",
    ("all_ac_matchs pat subj matches subj against pat, modulo AC, and displays the solutions",
     2,
     Generalized(
       fun sigma ->
	 Generalized(
	   fun vars ->
	     Non_generalized(
	       (arrow_type 
		  (term_type sigma vars)
		  (arrow_type (term_type sigma vars) unit_type))))),
     function
	 [ Term_value(sigma,vars,t1); Term_value(_,_,t2)] ->
	   Matching.all_matchs sigma.sig_val User_terms.user_hct 
	     Confluence.user_phct vars.vars_val t1 t2;
	   Unit_value
	     
       | _ -> raise (Eval_error "wrong arguments for ac_unify"))
    ;
     
    "ac_unify",
    ("ac_unify t1 t2 displays a complete set of AC unifiers for t1=t2",
     2,
     Generalized(
       fun sigma ->
	 Generalized(
	   fun vars ->
	     Non_generalized(
	       (arrow_type 
		  (term_type sigma vars)
		  (arrow_type (term_type sigma vars) unit_type))))),
     function
	 [ Term_value(sigma,vars,t1); Term_value(_,_,t2)] ->
	   Ac_unification.display_set_of_solutions 
	     sigma.sig_val User_terms.user_hct vars.vars_val t1 t2;
	   Unit_value
	     
       | _ -> raise (Eval_error "wrong arguments for ac_unify"))
    ;
    "unify",
    ("unify E t1 t2 displays a complete set of E-unifiers for t1=t2",
     3,
     Generalized(
       fun sigma ->
	 Generalized(
	   fun vars ->
	     Non_generalized(
	       (arrow_type 
		  (theo_type sigma)
		  (arrow_type 
		     (term_type sigma vars)
		     (arrow_type (term_type sigma vars) unit_type)))))),
     function
	 [Theo_value(sigma,th) ; Term_value(_,vars,t1) ; Term_value(_,_,t2)] ->
	   Unification.display_plain_set_of_solutions
	     sigma.sig_val User_terms.user_hct th vars.vars_val t1 t2;
	   Unit_value
       | _ -> raise (Eval_error "wrong arguments for unify"))
  ]
;;


(*

\subsection{Equational Logic}

*)
(*
let equational_logic_table =
  [
    "use_equation_at_position",
    ("use_equation_at_position e p t uses the equation e at position p in the term t",
     3,
     Generalized(
       fun sigma ->
	 Generalized(
	   fun vars ->
	     Non_generalized(
	       (arrow_type (term_type sigma vars) 
		  (arrow_type (term_type sigma vars) 
		     (arrow_type position_type 
			(arrow_type (term_type sigma vars) unit_type))))))),
     function
	 [Term_value(sigma,vars,l) ; Term_value(_,_,r) ; 
	  Position_value(p) ; Term_value (_,_,t)] ->
	    Equational_logic.apply_equation_at_position_to_term (l,r) p t;
	  Unit_value
	 | _ -> raise (Eval_error "wrong arguments for use_equation_at_position"))
  ]
;;

*)
(* 

\subsection{Knuth-Bendix completion}

*)

let kb_table =
  [
    "confluence",
    ("(confluence R) checks (local) confluence of the rewrite system R.",
     1,
     Generalized(
       fun sigma ->
	 Generalized(
	   fun vars ->
	     Non_generalized(
	       (arrow_type (trs_type sigma vars)
		  bool_type)))),
     function
	 [TRS_value(sigma,vars,rules,_)] ->
	   begin
	     if sigma.sig_val#contains_only_free_symbols
	     then
	       let b = 
		   Confluence.StandardConfluence.is_confluent
		     (sigma.sig_val :> 'symbol Signatures.signature) 
		     vars.vars_val rules
		 in
		 Bool_value(b)
             else
	       let b = 
		   Confluence.ACConfluence.is_confluent
		     (sigma.sig_val :> 'symbol Signatures.signature) 
		     vars.vars_val rules
		 in
		 Bool_value(b)
	   end
       | _ -> raise (Eval_error "wrong arguments for confluence"))
    
    ;
    
    "print_all_cps",
    ("(print_all_cps R) prints all non-joignable critical pairs of the rewrite system R.",
     1,
     Generalized(
       fun sigma ->
	 Generalized(
	   fun vars ->
	     Non_generalized(
	       (arrow_type (trs_type sigma vars)
		  unit_type)))),
     function
	 [TRS_value(sigma,vars,rules,_)] ->
	   begin
	     if sigma.sig_val#contains_only_free_symbols
	     then
	       let _ = 
		   Confluence.StandardConfluence.print_all_critical_pairs
		     (sigma.sig_val :> 'symbol Signatures.signature) 
		     vars.vars_val rules
		 in
		 Unit_value
             else
	       let _ = 
		   Confluence.ACConfluence.print_all_critical_pairs
		     (sigma.sig_val :> 'symbol Signatures.signature) 
		     vars.vars_val rules
		 in
		 Unit_value
	   end
       | _ -> raise (Eval_error "wrong arguments for confluence"))
    
    ;
    
    "complete_trs",
    ("(complete_trs o E) returns the completion (according to Knuth-Bendix method) of the equational presentation E, with term ordering o. It may raise an exception in case of failure",
     2,
     Generalized(
       fun sigma ->
	 Generalized(
	   fun vars ->
	     Non_generalized(
	       (arrow_type (term_ordering_type sigma)
		  (arrow_type (equations_type sigma vars)
		     (trs_type sigma vars)))))),

     function
	 [Term_ordering_value(sigma,ordering); Equations_value(_,vars,rules)] ->
	    begin
	      if sigma.sig_val#contains_only_free_symbols
	      then
		try
		  let l = 
		    Kb.StandardKBCompletion.complete
		      (sigma.sig_val :> 'symbol Signatures.signature) 
		      vars.vars_val ordering rules [] in
		    TRS_value(sigma,vars,l,Naive_dnet.compile sigma.sig_val l)
		with
		    Kb.StandardKBCompletion.KB_failed -> 
		      raise (Eval_error "Completion failed")
		  | Kb.StandardKBCompletion.KB_aborted -> 
		      raise (Eval_error "Completion aborted")
	      else 
		try
		  let l = 
		    Kb.ACCompletion.complete
		      (sigma.sig_val :> 'symbol Signatures.signature) 
		      vars.vars_val ordering rules [] in
		    TRS_value(sigma,vars,l, Naive_dnet.compile sigma.sig_val l)
		with 
		    Kb.ACCompletion.KB_failed -> 
		      raise (Eval_error "Completion failed")
		  | Kb.ACCompletion.KB_aborted -> 
		      raise (Eval_error "Completion aborted")
	    end
       | _ -> raise (Eval_error "wrong arguments for complete_trs"))
    ;

  
    "completion_coeff_size",
    ("not documented",
     1,
     Non_generalized(arrow_type int_type unit_type),
     function
       | [Int_value(n)] ->
	   let n = Numbers.to_int n in
	   Kb.coeff_size := n;
	   printf "Completion_coeff_size is now %d@." n;
	   if (!Kb.coeff_size + !Kb.coeff_age) = 0
	   then printf "Be careful, both completion coefficients are equal to %d@ now." 0;
	   Unit_value
       | _ -> raise (Eval_error "wrong arguments for completion_coeff_size"))
	     
    ;
    
    "completion_coeff_age",
    ("not documented",
     1,
     Non_generalized(arrow_type int_type unit_type),
     function
       | [Int_value(n)] -> 
	   let n = Numbers.to_int n in
	   Kb.coeff_age:= n;
	   printf "Completion_coeff_age is now %d@." n;
	   if (!Kb.coeff_size + !Kb.coeff_age) = 0
	   then printf "Be careful, both completion coefficients are equal to %d@ now." 0;
	   Unit_value
       | _ -> raise (Eval_error "wrong arguments for completion_coeff_age"))


    ;    
    "prove_conj_by_ordered_completion",
    ("(prove_conj_by_ordered_completion o R C) tries to prove the set of conjectures C modulo the equational theory defined by the term rewrite system R by completion (according to the term ordering o), and in case of success, returns true and false otherwise.",
     3,
     Generalized(
       fun sigma ->
	 Generalized(
	   fun vars ->
	     Non_generalized(
	       (arrow_type (term_ordering_type sigma)
		  (arrow_type (equations_type sigma vars)
		     (arrow_type (equations_type sigma vars)
			bool_type)))))),

     function
	 [Term_ordering_value(sigma,ordering); Equations_value(_,vars,rules); 
	  Equations_value(_,_,conjectures)] ->
	    begin
	      try
		let module TO = 
		    struct
		      type t = User_signatures.symbol_id Gen_terms.term
		      let o = ordering
		    end in
		  if sigma.sig_val#contains_only_free_symbols
		  then
		    let module OrderedKBCompletion =
		      Kb.OrderedKBCompletion (TO) in
		    let _ = 
		      OrderedKBCompletion.verbose := 
		      !Kb.StandardKBCompletion.verbose in
		      try
			let b = 
			  OrderedKBCompletion.prove_conj_by_comp
			    (sigma.sig_val :> 'symbol Signatures.signature) 
			    vars.vars_val ordering rules conjectures in
			  Bool_value b
		      with
			  OrderedKBCompletion.KB_failed -> 
			    raise (Eval_error "Completion failed")
			| OrderedKBCompletion.KB_aborted -> 
			    raise (Eval_error "Completion aborted")
			| OrderedKBCompletion.KB_conjecture_proved _ -> 
			    raise (Eval_error "All conjectures proved but completion halted")
			| Rewrite_rules.Rule_with_a_var_lhs ->
			    Format.print_string "Be carefull, all models are trivial.\n";
			    Bool_value true
		  else 
		    let module OrderedACCompletion =
		      Kb.OrderedACCompletion (TO) in
		    let _ = OrderedACCompletion.verbose :=
			    !Kb.StandardKBCompletion.verbose in
		      try
			let b = 
			  OrderedACCompletion.prove_conj_by_comp
			    (sigma.sig_val :> 'symbol Signatures.signature) 
			    vars.vars_val ordering rules conjectures in
			  Bool_value b
		      with 
			  OrderedACCompletion.KB_failed -> 
			    raise (Eval_error "Completion failed")
			| OrderedACCompletion.KB_aborted -> 
			    raise (Eval_error "Completion aborted")
			| OrderedACCompletion.KB_conjecture_proved _ -> 
			    raise (Eval_error "All conjectures proved but completion halted")	
			| Rewrite_rules.Rule_with_a_var_lhs ->
			    Format.print_string "Be carefull, all models are trivial.\n";
			    Bool_value true
  
	      with 
		  Failure s -> raise (Eval_error s)
		(* | _ -> raise (Eval_error "Problems in prove_conj_by_completion")*)
	      end
       | _ -> raise (Eval_error "wrong arguments for prove_conj_by_completion"))
  ]

(* 

\subsection{Predefined arithmetic operators}

*)


let arith_table =
  [
    "_minus",
    ("",
     1,
     Non_generalized(arrow_type int_type int_type),
     function 
	 [Int_value x] ->
	   Int_value(Numbers.minus x)
       | _ -> raise (Eval_error "wrong arguments for minus"))
    ;

    "+",
    ("",
     2,
     bin_arith_type,
     function 
	 [Int_value x;Int_value y] ->
	   Int_value(Numbers.add x y)
       | _ -> raise (Eval_error "wrong arguments for +"))

    ;
    "=",
    ("",
     2,
     bin_comp_type,
     function 
	 [Int_value x;Int_value y] ->
	   Bool_value(Numbers.eq x y)
       | _ -> raise (Eval_error "wrong arguments for ="))
    ;

    "<>",
    ("",
     2,
     bin_comp_type,
     function 
	 [Int_value x;Int_value y] ->
	   Bool_value(Numbers.neq x y)
       | _ -> raise (Eval_error "wrong arguments for <>"))
    ;

    ">",
    ("",
     2,
     bin_comp_type,
     function 
	 [Int_value x;Int_value y] ->
	   Bool_value(Numbers.gt x y)
       | _ -> raise (Eval_error "wrong arguments for >"))
    ;

    ">=",
    ("",
     2,
     bin_comp_type,
     function 
	 [Int_value x;Int_value y] ->
	   Bool_value(Numbers.ge x y)
       | _ -> raise (Eval_error "wrong arguments for >="))
    ;
    
    "<",
    ("",
     2,
     bin_comp_type,
     function 
	 [Int_value x;Int_value y] ->
	   Bool_value(Numbers.lt x y)
       | _ -> raise (Eval_error "wrong arguments for <"))
    ;

    "<=",
    ("",
     2,
     bin_comp_type,
     function 
	 [Int_value x;Int_value y] ->
	   Bool_value(Numbers.le x y)
       | _ -> raise (Eval_error "wrong arguments for <="))
    ;

    "-",
    ("",
     2,
     bin_arith_type,
     function 
	 [Int_value x;Int_value y] ->
	   Int_value(Numbers.sub x y)
       | _ -> raise (Eval_error "wrong arguments for -"))
    ;

    "*",
    ("",
     2,
     bin_arith_type,
     function 
	 [Int_value x;Int_value y] ->
	   Int_value(Numbers.mult x y)
       | _ -> raise (Eval_error "wrong arguments for *"))
    ;

    "and",
    ("",
     2,
     bin_bool_type,
     function 
	 [Bool_value x;Bool_value y] ->
	   Bool_value(x & y)
       | _ -> raise (Eval_error "wrong arguments for and"))
    ;

    "or",
    ("",
     2,
     bin_bool_type,
     function 
	 [Bool_value x;Bool_value y] ->
	   Bool_value(x or y)
       | _ -> raise (Eval_error "wrong arguments for or"))
    ;

    "not",
    ("",
     1,
     Non_generalized(arrow_type bool_type bool_type),
     function 
	 [Bool_value x] ->
	   Bool_value(not x)
       | _ -> raise (Eval_error "wrong arguments for not"))
    ;

    "^",
    ("",
     2,
      Non_generalized(arrow_type string_type (arrow_type string_type string_type)),
      function 
	 [String_value x;String_value y] ->
	   String_value(x ^ y)
       | _ -> raise (Eval_error "wrong arguments for ^"))
(*i

tests

    ;
    "first_ints",
    ("",
     1,
     Non_generalized(arrow_type int_type (set_type int_type)),
     let rec inter n m = 
       if n>m then [] else (Int_value(Num.num_of_int n))::(inter (succ n) m) 
     in
     function 
	 [Int_value n] ->
	   Set_value(inter 0 (Num.int_of_num n))
       | _ -> raise (Eval_error "wrong arguments for first_ints"))
    
    ; 
    "pair_ints",
    ("",
     2,
     Non_generalized(arrow_type int_type (arrow_type int_type (pair_type int_type int_type))),
     function 
	 [Int_value n;Int_value m] ->
	   Prod_value([Int_value n;Int_value m])
       | _ -> raise (Eval_error "wrong arguments for pair_ints"))
i*)
    ]
;;


let predef_table =
  [ "arith" , arith_table ;
    "dioph" , dioph_table ;
    "words" , words_table ;
    "pwords" , parameterized_words_table ;
    "terms" , terms_table ;
    "termination" , termination_table ;
    "unification" , unification_table ;
(*
    "equational logic" , equational_logic_table ;
*)
    "completion" , kb_table ; 
  ]
;;



let initial_env () =
  List.fold_right
    (fun (_,table) accu -> 
       List.fold_right
	 (fun (s,_) accu -> (s,{value=Predef(s,[])})::accu)
	 table
	 accu)
    predef_table
    []
;;

let initial_env_type () =
  List.fold_right
    (fun (_,table) accu -> 
       List.fold_right
	 (fun (s,(help,ar,t,c)) accu -> (s,t)::accu)
	 table
	 accu)
    predef_table
    []
;;

let code_table = Hashtbl.create 257;;

let _ =
  List.iter 
    (fun (_,table) -> 
       List.iter
	 (fun (s,(help,ar,t,c)) ->
	    Hashtbl.add code_table s (ar,c))
	 table)
    predef_table
;;


let get_code s =
  try
    Hashtbl.find code_table s
  with
      Not_found -> 
	raise (Eval_error ("name '"^s^"' not found in predefined environnement!"))
 

let get_predef_category_list () = List.map fst predef_table ;;

let get_predef_command_list category = 
  let t = List.assoc category predef_table in
  List.map fst t 

let get_help_of_command c =  
  List.fold_right
    (fun (cat,table) accu ->
       try
	 match List.assoc c table with (help,_,_,_) -> cat,help
       with
	   Not_found -> accu)
    predef_table
    ("Unknown category", "Unknown command.")
;;

let type_table = Hashtbl.create 257;;

let _ =
  List.iter 
    (fun (_,table) -> 
       List.iter
	 (fun (s,(help,ar,t,c)) ->
	    Hashtbl.add type_table s t)
	 table)
    predef_table
;;


let get_type_of_command c =  
  try
    Hashtbl.find type_table c
  with
      Not_found -> 
	raise (Eval_error ("name '"^c^"' not found in predefined environnement!"))





