(***************************************************************************

(This sentence has been added automatically,
it should be replaced by a description of the module)

CiME Project - Dmons research team - LRI - Universit Paris XI

$Id: typing.ml,v 1.26 2002/07/16 09:29:36 contejea Exp $

***************************************************************************)



open Abstract_syntax;;

open Format;;

exception Type_error of string;;


type depend = Depend_var | Anonymous | Depend_id of string ;;

type depend_name = depend ref;;

type types =
  | Base_type of depend_name list * types list * string 
  | Dep_arrow of depend_name * types * (depend_name -> types)
  | Type_variable of types option ref

and type_scheme =
    Generalized of (depend_name -> type_scheme)
  | Non_generalized of types

and typing_env_type = (string * type_scheme) list
;;




let generalize t = Non_generalized(t);;

    

let dep_of_expr global_env = function
    Var x -> 
      begin
	try 
	  let _ = List.assoc x global_env in ref (Depend_id x)
	with 
	    Not_found -> ref Anonymous
      end
  | _ -> ref (Depend_id "<unnamed>")
;;

let rec type_value v = 
  match v with
    | Type_variable x ->
	begin
	  match !x with
	      Some t -> 
		assert (not (t == v));
		type_value t
	    | None -> v
	end
    | _ -> v
;;

let rec specialize = function
    Non_generalized t -> t
  | Generalized f -> specialize (f (ref Depend_var))
;;


    
let unify_dep d1 d2 = 
  match (!d1,!d2) with
      (Depend_id id1,Depend_id id2) -> id1=id2
    | (Anonymous,Anonymous) -> true
    | (Depend_var,_) -> d1 := !d2; true
    | (_,Depend_var) -> d2 := !d1; true
    | (_,_) -> false
;;

let rec unify_types t1 t2 =
  match (t1,t2) with
    | (Type_variable x1,Type_variable x2) ->
	if x1==x2 then true else
	begin	  
	  match !x1 with
	      None -> x1 := Some t2; true
	    | Some t3 -> unify_types t3 t2
	end
    | (Type_variable x,_) ->
	begin
	  match !x with
	      None -> x := Some t2; true
	    | Some t3 -> unify_types t3 t2
	end
    | (_,Type_variable x) ->
	begin
	  match !x with
	      None -> x := Some t1; true
	    | Some t3 -> unify_types t1 t3
	end
    | (Base_type(l1,a1,s1),Base_type(l2,a2,s2)) -> 
	if s1=s2
	then 
	  try
	    List.for_all2 unify_dep l1 l2 
	    && List.for_all2 unify_types a1 a2 
	  with Invalid_argument _ -> false
	else false
    | (Dep_arrow(id1,t1,f1),Dep_arrow(id2,t2,f2)) ->
	if unify_types t1 t2
	then
	  if unify_dep id1 id2
	  then unify_types (f1 id1) (f2 id2)
	  else false
	else false
    | _ -> false
;;


let unit_type = Base_type([],[],"unit");;
let int_type = Base_type([],[],"int");;
let string_type = Base_type([],[],"string");;
let bool_type = Base_type([],[],"bool");;
let set_type t = Base_type([],[t],"set");;
let pair_type t1 t2 = Base_type([],[t1;t2],"");;

let dioph_constraint_type = Base_type([],[],"dioph_constraint");;
let signature_type = Base_type([],[],"signature");;
let word_signature_type = Base_type([],[],"word_signature");;
let theo_type sigma = Base_type([sigma],[],"theory");;
let variable_set_type = Base_type([],[],"variable_set");;

let parameter_set_type = Base_type([],[],"parameter_set");;
let pword_signature_type = Base_type([],[],"pword_signature");;
let pword_type sigma = Base_type([sigma],[],"pword");;
let psrs_type sigma = Base_type([sigma],[],"psrs");;
let psubst_type = Base_type([],[],"psubst");;

let word_type sigma = Base_type([sigma],[],"word");;
let word_precedence_type sigma = Base_type([sigma],[],"word_precedence");;
let precedence_type sigma = Base_type([sigma],[],"precedence");;
let status_type sigma = Base_type([sigma],[],"status");;
let word_ordering_type sigma = Base_type([sigma],[],"word_ordering");;
let term_ordering_type sigma = Base_type([sigma],[],"term_ordering");;
let srs_type sigma = Base_type([sigma],[],"SRS");;

let position_type = Base_type([],[],"position");;
let term_type sigma vars = Base_type([sigma;vars],[],"term");;
let trs_type sigma vars = Base_type([sigma;vars],[],"TRS");;
let equations_type sigma vars = Base_type([sigma;vars],[],"equations");;
let conjecture_type sigma vars = Base_type([sigma;vars],[],"conjecture");;
let htrs_type = Base_type([],[],"HTRS");;
let termination_constraint_type sigma vars = Base_type([sigma;vars],[],"termination_constraint");;
let order_constraint_type sigma vars = Base_type([sigma;vars],[],"order_constraint");;

let arrow_type t1 t2 = Dep_arrow(ref Anonymous,t1,fun _ -> t2);;
let bin_arith_type =
  Non_generalized(arrow_type int_type (arrow_type int_type int_type));;
let bin_comp_type =
  Non_generalized(arrow_type int_type (arrow_type int_type bool_type));;
let bin_bool_type =
  Non_generalized(arrow_type bool_type (arrow_type bool_type bool_type));;





let rec type_expr global_env env = function 
    Integer _ -> int_type
  | String _ -> string_type
  | Bool _ -> bool_type
  | Var x -> 
      begin
	try
	  type_value (specialize (List.assoc x env))
	with Not_found -> raise (Type_error ("undefined identifier "^x))
      end
  | Fun(x,e) ->
      let new_type_var = Type_variable (ref None) in
      let t2 = 
	type_expr global_env ((x,Non_generalized new_type_var)::env) e 
      in 
      begin
	match type_value new_type_var with
	    Type_variable _ -> raise (Type_error "polymorphic variable")
	  | t3 -> Dep_arrow (ref Anonymous,t3,fun _ -> t2)
      end
      
  | Apply(e1,e2) ->
      let t1 = type_expr global_env env e1
      and t2 = type_expr global_env env e2
      in 
      begin 
	match t1 with
	    Dep_arrow(id,t3,t4) ->
	      if unify_types t2 t3 then (t4 (dep_of_expr global_env e2))
	      else raise (Type_error "bad type argument in application")
	  | Type_variable _ ->
	      let t3 = Type_variable (ref None) in
	      if unify_types t1 (arrow_type t2 t3)
	      then t3
		else raise (Type_error "bad application")
	    | _ ->  raise (Type_error "bad application")
	end

  | If(e1,e2,e3) ->
      let t1 = type_expr global_env env e1
      and t2 = type_expr global_env env e2
      and t3 = type_expr global_env env e3
      in 
	if unify_types t1 bool_type & unify_types t2 t3 
	then t2
	else raise (Type_error "bad types in conditional")

  | Tuple(l) ->
      let l = List.map (type_expr global_env env) l in
      Base_type([],l,"")
	      
  | Set(l) ->
      let t = 
	List.fold_left
	  (fun t e -> 
	     let t' = type_expr global_env env e in
	     if unify_types t t' 
	     then t 
	     else 
	       raise 
		 (Type_error "sets must contains values of the same type"))
	  (Type_variable (ref None))
	  l
      in set_type t
;;




let print_dep k env dep = 
  match !dep with 
      Depend_id id -> 
	print_string id;
	(k,env)
    | Anonymous -> 
	print_string "<anon>";
	(k,env)
    | Depend_var -> 
	try
	  print_string ("'"^List.assq dep env);
	  (k,env)
	with 
	    Not_found -> 
	      let n = String.make 1 (Char.chr (k+Char.code 'A')) in
	      print_string ("'"^n);
	      (k+1,(dep,n)::env)
;;



let rec do_print_type n env k envd deep t = 
  match type_value t with
    | Base_type([],[],s) -> 
	print_string s;
	(n,env,k,envd)
    | Base_type([],t::l,s) -> 
	if deep & l=[] then print_string "(";
	if l<>[] then print_string "(";
	let deeper = l=[] & not deep in
	let sep = "," (*i if s="" then " * " else "," i*) in
	let (n1,env1,k1,envd1) =
	  List.fold_left
	    (fun (n,env,k,envd) t -> 
	       print_string sep; do_print_type n env k envd deeper t)
	    (do_print_type n env k envd deeper t)
	    l
	in
	if l<>[] then print_string ")";
	if s<> "" then begin print_string " "; print_string s end;
	if deep & l=[] then print_string ")";
	(n1,env1,k1,envd1)
    | Base_type([e],a,s) ->
	assert (a=[]);
	let (k1,envd1) = print_dep k envd e in
	printf " %s" s;
	(n,env,k1,envd1)
    | Base_type(e::l,a,s) ->
	assert (a=[]);
	print_string "(";
	let (k1,envd1) = print_dep k envd e in
	let (k2,envd2) = 
	  List.fold_left
	    (fun (k,envd) e -> print_string ","; print_dep k envd e)
	    (k1,envd1)
	    l
	in
	printf ") %s" s;
	(n,env,k2,envd2)
    | Dep_arrow(id,t,f) ->
	begin
	  match !id with
	      Anonymous ->
		if deep then print_string "(";
		let (n1,env1,k1,envd1) = do_print_type n env k envd true t in
		print_string " -> ";
		let (n2,env2,k2,envd2) = 
		  do_print_type n1 env1 k1 envd1 false (f id) 
		in
		if deep then print_string ")";
		(n2,env2,k2,envd2)
	    | _ ->
		print_string "(";
		let (k1,envd1) = print_dep k envd id in
		print_string ":";
		let (n2,env2,k2,envd2) = 
		  do_print_type n env k1 envd1 false t 
		in
		print_string ") -> ";
		do_print_type n2 env2 k2 envd2 false (f id)
	end;

  | Type_variable x ->
      try
	printf "'%s" (List.assq x env);
	(n,env,k,envd)
      with
	  Not_found -> 
	    let s = String.make 1 (Char.chr (n+Char.code 'a')) in
	    printf "'%s" s;
	    (n+1,(x,s)::env,k,envd)


let print_type n env k envd t = 
  do_print_type n env k envd false t;;
 
(*

\begin{verbatim}
let compute_dep k envd e =
  match !e with
      Some "" -> (k,envd)
    | _ ->
	try
	  let _ = List.assq e envd in
	  (k,envd)
	with Not_found -> 
	  print_string ((string_of_int k)^" ->\n");
	  (k+1,(e,String.make 1 (Char.chr (k+Char.code 'A')))::envd)
;;

let rec compute_var_names n env k envd t = 
  match (type_value t) with
    Base_type s -> (n,env,k,envd)
  | Term_type(e1,e2) -> 
      let (k1,env1) = compute_dep k envd e1 in
      let (k2,env2) = compute_dep k1 env1 e2 in
      (n,env,k2,env2)
  | TRS_type(e1,e2) -> 
      let (k1,env1) = compute_dep k envd e1 in
      let (k2,env2) = compute_dep k1 env1 e2 in
      (n,env,k2,env2)
  | Dep_arrow(t,id,f) ->
      let (n1,e1,k1,ed1) = compute_var_names n env k envd t in
      let (k2,env2) = compute_dep k1 ed1 id in
      compute_var_names n1 e1 k2 env2 (f id)
  | Type_variable x ->
      try
	let _ = List.assq x env in
	(n,env,k,envd)
      with
	  Not_found -> 
	    (n+1,(x,String.make 1 (Char.chr (n+Char.code 'a')))::env,k,envd)
;;
\end{verbatim}

  *)

let print_type t = 
(*
\begin{verbatim}
  let (_,e,_,ed) = compute_var_names 0 [] 0 [] t in
  print_string ("ed length = "^(string_of_int (List.length ed))^"\n");
\end{verbatim}
*)
  let _ = print_type 0 [] 0 [] t in ();;



(*i
let rec print_type_scheme s =
  match s with
      Non_generalized(t) -> print_type t 
    | Generalized(_,_,s) -> print_type_scheme s 
;;
i*)
