(***************************************************************************

  Unification in the free theory.

CiME Project - Dmons research team - LRI - Universit Paris XI

$Id: unif_free.ml,v 1.6 2002/07/11 13:26:28 pguillon Exp $

***************************************************************************)

open Variables
open Gen_terms
open Problems
open Oc
open Controle

type 'symbol pure_elem_pb =
    {
      map_of_var_var : (unit, var_id) VarMap.t;
      scanned_equations : ('symbol term * 'symbol term) list;
      other_equations : ('symbol term * 'symbol term) list
    }

(*===================================================================

  Deletion of trivial equations.

====================================================================*)

let delete pb = 
  match pb.other_equations with
    | (s,t) :: list_of_equations ->
	if s.Hcons.tag == t.Hcons.tag
	then 
	  {
	    map_of_var_var = pb.map_of_var_var;
	    scanned_equations = pb.scanned_equations;
	    other_equations = list_of_equations
	  }
	else raise Not_appliable
    | [] -> raise Not_appliable
	  
(*===================================================================

  Clash and decomposition for free symbols.

====================================================================*)

let clash_dec pb =
  match pb.other_equations with
      (t1,t2) :: list_of_equations ->
	begin
	  match t1.Hcons.node,t2.Hcons.node with
	    | Term (f,l), Term (g,k) ->
		if f == g
		then 
		  try
		    let new_equations = (List.combine l k) @ list_of_equations in
		      {
			map_of_var_var = pb.map_of_var_var;
			scanned_equations = pb.scanned_equations;
			other_equations = new_equations
		      }
		  with Invalid_argument "List.combine" -> raise No_solution
		else raise No_solution
	    | _ -> raise Not_appliable
	end
    | _ -> raise Not_appliable

(*===================================================================

  Coalesce (Replacement of a variable by a variable).

====================================================================*)

let representative x map_of_var_var = 
  try
    VarMap.find x map_of_var_var
  with 
      Not_found -> x


let coalesce sign hct pb = 
  match pb.other_equations with
    | (t1,t2) :: list_of_equations ->
	begin
	  match t1.Hcons.node,t2.Hcons.node with
	    | Var x,Var y ->
		if x == y 
		then 
		  {
		    map_of_var_var = pb.map_of_var_var;
		    scanned_equations = pb.scanned_equations;
		    other_equations = list_of_equations
		  }
		else 
		  let (var,rep) = 
		    let x' = representative x pb.map_of_var_var
		    and y' = representative y pb.map_of_var_var in
		      if (VarOrd.compare x' y') > 0
		      then (x',y')
		      else (y',x') in
		    let term_rep = make_var_term hct rep in
		    let new_map_of_var_var = 
		      VarMap.add var rep 
			(VarMap.map 
			   (function v -> if v == var then rep else v)
			   pb.map_of_var_var)
		    and reverted_new_list_of_equations = 
		      List.fold_left
			(fun partial_list_of_equations e ->
			   let e' = replace_a_var_by_a_var_in_an_eq sign hct var term_rep e in
			     e' :: partial_list_of_equations)
			(List.map 
			   (replace_a_var_by_a_var_in_an_eq sign hct var term_rep) 
			   pb.scanned_equations) 
			list_of_equations in
		      {
			map_of_var_var = new_map_of_var_var;
			scanned_equations = [];
			other_equations = List.rev reverted_new_list_of_equations
		      }
	    | _,_ -> raise Not_appliable
	end
    | [] -> raise Not_appliable


(*===================================================================

  Merge for two equations with the same variable left-hand side.

====================================================================*)

let rec find_a_value_of_var x list_of_equations = 
  match list_of_equations with
    | (t1,t2) :: list_of_other_eqs ->
	begin
	  match t1.Hcons.node,t2.Hcons.node with
	    | Var y, Term _ ->
		if x == y
		then t2 
		else find_a_value_of_var x list_of_other_eqs
	    | Term _, Var y ->
		if x == y
		then t1
		else find_a_value_of_var x list_of_other_eqs
	    | _, _ -> find_a_value_of_var x list_of_other_eqs
	end
    | [] -> raise Not_found


let remove_an_equation_in_a_list (s,t) list_of_equations =
  List.filter
    (function (t1,t2) -> 
       not (t1.Hcons.tag==s.Hcons.tag && t2.Hcons.tag==t.Hcons.tag) && 
       not (t1.Hcons.tag==t.Hcons.tag && t2.Hcons.tag==s.Hcons.tag))
    list_of_equations

let merge sign pb =
  match pb.other_equations with
      (t1,t2) :: list_of_equations ->
	let x, var_x, s =
	  match t1.Hcons.node,t2.Hcons.node with
	      Var x', Term _ -> x', t1, t2
	    | Term _, Var x' -> x', t2, t1
	    | _, _-> raise Not_appliable in
	  begin 
	    try 
	      let t = find_a_value_of_var x list_of_equations in
      		if  s.Hcons.tag == t.Hcons.tag
		then 
		  {
		    map_of_var_var = pb.map_of_var_var;
		    scanned_equations = pb.scanned_equations;
		    other_equations = list_of_equations
		  }
		else 
		  let v = if (size sign s) < (size sign t) then s else t in
		    {
		      map_of_var_var = pb.map_of_var_var;
		      scanned_equations = pb.scanned_equations;
		      other_equations =  
		       (s,t) :: (var_x,v) :: (remove_an_equation_in_a_list 
						(var_x,t) list_of_equations)
		    }
	    with Not_found -> 
	      try 
		let t = find_a_value_of_var x pb.scanned_equations in
      		  if  s.Hcons.tag == t.Hcons.tag
		  then 
		    {
		      map_of_var_var = pb.map_of_var_var;
		      scanned_equations = pb.scanned_equations;
		      other_equations = list_of_equations
		    }
		  else 
		    let v = if (size sign s) < (size sign t) then s else t in
		      {
			map_of_var_var = pb.map_of_var_var;
			scanned_equations = 
			 remove_an_equation_in_a_list 
			   (var_x, t) pb.scanned_equations;
			other_equations =  
			 (s,t) :: (var_x, v) :: list_of_equations
		      }
	      with Not_found -> raise Not_appliable
	  end
 
    | [] -> raise Not_appliable 


(*===================================================================

                             Controle

====================================================================*)

let rec repeat rule pb = 
  match pb.other_equations with
      [] -> pb.map_of_var_var, pb.scanned_equations
    | eq::list_of_equations -> 
	let pb' = 
	  try 
      	    rule pb
	  with 
	      Not_appliable ->
		{
		  map_of_var_var = pb.map_of_var_var;
		  scanned_equations = eq :: pb.scanned_equations;
		  other_equations = list_of_equations
		} in
	  repeat rule pb'

(*===================================================================

  [solve_without_marks sign list_of_equations] computes a unifier
  for [list_of_equations] modulo the empty equational theory.

====================================================================*)

let solve_without_marks sign hct (*i vars i*) list_of_equations =
(*
  let _ =
    begin
      Printf.printf "La liste des equations a resoudre\n";
      List.iter
	(fun (s,t) ->
	   print_term sign vars s;
	   Format.print_flush ();
	   Printf.printf " = ";
	   print_term sign vars t;
	   Format.print_flush ();
	   Printf.printf "\n")
	list_of_equations
    end in
*)
  let map_of_var_var, quasi_solved_form = 
    repeat
      (orelse (orelse (orelse delete clash_dec) (coalesce sign hct)) (merge sign))
      {
	map_of_var_var = VarMap.empty;
	scanned_equations = [];
	other_equations = list_of_equations
      } in
(*
  let _ =
    begin
      Printf.printf "La forme quasi-resolue avant OC\n";
      VarMap.iter
	(fun x y ->
	   print_term sign vars (Var x);
	   Format.print_flush ();
	   Printf.printf " = ";
	   print_term sign vars (Var y);
	   Format.print_flush ();
	   Printf.printf "\n")
	map_of_var_var;
      List.iter
	(function (s,t) ->
	   print_term sign vars s;
	   Format.print_flush ();
	   Printf.printf " = ";
	   print_term sign vars t;
	   Format.print_flush ();
	   Printf.printf "\n")
	quasi_solved_form
    end in
*)
  occur_check sign hct (*i vars i*) map_of_var_var quasi_solved_form
       

(*

  [solve sign elem_pb] computes a unifier for [elem_pb] modulo the
  empty equational theory, that is, solves the equations of the
  problem and checks that the constraints of marks are satisfied.

*)

let solve sign hct (*i vars i*) elem_pb =
  let solved_form = solve_without_marks sign hct (*i vars i*) elem_pb.equations in
  let _ = 
    List.iter 
      (function (t1,t2) ->
	 match t1.Hcons.node,t2.Hcons.node with
	   | (Var x, Term _) ->
	       if VarMap.mem x elem_pb.marked_variables
	       then raise No_solution
	   | (Term _, Var x) ->
	       if VarMap.mem x elem_pb.marked_variables
	       then raise No_solution
	   | _ -> ())
      solved_form in
    [solved_form]









