(***************************************************************************

This module provides functions to complete a string
rewriting system.

CiME Project - Dmons research team - LRI - Universit Paris XI

$Id: srs_completion.ml,v 1.16 2001/12/13 14:29:58 monate Exp $

***************************************************************************)

open Format

let verbose = ref 0 

type 'a tagged_content = {tag: int ; content: 'a}

let normalize_rhs_of_rule compiled_rules ({tag = i ; content = (r1,r2)} as r)= 
  {r with content = r1,(String_rewriting.compiled_normalize compiled_rules r2)}

let normalize_pair compiled_rules (r1,r2) = 
  (String_rewriting.compiled_normalize compiled_rules r1),
  (String_rewriting.compiled_normalize compiled_rules r2)

let trivialize compiled_rules l = 
  List.fold_left (fun acc x -> 
		    let x1,x2 as x = normalize_pair compiled_rules x in
		      if x1<>x2 
		      then x::acc 
		      else acc ) [] l

let rec all_critical_pairs srs r = 
  List.fold_left
    (fun acc x -> 
       (String_unification.superpose x r)@
       (String_unification.superpose r x)@acc)
    [] srs

let size_of_word_couple (x,y) = Words.length x + Words.length y ;;

module Eq_Set = 
  Ordered_sets.MakePoly
    (struct
       type 'symbol t = ('symbol Words.word * 'symbol Words.word) 
		  tagged_content
       let compare x y = 
	 let c = x.tag - y.tag in 
	 if c=0 then compare x y else c
     end)

module Oriented_Eq_Set = 
  Ordered_sets.MakePoly
    (struct 
       type 'symbol t = ('symbol Words.word * 'symbol Words.word)
			  tagged_content
       let compare x y = 
	 let c = x.tag - y.tag in 
	 if c=0 then compare x y else c
     end)	
    
type 'symbol rule = 
    {number: int ; 
     rule: 'symbol Words.word * 'symbol Words.word}
    
module Rule_Map = 
  Map.Make(struct type t = int
		  let compare = compare 
	   end)
    
type rule_counter = {max: int}

let print_rule sigma number {tag=i;content=(x,y)} =
  printf "[%i] " i ;
  Words.pretty_print sigma x ; 
  printf "@  -> ";
  Words.pretty_print sigma y ;
  printf ";@\n"

let print_raw_rule sigma {tag=_;content=(x,y)} =
  Words.pretty_print sigma x ; 
  printf "@  -> ";
  Words.pretty_print sigma y ;
  printf ";@\n"
		      
let print_equation sigma {tag=i ; content= x,y} = 
  printf "[%i] " i ;
  Words.pretty_print sigma x ; printf "@  = ";
  Words.pretty_print sigma y ;printf ";@\n"

let print_oriented_equation sigma {tag=i ; content= x,y} =
  printf "[%i] " i ;
  Words.pretty_print sigma x ; printf "@  > ";
  Words.pretty_print sigma y ;printf ";@\n"

let pretty_print sigma eq_set eq_oriented_set rule_map rule_counter =
  if !verbose > 1 then
    let max_eq_size = 
      Eq_Set.fold (fun s acc -> max acc s.tag) eq_set 0 in
    let nb_eq = Eq_Set.cardinal eq_set in
    let nb_oriented_eq = Oriented_Eq_Set.cardinal eq_oriented_set in 
    let nb_rules = rule_counter.max in
    begin
      printf 
	"@[***********************@\n\
         State:@\nEquations = %i ; Oriented Equations = %i ;\
         Rules counter = %i ;@ Max size equation : %i@]@\n" 
(*i oh poor tuareg...""i*)
	nb_eq nb_oriented_eq nb_rules max_eq_size ;
      if !verbose > 2 then 
	begin
          printf "@[<hov 2>Rules {@\n";
	  Rule_Map.iter (print_rule sigma) rule_map;
	  printf "}@]@\n@[<hov 2>Equations {@\n";
	  Eq_Set.iter (print_equation sigma) eq_set;
	  printf "}@]@\n";
	  printf "@[<hov 2>Oriented equations {@\n";
	  Oriented_Eq_Set.iter (print_oriented_equation sigma) eq_oriented_set;
	  printf "}@]@\n";
	end
    end


let rec cross x m  = 
    Rule_Map.fold (fun i d acc -> Stream.icons (x,i) acc) m Stream.sempty   

let add_rule_cp nb tagged_rule cp_stream rule_map = 
  let new_rule_map = Rule_Map.add nb tagged_rule rule_map in
  let new_cp_stream = Stream.iapp cp_stream (cross nb new_rule_map) in
  new_cp_stream,new_rule_map

let rule_map_to_list s = Rule_Map.fold (fun i r acc -> r.content::acc) s []

let compile_set_of_rules s = String_rewriting.compile_srs (rule_map_to_list s)

let normalize_eq_set compiled_rules eq_set = 
  let card = Eq_Set.cardinal eq_set in
  if card > 3  then
    begin
      
      let normalization_function x = 
	let x1,x2 as x' = normalize_pair compiled_rules x.content 
	in
	if x1<>x2 then Some {tag= size_of_word_couple x';content=x'}
	else None
      in
      let eq_stream = Eq_Set.fold (fun x s -> Stream.icons x s ) 
			eq_set 
			Stream.sempty
      in
      Parallel.M.compute normalization_function eq_stream
	(fun x acc -> match x with
	   | None -> acc
	   | Some r -> Eq_Set.add r acc) Eq_Set.empty
    end
  else
    begin
      Eq_Set.fold (fun x acc -> 
		     let x1,x2 as x' = 
		       normalize_pair compiled_rules x.content 
		     in
		     if x1<>x2 
		     then Eq_Set.add 
		       {tag= size_of_word_couple x';content=x'} acc 
		     else acc ) eq_set Eq_Set.empty 
    end

let rec set_of_critical_pairs r1 r2 =
  let t1,t2= snd r1.content , snd r2.content in
  let s1 = String_unification.superpose r1.content r2.content in
  let s2 = 
    if r1.content==r2.content then []
    else String_unification.superpose r2.content r1.content 
  in
  List.fold_left 
    (fun acc (w1,w2) -> 
       let cp_eq = w1,w2 in
       Eq_Set.add {tag= size_of_word_couple cp_eq ; content= cp_eq} acc)
    (
      (List.fold_left 
	 (fun acc (w1,w2) -> 
	    let cp_eq = w1,w2 in
	    Eq_Set.add  
	      {tag= size_of_word_couple cp_eq ; 
	       content=cp_eq} 
	      acc
	 ) 
	 Eq_Set.empty) 
      s1
    )
    s2

let set_of_essential_critical_pairs compiled_rules r1 r2 = 
  let s = set_of_critical_pairs r1 r2 in
  if Eq_Set.is_empty s then s else
  normalize_eq_set compiled_rules s
  
let rec collapse sigma rules r =
  Rule_Map.fold
    (fun number ({tag=i;content=(lhs,rhs)} as may_collapse) (accr,acceq)   -> 
       if String_rewriting.is_nf_compiled (String_rewriting.compile_srs [r]) 
	   lhs  
       then 
	 Rule_Map.add number 
	   {tag= size_of_word_couple (lhs,rhs) ; content = (lhs,rhs)} 
	   accr,
	   acceq
       else (
	 if !verbose > 0 then begin
	   (printf "Deleting one rule :@ " ; 
	    print_raw_rule sigma may_collapse) ;
	   if !verbose > 3
	   then printf "@[New equation from collapse@]@\n" ; 
	 end;
	 accr,
	 Eq_Set.add 
	   {tag= size_of_word_couple (lhs,rhs) ; content = (lhs,rhs)}  
	   acceq))
    rules
    (Rule_Map.empty,Eq_Set.empty)
    
let normalize_rules_rhs compiled_rules all_rules = 
  Rule_Map.fold 
    (fun number {tag=i;content=(x,y)} acc -> 
       Rule_Map.add number
	 (let v = (x,String_rewriting.compiled_normalize compiled_rules y) 
	  in {tag= size_of_word_couple v;content=v})
	 acc)
    all_rules Rule_Map.empty
    
let rec collapse_oriented_eq compiled_rules o_eq =
  Oriented_Eq_Set.fold
    (fun {tag=i;content=(lhs,rhs)} (acc_oeq,acc_eq)  -> 
       if String_rewriting.is_nf_compiled compiled_rules lhs  
       then 
	 (Oriented_Eq_Set.add 
	    (let v = (lhs,String_rewriting.compiled_normalize compiled_rules rhs) 
	     in {tag= size_of_word_couple v;content=v})
	    acc_oeq)
	 ,acc_eq
       else begin
	 if !verbose > 3 then
	   printf "@[New equation from collapse oriented equation@]@\n" ; 
	 acc_oeq,Eq_Set.add 
	   (let v = (lhs,rhs) in {tag= size_of_word_couple v;content=v}) 
	   acc_eq
       end)
    o_eq
    (Oriented_Eq_Set.empty,Eq_Set.empty)
    
type choice = ORIENT | ORIENT_ALL | COMPUTE_CRITICAL_PAIR | MAKE_RULE | DONE
    
let strategy eq_set eq_oriented_set rule_map cp_stream rule_counter = 
  let veq_set = Eq_Set.is_empty eq_set in 
  let veq_oriented_set = Oriented_Eq_Set.is_empty eq_oriented_set in
  let vcp_stream = 
    try 
      Stream.empty cp_stream;
      true 
    with Stream.Failure -> false 
  in
  let _ = if !verbose>1 then printf "@[Next step is : " in
  match (veq_set,veq_oriented_set,vcp_stream) with
    | false,_,_ ->if !verbose>1 then printf "ORIENT_ALL@]@\n";ORIENT_ALL
    | _,false,_ -> if !verbose>1 then printf "MAKE_RULE@]@\n";MAKE_RULE
    | true,true,true -> if !verbose>1 then printf "DONE@]@\n";DONE
    | true,true,false -> 
	if !verbose>1 then 
	  printf "COMPUTE_CRITICAL_PAIR@]@\n"; COMPUTE_CRITICAL_PAIR
	    
let complete sigma initial_srs ordering =
  let (eq_set,eq_oriented_set,rule_map,rule_counter) = 
    (List.fold_left 
       (fun acc x -> Eq_Set.add {tag= size_of_word_couple x ;content=x} acc) 
       Eq_Set.empty 
       initial_srs),
    Oriented_Eq_Set.empty,
    Rule_Map.empty,
    {max=1}
  in
  let rec internal_completion compiled_rules eq_set eq_oriented_set rule_map 
    cp_stream rule_counter =
    pretty_print  sigma eq_set eq_oriented_set rule_map rule_counter;  
    match 
      (strategy eq_set eq_oriented_set rule_map cp_stream rule_counter) 
    with 
	DONE -> rule_map_to_list rule_map
      | ORIENT -> 
	  let {tag=i;content=(r1,r2)} as eq_to_orient = 
	    try Eq_Set.min_elt eq_set 
	    with Not_found -> 
	      failwith "complete: cannot ORIENT. Bug in strategy"
	  in
	  let oriented_eq =
	    {tag=i;content=
	       match ordering r1 r2 with 
		   Orderings_generalities.Greater_than -> (r1,r2)
		 | Orderings_generalities.Less_than -> (r2,r1)
		 | _ -> invalid_arg "complete: cannot order words"}
	  in
	  if !verbose > 1 then 
	    begin
	      printf "Orienting equation :@ ";
	      print_oriented_equation sigma oriented_eq
	    end;
	  internal_completion 
	    compiled_rules
	    (Eq_Set.remove eq_to_orient eq_set)
	    (Oriented_Eq_Set.add  oriented_eq eq_oriented_set)
	    rule_map cp_stream rule_counter
      | ORIENT_ALL ->
	  let rec orient_all 
	    ({tag=i;content=(r1,r2)} as eq_to_orient) eq_oriented_set =
	    let oriented_eq =
	      {tag=i;content=
		 match ordering r1 r2 with 
		     Orderings_generalities.Greater_than -> (r1,r2)
		   | Orderings_generalities.Less_than -> (r2,r1)
		   | _ -> invalid_arg "complete: cannot order words"}
	    in
	    if !verbose > 1 then 
	      begin
		printf "Orienting equation :@ ";
		print_oriented_equation sigma oriented_eq
	      end;
	    Oriented_Eq_Set.add oriented_eq eq_oriented_set
	  in
	  let new_oriented_eq_set =
	    Eq_Set.fold orient_all eq_set eq_oriented_set
	  in
	  internal_completion 
	    compiled_rules
	    Eq_Set.empty
	    new_oriented_eq_set
	    rule_map cp_stream rule_counter
      | MAKE_RULE ->
	  let {tag=i;content=(r1,r2)} as new_rule =
	    try Oriented_Eq_Set.min_elt eq_oriented_set 
	    with Not_found ->
	      failwith "complete: cannot MAKE_RULE. Bug in strategy"
	  in
	  let collapsed_rule_set,created_equation_set = 
	    collapse sigma rule_map  new_rule.content 
	  in
	  let new_rule_counter = 
	    {max = rule_counter.max + 1}  
	  in
	  let cp_stream,new_rule_set = add_rule_cp
					 new_rule_counter.max 
					 new_rule cp_stream collapsed_rule_set 
	  in
	  let new_compiled_rules = compile_set_of_rules new_rule_set in
	  if !verbose > 0 then 
	    begin
	      printf "@[Adding rule :@ " ;
	      print_raw_rule sigma
		(normalize_rhs_of_rule new_compiled_rules new_rule);
	      printf "@]"
	    end;
	  let simplified_rules = 
	    normalize_rules_rhs new_compiled_rules new_rule_set in
	  let recompiled_rules =  compile_set_of_rules simplified_rules in
	  let new_ordered_eq,eq_from_collapse_ord = 
	    collapse_oriented_eq recompiled_rules  
	      (Oriented_Eq_Set.remove new_rule eq_oriented_set) 
	  in
	  let unnormalized_eq_set = (Eq_Set.union eq_from_collapse_ord 
				   (Eq_Set.union created_equation_set eq_set))
	  in
	  let normalized_eq = 
	    if Eq_Set.is_empty unnormalized_eq_set
	    then unnormalized_eq_set
	    else normalize_eq_set recompiled_rules unnormalized_eq_set
	  in
	  internal_completion
	    recompiled_rules
	    normalized_eq
	    new_ordered_eq
	    simplified_rules
	    cp_stream 
	    new_rule_counter
      | COMPUTE_CRITICAL_PAIR ->
	  let rec find_next_cp () = 
	    let next_cp = Stream.peek cp_stream in
	    match next_cp with 
		Some(cp1,cp2) ->
		  Stream.junk cp_stream ;
		  (try 
		     (cp1,Rule_Map.find cp1 rule_map) ,
		     (cp2,Rule_Map.find cp2 rule_map) 
		   with Not_found -> find_next_cp ()) 
	      | None -> failwith 
		    "complete: cannot COMPUTE_CRITICAL_PAIR. Bug in strategy"
	  in
	  let(cp1,r1),(cp2,r2) = find_next_cp () in
	  let _ = if !verbose>3 then 
	    printf "@[CP of %i and %i@]@\n" cp1 cp2 
	  in
	  let new_eq_set = set_of_essential_critical_pairs compiled_rules r1 r2
	  in
	  internal_completion 
	    compiled_rules 
	    (Eq_Set.union new_eq_set eq_set) 
	    eq_oriented_set 
	    rule_map 
	    cp_stream 
	    rule_counter
  in
  internal_completion 
    (String_rewriting.compile_srs []) 
    eq_set eq_oriented_set 
    rule_map 
    Stream.sempty 
    {max=0}
    
let complete_srs_std_strategy sigma order srs =
  let completed = complete sigma srs order
  in 
  completed


(* This is a testing purpose function which embeds a lrlex order*)
let complete_srs_alphabetic_order_std_strategy sigma srs =
  let prec x y = match compare x y with (* Inverse alphabetic order *)
      l when l>0  -> Orderings_generalities.Less_than
    | l when l<0 -> Orderings_generalities.Greater_than
    | l -> Orderings_generalities.Equivalent 
  in 
  let order = Word_orderings.length_lex prec in
  complete_srs_std_strategy sigma order srs

