;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; cime2.el - inspired from tuareg.el

(defconst cime2-mode-version "Cime2 Version 1.0"
  " ")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                        Emacs versions support

(defconst cime2-with-xemacs (string-match "XEmacs" emacs-version))

(if (not cime2-with-xemacs)
    (defun match-string (num &optional string)
      "Return string of text matched by last search.

NUM specifies which parenthesized expression in the last regexp.
Value is nil if NUMth pair didn't match, or there were less than NUM
pairs.  Zero means the entire text matched by the whole regexp or
whole string.  STRING should be given if the last search was by
`string-match' on STRING."
      (let* ((data (match-data))
	     (begin (nth (* 2 num) data))
	     (end (nth (1+ (* 2 num)) data)))
	(if string (substring string begin end)
	  (buffer-substring-no-properties begin end)))))

(if (fboundp 'caar) ()
  (defun caar (l) (car (car l)))
  (defun cdar (l) (cdr (car l)))
  (defun cadr (l) (car (cdr l)))
  (defun cddr (l) (cdr (cdr l))))

(if (fboundp 'cadar) ()
  (defun cadar (l) (car (cdar l)))
  (defun cddar (l) (cdr (cdar l))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                       User customizable variables

;; use `cime2-mode-hook' to configure these variables, instead of patching
;; the mode itself!

;; comments

(defvar cime2-indent-leading-comments t
  "*If true, indent leading comment lines (starting with `(*') like others.")
(defvar cime2-indent-comments t
  "*If true, automatically align multi-line comments.")
(defvar cime2-comment-end-extra-indent 0
  "*How many spaces to indent a leading comment end `*)'.
If you expect comments to be indented like
	(*
          ...
	 *)
even without leading `*', use `cime2-comment-end-extra-indent' = 1.")
(defvar cime2-support-leading-star-comments nil
  "*If true, allow automatic intentation of comments of the form
        (*
         * ...
         *)
If you still expect comments to be indented like
	(*
          ...
	 *)
without leading `*', use `cime2-comment-end-extra-indent' = 1.")

;; indentation defaults

(defvar cime2-default-indent 2
  "*Default indentation.

Global indentation variable (large values may lead to indentation overflows).
When no governing keyword is found, this value is used to indent the line
if it has to.")

(defvar cime2-lazy-= nil
  "*If true, indent `=' like a standard keyword (not `:=', `<='...).")

(defvar cime2-lazy-paren nil
  "*If true, indent parentheses like a standard keyword.")

(defvar cime2-let-always-indent t
  "*If true, enforce indentation is at least default after a let keyword.

As an example, set it to false when you have `cime2-with-indent' set to 0,
and you want `let x = match ... with' and `match ... with' indent the
same way.")

(defvar cime2-|-extra-unindent cime2-default-indent
  "*Extra backward indent for CiME2 lines starting with the `|' operator.

It is automatically added to `function', `with', `parse' and some cases
of `type' keywords to leave enough space for `|' backward indentation.")

(defvar cime2-class-indent cime2-default-indent
  "*How many spaces to indent from a `class' keyword.")

(defvar cime2-sig-struct-align t
  "*Align `sig' and `struct' keywords with `module'.")

(defvar cime2-sig-struct-indent cime2-default-indent
  "*How many spaces to indent from a `sig' or `struct' keyword.")

(defvar cime2-method-indent cime2-default-indent
  "*How many spaces to indent from a `method' keyword.")

(defvar cime2-begin-indent cime2-default-indent
  "*How many spaces to indent from a `begin' keyword.")

(defvar cime2-for-while-indent cime2-default-indent
  "*How many spaces to indent from a `for' or `while' keyword.")

(defvar cime2-do-indent cime2-default-indent
  "*How many spaces to indent from a `do' keyword.")

(defvar cime2-fun-indent cime2-default-indent
  "*How many spaces to indent from a `fun' keyword.")

(defvar cime2-function-indent cime2-default-indent
  "*How many spaces to indent from a `function' keyword.")

(defvar cime2-if-then-else-indent cime2-default-indent
  "*How many spaces to indent from an `if', `then' or `else' keyword
in CiME2 mode.")

(defvar cime2-let-indent cime2-default-indent
  "*How many spaces to indent from a `let' keyword.")

(defvar cime2-in-indent cime2-default-indent
  "*How many spaces to indent from a `in' keyword.
A lot of people like formatting `let' ... `in' expressions whithout
indentation:
        let x = 0 in
        blah x
Set this variable to 0 to get this behaviour.
However, nested declarations are always correctly handled:
        let x = 0 in                             let x = 0
        let y = 0 in              or             in let y = 0
        let z = 0 ...                            in let z = 0 ...")

(defvar cime2-match-indent cime2-default-indent
  "*How many spaces to indent from a `match' keyword.")

(defvar cime2-try-indent cime2-default-indent
  "*How many spaces to indent from a `try' keyword.")

(defvar cime2-with-indent cime2-default-indent
  "*How many spaces to indent from a `with' keyword.")

(defvar cime2-rule-indent cime2-default-indent
  "*How many spaces to indent from a `rule' keyword.")

(defvar cime2-parse-indent cime2-default-indent
  "*How many spaces to indent from a `parse' keyword.")

(defvar cime2-parser-indent cime2-default-indent
  "*How many spaces to indent from a `parser' keyword.")

(defvar cime2-type-indent cime2-default-indent
  "*How many spaces to indent from a `type' keyword.")

(defvar cime2-val-indent cime2-default-indent
  "*How many spaces to indent from a `val' keyword.")

;; automatic indentation
;; using abbrev-mode and electric keys

(defvar cime2-use-abbrev-mode t
  "*Non-nil means electrically indent lines starting with leading keyword
such as `end', `done', `else' etc. It makes use of abbrev-mode.

Many people find eletric keywords irritating, so you can disable them in
setting this variable to nil.")
(make-variable-buffer-local 'cime2-use-abbrev-mode)
(defvar cime2-electric-indent t
  "*Non-nil means electrically indent lines starting with `|', `)', `]' or `}'.

Many people find eletric keys irritating, so you can disable them in
setting this variable to nil.")
(make-variable-buffer-local 'cime2-electric-indent)
(defvar cime2-electric-close-vector t
  "*Non-nil means electrically insert a `|' before a vector-closing `]',
a `>' before a stream-closing `]' or before an object-closing `}'.

Many people find eletric keys irritating, so you can disable them in
setting this variable to nil. You should probably have this on,
though, if you also have cime2-electric-indent on.")
(make-variable-buffer-local 'cime2-electric-close-vector)

;; CiME2-Interactive
;; configure via `cime2-mode-hook'

(defvar cime2-skip-after-eval-phrase t
  "*Non-nil means skip to the end of the phrase after evaluation in the
CiME2 toplevel.")

(defvar cime2-interactive-read-only-input cime2-with-xemacs
  "*Non-nil means input send to the CiME2 toplevel is read-only.")

(defvar cime2-interactive-echo-phrase t
  "*Non-nil means echo phrases in the toplevel buffer when sending
them to the CiME2 toplevel.")

(defvar cime2-interactive-input-font-lock t
  "*Non nil means Font-Lock for toplevel input phrases.")

(defvar cime2-interactive-output-font-lock t
  "*Non nil means Font-Lock for toplevel output messages.")

(defvar cime2-interactive-error-font-lock t
  "*Non nil means Font-Lock for toplevel error messages.")

(defvar cime2-manual-url "http://www.lri.fr/~demons/cime.html"
  "*URL to the CiME2 reference manual.")

(defvar cime2-browser 'cime2-netscape-manual
  "*Name of function that displays the CiME2 reference manual.
Valid names are `cime2-netscape-manual', `cime2-mmm-manual'
and `cime2-xemacs-w3-manual' (XEmacs only).")

(defvar cime2-library-path "/usr/local/lib/ocaml/"
  "*Path to the CiME2 library.")

(defvar cime2-definitions-max-items 30
  "*Maximum number of items a definitions menu can contain.")

(defvar cime2-options-list
  '(("Lazy Parentheses Indentation" . 'cime2-lazy-paren)
    ("Lazy `=' Indentation" . 'cime2-lazy-=)
    ("Force indentation after `let'" . 'cime2-let-always-indent)
    "---"
    ("Indent body of comments" . 'cime2-indent-comments)
    ("Indent first line of comments" . 'cime2-indent-leading-comments)
    ("Leading-`*' comment style" . 'cime2-support-leading-star-comments)
    "---"
    ("Objective Labl Syntax" . 'cime2-labl-support))
  "*List of menu-configurable CiME2 options")

(defvar cime2-interactive-options-list
  '(("Skip phrase after evaluation" . 'cime2-skip-after-eval-phrase)
    ("Echo phrase in interactive buffer" . 'cime2-interactive-echo-phrase)
    "---"
    ("Font-lock interactive input" . 'cime2-interactive-input-font-lock)
    ("Font-lock interactive output" . 'cime2-interactive-output-font-lock)
    ("Font-lock interactive error" . 'cime2-interactive-error-font-lock)
    "---"
    ("Read only input (XEmacs)" . 'cime2-interactive-read-only-input))
  "*List of menu-configurable CiME2 options")

(defvar cime2-interactive-program "cime2"
  "*Default program name for invoking a CiME2 toplevel from Emacs.")
(make-variable-buffer-local 'cime2-interactive-program)

;; customizable faces for Font-Lock mode
;; configure via `cime2-mode-hook'

(defvar cime2-font-lock-documentation '("goldenrod4" "goldenrod" t nil t nil)
  "Face description: LIGHT DARK col-ITALIC col-BOLD font-ITALIC font-BOLD.")
(defvar cime2-font-lock-comment '("pink4" "mocassin" t nil t nil)
  "Face description: LIGHT DARK col-ITALIC col-BOLD font-ITALIC font-BOLD.")
(defvar cime2-font-lock-string '("darkgreen" "aquamarine2" nil nil nil nil)
  "Face description: LIGHT DARK col-ITALIC col-BOLD font-ITALIC font-BOLD.")
(defvar cime2-font-lock-module '("midnightblue" "lightblue" nil t nil t)
  "Face description: LIGHT DARK col-ITALIC col-BOLD font-ITALIC font-BOLD.")
(defvar cime2-font-lock-governing '("darkorange2" "orange" nil t nil t)
  "Face description: LIGHT DARK col-ITALIC col-BOLD font-ITALIC font-BOLD.")
(defvar cime2-font-lock-exception '("red" "red" nil t nil t)
  "Face description: LIGHT DARK col-ITALIC col-BOLD font-ITALIC font-BOLD.")
(defvar cime2-font-lock-preprocessor '("darkorchid" "yellow" nil nil nil nil)
  "Face description: LIGHT DARK col-ITALIC col-BOLD font-ITALIC font-BOLD.")
(defvar cime2-font-lock-keyword '("blue3" "skyblue" nil t nil t)
  "Face description: LIGHT DARK col-ITALIC col-BOLD font-ITALIC font-BOLD.")
(defvar cime2-font-lock-function '("blue2" "cyan" nil t nil t)
  "Face description: LIGHT DARK col-ITALIC col-BOLD font-ITALIC font-BOLD.")
(defvar cime2-font-lock-operator '("brown" "lightgray" nil nil nil nil)
  "Face description: LIGHT DARK col-ITALIC col-BOLD font-ITALIC font-BOLD.")
(defvar cime2-font-lock-reference '("firebrick" "plum1" nil nil t t)
  "Face description: LIGHT DARK col-ITALIC col-BOLD font-ITALIC font-BOLD.")
(defvar cime2-font-lock-function-name '("blue2" "cyan" nil nil t t)
  "Face description: LIGHT DARK col-ITALIC col-BOLD font-ITALIC font-BOLD.")
(defvar cime2-font-lock-exception-name '("red" "red" nil nil t t)
  "Face description: LIGHT DARK col-ITALIC col-BOLD font-ITALIC font-BOLD.")
(defvar cime2-font-lock-interactive-output '("blue4" "lightblue" t nil t nil)
  "Face description: LIGHT DARK col-ITALIC col-BOLD font-ITALIC font-BOLD.")
(defvar cime2-font-lock-interactive-error '("firebrick" "plum1" t t t t)
  "Face description: LIGHT DARK col-ITALIC col-BOLD font-ITALIC font-BOLD.")

;; end of customizable variables

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                                  Font-Lock

(defvar cime2-window-system
  (or (and (boundp 'window-system) window-system)
      (and (fboundp 'console-type) (or (eq (console-type) 'x)
				       (eq (console-type) 'win32))))
  "Are we running under a window system?")

(if (and (featurep 'font-lock)
	 (or cime2-window-system cime2-with-xemacs))
    (progn
      (if (boundp 'font-lock-use-colors)
	  (defvar cime2-font-lock-use-colors font-lock-use-colors))
      (if (boundp 'font-lock-use-fonts)
	  (defvar cime2-font-lock-use-fonts font-lock-use-fonts))

      (if (not cime2-with-xemacs) (transient-mark-mode t)
	(defvar cime2-ext-start nil "Temporary for fontification.")
	(make-variable-buffer-local 'cime2-ext-start)
	(defvar cime2-ext-end nil "Temporary for fontification.")
	(make-variable-buffer-local 'cime2-ext-end)

	(defun cime2-pre-idle-hook-first ()
	  (condition-case nil
	      (if font-lock-old-extent
		  (setq cime2-ext-start
			(extent-start-position font-lock-old-extent)
			cime2-ext-end
			(extent-end-position font-lock-old-extent))
		(setq cime2-ext-start nil))
	    (error (warn "Error caught in `cime2-pre-idle-hook-first'"))))

	(defun cime2-pre-idle-hook-last ()
	  (condition-case nil
	      (if (and (eq major-mode 'cime2-mode)
		       cime2-ext-start)
		  (cime2-fontify cime2-ext-start cime2-ext-end))
	    (error (warn "Error caught in `cime2-pre-idle-hook-last'"))))

	(add-hook 'font-lock-after-fontify-buffer-hook 'cime2-fontify-buffer)

	(defun cime2-fontify-buffer ()
	  (if (eq major-mode 'cime2-mode)
	      (cime2-fontify (point-min) (point-max))))

	(defun cime2-fontify (start end)
	  (save-excursion
	    (goto-char start) (beginning-of-line) (setq start (point))
	    (goto-char end) (end-of-line) (setq end (point))
	    (while (> end start)
	      (goto-char (1- end))
	      (cime2-in-literal-or-comment)
	      (cond
	       ((cdr cime2-last-loc)
		(cime2-beginning-of-literal-or-comment)
		(font-lock-set-face (max start (point)) end
				    (if (looking-at "(\\*[tT][eE][xX]")
					'cime2-font-lock-documentation-face
				      'cime2-font-lock-comment-face))
		(setq end (1- (point))))
	       ((car cime2-last-loc)
		(cime2-beginning-of-literal-or-comment)
		(font-lock-set-face (max start (point)) end
				    'cime2-font-lock-string-face)
		(setq end (point)))
	       (t (while (and cime2-cache-local
			      (or (> (caar cime2-cache-local) end)
				  (eq 'b (cadar cime2-cache-local))))
		    (setq cime2-cache-local (cdr cime2-cache-local)))
		  (setq end (if cime2-cache-local
				(caar cime2-cache-local) start))))))))

      (defun cime2-font-lock-hook ()
	"Function called by `font-lock-mode' for initialization purposes."
	(if (eq major-mode 'cime2-mode)
	    (progn
	      (if (not cime2-with-xemacs) ()
		(add-hook 'pre-idle-hook 'cime2-pre-idle-hook-first)
		(add-hook 'pre-idle-hook 'cime2-pre-idle-hook-last t)
		(if (or (and cime2-font-lock-use-colors
			     font-lock-use-colors)
			(and cime2-font-lock-use-fonts
			     font-lock-use-fonts)) ()
		  (setq cime2-font-lock-use-colors font-lock-use-colors)
		  (setq cime2-font-lock-use-fonts font-lock-use-fonts)
		  (cime2-set-font-lock-faces))))))

      (defun cime2-make-face-italic (face)
	(condition-case nil (make-face-italic face) (error nil)))
      (defun cime2-make-face-bold (face)
	(condition-case nil (make-face-bold face) (error nil)))
      (defun cime2-make-face-unitalic (face)
	(condition-case nil (make-face-unitalic face) (error nil)))
      (defun cime2-make-face-unbold (face)
	(condition-case nil (make-face-unbold face) (error nil)))

      (defun cime2-make-face (name)
	(let ((desc (eval (intern name)))
	      (face (intern (concat name "-face"))))
	  (if (not (facep face)) (make-face face))
	  (set-face-foreground face
			       (if use-fonts default-color
				 (if light-bg (nth 0 desc) (nth 1 desc))))
	  (if use-fonts
	      (progn
		(if (nth 4 desc) (cime2-make-face-italic face)
		  (cime2-make-face-unitalic face))
		(if (nth 5 desc) (cime2-make-face-bold face)
		  (cime2-make-face-unbold face)))
	    (if (nth 2 desc) (cime2-make-face-italic face)
	      (cime2-make-face-unitalic face))
	    (if (nth 3 desc) (cime2-make-face-bold face)
	      (cime2-make-face-unbold face)))))

      (defun cime2-set-font-lock-faces ()
	"Set faces for Font-Lock mode."
	(let* ((use-fonts
		(or (and (boundp 'font-lock-use-fonts)
			 font-lock-use-fonts
			 (not (and (boundp 'font-lock-use-colors)
				   font-lock-use-colors)))
		    (and (fboundp 'device-class)
			 (eq (device-class) 'mono))
		    (and (not (fboundp 'device-class))
			 (fboundp 'x-display-color-p)
			 (not (x-display-color-p)))))
	       (light-bg
		(if cime2-window-system
		    (if cime2-with-xemacs
			(if (and (fboundp 'color-rgb-components)
				 (< (apply '+ (color-rgb-components
					       (make-color-specifier
						[default background])))
				    (* (apply '+
					      (color-rgb-components
					       (make-color-specifier "white")))
				       0.6))) nil t)
		      (let ((param (cdr (assq 'background-color
					      (frame-parameters)))))
			(cond
			 ((boundp 'font-lock-background-mode)
			  (if (eq font-lock-background-mode 'dark) nil t))
			 ((eq system-type 'ms-dos)
			  (if (string-match "light" param) t nil))
			 ((and param (fboundp 'x-color-values)
			       (< (apply '+ (x-color-values param))
				  (* (apply '+ (x-color-values "white"))
				     0.6))) nil)
			 (t t))))))
	       (default-color (if light-bg "black" "white")))
	  (mapcar 'cime2-make-face
		  '("cime2-font-lock-documentation"
		    "cime2-font-lock-comment"
		    "cime2-font-lock-string"
		    "cime2-font-lock-module"
		    "cime2-font-lock-governing"
		    "cime2-font-lock-preprocessor"
		    "cime2-font-lock-keyword"
		    "cime2-font-lock-function"
		    "cime2-font-lock-exception"
		    "cime2-font-lock-reference"
		    "cime2-font-lock-operator"
		    "cime2-font-lock-function-name"
		    "cime2-font-lock-exception-name"
		    "cime2-font-lock-interactive-output"
		    "cime2-font-lock-interactive-error"))))

      (defvar cime2-font-lock-keywords
	(append
	 (if cime2-with-xemacs ()
	   '(("\\(^\\|[^'`]\\)\"\\([^\"\n]\\|\\\\\"\\)*\\((\\*\\|\\*)\\)\\([^\"\n]\\|\\\\\"\\)*\""
	      0 'cime2-font-lock-string-face t) ; big hack, especially the \n's
;	     ("(\\*[tT][eE][xX]\\>\\([^*]\\|\\*+[^*)]\\)*\\(\\*+)\\|\\**$\\)"
;	      0 'cime2-font-lock-documentation-face t)
	     ("(\\*\\([^*]\\|\\*+[^*)]\\)*\\(\\*+)\\|\\**$\\)"
	      0 'cime2-font-lock-comment-face t)
	     ("\\(^\\|[^'`]\\)\"\\([^\\\"]\\|\\\\.\\)*\"\\|'\\([^\\\n']\\|\\\\[\\a-z']\\|\\\\[0-9][0-9][0-9]\\)'\\|`\\([^\\\n`]\\|\\\\[\\a-z`]\\|\\\\[0-9][0-9][0-9]\\)`"
	      0 'cime2-font-lock-string-face t)))
	 '(("^#[ \t]*[a-z][_a-z]*\\>\\|;;\\|\\<external\\>"
	    0 'cime2-font-lock-preprocessor-face nil)
;	   ("\\<\\(open\\|s\\(ig\\|truct\\)\\|module\\([ \t\n]+type\\)?\\|functor\\|\\(with\\|and\\|let\\)[ \t\n]+\\(type\\|module\\)\\)\\>"
;	    0 'cime2-font-lock-module-face nil)
;	   ("\\<\\(v\\(al\\|irtual\\)\\|type\\|method\\|c\\(onstraint\\|lass\\)\\|in\\(herit\\|itializer\\)?\\|let\\|rec\\|and\\|begin\\|object\\|end\\)\\>"
;	    0 'cime2-font-lock-governing-face nil)
;	   ("\\<\\(exception\\|raise\\|failwith\\|assert\\)\\>"
;	    0 'cime2-font-lock-exception-face nil)
;	   ("\\<fun\\(ction\\)?\\>"
;	    0 'cime2-font-lock-function-face nil)
;	   ("\\<\\(as\\|do\\(ne\\|wnto\\)?\\|else\\|for\\|if\\|let\\|m\\(atch\\|utable\\)\\|new\\|p\\(arser\\|rivate\\)\\|t\\(hen\\|o\\|ry\\)\\|w\\(h\\(en\\|ile\\)\\|ith\\)\\|lazy\\)\\>"
;	    0 'cime2-font-lock-keyword-face nil)
;	   ("\\<open\\>[ \t\n]*\\([_A-Za-z\277-\377]\\(\\w\\|\\.\\)*\\)"
;	    1 'cime2-font-lock-function-name-face nil)
;	   ("\\<\\([_A-Za-z\277-\377]\\w*\\)\\>[ \t\n]*:\\<[^:=>]"
;	    1 'cime2-font-lock-function-name-face nil)
;	   ("\\<\\(virtual\\|val\\([ \t\n]+mutable\\)?\\|inherit\\|private\\|constraint\\|external\\|method\\|module\\([ \t\n]+type\\)?\\|and\\|type\\|class\\|let\\([ \t\n]+rec\\)?\\)\\>[ \t]*\\(\\(\\<['_A-Za-z\277-\377]\\w*\\>[ \t]*\\)*\\)\\>"
;	    5 'cime2-font-lock-function-name-face nil)
;	   ("\\<exception\\>[ \t]*\\(\\<[_A-Za-z\277-\377]\\w*\\>\\)"
;	    1 'cime2-font-lock-exception-name-face nil)
;	   ("\\<\\(as[lr]\\|false\\|l\\(and\\|xor\\|or\\|s[lr]\\)mod\\|not\\|ref\\|o[fr]\\|true\\|unit\\)\\>"
;	    0 'cime2-font-lock-reference-face nil)
	   ("[][;,()|{}@^!:#*=<>&/%+---]\\.?\\|\\.\\."
	    0 'cime2-font-lock-operator-face nil)))
	"Font-Lock patterns for CiME2 mode.")

      (if (featurep 'sym-lock)
	  ;; to change this table, xfd -fn '-adobe-symbol-*--12-*' may be
	  ;; used to determine the symbol character codes.
	  (defvar cime2-sym-lock-keywords
	    '(("<-" 0 1 172) ("->" 0 1 174)
	      (":=" 0 1 220) ("<=" 0 1 163)
	      (">=" 0 1 179) ("<>" 0 1 185)
	      ("==" 0 1 186) ("||" 0 1 218)
	      ("&&" 0 1 217) ("[^*]\\(\\*\\)\\." 1 8 180)
	      ("\\(/\\)\\." 1 3 184) (":>" 0 1 202)
	      (";;" 0 1 191) ("\\<_\\>" 0 3 188)
	      ("\\<sqrt\\>" 0 3 214) ("\\<unit\\>" 0 3 198)
	      ;; ("\\<fun\\>" 0 3 108)
	      ("\\<or\\>" 0 3 218) ("\\<not\\>" 0 3 216))
	    "If non nil: Overrides default Sym-Lock patterns for CiME2."))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                                    Keymap

(defvar cime2-labl-support nil
  "Do not modify this variable directly,
use function `cime2-do-support-labl'.")

(defun cime2-do-support-labl (&optional switch)
  "Switch from OCiME2 to OLabl syntax in modifying the syntax of operator `:'.
You may also use an argument `0' (resp. `1') to disable (resp. enable)
OLabl syntax."
  (interactive)
  (cond ((not switch) (setq cime2-labl-support (not cime2-labl-support)))
	((>= switch 1) (setq cime2-labl-support t))
	(t (setq cime2-labl-support nil)))
  (if cime2-labl-support
      (progn
	(modify-syntax-entry ?: "w" cime2-mode-syntax-table)
	(modify-syntax-entry ?` "w" cime2-mode-syntax-table)
	(message "Objective Labl syntax now supported."))
    (modify-syntax-entry ?: "." cime2-mode-syntax-table)
    (modify-syntax-entry ?` "\"" cime2-mode-syntax-table)
    (message "Objective CiME2 syntax.")))

(defvar cime2-mode-map nil
  "Keymap used in CiME2 mode.")
(setq cime2-mode-map (make-sparse-keymap))
(define-key cime2-mode-map "#" 'cime2-electric)
(define-key cime2-mode-map "%" 'cime2-electric)
(define-key cime2-mode-map "|" 'cime2-electric)
(define-key cime2-mode-map ")" 'cime2-electric)
(define-key cime2-mode-map "}" 'cime2-electric-rc)
(define-key cime2-mode-map "]" 'cime2-electric-rb)
(define-key cime2-mode-map "\t" 'cime2-indent-command)
(define-key cime2-mode-map "\M-\C-h" 'cime2-mark-phrase)
(define-key cime2-mode-map "\M-q" 'cime2-indent-phrase)
(define-key cime2-mode-map "\C-c\C-q" 'cime2-indent-phrase)
(define-key cime2-mode-map "\C-c\C-a" 'cime2-find-alternate-file)
(define-key cime2-mode-map "\C-c\C-c" 'compile)
(define-key cime2-mode-map "\M-\C-x" 'cime2-eval-phrase)
(define-key cime2-mode-map "\C-x\C-e" 'cime2-eval-phrase)
(define-key cime2-mode-map "\C-c\C-e" 'cime2-eval-phrase)
(define-key cime2-mode-map "\C-c\C-r" 'cime2-eval-region)
(define-key cime2-mode-map "\C-c\C-b" 'cime2-eval-buffer)
(define-key cime2-mode-map "\C-c\C-s" 'cime2-run-caml)
(define-key cime2-mode-map "\C-c\C-i" 'cime2-interrupt-caml)
(define-key cime2-mode-map "\C-c\C-k" 'cime2-kill-caml)
(define-key cime2-mode-map "\C-c\C-n" 'cime2-next-phrase)
(define-key cime2-mode-map "\C-c\C-p" 'cime2-previous-phrase)
(define-key cime2-mode-map [(meta control down)]  'cime2-next-phrase)
(define-key cime2-mode-map [(meta control up)] 'cime2-previous-phrase)
(define-key cime2-mode-map "\C-c`" 'cime2-interactive-next-error)
(define-key cime2-mode-map "\C-cb" 'cime2-insert-begin-form)
(define-key cime2-mode-map "\C-cf" 'cime2-insert-for-form)
(define-key cime2-mode-map "\C-cw" 'cime2-insert-while-form)
(define-key cime2-mode-map "\C-ci" 'cime2-insert-if-form)
(define-key cime2-mode-map "\C-cl" 'cime2-insert-let-form)
(define-key cime2-mode-map "\C-cm" 'cime2-insert-match-form)
(define-key cime2-mode-map "\C-ct" 'cime2-insert-try-form)

(defvar cime2-mode-syntax-table ()
  "Syntax table in use in CiME2 mode buffers.")
(setq cime2-mode-syntax-table (make-syntax-table))
(modify-syntax-entry ?_ "w" cime2-mode-syntax-table)
(modify-syntax-entry ?: "." cime2-mode-syntax-table)
(modify-syntax-entry ?' "w" cime2-mode-syntax-table)
;; ' is part of words (for primes)
(modify-syntax-entry ?` "\"" cime2-mode-syntax-table)
;; ` is a string delimiter (camllight compatibility)
(modify-syntax-entry ?\" "\"" cime2-mode-syntax-table)
;; " is a string delimiter
(modify-syntax-entry ?\\ "\\" cime2-mode-syntax-table)
(modify-syntax-entry ?\( "()1" cime2-mode-syntax-table)
(modify-syntax-entry ?*  ".23" cime2-mode-syntax-table)
(modify-syntax-entry ?\) ")(4" cime2-mode-syntax-table)
(let ((i 192))
  (while (< i 256)
    (modify-syntax-entry i "w" cime2-mode-syntax-table)
    (setq i (1+ i))))

(defconst cime2-font-lock-syntax
  '((?` . ".") (?\" . "."))
  "Syntax changes for Font-Lock.")

(defvar cime2-mode-abbrev-table ()
  "Abbrev table used for CiME2 mode buffers.")
(defun cime2-define-abbrev (keyword)
  (define-abbrev cime2-mode-abbrev-table keyword keyword 'cime2-abbrev-hook))
(if cime2-mode-abbrev-table ()
  (setq cime2-mode-abbrev-table (make-abbrev-table))
  (mapcar 'cime2-define-abbrev
	  '("module" "class" "object" "let" "type" "val" "inherit" "virtual"
	    "constraint" "exception" "external" "open" "method" "and"
	    "initializer" "to" "downto" "do" "done" "else" "end" "in"
	    "then" "with"))
  (setq abbrevs-changed nil))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                              The major mode

(defun cime2-mode ()
  "Major mode for editing CiME2 code.

Dedicaced to Emacs and XEmacs. Works with Emacs version 19.30 and
higher, XEmacs version 19.13 and higher. Provides automatic
indentation and compilation interface. Performs font/color highlighting
using Font-Lock. It is designed for Objective CiME2 but handles
Objective Labl and CiME2light as well.

Report bugs, remarks and questions to Albert.Cohen@prism.uvsq.fr.

If you use XEmacs, the Font-Lock minor-mode is used accordingly to
your customization options. Within Emacs, you may want to use
Font-Lock in addding the following lines to the configuration file:

  (if (and (boundp 'window-system) window-system)
      (require 'font-lock))

Within XEmacs (more generally, if you have variable-sized fonts
and `atomic-extents' supported) you may also want to use Sym-Lock:

  (if (and (boundp 'window-system) window-system
	   (string-match \"XEmacs\" emacs-version))
      (require 'sym-lock))

Emacs supports neither variable-sized fonts nor atomic sequences, perhaps
it is a good occasion for you to try XEmacs!

You have better byte-compile cime2.el (and sym-lock.el if you use it)
because symbol highlighting is very time consuming.

For customization purposes, you should use `cime2-mode-hook'
(run for every file) or `cime2-load-hook' (run once) and not patch
the mode itself. You should add to your configuration file something like:
  (add-hook 'cime2-mode-hook
            '(lambda ()
               ... ; your customization code
             ))
For example you can change the indentation of some keywords, the
`electric' flags, Font-Lock colors... Every customizable variable is
documented, use `C-h-v' or look at the mode's source code.

A special case is Sym-Lock customization: You may set
`cime2-sym-lock-keywords' in your `.emacs' configuration file
to override default Sym-Lock patterns.

`custom-cime2.el' is a sample customization file for standard changes.
You can append it to your `.emacs' or use it as a tutorial.

To enable OLabl support call `cime2-do-support-labl', either
interactively or within `cime2-mode-hook'. It is a switch when supplied with
no arguments, but you may use `0' (resp. `1') to disable (resp. enable)
OLabl syntax.

`M-x camldebug' FILE starts the CiME2 debugger camldebug on the executable
FILE, with input and output in an Emacs buffer named *camldebug-FILE*.

A CiME2 Interactive Mode to evaluate expressions in a toplevel is included.
Type `M-x cime2-run-caml' or see special-keys below.

Some elementary rules have to be followed in order to get the best of
indentation facilities.
  - Because the `function' keyword has a special indentation (to handle
    case matchs) use the `fun' keyword when no case match is done.
  - Prefer the `or' keyword to `||' (they are semantically equivalent),
    it avoids some unwanted electric indentations.
  - In OCiME2, `;;' is no longer necessary for correct indentation,
    except before top level phrases not introduced by `type', `val', `let'
    etc. (i.e. phrases used for their side-effects or to be executed
    in a top level.)
  - If you keep the default value for `cime2-indent-leading-comments',
    never write comments at the beginning of a line followed by some code,
    since the line will not be automatically indented. This recommendation
    does not hold if `cime2-indent-leading-comments' is true.
  - Long sequences of `and's may slow down indentation slightly, since
    some computations (few actually) require to go back to the
    beginning of the sequence. Some very long nested blocks may also lead
    to slow processing of `end's, `else's, `done's...
  - Multiline strings are handled properly, but the string concatenation `^'
    is preferred to break long strings (the C-j keystroke can help in some
    cases).

Known bugs:
  - With Emacs: Text highlighting may be incorrect in some tricky examples
    (nested comments and strings), but there exists no proper solution
    to this without rewritting Font-Lock algorithms. This has been done
    with the improved XEmacs Font-Lock interface: Use XEmacs!
  - When writting a line with mixed code and comments, avoid putting
    comments at the beginning or middle of the text. More precisely, 
    writing comments immediately after `=' or parentheses then writing
    some more code on the line leads to indentation errors: You may write
    `let x (* blah *) = blah' but should avoid `let x = (* blah *) blah'.
    I'll try to fix this, even if I did not get any bug report about
    it so far...

Special keys for CiME2 mode:\\{cime2-mode-map}"
  (interactive)
  (kill-all-local-variables)
  (setq case-fold-search nil)
  (setq major-mode 'cime2-mode)
  (setq mode-name "CiME2")
  (use-local-map cime2-mode-map)
  (set-syntax-table cime2-mode-syntax-table)
  (setq local-abbrev-table cime2-mode-abbrev-table)

  (if cime2-window-system (cime2-build-menu))

  (make-local-variable 'paragraph-start)
  (setq paragraph-start (concat "^$\\|\\*)$\\|" page-delimiter))
  (make-local-variable 'paragraph-separate)
  (setq paragraph-separate paragraph-start)
  (make-local-variable 'require-final-newline)
  (setq require-final-newline t)
  (make-local-variable 'comment-start)
  (setq comment-start "(* ")
  (make-local-variable 'comment-end)
  (setq comment-end " *)")
  (make-local-variable 'comment-column)
  (setq comment-column 40)
  (make-local-variable 'comment-start-skip)
  (setq comment-start-skip "(\\*+[ \t]*")
  (make-local-variable 'parse-sexp-ignore-comments)
  (setq parse-sexp-ignore-comments nil)
  (make-local-variable 'indent-line-function)
  (setq indent-line-function 'cime2-indent-command)
  (make-variable-buffer-local 'before-change-functions)
  (add-hook 'before-change-functions 'cime2-before-change-function)
  (make-variable-buffer-local 'normal-auto-fill-function)
  (setq normal-auto-fill-function 'cime2-auto-fill-function)

  ;; hooks for cime2-mode, use them for cime2-mode configuration
  (run-hooks 'cime2-mode-hook)

  (cime2-install-font-lock)

  (if cime2-use-abbrev-mode (abbrev-mode 1))
  (message (concat "Major mode for CiME2 programs, " cime2-mode-version ".")))

(defun cime2-install-font-lock ()
  (if (and (featurep 'font-lock)
	   (or cime2-window-system cime2-with-xemacs))
      (progn
	(cime2-set-font-lock-faces)
	(if (featurep 'sym-lock) ; needed AFTER cime2-set-font-lock-faces
	    (progn
	      (setq sym-lock-color
		    (face-foreground 'cime2-font-lock-operator-face))
	      (if (not sym-lock-keywords)
		  (sym-lock cime2-sym-lock-keywords))))
	(add-hook 'font-lock-mode-hook 'cime2-font-lock-hook)
	(make-variable-buffer-local 'font-lock-defaults)
	(setq font-lock-defaults
	      (list 'cime2-font-lock-keywords t nil cime2-font-lock-syntax))
	(font-lock-set-defaults)
	(if (not (or cime2-with-xemacs font-lock-mode)) (font-lock-mode 1))
	'font-lock)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                               Error processing

(require 'compile)

;; In Emacs 19, the regexps in compilation-error-regexp-alist do not
;; match the error messages when the language is not English.
;; Hence we add a regexp.

(defconst cime2-error-regexp
  "^[A-\377]+ \"\\([^\"\n]+\\)\", [A-\377]+ \\([0-9]+\\)[-,:]"
  "Regular expression matching the error messages produced by (o)camlc.")

(if (boundp 'compilation-error-regexp-alist)
    (or (assoc cime2-error-regexp
               compilation-error-regexp-alist)
        (setq compilation-error-regexp-alist
              (cons (list cime2-error-regexp 1 2)
               compilation-error-regexp-alist))))

;; A regexp to extract the range info.

(defconst cime2-error-chars-regexp
  ".*, .*, [A-\377]+ \\([0-9]+\\)-\\([0-9]+\\):"
  "Regular expression extracting the character numbers
from an error message produced by (o)camlc.")

;; Wrapper around next-error.

;; itz 04-21-96 somebody didn't get the documentation for next-error
;; right. When the optional argument is a number n, it should move
;; forward n errors, not reparse.

;; itz 04-21-96 instead of defining a new function, use defadvice
;; that way we get our effect even when we do \C-x` in compilation buffer  

(defadvice next-error (after cime2-next-error activate)
 "Reads the extra positional information provided by the CiME2 compiler.

Puts the point and the mark exactly around the erroneous program
fragment. The erroneous fragment is also temporarily highlighted if
possible."
 (if (eq major-mode 'cime2-mode)
     (let ((beg nil) (end nil))
       (save-excursion
	 (set-buffer compilation-last-buffer)
	 (save-excursion
	   (goto-char (window-point (get-buffer-window (current-buffer))))
	   (if (looking-at cime2-error-chars-regexp)
	       (setq beg (string-to-int (match-string 1))
		     end (string-to-int (match-string 2))))))
       (beginning-of-line)
       (if beg
	   (progn
	     (setq beg (+ (point) beg) end (+ (point) end))
	     (goto-char beg) (push-mark end t t))))))

(defvar cime2-interactive-error-regexp
  (concat "\\(\\("
	  "Toplevel input:"
	  "\\|Entr.e interactive:"
	  "\\|Characters [0-9-]*:"
	  "\\|The global value [^ ]* is referenced before being defined."
	  "\\|La valeur globale [^ ]* est utilis.e avant d'.tre d.finie."
	  "\\|Reference to undefined global"
	  "\\|The C primitive \"[^\"]*\" is not available."
	  "\\|La primitive C \"[^\"]*\" est inconnue."
	  "\\|Cannot find \\(the compiled interface \\)?file"
	  "\\|L'interface compil.e [^ ]* est introuvable."
	  "\\|Le fichier [^ ]* est introuvable."
	  "\\|Exception non rattrap.e:"
	  "\\|Uncaught exception:"
	  "\\)[^#]*\\)" )
  "Regular expression matching the error messages produced by CiME2.")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                            Support definitions

(defun cime2-auto-fill-function ()
  (if (not (cime2-in-literal-p))
      (do-auto-fill)))

(defun cime2-forward-char (&optional step)
  (if step (goto-char (+ (point) step))
    (goto-char (1+ (point)))))

(defun cime2-backward-char (&optional step)
  (if step (goto-char (- (point) step))
    (goto-char (1- (point)))))

(defun cime2-in-indentation-p ()
  "Tests whether all characters between beginning of line and point
are blanks."
  (save-excursion
    (skip-chars-backward " \t")
    (bolp)))

(defvar cime2-cache-stop (point-min))
(make-variable-buffer-local 'cime2-cache-stop)
(defvar cime2-cache nil)
(make-variable-buffer-local 'cime2-cache)
(defvar cime2-cache-local nil)
(make-variable-buffer-local 'cime2-cache-local)
(defvar cime2-cache-last-local nil)
(make-variable-buffer-local 'cime2-cache-last-local)
(defvar cime2-last-loc (cons nil nil))

(defun cime2-before-change-function (begin end)
;;  (message (concat (number-to-string begin) "-" (number-to-string end)))
  (setq cime2-cache-stop (min cime2-cache-stop (1- begin))))

(defun cime2-in-literal-p ()
  "Returns non-nil if point is inside a CiME2 literal."
  (car (cime2-in-literal-or-comment)))
(defun cime2-in-comment-p ()
  "Returns non-nil if point is inside a CiME2 comment."
  (cdr (cime2-in-literal-or-comment)))
(defun cime2-in-literal-or-comment-p ()
  "Returns non-nil if point is inside a CiME2 literal or comment."
  (cime2-in-literal-or-comment)
  (or (car cime2-last-loc) (cdr cime2-last-loc)))
(defun cime2-in-literal-or-comment ()
  "Returns the pair `((cime2-in-literal-p) . (cime2-in-comment-p))'."
  (if (and (<= (point) cime2-cache-stop) cime2-cache)
      (progn
	(if (or (not cime2-cache-local) (not cime2-cache-last-local)
		(and (>= (point) (caar cime2-cache-last-local))))
	    (setq cime2-cache-local cime2-cache))
	(while (and cime2-cache-local (< (point) (caar cime2-cache-local)))
	  (setq cime2-cache-last-local cime2-cache-local
		cime2-cache-local (cdr cime2-cache-local)))
	(setq cime2-last-loc
	      (if cime2-cache-local
		  (cons (eq (cadar cime2-cache-local) 'b)
			(> (cddar cime2-cache-local) 0))
		(cons nil nil))))
    (let ((flag t) (op (point)) (mp (min (point) (1- (point-max))))
	  (literal nil) (balance 0) (end-of-comment nil))
      (if cime2-cache
	  (progn
	    (while (and cime2-cache (< cime2-cache-stop (caar cime2-cache)))
	      (setq cime2-cache (cdr cime2-cache)))
	    (if cime2-cache
		(prog2
		    (if (eq (cadar cime2-cache) 'b)
			(setq cime2-cache-stop (1- (caar cime2-cache))))
		    (setq balance (cddar cime2-cache))))))
      (goto-char cime2-cache-stop)
      (if (char-equal ?* (preceding-char)) (backward-char))
      (while flag
	(setq literal nil)
	(if end-of-comment (setq balance 0 end-of-comment nil))
	(skip-chars-forward "^\\\\'`\"(\\*")
	(cond
	 ((looking-at "\\\\")
	  (cime2-forward-char 2))
	 ((looking-at "'\\([^\n']\\|\\\\..?.?\\)'")
	  (cime2-forward-char)
	  (setq cime2-cache (cons (cons (point) (cons 'b balance))
				   cime2-cache))
	  (skip-chars-forward "^'") (cime2-forward-char)
	  (setq cime2-cache (cons (cons (point) (cons 'e balance))
				   cime2-cache))
	  (setq literal t))
	 ((looking-at "`\\([^\n']\\|\\\\..?.?\\)`")
	  (cime2-forward-char)
	  (setq cime2-cache (cons (cons (point) (cons 'b balance))
				   cime2-cache))
	  (skip-chars-forward "^`") (cime2-forward-char)
	  (setq cime2-cache (cons (cons (point) (cons 'e balance))
				   cime2-cache))
	  (setq literal t))
	 ((looking-at "\"")
	  (cime2-forward-char)
	  (setq cime2-cache (cons (cons (point) (cons 'b balance))
				   cime2-cache))
	  (skip-chars-forward "^\\\\\"")
	  (while (looking-at "\\\\")
	    (cime2-forward-char 2) (skip-chars-forward "^\\\\\""))
	  (cime2-forward-char)
	  (setq cime2-cache (cons (cons (point) (cons 'e balance))
				   cime2-cache))
	  (setq literal t))
	 ((looking-at "(\\*")
	  (setq balance (1+ balance))
	  (setq cime2-cache (cons (cons (point) (cons nil balance))
				   cime2-cache))
	  (cime2-forward-char 2))
	 ((looking-at "\\*)")
	  (cime2-forward-char 2)
	  (if (> balance 1)
	      (prog2
		  (setq balance (1- balance))
		  (setq cime2-cache (cons (cons (point) (cons nil balance))
					   cime2-cache)))
	    (setq end-of-comment t)
	    (setq cime2-cache (cons (cons (point) (cons nil 0))
				     cime2-cache))))
	 (t (cime2-forward-char)))
	(setq flag (<= (point) mp)))
      (setq cime2-cache-local cime2-cache
	    cime2-cache-stop (point))
      (goto-char op)
      (if cime2-cache (cime2-in-literal-or-comment) (cons nil nil)))))

(defun cime2-beginning-of-literal-or-comment ()
  "Skips to the beginning of the current literal or comment (or buffer)."
  (interactive)
  (if (cime2-in-literal-or-comment-p)
      (cime2-beginning-of-literal-or-comment-fast)))

(defun cime2-beginning-of-literal-or-comment-fast ()
  (while (and cime2-cache-local
	      (or (eq 'b (cadar cime2-cache-local))
		  (> (cddar cime2-cache-local) 0)))
    (setq cime2-cache-last-local cime2-cache-local
	  cime2-cache-local (cdr cime2-cache-local)))
  (goto-char (caar cime2-cache-last-local))
  (if (eq 'b (cadar cime2-cache-last-local)) (cime2-backward-char)))

(defun cime2-false-=-p ()
  "Is the underlying `=' the second letter of an operator?"
  (or (char-equal ?: (preceding-char))
      (char-equal ?> (preceding-char))
      (char-equal ?< (preceding-char))
      (char-equal ?= (preceding-char))))

(defun cime2-at-phrase-break-p ()
  "Is the underlying `;' a phrase break?"
  (and (char-equal ?\; (following-char))
       (or (and (not (eobp))
		(char-equal ?\; (char-after (1+ (point)))))
	   (char-equal ?\; (preceding-char)))))

(defun cime2-backward-up-list ()
  "Safe up-list regarding comments, literals and errors."
  (let ((balance 1) (op (point)) (oc nil))
    (cime2-in-literal-or-comment)
    (while (and (> (point) (point-min)) (> balance 0))
      (setq oc (if cime2-cache-local (caar cime2-cache-local) (point-min)))
      (condition-case nil (up-list -1) (error (goto-char (point-min))))
      (if (>= (point) oc) (setq balance (1- balance))
	(goto-char op)
	(skip-chars-backward "^[]{}()") (cime2-backward-char)
	(if (not (cime2-in-literal-or-comment-p))
	    (cond
	     ((looking-at "[[{(]")
	      (setq balance (1- balance)))
	     ((looking-at "[]})]")
	      (setq balance (1+ balance))))
	  (cime2-beginning-of-literal-or-comment-fast)))
      (setq op (point)))))

(defun cime2-assoc-indent (kwop &optional looking-let)
  "Returns relative indentation of the keyword given in argument."
  (let ((ind (symbol-value (cdr (assoc kwop cime2-keyword-alist)))))
    (if (string-match "\\<\\(with\\|function\\|parser?\\)\\>" kwop)
	(+ (if (and cime2-let-always-indent
		    looking-let (< ind cime2-default-indent))
	       cime2-let-indent ind) cime2-|-extra-unindent) ind)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                               Indentation stuff

(defconst cime2-keyword-regexp "\\<\\(object\\|initializer\\|and\\|c\\(onstraint\\|lass\\)\\|m\\(atch\\|odule\\|ethod\\|utable\\)\\|s\\(ig\\|truct\\)\\|begin\\|e\\(lse\\|x\\(ception\\|ternal\\)\\)\\|t\\(o\\|hen\\|ry\\|ype\\)\\|v\\(irtual\\|al\\)\\|w\\(h\\(ile\\|en\\)\\|ith\\)\\|i\\(f\\|n\\(herit\\)?\\)\\|f\\(or\\|un\\(ct\\(or\\|ion\\)\\)?\\)\\|let\\|do\\(wnto\\)?\\|parser?\\|rule\\|of\\)\\>\\|->\\|[;,|]"
  "Regexp for all recognized keywords.")

(defconst cime2-match-|-keyword-regexp
  "\\<\\(and\\|function\\|type\\|with\\|parser?\\)\\>\\|[|=]"
  "Regexp for keywords supporting case match.")

(defconst cime2-operator-regexp "[---+*/=<>@^:&|]\\|\\<\\(or\\|l\\(and\\|x?or\\|s[lr]\\)\\|as[lr]\\|mod\\)\\>"
  "Regexp for all operators.")

(defconst cime2-kwop-regexp (concat cime2-keyword-regexp "\\|=")
  "Regexp for all keywords, and the = operator which is generally
considered as a special keyword.")

(defconst cime2-matching-keyword-regexp
  "\\<\\(and\\|do\\(ne\\)?\\|e\\(lse\\|nd\\)\\|in\\|then\\|with\\|\\(down\\)?to\\)\\>"
  "Regexp matching CiME2 keywords which act as end block delimiters.")

(defconst cime2-leading-kwop-regexp
  (concat cime2-matching-keyword-regexp "\\|\\<and\\>\\|[|>]?\\]\\|>?}\\|[|)#%]\\|;;")
  "Regexp matching CiME2 keywords which need special indentation.")

(defconst cime2-governing-phrase-regexp
  "\\<\\(v\\(al\\|irtual\\)\\|type\\|m\\(ethod\\|odule\\)\\|c\\(onstraint\\|lass\\)\\|in\\(herit\\|itializer\\)\\|ex\\(ternal\\|ception\\)\\|open\\|let\\|object\\)\\>"
  "Regexp matching cime2 phrase delimitors.")

(defconst cime2-keyword-alist
  '(("module" . cime2-default-indent)
    ("class" . cime2-class-indent)
    ("sig" . cime2-sig-struct-indent)
    ("struct" . cime2-sig-struct-indent)
    ("method" . cime2-method-indent)
    ("object" . cime2-begin-indent)
    ("begin" . cime2-begin-indent)
    ("for" . cime2-for-while-indent)
    ("while" . cime2-for-while-indent)
    ("do" . cime2-do-indent)
    ("type" . cime2-type-indent) ; in some cases, `type' acts like a match
    ("val" . cime2-val-indent)
    ("fun" . cime2-fun-indent)
    ("if" . cime2-if-then-else-indent)
    ("then" . cime2-if-then-else-indent)
    ("else" . cime2-if-then-else-indent)
    ("let" . cime2-let-indent)
    ("match" . cime2-match-indent)
    ("try" . cime2-try-indent)
    ("rule" . cime2-rule-indent)

    ;; case match keywords
    ("function" . cime2-function-indent)
    ("with" . cime2-with-indent)
    ("parse" . cime2-parse-indent)
    ("parser" . cime2-parser-indent)

    ;; default indentation keywords
    ("when" . cime2-default-indent)
    ("functor" . cime2-default-indent)
    ("exception" . cime2-default-indent)
    ("inherit" . cime2-default-indent)
    ("initializer" . cime2-default-indent)
    ("constraint" . cime2-default-indent)
    ("virtual" . cime2-default-indent)
    ("mutable" . cime2-default-indent)
    ("external" . cime2-default-indent)
    ("in" . cime2-in-indent)
    ("of" . cime2-default-indent)
    ("to" . cime2-default-indent)
    ("downto" . cime2-default-indent)
    ("->" . cime2-default-indent)
    ("|" . cime2-default-indent))
"Association list of indentation values based on governing keywords.")

(defconst cime2-leading-kwop-alist
  '(("|" . cime2-find-|-match)
    ("}" . cime2-find-match)
    (">}" . cime2-find-match)
    (")" . cime2-find-match)
    ("]" . cime2-find-match)
    ("|]" . cime2-find-match)
    (">]" . cime2-find-match)
    ("end" . cime2-find-match)
    ("done" . cime2-find-done-match)
    ("in"  . cime2-find-in-match)
    ("with" . cime2-find-with-match)
    ("else" . cime2-find-else-match)
    ("then" . cime2-find-match)
    ("do" . cime2-find-do-match)
    ("to" . cime2-find-match)
    ("downto" . cime2-find-match)
    ("and" . cime2-find-and-match))
  "Association list used in CiME2 mode for skipping back over nested blocks.")

(defun cime2-find-meaningful-word ()
  "Look back for a word, skipping comments and blanks.
Returns the actual text of the word, if found."
  (let ((found nil) (kwop nil))
    (while (and (not found)
		(re-search-backward
		 "[^ \t\n'_A-Za-z\277-\377]\\|\\<\\w+\\>\\|\\*)"
		 (point-min) t))
      (setq kwop (match-string 0))
      (if kwop
	  (if (cime2-in-comment-p)
	      (cime2-beginning-of-literal-or-comment-fast)
	    (setq found t))
	(setq found t)))
    (if found kwop (goto-char (point-min)) nil)))

(defconst cime2-find-kwop-regexp
  (concat cime2-matching-keyword-regexp "\\|\\<\\(for\\|while\\|do\\|if\\|try\\|match\\|begin\\|s\\(ig\\|truct\\)\\|class\\|object\\)\\>\\|[][(){}]\\|\\*)"))
(defun cime2-make-find-kwop-regexp (kwop-regexp)
  (concat cime2-find-kwop-regexp "\\|" kwop-regexp))

(defun cime2-find-kwop (kwop-regexp &optional do-not-skip-regexp)
  "Look back for a CiME2 keyword or operator matching KWOP-REGEXP.
Skips blocks etc...

Ignore occurences inside literals and comments.
If found, return the actual text of the keyword or operator."
  (let ((found nil) (kwop nil))
    (while (and (not found)
		(re-search-backward kwop-regexp (point-min) t)
		(setq kwop (match-string 0)))
      (cond
       ((cime2-in-literal-or-comment-p)
	(cime2-beginning-of-literal-or-comment-fast))
       ((looking-at "[]})]")
	(cime2-backward-up-list))
       ((cime2-at-phrase-break-p)
	(setq found t))
       ((and do-not-skip-regexp (looking-at do-not-skip-regexp))
	(if (and (string= kwop "|") (char-equal ?| (preceding-char)))
	    (backward-char)
	  (setq found t)))
       ((looking-at cime2-matching-keyword-regexp)
	(funcall (cdr (assoc (match-string 0)
			     cime2-leading-kwop-alist))))
       (t (setq found t))))
    (if found kwop (goto-char (point-min)) nil)))

(defun cime2-find-match ()
  (cime2-find-kwop cime2-find-kwop-regexp))

(defconst cime2-find-with-match-regexp
  (cime2-make-find-kwop-regexp "\\<\\(begin\\|module\\)\\>\\|[({[]"))
(defun cime2-find-with-match ()
  (let ((kwop (cime2-find-kwop cime2-find-with-match-regexp)))
    (cond
     ((looking-at "([ \t]*\\((\\*\\|$\\)") "!with")
     ((looking-at "[[{(]") "!with")
     ((looking-at "\\<begin\\>") "!with")
     ((looking-at "\\<module\\>") "!with")
     (t kwop))))

(defconst cime2-find-in-match-regexp
  (cime2-make-find-kwop-regexp "\\<let\\>"))
(defun cime2-find-in-match ()
  (cime2-find-kwop cime2-find-in-match-regexp "\\<and\\>"))

(defconst cime2-find-else-match-regexp
  (cime2-make-find-kwop-regexp ";"))
(defun cime2-find-else-match ()
  (let ((kwop (cime2-find-kwop cime2-find-else-match-regexp
				     "\\<then\\>")))
    (cond ((string= kwop "then")
	   (cime2-find-match) kwop)
	  ((string= kwop ";")
	   (cime2-find-semi-colon-match)
	   (cime2-find-else-match) kwop))))

(defun cime2-find-do-match ()
  (let ((kwop (cime2-find-kwop cime2-find-kwop-regexp
				   "\\<\\(down\\)?to\\>")))
    (if (or (string= kwop "to") (string= kwop "downto"))
	(cime2-find-match) kwop)))

(defun cime2-find-done-match ()
  (let ((kwop (cime2-find-kwop cime2-find-kwop-regexp "\\<do\\>")))
    (if (string= kwop "do")
	(cime2-find-do-match) kwop)))

(defconst cime2-find-and-match-regexp
  "\\<\\(do\\(ne\\)?\\|e\\(lse\\|nd\\)\\|in\\|then\\|with\\|\\(down\\)?to\\)\\>\\|\\<\\(for\\|while\\|do\\|if\\|try\\|match\\|begin\\|s\\(ig\\|truct\\)\\|class\\)\\>\\|[][(){}]\\|\\*)\\|\\<\\(rule\\|exception\\|let\\|in\\|type\\|val\\|module\\)\\>")
(defconst cime2-find-and-match-regexp-dnr
  (concat cime2-find-and-match-regexp "\\|\\<and\\>"))
(defun cime2-find-and-match (&optional do-not-recurse)
  (let* ((kwop (cime2-find-kwop (if do-not-recurse
					  cime2-find-and-match-regexp-dnr
					  cime2-find-and-match-regexp)
				      "\\<and\\>"))
	 (old-point (point)))
    (cond ((or (string= kwop "type") (string= kwop "module"))
	   (let ((kwop2 (cime2-find-meaningful-word)))
	     (cond ((string= kwop2 "with") kwop2)
		   ((string= kwop2 "and") (cime2-find-and-match))
		   ((string= kwop "module") (string= kwop2 "let") kwop2)
		   (t (goto-char old-point) kwop))))
	  (t kwop))))

(defconst cime2-find-=-match-regexp
  (cime2-make-find-kwop-regexp "\\<\\(let\\|m\\(ethod\\|odule\\)\\|type\\|class\\|when\\|i[fn]\\)\\>"))
(defun cime2-find-=-match ()
  (let ((kwop (cime2-find-kwop cime2-find-=-match-regexp
				     "\\<and\\|in\\>")))
    (if (string= kwop "and") (cime2-find-and-match) kwop)))

(defconst cime2-find-|-match-regexp
  (cime2-make-find-kwop-regexp
   "\\<\\(function\\|type\\|parser?\\)\\>\\|[=|]"))
(defun cime2-find-|-match ()
  (let* ((kwop (cime2-find-kwop cime2-find-|-match-regexp
				    "\\<\\(and\\|with\\)\\>\\||"))
	 (old-point (point)))
    (cond ((string= kwop "and")
	   (setq old-point (point))
	   (setq kwop (cime2-find-and-match))
	   (goto-char old-point)
	   kwop)
	  ((and (string= kwop "|")
		(looking-at "|[^|]")
		(cime2-in-indentation-p)) kwop)
	  ((string= kwop "|") (cime2-find-|-match))
	  ((and (string= kwop "=") (or (looking-at "=[ \t]*\\((\\*\\|$\\)")
				       (cime2-false-=-p)
				       (not (string= (save-excursion
						       (cime2-find-=-match))
						     "type"))))
	   (cime2-find-|-match))
	  ((string= kwop "parse")
	   (if (and (string-match "\\.mll" (buffer-name))
		    (save-excursion
		      (string= (cime2-find-meaningful-word) "=")))
	       kwop (cime2-find-|-match)))
	  (t kwop))))

(defconst cime2-find-->-match-regexp
  (cime2-make-find-kwop-regexp "\\<\\(with\\|fun\\(ction\\)?\\|parser\\)\\>\\|[|;]"))
(defun cime2-find-->-match ()
  (cime2-find-kwop cime2-find-->-match-regexp "\\<with\\>"))

(defconst cime2-find-semi-colon-match-regexp
  (cime2-make-find-kwop-regexp ";[ \t]*\\((\\*\\|$\\)\\|->\\|\\<\\(let\\|method\\)\\>"))
(defun cime2-find-semi-colon-match ()
  (cime2-find-kwop cime2-find-semi-colon-match-regexp
			 "\\<\\(in\\|and\\|do\\|with\\)\\>")
  ;; we don't need to find the keyword matching `and' since we know it's `let'!
  (cond
   ((looking-at ";[ \t]*\\((\\*\\|$\\)")
    (forward-line 1)
    (while (or (cime2-in-comment-p)
	       (looking-at "^[ \t]*\\((\\*\\|$\\)"))
      (forward-line 1))
    (back-to-indentation)
    (current-column))
   ((looking-at "\\((\\|\\[[<|]?\\|{<?\\)[ \t]*\\((\\*\\|$\\)")
    (cime2-back-to-paren-or-indentation t)
    (+ (current-column) cime2-default-indent))
   ((looking-at "\\((\\|\\[[<|]?\\|{<?\\)[ \t]*[^ \t\n]")
    (cime2-search-forward-paren)
    (current-column))
   ((looking-at "\\<method\\>[ \t]*\\((\\*\\|$\\)")
    (cime2-back-to-paren-or-indentation)
    (+ (current-column) cime2-method-indent))
   ((looking-at "\\<begin\\>[ \t]*\\((\\*\\|$\\)")
    (cime2-back-to-paren-or-indentation t)
    (+ (current-column) cime2-begin-indent))
   ((looking-at "->")
    (if (save-excursion
	  (cime2-find-->-match)
	  (looking-at "\\<\\(with\\|fun\\(ction\\)?\\|parser\\)\\>\\||"))
	(prog2
	    (cime2-back-to-paren-or-indentation)
	    (+ (current-column) cime2-default-indent))
      (cime2-find-semi-colon-match)))
   ((looking-at "\\<in\\>")
    (cime2-find-in-match)
    (cime2-back-to-paren-or-indentation)
    (+ (current-column) cime2-in-indent))
   (t (cime2-back-to-paren-or-indentation t)
      (+ (current-column) cime2-default-indent))))

(defconst cime2-find-phrase-indentation-regexp
  (cime2-make-find-kwop-regexp (concat cime2-governing-phrase-regexp
					"\\|\\<and\\>")))
(defconst cime2-find-phrase-indentation-regexp-pb
  (concat cime2-find-phrase-indentation-regexp "\\|;;"))
(defconst cime2-find-phrase-indentation-class-regexp
  (concat cime2-matching-keyword-regexp "\\|\\<class\\>"))
(defun cime2-find-phrase-indentation (&optional phrase-break)
  (if (and (looking-at "\\<\\(type\\|module\\)\\>") (> (point) (point-min))
	   (save-excursion
	     (cime2-find-meaningful-word)
	     (looking-at "module\\|with\\|and\\|let")))
      (prog2
	  (cime2-find-meaningful-word)
	  (+ (current-column) cime2-default-indent))
    (let ((looking-at-and (looking-at "\\<and\\>"))
	  (kwop (cime2-find-kwop
		 (if phrase-break
		     cime2-find-phrase-indentation-regexp-pb
		   cime2-find-phrase-indentation-regexp)
		 "\\<\\(end\\|and\\|with\\|in\\)\\>"))
	  (tmpkwop nil) (curr nil)
	  (old-point (save-excursion (beginning-of-line) (point))))
      (if (and kwop (string= kwop "and"))
	  (setq kwop (cime2-find-and-match)))
      (if (not kwop) (current-column)
	(cond
	 ((string= kwop "end")
	  (if (not (save-excursion
		     (setq tmpkwop (cime2-find-match))
		     (setq curr (point))
		     (string= tmpkwop "object")))
	      (prog2
		  (cime2-find-match)
		  (cime2-find-phrase-indentation phrase-break))
	    (cime2-find-kwop cime2-find-phrase-indentation-class-regexp)
	    (current-column)))
	 ((and (string= kwop "with")
	       (not (save-excursion
		      (setq tmpkwop (cime2-find-with-match))
		      (setq curr (point))
		      (and (string= tmpkwop "!with")
			   (looking-at "\\<module\\>")))))
	  (goto-char curr)
	  (cime2-find-phrase-indentation phrase-break))
	 ((and (string= kwop "in")
	       (not (save-excursion
		      (setq tmpkwop (cime2-find-in-match))
		      (if (string= tmpkwop "and")
			  (setq tmpkwop (cime2-find-and-match)))
		      (setq curr (point))
		      (and (string= tmpkwop "let")
			   (not (cime2-looking-at-expression-let))))))
	  (goto-char curr)
	  (cime2-find-phrase-indentation phrase-break))
	 ((cime2-at-phrase-break-p)
	  (end-of-line)
	  (cime2-skip-blank-and-comments)
	  (current-column))
	 ((string= kwop "let")
	  (if (cime2-looking-at-expression-let)
	      (cime2-find-phrase-indentation phrase-break)
	    (current-column)))
	 ((string= kwop "with")
	  (current-column))
	 ((string= kwop "end")
	  (current-column))
	 ((string= kwop "in")
	  (cime2-find-in-match)
	  (current-column))
	 ((looking-at "\\<\\(class\\|object\\|s\\(ig\\|truct\\)\\)\\>")
	  (cime2-back-to-paren-or-indentation)
	  (+ (cime2-assoc-indent kwop) (current-column)))
	 ((string= kwop "virtual")
	  (if (save-excursion
		(setq kwop (cime2-find-meaningful-word))
		(setq curr (current-column))
		(setq old-point (point))
		(string= kwop "class"))
	      (prog2 (goto-char old-point) (+ curr cime2-class-indent))
	    (current-column)))
	 ((or (string= kwop "type") (string= kwop "module"))
	  (if (or (cime2-looking-at-false-type)
		  (cime2-looking-at-false-module))
	      (if looking-at-and (current-column)
		(cime2-find-meaningful-word)
		(if (looking-at "\\<and\\>")
		    (prog2
			(cime2-find-and-match)
			(cime2-find-phrase-indentation phrase-break))
		  (current-column)))
	    (current-column)))
	 ((looking-at
	   "\\((\\|\\[[<|]?\\|{<?\\)[ \t]*\\((\\*\\|$\\)")
	  (cime2-back-to-paren-or-indentation)
	  (+ (current-column) cime2-default-indent))
	 ((looking-at "\\((\\|\\[[<|]?\\|{<?\\)[ \t]*[^ \t\n]")
	  (cime2-search-forward-paren)
	  (current-column))
	 ((string= kwop "open") ; compatibility with camllight `#open'
	  (cime2-back-to-paren-or-indentation) (current-column))
	 (t (current-column)))))))

(defconst cime2-back-to-paren-or-indentation-regexp
  "[][(){}]\\|\\*)\\|^[ \t]*\\(.\\|\n\\)")
(defconst cime2-back-to-paren-or-indentation-in-regexp
  (concat "\\<in\\>\\|" cime2-back-to-paren-or-indentation-regexp))
(defconst cime2-back-to-paren-or-indentation-lazy-regexp
  "[])}]\\|\\*)\\|^[ \t]*\\(.\\|\n\\)")
(defconst cime2-back-to-paren-or-indentation-lazy-in-regexp
  (concat "\\<in\\>\\|" cime2-back-to-paren-or-indentation-regexp))
(defun cime2-back-to-paren-or-indentation (&optional forward-in)
  "Searches backwards for the first open paren in line, or skip to indentation.
Returns t iff skipped to indentation."
  (if (or (bolp) (cime2-in-indentation-p)) (prog2 (back-to-indentation) t)
    (let ((kwop (cime2-find-kwop
		 (if cime2-lazy-paren
		     (if forward-in
			 cime2-back-to-paren-or-indentation-lazy-in-regexp
		       cime2-back-to-paren-or-indentation-lazy-regexp)
		   (if forward-in
		       cime2-back-to-paren-or-indentation-in-regexp
		     cime2-back-to-paren-or-indentation-regexp))
		 "\\<and\\|with\\|in\\>"))
	  (retval))
      (if (string= kwop "with")
	  (let ((with-point (point)))
	    (setq kwop (cime2-find-with-match))
	    (if (or (string= kwop "match") (string= kwop "try"))
		(cime2-find-kwop
		 cime2-back-to-paren-or-indentation-regexp
		 "\\<and\\>")
	      (setq kwop "with") (goto-char with-point))))
      (setq retval
	    (cond
	     ((string= kwop "with") nil)
	     ((string= kwop "in") (cime2-in-indentation-p))
	     ((looking-at "[[{(]")
	      (prog2 (cime2-search-forward-paren) nil))
	     (t (back-to-indentation) t)))
      (cond
       ((looking-at "|[^|]")
	(prog2 (re-search-forward "|[^|][ \t]*") nil))
       ((and forward-in (string= kwop "in"))
	(cime2-find-in-match)
	(cime2-back-to-paren-or-indentation forward-in)
	(if (looking-at "\\<\\(let\\|and\\)\\>")
	    (forward-char cime2-let-indent)) nil)
       (t retval)))))

(defun cime2-search-forward-paren ()
  (if cime2-lazy-paren (cime2-back-to-paren-or-indentation)
    (re-search-forward "\\((\\|\\[[<|]?\\|{<?\\)[ \t]*")))

(defun cime2-add-default-indent (leading-operator)
  (if leading-operator 0 cime2-default-indent))

(defconst cime2-compute-argument-indent-regexp
  (cime2-make-find-kwop-regexp cime2-kwop-regexp))
(defun cime2-compute-argument-indent (leading-operator)
  (let ((old-point (save-excursion (beginning-of-line) (point)))
	(match-end-point) (kwop))
    (setq kwop (cime2-find-kwop cime2-compute-argument-indent-regexp
				 cime2-keyword-regexp))
    (setq match-end-point (+ (point) (length kwop))) ; match-end is invalid !
    (if (and (if cime2-lazy-= (looking-at "[[{(]") (looking-at "[[{(=]"))
	     (not (looking-at "=[ \t]*\\((\\*.*\\)?$")))
	(prog2
	    (if (and (not cime2-lazy-=)
		     (looking-at "=")) (re-search-forward "=[ \t]*")
	      (cime2-search-forward-paren))
	    (+ (cime2-add-default-indent leading-operator)
	       (current-column)))
      (if (<= old-point (point))
	  (+ (cime2-add-default-indent leading-operator) (current-column))
	(forward-line 1)
	(beginning-of-line)
	(while (or (cime2-in-comment-p)
		   (looking-at "^[ \t]*\\((\\*.*\\)?$"))
	  (forward-line 1))
	(cime2-back-to-paren-or-indentation)
	(if (save-excursion (goto-char match-end-point)
			    (looking-at "[ \t]*\\((\\*.*\\)?$"))
	    (+ (cime2-add-default-indent leading-operator)
	       (current-column))
	  (current-column))))))

(defconst cime2-compute-indent-regexp
  (concat cime2-compute-argument-indent-regexp "\\|^[ \t]*[^ \t\n]"))
(defun cime2-compute-indent ()
  (let ((leading-operator (looking-at cime2-operator-regexp)))
    (beginning-of-line)
    (save-excursion
      (cime2-find-meaningful-word)
      (if (looking-at cime2-operator-regexp) (setq leading-operator t)))
    (save-excursion
      (let ((kwop (cime2-find-kwop (if leading-operator
					cime2-compute-argument-indent-regexp
				      cime2-compute-indent-regexp)
				    cime2-keyword-regexp))
	    (leading-special 0))
	(if (string= kwop "and") (setq kwop (cime2-find-and-match)))
	(while (and (string= kwop "=") (cime2-false-=-p))
	  (setq kwop (cime2-find-kwop cime2-compute-indent-regexp
				       cime2-keyword-regexp))
	  (if (string= kwop "and") (setq kwop (cime2-find-and-match))))
	(if (not kwop) (current-column)
	  (cond
	   ((cime2-at-phrase-break-p)
	    (cime2-find-phrase-indentation t))
	   ((and (string= kwop "|") (not  (char-equal ?\[ (preceding-char))))
	    (cime2-backward-char)
	    (cime2-back-to-paren-or-indentation)
	    (+ (current-column) cime2-default-indent
	       (cime2-add-default-indent leading-operator)))
	   ((or (looking-at "[[{(]")
		(and (looking-at "[<|]")
		     (char-equal ?\[ (preceding-char))
		     (prog2 (cime2-backward-char) t))
		(and (looking-at "<")
		     (char-equal ?\{ (preceding-char))
		     (prog2 (cime2-backward-char) t)))
	    (if (looking-at "\\((\\|\\[[<|]?\\|{<?\\)[ \t]*\\((\\*\\|$\\)")
		(prog2
		    (cime2-back-to-paren-or-indentation t)
		    (+ cime2-default-indent
		       (current-column))) ; parens do not operate
	      (cime2-search-forward-paren)
	      (+ (cime2-add-default-indent leading-operator)
		 (current-column))))
	   ((looking-at cime2-keyword-regexp)
	    (cond ((or (string= kwop ";") (string= kwop ","))
		   (if (looking-at ";[ \t]*\\((\\*\\|$\\)")
		       (cime2-find-semi-colon-match)
		     (if (looking-at ",[ \t]*\\((\\*\\|$\\)")
			 (progn
			   (cime2-back-to-paren-or-indentation t)
			   (current-column))
		       (cime2-back-to-paren-or-indentation t)
		       (+ (current-column) cime2-default-indent))))
		  ((and (looking-at "\\<\\(in\\|begin\\|do\\)\\>\\|->")
			(not (looking-at
			      "\\([a-z]+\\|->\\)[ \t]*\\((\\*\\|$\\)")))
		   (if (string= kwop "in")
		       (re-search-forward "\\<in\\>[ \t]*")
		     (cime2-back-to-paren-or-indentation t))
		   (+ (current-column)
		      (cime2-add-default-indent leading-operator)
		      (if (string= kwop "in") 0 ; aligned, do not indent
			(cime2-assoc-indent kwop))))
		  ((string= kwop "with")
		   (if (string= (save-excursion (cime2-find-with-match))
				"!with")
		       (prog2
			   (cime2-back-to-paren-or-indentation)
			   (+ (current-column) cime2-default-indent))
		     (prog2
			 (cime2-back-to-paren-or-indentation)
			 (+ (current-column)
			    (cime2-assoc-indent kwop
						 (looking-at "\\<let\\>"))))))
		  ((string= kwop "in")
		   (cime2-find-in-match)
		   (cime2-back-to-paren-or-indentation)
		   (+ (current-column) cime2-in-indent))
		  (t (cime2-back-to-paren-or-indentation)
		     (+ (current-column)
			(cime2-assoc-indent kwop
					     (looking-at "\\<let\\>"))))))
	   ((looking-at "=")
	    (if (or cime2-lazy-=
		    (looking-at "=[ \t]*\\((\\*\\|$\\)")) ; not perfect...
		(let ((current-column-module-type nil))
		  (+
		   (progn
		     (cime2-find-=-match)
		     (save-excursion
		       (if (looking-at "\\<and\\>") (cime2-find-and-match))
		       (cond
			((looking-at "\\<type\\>")
			 (cime2-find-meaningful-word)
			 (if (looking-at "\\<module\\>")
			     (progn
			       (setq current-column-module-type
				     (current-column))
			       cime2-default-indent)
			   (if (looking-at "\\<\\(with\\|and\\)\\>")
			       (progn
				 (setq current-column-module-type
				       (current-column))
				 cime2-default-indent)
			     (re-search-forward "\\<type\\>")
			     (beginning-of-line)
			     (+ cime2-type-indent
				cime2-|-extra-unindent))))
			((looking-at
			  "\\<\\(let\\|m\\(ethod\\|odule\\)\\|class\\|when\\|\\|for\\|if\\)\\>")
			 (let ((matched-string (match-string 0)))
			   (cime2-back-to-paren-or-indentation t)
			   (setq current-column-module-type (current-column))
			   (cime2-assoc-indent matched-string)))
			(t (cime2-search-forward-paren)
			   (setq current-column-module-type (current-column))
			   cime2-default-indent))))
		   (if current-column-module-type
		       current-column-module-type
		     (current-column))))
	      (+ (cime2-add-default-indent leading-operator)
		 (prog2
		     (re-search-forward "=[ \t]*")
		     (current-column)))))
	   (nil 0)
	   (t (cime2-compute-argument-indent leading-operator))))))))

(defun cime2-looking-at-expression-let ()
  (save-excursion
    (and (cime2-find-meaningful-word)
	 (not (cime2-at-phrase-break-p))
	 (or (looking-at "[[({;=]\\|\\<\\(begin\\|i[fn]\\|do\\|t\\(ry\\|hen\\)\\|else\\|match\\|wh\\(ile\\|en\\)\\)\\>")
	     (looking-at cime2-operator-regexp)))))

(defun cime2-looking-at-false-module ()
  (save-excursion (cime2-find-meaningful-word)
		  (looking-at "\\<\\(let\\|with\\|and\\)\\>")))

(defun cime2-looking-at-false-sig-struct ()
  (save-excursion (cime2-find-module)
		  (looking-at "\\<module\\>")))

(defun cime2-looking-at-false-type ()
  (save-excursion (cime2-find-meaningful-word)
		  (looking-at "\\<\\(with\\|module\\|and\\)\\>")))

(defun cime2-looking-at-in-let ()
  (save-excursion (string= (cime2-find-meaningful-word) "in")))

(defconst cime2-find-module-regexp
  (cime2-make-find-kwop-regexp "\\<module\\>"))
(defun cime2-find-module ()
  (cime2-find-kwop cime2-find-module-regexp))

(defun cime2-indent-command ()
  "Indent the current line in CiME2 mode.

Compute new indentation based on CiME2 syntax."
  (interactive "*")
  (save-excursion
    (setq case-fold-search nil)
    (back-to-indentation)
    (indent-line-to
     (save-excursion
       (cond
	((and (cime2-in-comment-p) (not (and (looking-at "(\\*")
					      cime2-indent-leading-comments)))
	 (cond
	  ((looking-at "(\\*") (current-column))
	  ((looking-at "\\*\\**)")
	   (cime2-beginning-of-literal-or-comment-fast)
	   (if cime2-support-leading-star-comments
	       (+ (current-column)
		  (if (save-excursion
			(forward-line 1)
			(back-to-indentation)
			(looking-at "*")) 1
		    cime2-comment-end-extra-indent))
	     (+ (current-column) cime2-comment-end-extra-indent)))
	  (cime2-indent-comments
	   (let ((star (and cime2-support-leading-star-comments
			    (looking-at "\\*"))))
	     (cime2-beginning-of-literal-or-comment-fast)
	     (if star (re-search-forward "(") (re-search-forward "(\\*[ \t]*"))
	     (current-column)))))
	((cime2-in-literal-p)
	 (current-column))
	((looking-at "\\<let\\>")
	 (if (cime2-looking-at-expression-let)
	     (if (cime2-looking-at-in-let)
		 (progn
		   (cime2-find-meaningful-word)
		   (cime2-find-in-match)
		   (if (cime2-looking-at-in-let)
		       (prog2 (cime2-back-to-paren-or-indentation)
			   (current-column))
		     (current-column)))
	       (cime2-compute-indent))
	   (cime2-find-phrase-indentation)))
	((looking-at cime2-governing-phrase-regexp)
	 (cime2-find-phrase-indentation))
	((and cime2-sig-struct-align (looking-at "\\<\\(sig\\|struct\\)\\>"))
	 (if (string= (cime2-find-module) "module") (current-column)
	   (cime2-back-to-paren-or-indentation)
	   (+ cime2-default-indent (current-column))))
	((looking-at cime2-leading-kwop-regexp)
	 (let ((kwop (match-string 0)))
	   (if (looking-at "[#%]\\|;;") 0
	     (let* ((old-point (point))
		    (need-not-back-kwop (string= kwop "and"))
		    (real-| (looking-at "|\\([^|]\\|$\\)"))
		    (matching-kwop
		     (if (string= kwop "and") (cime2-find-and-match t)
		       (funcall (cdr (assoc kwop cime2-leading-kwop-alist)))))
		    (match-|-keyword-p
		     (looking-at cime2-match-|-keyword-regexp)))
	       (if (string= matching-kwop "!with")
		   (prog2
		       (cond
			((looking-at "[[{(]")
			 (cime2-search-forward-paren))
			((looking-at "\\<begin\\>")
			 (forward-char cime2-begin-indent))
			((looking-at "\\<module\\>")
			 (forward-char cime2-default-indent)))
		       (current-column))
		 (if (and (looking-at "\\(\\[|?\\|{<?\\|(\\)[ \t]*[^ \t\n]")
			  (not (looking-at
				"[[{(][|<]?[ \t]*\\((\\*\\|$\\)")))
		     (prog2
			 (if cime2-lazy-paren
			     (cime2-back-to-paren-or-indentation))
			 (current-column))
		   (if (string= kwop "|")
		       (if real-|
			   (cond
			    ((string= matching-kwop "|")
			     (if (not need-not-back-kwop)
				 (cime2-back-to-paren-or-indentation))
			     (current-column))
			    ((and (string= matching-kwop "=")
				  (not (cime2-false-=-p)))
			     (re-search-forward "=[ \t]*")
			     (current-column))
			    (match-|-keyword-p
			     (if (not need-not-back-kwop)
				 (cime2-back-to-paren-or-indentation))
			     (- (+ (cime2-assoc-indent
				    matching-kwop (looking-at "\\<let\\>"))
				   (current-column))
				(if (string= matching-kwop "type") 0
				  cime2-|-extra-unindent)))
			    (t (goto-char old-point)
			       (cime2-compute-indent)))
			 (goto-char old-point)
			 (cime2-compute-indent))
		     (if (not need-not-back-kwop)
			 (cime2-back-to-paren-or-indentation
			  (not (string= kwop "in"))))
		     (current-column))))))))
	(t (cime2-compute-indent))))))
  (if (cime2-in-indentation-p) (back-to-indentation)))

(defun cime2-split-string ()
  "Called whenever a line is broken inside a CiME2 string literal."
  (insert-before-markers "\" ^\"")
  (cime2-backward-char))

(defadvice newline-and-indent (around
			       cime2-newline-and-indent
			       activate)
  "Handle multi-line strings in CiME2 mode."
    (let ((hooked (and (eq major-mode 'cime2-mode) (cime2-in-literal-p)))
	  (split-mark))
      (if (not hooked) nil
	(setq split-mark (set-marker (make-marker) (point)))
	(cime2-split-string))
      ad-do-it
      (if (not hooked) nil
	(goto-char split-mark)
	(set-marker split-mark nil))))

(defun cime2-electric ()
  "If inserting a |, ) operator at beginning of line, reindent the line."
  (interactive "*")
  (let ((electric (and cime2-electric-indent
		       (cime2-in-indentation-p)
		       (not (cime2-in-literal-or-comment-p)))))
    (self-insert-command 1)
    (if (and electric
	     (not (and (char-equal ?| (preceding-char))
		       (save-excursion
			 (cime2-backward-char)
			 (cime2-find-|-match)
			 (not (looking-at cime2-match-|-keyword-regexp))))))
	(cime2-indent-command))))

(defun cime2-electric-rc ()
  "If inserting a } operator at beginning of line, reindent the line.

Reindent also if } is inserted after a > operator at beginning of line.
Also, if the matching { is followed by a < and this } is not preceded
by >, insert one >."
  (interactive "*")
  (let* ((prec (preceding-char))
	 (look-bra (and cime2-electric-close-vector
			(not (cime2-in-literal-or-comment-p))
			(not (char-equal ?> prec))))
	 (electric (and cime2-electric-indent
			(or (cime2-in-indentation-p)
			    (and (char-equal ?> prec)
				 (save-excursion (cime2-backward-char)
						 (cime2-in-indentation-p))))
			(not (cime2-in-literal-or-comment-p)))))
    (self-insert-command 1)
    (if look-bra
	(save-excursion
	  (let ((inserted-char
		 (save-excursion
		   (cime2-backward-char)
		   (cime2-backward-up-list)
		   (cond ((looking-at "{<") ">")
			 (t "")))))
	    (cime2-backward-char)
	    (insert inserted-char))))
    (if electric (cime2-indent-command))))

(defun cime2-electric-rb ()
  "If inserting a ] operator at beginning of line, reindent the line.

Reindent also if ] is inserted after a | or > operator at beginning of line.
Also, if the matching [ is followed by a | (resp. <) and this ] is not
preceded by | (resp. >), insert one | (resp. >)."
  (interactive "*")
  (let* ((prec (preceding-char))
	 (look-|-or-bra (and cime2-electric-close-vector
			     (not (cime2-in-literal-or-comment-p))
			     (not (or (and (char-equal ?| prec)
					   (not (char-equal
						 (save-excursion
						   (cime2-backward-char)
						   (preceding-char)) ?\[)))
				      (char-equal ?> prec)))))
	 (electric (and cime2-electric-indent
			(or (cime2-in-indentation-p)
			    (and (or (char-equal ?| prec) (char-equal ?> prec))
				 (save-excursion (cime2-backward-char)
						 (cime2-in-indentation-p))))
			(not (cime2-in-literal-or-comment-p)))))
    (self-insert-command 1)
    (if look-|-or-bra
	(save-excursion
	  (let ((inserted-char
		 (save-excursion
		   (cime2-backward-char)
		   (cime2-backward-up-list)
		   (cond ((looking-at "\\[|") "|")
			 ((looking-at "\\[<") ">")
			 (t "")))))
	    (cime2-backward-char)
	    (insert inserted-char))))
    (if electric (cime2-indent-command))))

(defun cime2-abbrev-hook ()
  "If inserting a leading keyword at beginning of line, reindent the line."
  (if (not (cime2-in-literal-or-comment-p))
      (let* ((bol (save-excursion (beginning-of-line) (point)))
	     (kw (save-excursion
		   (and (re-search-backward "^[ \t]*\\(\\w+\\)\\=" bol t)
			(match-string 1)))))
	(if kw (progn
		   (insert " ")
		   (cime2-indent-command)
		   (backward-delete-char-untabify 1))))))

(defun cime2-skip-to-end-of-phrase ()
  (let ((old-point (point)))
    (if (and (string= (cime2-find-meaningful-word) ";")
	     (char-equal (preceding-char) ?\;))
	(setq old-point (1- (point))))
    (goto-char old-point)
    (let ((kwop (cime2-find-meaningful-word)))
      (goto-char (+ (point) (length kwop))))))

(defun cime2-skip-blank-and-comments ()
  (skip-chars-forward " \t\n")
  (while (and (< (point) (point-max)) (cime2-in-comment-p))
    (re-search-forward "\\*)") (skip-chars-forward " \t\n")))

(defun cime2-skip-back-blank-and-comments ()
  (skip-chars-backward " \t\n")
  (while (save-excursion (cime2-backward-char)
			 (and (> (point) (point-min)) (cime2-in-comment-p)))
    (cime2-backward-char)
    (cime2-beginning-of-literal-or-comment) (skip-chars-backward " \t\n")))

(defconst cime2-beginning-phrase-regexp
  "^#[ \t]*[a-z][_a-z]*\\>\\|\\<\\(let\\)\\>\\|;"
  "Regexp matching cime2 phrase delimitors.")
(defun cime2-find-phrase-beginning ()
  "Find `real' phrase beginning and returns point."
  (beginning-of-line)
  (cime2-skip-blank-and-comments)
  (end-of-line)
  (cime2-skip-to-end-of-phrase)
  (let ((old-point (point)))
    (cime2-find-kwop cime2-beginning-phrase-regexp)
    (while (and (> (point) (point-min)) (< (point) old-point)
		(or (not (looking-at cime2-beginning-phrase-regexp))
		    (and (looking-at "\\<let\\>")
			 (cime2-looking-at-expression-let))
		    (and (looking-at "\\<module\\>")
			 (cime2-looking-at-false-module))
		    (and (looking-at "\\<\\(sig\\|struct\\)\\>")
			 (cime2-looking-at-false-sig-struct))
		    (and (looking-at "\\<type\\>")
			 (cime2-looking-at-false-type))))
      (if (looking-at "\\<end\\>")
	  (cime2-find-match)
	(if (not (bolp)) (cime2-backward-char))
	(setq old-point (point))
	(cime2-find-kwop cime2-beginning-phrase-regexp)))
    (if (cime2-at-phrase-break-p)
	(prog2 (end-of-line) (cime2-skip-blank-and-comments)))
    (back-to-indentation)
    (point)))

(defun cime2-search-forward-end ()
  (re-search-forward "\\<\\(end\\|begin\\)\\>" (point-max) t)
  (if (string= (match-string 0) "\\<begin\\>") (cime2-search-forward-end)))

(defconst cime2-inside-block-regexp
  (concat cime2-matching-keyword-regexp
	  "\\|\\<\\(class\\|sig\\|struct\\)\\>"))
(defun cime2-inside-block-find-kwop ()
  (let ((kwop (cime2-find-kwop cime2-inside-block-regexp
				"\\<\\(and\\|end\\)\\>")))
    (if (string= kwop "and") (setq kwop (cime2-find-and-match)))
    (if (string= kwop "end")
	(progn
	  (cime2-find-match)
	  (cime2-find-kwop cime2-inside-block-regexp)
	  (cime2-inside-block-find-kwop))
      kwop)))
(defun cime2-inside-block-p ()
  (let ((begin) (end) (and-end) (kwop t))
    (save-excursion
      (if (looking-at "\\<and\\>")
	  (cime2-find-and-match))
      (if (not (looking-at "\\<\\(class\\|sig\\|struct\\)\\>"))
	  (while (and (setq kwop (cime2-inside-block-find-kwop))
		      (not (looking-at "\\<\\(class\\|sig\\|struct\\)\\>")))))
      (if (not kwop) ()
	(setq begin (point))
	(while (and (cime2-search-forward-end)
		    (save-excursion
		      (cime2-backward-char 3)
		      (cime2-find-match)
		      (if (looking-at "\\<object\\>")
			  (cime2-inside-block-find-kwop))
		      (> (point) begin))))
	(cime2-backward-char 3)
	(if (not (looking-at "\\<end\\>")) ()
	  (cime2-forward-char 3)
	  (setq end (point))
	  (setq and-end (point))
	  (cime2-skip-blank-and-comments)
	  (while (looking-at "\\<and\\>")
	    (setq and-end (point))
	    (while (and (cime2-search-forward-end)
			(save-excursion
			  (cime2-backward-char 3)
			  (cime2-find-match)
			  (if (looking-at "\\<object\\>")
			      (cime2-inside-block-find-kwop))
			  (> (point) and-end))))
	    (cime2-backward-char 3)
	    (if (not (looking-at "\\<end\\>")) ()
	      (cime2-forward-char 3)
	      (setq and-end (point))
	      (cime2-skip-blank-and-comments)))
	  (list begin end and-end))))))
	    
(defun cime2-discover-phrase (&optional quiet)
  (end-of-line)
  (let ((end (point)))
    (cime2-find-phrase-beginning)
    (if (> (point) end) (setq end (point)))
    (save-excursion
      (let ((begin (point)) (cpt 0) (lines-left 0) (stop)
	    (inside-block (cime2-inside-block-p))
	    (looking-block (looking-at "\\<class\\>")))
	(if (and looking-block inside-block)
	    (progn
	      (setq begin (nth 0 inside-block))
	      (setq end (nth 2 inside-block))
	      (goto-char end))
	  (if inside-block
	      (progn
		(setq stop (save-excursion (goto-char (nth 1 inside-block))
					   (beginning-of-line) (point)))
		(if (< stop end) (setq stop (point-max))))
	    (setq stop (point-max)))
	  (save-restriction
	    (goto-char end)
	    (while (and (= lines-left 0)
			(or (not inside-block) (< (point) stop))
			(<= (save-excursion
			      (cime2-find-phrase-beginning)) end))
	      (if (not quiet)
		  (prog2
		      (setq cpt (1+ cpt))
		      (if (= 8 cpt)
			  (message "Looking for enclosing phrase..."))))
	      (setq end (point))
	      (cime2-skip-to-end-of-phrase)
	      (beginning-of-line)
	      (narrow-to-region (point) (point-max))
	      (goto-char end)
	      (setq lines-left (forward-line 1)))))
	(if (>= cpt 8) (message "Looking for enclosing phrase... done."))
	(save-excursion (cime2-skip-blank-and-comments) (setq end (point)))
	(cime2-skip-back-blank-and-comments)
	(list begin (point) end)))))

(defun cime2-mark-phrase ()
  "Put mark at end of this CiME2 phrase, point at beginning.
The CiME2 phrase is the phrase just before the point."
  (interactive)
  (let ((pair (cime2-discover-phrase)))
    (goto-char (nth 1 pair)) (push-mark (nth 0 pair) t t)))

(defun cime2-next-phrase (&optional quiet)
  "Skip to the beginning of the next phrase."
  (interactive "i")
  (goto-char (save-excursion (nth 2 (cime2-discover-phrase quiet))))
  (if (looking-at "\\<end\\>") (cime2-next-phrase quiet)))

(defun cime2-previous-phrase ()
  "Skip to the beginning of the previous phrase."
  (interactive)
  (beginning-of-line)
  (cime2-skip-to-end-of-phrase)
  (cime2-discover-phrase))

(defun cime2-indent-phrase ()
  "If inside a comment call `fill-paragraph-or-region'; Otherwise
indent all lines in the current phrase."
  (interactive)
  (save-excursion
    (if (and (cime2-in-comment-p)
	     (save-excursion (beginning-of-line) (cime2-in-comment-p)))
	(let ((begpoint (save-excursion
			  (cime2-beginning-of-literal-or-comment) (point)))
	      (endpoint (save-excursion
			  (while (cime2-in-comment-p)
			    (re-search-forward "*)")) (point))))
	  (fill-region begpoint endpoint))
      (let ((pair (cime2-discover-phrase)))
	(indent-region (nth 0 pair) (nth 1 pair) nil)))))

;; Auxiliary function (from J. Garrigue)
(defun cime2-find-alternate-file ()
  "Switch Implementation/Interface."
  (interactive)
  (let ((name (buffer-file-name)))
    (if (string-match "^\\(.*\\)\\.\\(ml\\|mli\\)$" name)
	(find-file (concat (match-string 1 name)
			   (if (string= "ml" (match-string 2 name))
			       ".mli" ".ml"))))))

(defun cime2-insert-begin-form ()
  "Inserts a nicely formatted begin-end form, leaving a mark after end."
  (interactive "*")
  (let ((prec (preceding-char)))
    (if (and prec (not (char-equal ?\  (char-syntax prec))))
	(insert " ")))
  (let ((old (point)))
    (insert "begin\n\nend\n")
    (end-of-line)
    (indent-region old (point) nil)
    (push-mark)
    (forward-line -2)
    (cime2-indent-command)))

(defun cime2-insert-for-form ()
  "Inserts a nicely formatted for-to-done form, leaving a mark after done."
  (interactive "*")
  (let ((prec (preceding-char)))
    (if (and prec (not (char-equal ?\  (char-syntax prec))))
	(insert " ")))
  (let ((old (point)))
    (insert "for  do\n\ndone\n")
    (end-of-line)
    (indent-region old (point) nil)
    (push-mark)
    (forward-line -2)
    (cime2-indent-command)
    (beginning-of-line 1)
    (backward-char 4)))

(defun cime2-insert-while-form ()
  "Inserts a nicely formatted for-to-done form, leaving a mark after done."
  (interactive "*")
  (let ((prec (preceding-char)))
    (if (and prec (not (char-equal ?\  (char-syntax prec))))
	(insert " ")))
  (let ((old (point)))
    (insert "while  do\n\ndone\n")
    (end-of-line)
    (indent-region old (point) nil)
    (push-mark)
    (forward-line -2)
    (cime2-indent-command)
    (beginning-of-line 1)
    (backward-char 4)))

(defun cime2-insert-if-form ()
  "Inserts a nicely formatted if-then-else form, leaving a mark after else."
  (interactive "*")
  (let ((prec (preceding-char)))
    (if (and prec (not (char-equal ?\  (char-syntax prec))))
	(insert " ")))
  (let ((old (point)))
    (insert "if\n\nthen\n\nelse\n")
    (end-of-line)
    (indent-region old (point) nil)
    (cime2-indent-command)
    (push-mark)
    (forward-line -2)
    (cime2-indent-command)
    (forward-line -2)
    (cime2-indent-command)))

(defun cime2-insert-match-form ()
  "Inserts a nicely formatted math-with form, leaving a mark after with."
  (interactive "*")
  (let ((prec (preceding-char)))
    (if (and prec (not (char-equal ?\  (char-syntax prec))))
	(insert " ")))
  (let ((old (point)))
    (insert "match\n\nwith\n")
    (end-of-line)
    (indent-region old (point) nil)
    (cime2-indent-command)
    (push-mark)
    (forward-line -2)
    (cime2-indent-command)))

(defun cime2-insert-let-form ()
  "Inserts a nicely formatted let-in form, leaving a mark after in."
  (interactive "*")
  (let ((prec (preceding-char)))
    (if (and prec (not (char-equal ?\  (char-syntax prec))))
	(insert " ")))
  (let ((old (point)))
    (insert "let  in\n")
    (end-of-line)
    (indent-region old (point) nil)
    (cime2-indent-command)
    (push-mark)
    (beginning-of-line)
    (backward-char 4)
    (cime2-indent-command)))

(defun cime2-insert-try-form ()
  "Inserts a nicely formatted try-with form, leaving a mark after with."
  (interactive "*")
  (let ((prec (preceding-char)))
    (if (and prec (not (char-equal ?\  (char-syntax prec))))
	(insert " ")))
  (let ((old (point)))
    (insert "try\n\nwith\n")
    (end-of-line)
    (indent-region old (point) nil)
    (cime2-indent-command)
    (push-mark)
    (forward-line -2)
    (cime2-indent-command)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                            CiME2 interactive mode

;; Augment CiME2 mode, so you can process CiME2 code in the source files.

(require 'comint)

(defvar cime2-interactive-mode-map nil)
(if cime2-interactive-mode-map nil
  (setq cime2-interactive-mode-map
        (copy-keymap comint-mode-map)))
(define-key cime2-interactive-mode-map "#" 'cime2-electric)
(define-key cime2-interactive-mode-map "%" 'cime2-electric)
(define-key cime2-interactive-mode-map "|" 'cime2-electric)
(define-key cime2-interactive-mode-map ")" 'cime2-electric)
(define-key cime2-interactive-mode-map "}" 'cime2-electric-rc)
(define-key cime2-interactive-mode-map "]" 'cime2-electric-rb)
(define-key cime2-interactive-mode-map "\t" 'cime2-indent-command)
(define-key cime2-interactive-mode-map "\C-c\C-i" 'cime2-interrupt-caml)
(define-key cime2-interactive-mode-map "\C-c\C-k" 'cime2-kill-caml)
(define-key cime2-interactive-mode-map "\C-m"
  'cime2-interactive-send-input)
(define-key cime2-interactive-mode-map "\C-j"
  'cime2-interactive-send-input-or-indent)
(define-key cime2-interactive-mode-map "\M-\C-m"
  'comint-send-input)
(define-key cime2-interactive-mode-map (read-kbd-macro "<kp-enter>")
  'comint-send-input)

(defconst cime2-interactive-buffer-name "*caml-toplevel*")

(defconst cime2-interactive-toplevel-error-regexp
  "Characters \\([0-9]+\\)-\\([0-9]+\\):"
  "Regular expression extracting the character numbers
from an error message produced by ocaml toplevel.")
(defvar cime2-interactive-last-phrase-pos-in-source 0)
(defvar cime2-interactive-last-phrase-pos-in-toplevel 0)

(defun cime2-interactive-filter (text)
  (if (< comint-last-input-end comint-last-input-start) ()
    (if (and cime2-with-xemacs cime2-interactive-read-only-input)
	(add-text-properties comint-last-input-start comint-last-input-end
			     (list 'read-only t)))
    (if (and (or cime2-window-system cime2-with-xemacs)
	     (featurep 'font-lock)
	     cime2-interactive-input-font-lock)
	(progn
	  (font-lock-fontify-region comint-last-input-start
				    comint-last-input-end)
	  (if (featurep 'sym-lock)
	      (sym-lock-make-symbols-atomic comint-last-input-start
					    comint-last-input-end)))))
  (if cime2-interactive-output-font-lock
      (save-excursion
	(add-text-properties
	 comint-last-input-end (point-max)
	 '(face cime2-font-lock-interactive-output-face))))
  (if cime2-interactive-error-font-lock
      (save-excursion
	(goto-char comint-last-input-end)
	(while (re-search-forward cime2-interactive-error-regexp nil t)
	  (add-text-properties
	   (match-beginning 1) (match-end 1)
	   '(face cime2-font-lock-interactive-error-face))))))

(defun cime2-interactive-mode ()
  "Major mode for interacting with a CiME2 process.
Runs a CiME2 toplevel as a subprocess of Emacs, with I/O through an
Emacs buffer. A history of input phrases is maintained. Phrases can
be sent from another buffer in CiME2 mode.

Special keys for CiME2 interactive mode:\\{cime2-interactive-mode-map}"
  (interactive)
  (comint-mode)
  (if (not (eq (cime2-install-font-lock) 'font-lock))
      ()
    (add-hook 'comint-output-filter-functions 'cime2-interactive-filter)
    (if (not (boundp 'after-change-functions))
	()
      (make-local-hook 'after-change-functions)
      (put 'after-change-functions 'permanent-local t)
      (remove-hook 'after-change-functions
		   'font-lock-after-change-function t))
    (if (not (boundp 'pre-idle-hook))
	()
      (make-local-hook 'pre-idle-hook)
      (put 'pre-idle-hook 'permanent-local t)
      (remove-hook 'pre-idle-hook
		   'font-lock-pre-idle-hook t)))
  (setq comint-prompt-regexp "^#")
  (setq major-mode 'cime2-interactive-mode)
  (setq mode-name "CiME2-Interactive")
  (setq comint-scroll-to-bottom-on-output t)
  (use-local-map cime2-interactive-mode-map)
  (set-syntax-table cime2-mode-syntax-table)
  (setq local-abbrev-table cime2-mode-abbrev-table)

  (if cime2-window-system (cime2-interactive-build-menu))

  ;; hooks for cime2-interactive-mode
  (run-hooks 'cime2-interactive-mode-hook))

(defun cime2-run-caml (&optional cmd)
  "Run a CiME2 toplevel process. I/O via buffer `*caml-toplevel*'."
  (interactive
   (list (if (not (comint-check-proc cime2-interactive-buffer-name))
	     (read-from-minibuffer "CiME2 toplevel to run: "
				   cime2-interactive-program))))
  (cime2-run-process-if-needed cmd)
  (switch-to-buffer-other-window cime2-interactive-buffer-name))

(defun cime2-run-process-if-needed (&optional cmd)
  (if (not cmd)
      (if (comint-check-proc cime2-interactive-buffer-name)
	  (setq cmd cime2-interactive-program)
	(setq cmd (read-from-minibuffer "CiME2 toplevel to run: "
					cime2-interactive-program))))
  (setq cime2-interactive-program cmd)
  (if (not (comint-check-proc cime2-interactive-buffer-name))
      (let ((cmdlist (cime2-args-to-list cmd))
            (process-connection-type nil))
	(set-buffer (apply (function make-comint) "caml-toplevel"
			   (car cmdlist) nil (cdr cmdlist)))
	(cime2-interactive-mode)
	(sleep-for 1))))

(defun cime2-args-to-list (string)
  (let ((where (string-match "[ \t]" string)))
    (cond ((null where) (list string))
	  ((not (= where 0))
	   (cons (substring string 0 where)
		 (cime2-args-to-list (substring string (+ 1 where)
						 (length string)))))
	  (t (let ((pos (string-match "[^ \t]" string)))
	       (if (null pos)
		   nil
		 (cime2-args-to-list (substring string pos
						 (length string)))))))))

(defun cime2-interactive-end-of-phrase ()
  (save-excursion
    (end-of-line)
    (cime2-find-meaningful-word)
    (cime2-find-meaningful-word)
    (looking-at ";;")))

(defun cime2-interactive-send-input ()
  "Process if the current line ends with `;;' then send the
current phrase else insert a newline."
  (interactive)
  (if (cime2-interactive-end-of-phrase)
      (comint-send-input)
    (insert "\n")
    (message "Phrase must end with `;;' to be processed by CiME2 toplevel...")))

(defun cime2-interactive-send-input-or-indent ()
  "Process if the current line ends with `;;' then send the
current phrase else insert a newline and indent."
  (interactive)
  (if (cime2-interactive-end-of-phrase)
      (comint-send-input)
    (insert "\n")
    (cime2-indent-command)
    (message "Phrase must end with `;;' to be processed by CiME2 toplevel...")))

(defun cime2-eval-region (start end)
  "Eval the current region in the CiME2 toplevel."
  (interactive "r")
  (save-excursion (cime2-run-process-if-needed))
  (comint-preinput-scroll-to-bottom)
  (setq cime2-interactive-last-phrase-pos-in-source start)
  (save-excursion
    (let ((text (buffer-substring-no-properties start end)))
      (goto-char end)
      (if (progn (cime2-find-meaningful-word) (cime2-find-meaningful-word)
		 (looking-at ";;")) ()
	(setq text (concat text "")))
      (if (string= text "")
	  (message "Cannot send empty commands to CiME2 toplevel!")
	(set-buffer cime2-interactive-buffer-name)
	(goto-char (point-max))
	(setq cime2-interactive-last-phrase-pos-in-toplevel (point))
	(if cime2-interactive-echo-phrase (insert text)
	  (comint-send-string cime2-interactive-buffer-name text)
	  (comint-send-string cime2-interactive-buffer-name "\n"))
	(comint-send-input)))
    (display-buffer cime2-interactive-buffer-name t)))

(defun cime2-eval-phrase ()
  "Eval the surrounding CiME2 phrase in the CiME2 toplevel."
  (interactive)
  (let ((end))
    (save-excursion
      (let ((pair (cime2-discover-phrase)))
	(setq end (nth 2 pair))
	(cime2-eval-region (nth 0 pair) (nth 1 pair))))
    (if cime2-skip-after-eval-phrase
	(goto-char end))))

(defun cime2-eval-buffer ()
  "Send the buffer to the inferior CiME2 process."
  (interactive)
  (cime2-eval-region (point-min) (point-max)))

(defun cime2-interactive-next-error ()
  (interactive)
  (let ((error-pos) (beg 0) (end 0))
    (save-excursion
      (set-buffer cime2-interactive-buffer-name)
      (goto-char cime2-interactive-last-phrase-pos-in-toplevel)
      (setq error-pos
	    (re-search-forward cime2-interactive-toplevel-error-regexp
			       (point-max) t))
      (if error-pos
	  (setq beg (string-to-int (match-string 1))
		end (string-to-int (match-string 2)))))
    (if (not error-pos)
	(message "No syntax or typing error in last phrase.")
      (setq beg (+ cime2-interactive-last-phrase-pos-in-source beg)
	    end (+ cime2-interactive-last-phrase-pos-in-source end))
      (goto-char beg) (push-mark end t t))))

(defun cime2-interrupt-caml ()
  (interactive)
  (if (comint-check-proc cime2-interactive-buffer-name)
      (save-excursion
	(set-buffer cime2-interactive-buffer-name)
	(comint-interrupt-subjob))))

(defun cime2-kill-caml ()
  (interactive)
  (if (comint-check-proc cime2-interactive-buffer-name)
      (save-excursion
	(set-buffer cime2-interactive-buffer-name)
	(comint-kill-subjob))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                               Menu support

(defun cime2-about () (interactive)
  (describe-variable 'cime2-mode-version))
(defun cime2-help () (interactive)
  (describe-function 'cime2-mode))
(defun cime2-interactive-help () (interactive)
  (describe-function 'cime2-interactive-mode))

(defvar cime2-definitions-menu-last-buffer nil)
(defvar cime2-definitions-keymaps nil)

(defun cime2-build-menu ()
  (if (condition-case nil (prog2 (require 'easymenu) nil) (error t))
      ()
    (easy-menu-define
     cime2-mode-menu (list cime2-mode-map)
     "CiME2 Mode Menu."
     '("CiME2"
       ("Interactive Mode"
	["Run CiME2 Toplevel" cime2-run-caml t]
	["Interrupt CiME2 Toplevel" cime2-interrupt-caml
	 :active (comint-check-proc cime2-interactive-buffer-name)]
	["Kill CiME2 Toplevel" cime2-kill-caml
	 :active (comint-check-proc cime2-interactive-buffer-name)]
	["Evaluate Region" cime2-eval-region :active (region-active-p)]
	["Evaluate Phrase" cime2-eval-phrase t]
	["Evaluate Buffer" cime2-eval-buffer t])
       ("CiME2 Forms"
	 ["try .. with .." cime2-insert-try-form t]
	 ["match .. with .." cime2-insert-match-form t]
	 ["let .. in .." cime2-insert-let-form t]
	 ["if .. then .. else .." cime2-insert-if-form t]
	 ["while .. do .. done" cime2-insert-while-form t]
	 ["for .. do .. done" cime2-insert-for-form t]
	 ["begin .. end" cime2-insert-begin-form t])
       ["Switch .ml/.mli" cime2-find-alternate-file t]
       "---"
       ["Compile..." compile t]
       ["Reference manual..." cime2-browse-manual t]
       ["CiME2 library..." cime2-browse-library t]
       ("Definitions"
	:filter cime2-update-definitions-menu
	["Scan..." cime2-list-definitions t])
       "---"
       ("CiME2 Options" ["Dummy" nil t])
       ("CiME2 Interactive Options" ["Dummy" nil t])
       "---"
       ["About" cime2-about t]
       ["Help" cime2-help t]))
    (easy-menu-add cime2-mode-menu)
    (cime2-update-options-menu)
    ;; save and update definitions menu
    (make-local-variable 'cime2-definitions-menu)
    (if (or cime2-with-xemacs
	    (not (functionp 'easy-menu-create-keymaps))) ()
      ;; patch for Emacs 20.2
      (add-hook 'menu-bar-update-hook
		'cime2-with-emacs-update-definitions-menu)
      (make-local-variable 'cime2-definitions-keymaps)
      (setq cime2-definitions-keymaps
	    (cdr (easy-menu-create-keymaps
		  "Definitions" cime2-definitions-menu)))
      (setq cime2-definitions-menu-last-buffer nil))))

(defun cime2-interactive-build-menu ()
  (if (condition-case nil (prog2 (require 'easymenu) nil) (error t))
      ()
    (easy-menu-define
     cime2-interactive-mode-menu (list cime2-interactive-mode-map)
     "CiME2 Interactive Mode Menu."
     '("CiME2"
       ("Interactive Mode"
	["Run CiME2 Toplevel" cime2-run-caml t]
	["Interrupt CiME2 Toplevel" cime2-interrupt-caml
	 :active (comint-check-proc cime2-interactive-buffer-name)]
	["Kill CiME2 Toplevel" cime2-kill-caml
	 :active (comint-check-proc cime2-interactive-buffer-name)]
	["Evaluate Region" cime2-eval-region :active (region-active-p)]
	["Evaluate Phrase" cime2-eval-phrase t]
	["Evaluate Buffer" cime2-eval-buffer t])
       "---"
       ("CiME2 Options" ["Dummy" nil t])
       ("CiME2 Interactive Options" ["Dummy" nil t])
       "---"
       ["Help" cime2-interactive-help t]))
    (easy-menu-add cime2-interactive-mode-menu)
    (cime2-update-options-menu)))

(defun cime2-update-definitions-menu (list)
   cime2-definitions-menu)

(defun cime2-with-emacs-update-definitions-menu ()
  (if (current-local-map)
      (let ((keymap
	     (lookup-key (current-local-map) [menu-bar CiME2 Definitions])))
	(if (and
	     (keymapp keymap)
	     (not (eq cime2-definitions-menu-last-buffer (current-buffer))))
	    (setcdr keymap cime2-definitions-keymaps)
	  (setq cime2-definitions-menu-last-buffer (current-buffer))))))

(defun cime2-toggle-option (symbol)
  (interactive)
  (cond
   ((eq symbol 'cime2-labl-support) (cime2-do-support-labl))
   (t (set symbol (not (symbol-value symbol)))))
  (if cime2-with-xemacs nil (cime2-update-options-menu)))

(defun cime2-update-options-menu ()
  (easy-menu-change
   '("CiME2") "CiME2 Options"
   (mapcar (lambda (pair)
	     (if (consp pair)
		 (vector (car pair)
			 (list 'cime2-toggle-option (cdr pair))
			 :style 'toggle
			 :selected (nth 1 (cdr pair))
			 :active t)
	       pair)) cime2-options-list))
  (easy-menu-change
   '("CiME2") "CiME2 Interactive Options"
   (mapcar (lambda (pair)
	     (if (consp pair)
		 (vector (car pair)
			 (list 'cime2-toggle-option (cdr pair))
			 :style 'toggle
			 :selected (nth 1 (cdr pair))
			 :active t)
	       pair)) cime2-interactive-options-list)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                             Browse Manual

;; From M. Quercia

(defun cime2-browse-manual ()
  "*Browse CiME2 reference manual."
  (interactive)
  (setq cime2-manual-url (read-from-minibuffer "URL: " cime2-manual-url))
  (funcall cime2-browser cime2-manual-url))

(defun cime2-xemacs-w3-manual (url)
  "*Browse CiME2 reference manual."
  (w3-fetch-other-frame url))

(defun cime2-netscape-manual (url)
  "*Browse CiME2 reference manual."
  (start-process-shell-command
   "netscape" nil
   (concat "netscape -remote 'openURL ("
	   url ", newwindow)' || netscape " url)))

(defun cime2-mmm-manual (url)
  "*Browse CiME2 reference manual."
  (start-process-shell-command
   "mmm" nil
   (concat "mmm_remote " url " || mmm -external " url)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                             Browse Library

;; From M. Quercia

(defun cime2-browse-library()
  "Browse the CiME2 library."
  (interactive)
  (let ((buf-name "*caml-library*") (opoint)
	(dir (read-from-minibuffer "Library path: " cime2-library-path)))
    (if (and (file-directory-p dir) (file-readable-p dir))
	(progn
	  (setq cime2-library-path dir)
	  ;; list *.ml and *.mli files
	  (with-output-to-temp-buffer buf-name
	    (buffer-disable-undo standard-output)
	    (save-excursion
	      (set-buffer buf-name)
	      (kill-all-local-variables)
	      (make-local-variable 'cime2-library-path)
	      (setq cime2-library-path dir)
	      ;; help
	      (insert "Directory \"" dir "\".\n") 
	      (insert "Select a file with middle mouse button or RETURN.\n\n")
	      (insert "Interface files (.mli):\n\n")
	      (insert-directory (concat dir "/*.mli") "-C" t nil)
	      (insert "\n\nImplementation files (.ml):\n\n")
	      (insert-directory (concat dir "/*.ml") "-C" t nil)
	      ;; '.', '-' and '_' are now letters
	      (modify-syntax-entry ?. "w")
	      (modify-syntax-entry ?- "w")
	      (modify-syntax-entry ?_ "w")
	      ;; every file name is now mouse-sensitive
	      (goto-char (point-min))
	      (while (< (point) (point-max))
		(re-search-forward "\\.ml.?\\>")
		(setq opoint (point))
		(re-search-backward "\\<" (point-min) 1)
		(put-text-property (point) opoint 'mouse-face 'highlight)
		(goto-char (+ 1 opoint)))
	      ;; activate cime2-library mode
	      (setq major-mode 'cime2-library-mode)
	      (setq mode-name "cime2-library")
	      (use-local-map cime2-library-mode-map)
	      (setq buffer-read-only t)))))))
  
(setq cime2-library-mode-map (make-keymap))
(suppress-keymap cime2-library-mode-map)
(define-key cime2-library-mode-map [return] 'cime2-library-find-file)
(define-key cime2-library-mode-map [mouse-2] 'cime2-library-mouse-find-file)
  
(defun cime2-library-find-file ()
  "Load the file whose name is near point."
  (interactive)
  (save-excursion
    (if (text-properties-at (point))
	(progn
	  (re-search-backward "\\<") (setq beg (point))
	  (re-search-forward "\\>")
	  (find-file-read-only (concat cime2-library-path "/"
				       (buffer-substring-no-properties
					beg (point))))))))

(defun cime2-library-mouse-find-file (event)
  "Visit the file name you click on."
  (interactive "e")
  (let ((owindow (selected-window)))
    (mouse-set-point event)
    (cime2-library-find-file)
    (select-window owindow)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                             Definitions List

;; Designed from original code by M. Quercia

(defconst cime2-definitions-regexp
  "\\<\\(and\\|val\\|value\\|type\\|module\\|class\\|exception\\|let\\)\\>"
  "Regexp matching definition phrases.")

(defconst cime2-definitions-bind-skip-regexp
  "\\<\\(rec\\|type\\|virtual\\)\\>\\|'[A-Za-z\300-\377][0-9_'A-Za-z\300-\377]*\\|('.*)"
  "Regexp matching stuff to ignore after a binding keyword.")

(defvar cime2-definitions-menu (list ["Scan..." cime2-list-definitions t])
  "Initial content of the definitions menu.")

(defun cime2-list-definitions()
  "Parses the buffer and gathers toplevel definitions for quick
jump via the definitions menu'."
  (interactive)
  (message "Searching definitions...")
  (save-excursion
    (let ((cpt 0) (kw) (menu)
	  (value-list) (type-list) (module-list) (class-list) (misc-list))
      (goto-char (point-min))
      (cime2-skip-blank-and-comments)
      (while (< (point) (point-max))
	(if (looking-at cime2-definitions-regexp)
	    (progn
	      (setq kw (match-string 0))
	      (if (string= kw "and")
		  (setq kw (save-match-data
			     (save-excursion (cime2-find-and-match)))))
	      (if (or (string= kw "exception")
		      (string= kw "val")
		      (string= kw "value")) (setq kw "let"))
	      ;; skip optional elements
	      (goto-char (match-end 0))
	      (cime2-skip-blank-and-comments)
	      (if (looking-at cime2-definitions-bind-skip-regexp)
		  (goto-char (match-end 0)))
	      (cime2-skip-blank-and-comments)
	      (if (looking-at "\\<[A-Za-z\300-\377][0-9_'A-Za-z\300-\377]*\\>")
		  ;; menu item : [name (goto-char ...) t]
		  (let* ((p (make-marker))
			 (ref (vector (match-string 0)
				      (list 'cime2-goto p) t)))
		    (setq cpt (1+ cpt))
		    (message (concat "Searching definitions... ("
				     (number-to-string cpt) ")"))
		    (set-marker p (point))
		    (cond
		     ((string= kw "let")
		      (setq value-list (cons ref value-list)))
		     ((string= kw "type")
		      (setq type-list (cons ref type-list)))
		     ((string= kw "module")
		      (setq module-list (cons ref module-list)))
		     ((string= kw "class")
		      (setq class-list (cons ref class-list)))
		     (t (setq misc-list (cons ref misc-list))))))))
	;; skip to next phrase or next top-level `and'
	(cime2-forward-char)
	(let ((old-point (point)) (last-and))
	  (cime2-next-phrase t)
	  (setq last-and (point))
	  (save-excursion
	    (while (and (re-search-backward "\\<and\\>" old-point t)
			(not (cime2-in-literal-or-comment-p))
			(save-excursion (cime2-find-and-match)
					(>= old-point (point))))
	      (setq last-and (point))))
	  (goto-char last-and)))
      ;; sort and build lists
      (mapcar (lambda (pair)
		(if (cdr pair)
		    (setq menu
			  (append (cime2-split-long-list
			    (car pair) (cime2-sort-definitions (cdr pair)))
				  menu))))
	      (list (cons "Miscellaneous" misc-list)
		    (cons "Values" value-list)
		    (cons "Classes" class-list)
		    (cons "Types" type-list)
		    (cons "Modules" module-list)))
      ;; update definitions menu
      (setq cime2-definitions-menu
	    (append menu (list "---" ["Rescan..." cime2-list-definitions t])))
      (if (or cime2-with-xemacs
	      (not (functionp 'easy-menu-create-keymaps))) ()
	;; patch for Emacs 20.2
	(setq cime2-definitions-keymaps
	      (cdr (easy-menu-create-keymaps 
		    "Definitions" cime2-definitions-menu)))
	(setq cime2-definitions-menu-last-buffer nil))
      (message "Searching definitions... done"))))

(defun cime2-goto (pos)
  (goto-char pos)
  (recenter))

(defun cime2-sort-definitions (list)
  (let* ((last "") (cpt 1)
	 (list (sort (nreverse list)
		     (lambda (p q) (string< (elt p 0) (elt q 0)))))
	 (tail list))
    (while tail
      (if (string= (elt (car tail) 0) last)
	  (prog2
	      (setq cpt (1+ cpt))
	      (aset (car tail) 0 (format "%s (%d)" last cpt)))
	(setq cpt 1)
	(setq last (elt (car tail) 0)))
      (setq tail (cdr tail)))
    list))

;; look for the (n-1)th or last element of a list
(defun cime2-nth (n list)
  (if (or (<= n 1) (null list) (null (cdr list))) list
    (cime2-nth (1- n) (cdr list))))
    
;; split a definition list if it is too long
(defun cime2-split-long-list (title list)
  (let ((tail (cime2-nth cime2-definitions-max-items list)))
    (if (or (null tail) (null (cdr tail)))
        ;; list not too long, cons the title
        (list (cons title list))
      ;; list too long, split and add initials to the title
      (let (lists)
        (while list
          (let ((beg (substring (elt (car list) 0) 0 1))
                (end (substring (elt (car tail) 0) 0 1)))
            (setq lists (cons
                         (cons (format "%s %s-%s" title beg end) list)
                         lists))
            (setq list (cdr tail))
            (setcdr tail nil)
            (setq tail (cime2-nth cime2-definitions-max-items list))))
        (nreverse lists)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                             Hooks and Exit

(defvar cime2-load-hook nil
  "This hook is run when CiME2 is loaded in. It is a good place to put
key-bindings or hack Font-Lock keywords...")

(run-hooks 'cime2-load-hook)

(provide 'cime2)
